// -*- c -*- 
// (setq c-basic-offset 4)
// "clsConfItem.h"

#include "clsConfItem.h"

#include <fstream>
#include <sstream>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "../COMMON/CommonDefinitions.h"

#ifndef _ASSERT
#define _ASSERT assert
#endif

ConfItem_cls::
ConfItem_cls(int type){
    mstrHeader = wxT("NONE");
    mstrBody   = wxT("NONE");
    mType = type;
}

void ConfItem_cls::
ReplaceSeparator(){
    for (unsigned int i = 0; i < mstrBody.size(); i++){
	if (mstrBody[i] == '\\') mstrBody[i] = '/';
	if (mstrBody[i] == ' '){
	    mstrBody.insert(i, L"\\"); 
	    i++; }
    }
}

bool ConfItem_cls::
ReadStrLineBase(wxString strLine){
    int loc = strLine.Find(mstrHeader);
    if (loc != wxNOT_FOUND){
	mstrBody = strLine.Mid(mstrHeader.Len());
	mstrBody.Replace(wxT("\\n"), wxT("\n"), TRUE);
	return TRUE;
    }
    return FALSE;
}

//----------------------------------

unsigned int ConfArray_cls::
CountDimension(wxString strBody){
    unsigned int dimension = 0;
    std::wstring::size_type pos = 0;
    while (pos != std::wstring::npos){
	pos = strBody.find(L" ", pos + 1);
	dimension++;
    }
    _ASSERT(dimension <= _CONFARRAY_MAX);
    return dimension;
}

//----------------------------------

ConfItemSTR_cls::
ConfItemSTR_cls() : ConfItem_cls (_CONF_ITEM_STR){
}

ConfItemSTR_cls::
ConfItemSTR_cls(wxString strHeader, wxString strBody) : 
ConfItem_cls (_CONF_ITEM_STR)
{
    SetItem(strHeader, strBody);
}

void ConfItemSTR_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody = strBody;
}

bool ConfItemSTR_cls::
ReadStrLine(wxString strLine){
    return ReadStrLineBase(strLine);
}

wxString ConfItemSTR_cls::
Serialize(){
    if (mstrBody == wxT("!!SKIP!!")) return wxT("");
    wxString strCompact = mstrBody;
    strCompact.Replace(wxT("\n"), wxT("\\n"), TRUE);
    return mstrHeader + strCompact + ENDL;
}

wxString ConfItemSTR_cls::
GetItemStr(){
    return mstrBody;
}

//------------------------------------

ConfItemINT_cls::
ConfItemINT_cls():
ConfItem_cls(_CONF_ITEM_INT){}

ConfItemINT_cls::
ConfItemINT_cls(wxString strHeader, wxString strBody) :
ConfItem_cls(_CONF_ITEM_INT)
{
    SetItem(strHeader, strBody);
}

void ConfItemINT_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody = strBody;
#ifdef WIN32
    mBody = _wtoi(mstrBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mBody = atoi(wxBody.ToAscii());
#endif
}

bool ConfItemINT_cls::
ReadStrLine(wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mBody = _wtoi(mstrBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mBody = atoi(wxBody.ToAscii());
#endif
    return TRUE;
}

wxString ConfItemINT_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mBody << ENDL;
    return sOut;
}

wxString ConfItemINT_cls::
GetItemStr(){
    wxString sOut;
    sOut << mBody ;
    return sOut;
}

// -------------

ConfItemINTARRAY_cls::
ConfItemINTARRAY_cls() : ConfItem_cls(_CONF_ITEM_INTARRAY){
}

void ConfItemINTARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;   
    mstrBody = strBody;
    mDimension = CountDimension(strBody);
    wxString strTmp;
    strTmp << mstrHeader.c_str() << strBody.c_str();
    ReadStrLine(strTmp);
}

bool ConfItemINTARRAY_cls::
ReadStrLine(wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::string strBuffer = (char*)mstrBody.char_str();
    std::stringstream strTmp(strBuffer/*wxConvFile.cWX2WC(mstrBody)*/);
    mDimension = 0;
    while (!strTmp.eof() || strTmp.fail()){
	strTmp >> mpBody[mDimension];
	mDimension++;
    }
    /*
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpBody[i];
    */
    return TRUE;
}

wxString ConfItemINTARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mpBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    sOut << ENDL;
    return sOut;
}

wxString ConfItemINTARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpBody[0];
    for (unsigned int i = 0; i < mDimension; i++) 
	sOut << wxT(" ") << mpBody[i];
    return sOut;
}

//------------------------------------

ConfItemUINT_cls::
ConfItemUINT_cls() : ConfItem_cls (_CONF_ITEM_UINT) {
}

ConfItemUINT_cls::
ConfItemUINT_cls(wxString strHeader, wxString strBody) :
ConfItem_cls (_CONF_ITEM_UINT)
{
    SetItem(strHeader, strBody);
}

void ConfItemUINT_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody   = strBody;
#ifdef WIN32
    mBody = _wtoi(mstrBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mBody = atoi(wxBody.ToAscii());
#endif
}

bool ConfItemUINT_cls::
ReadStrLine(/*std::wstring*/wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mBody = _wtoi(mstrBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mBody = atoi(wxBody.ToAscii());
#endif
    return TRUE;
}

wxString ConfItemUINT_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mBody << ENDL;
    return sOut;
}

wxString ConfItemUINT_cls::
GetItemStr(){
    wxString sOut;
    sOut << mBody ;
    return sOut;
}

//-------------------------------------

void ConfItemUINTARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody = strBody;
    mDimension = CountDimension(strBody);
    wxString strTmp;
    strTmp << mstrHeader.c_str() << strBody.c_str();
    ReadStrLine(strTmp);
}

bool ConfItemUINTARRAY_cls::
ReadStrLine(/*std::wstring*/wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::wstringstream strTmp(wxConvFile.cWX2WC(mstrBody));
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpBody[i];
    return TRUE;
}

wxString ConfItemUINTARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mpBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    sOut << ENDL;
    return sOut;
}

wxString ConfItemUINTARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpBody[0];
    for (unsigned int i = 0; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    return sOut;
}

//-------------------------------------

ConfItemFLOAT_cls::
ConfItemFLOAT_cls():
ConfItem_cls(_CONF_ITEM_FLOAT){}

ConfItemFLOAT_cls::
ConfItemFLOAT_cls(wxString strHeader, wxString strBody) : 
ConfItem_cls(_CONF_ITEM_FLOAT)
{
    SetItem(strHeader, strBody);
}

void ConfItemFLOAT_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody   = strBody;
#ifdef WIN32
    mBody = _wtof(mstrBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mBody = atof(wxBody.ToAscii());
#endif
}

bool ConfItemFLOAT_cls::
ReadStrLine(/*std::wstring*/wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mBody = _wtof(mstrBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mBody = atof(wxBody.ToAscii());
#endif
    return TRUE;
}

wxString ConfItemFLOAT_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mBody << ENDL;
    return sOut;
}

wxString ConfItemFLOAT_cls::
GetItemStr(){
    wxString sOut;
    sOut << mBody ;
    return sOut;
}

// ------------------------------------

ConfItemFLOATARRAY_cls::
ConfItemFLOATARRAY_cls() : ConfItem_cls (_CONF_ITEM_FLOATARRAY) {
}

ConfItemFLOATARRAY_cls::
ConfItemFLOATARRAY_cls(wxString strHeader, wxString strBody) : 
ConfItem_cls (_CONF_ITEM_FLOATARRAY)
{
    SetItem(strHeader, strBody);
}

void ConfItemFLOATARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody = strBody;
    mDimension = CountDimension(strBody);
    wxString strTmp;
    strTmp << mstrHeader.c_str() << strBody.c_str();
    ReadStrLine(strTmp);
}

bool ConfItemFLOATARRAY_cls::
ReadStrLine(wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::wstringstream strTmp(wxConvFile.cWX2WC(mstrBody));
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpBody[i];
    return TRUE;
}

wxString ConfItemFLOATARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mpBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    sOut << ENDL;
    return sOut;
}

wxString ConfItemFLOATARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpBody[0];
    for (unsigned int i = 0; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    return sOut;
}

// ------------------------------------

ConfItemDOUBLEARRAY_cls::
ConfItemDOUBLEARRAY_cls() : ConfItem_cls(_CONF_ITEM_DOUBLEARRAY) {
}

ConfItemDOUBLEARRAY_cls::
ConfItemDOUBLEARRAY_cls(wxString strHeader, wxString strBody) :
ConfItem_cls(_CONF_ITEM_DOUBLEARRAY)
{
    SetItem(strHeader, strBody);
}

void ConfItemDOUBLEARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody = strBody;
    mDimension = CountDimension(strBody);
    wxString strTmp;
    strTmp << mstrHeader.c_str() << strBody.c_str();
    ReadStrLine(strTmp);
}

bool ConfItemDOUBLEARRAY_cls::
ReadStrLine(wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::wstringstream strTmp(wxConvFile.cWX2WC(mstrBody));
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpBody[i];
    return TRUE;
}

wxString ConfItemDOUBLEARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mpBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    sOut << ENDL;
    return sOut;
}

wxString ConfItemDOUBLEARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpBody[0];
    for (unsigned int i = 0; i < mDimension; i++)
	sOut << wxT(" ") << mpBody[i];
    return sOut;
}

// ------------------------------------

ConfItemDOUBLE_cls::
ConfItemDOUBLE_cls() : ConfItem_cls(_CONF_ITEM_DOUBLE){
}

void ConfItemDOUBLE_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody   = strBody;
#ifdef WIN32
    mBody = _wtof(mstrBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mBody = atof(wxBody.ToAscii());
#endif
}

bool ConfItemDOUBLE_cls::
ReadStrLine(/*std::wstring*/wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mBody = _wtof(mstrBody.c_str());
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mBody = atof(wxBody.ToAscii());
#endif
    return TRUE;
}

wxString ConfItemDOUBLE_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mBody << ENDL;
    return sOut;
}

wxString ConfItemDOUBLE_cls::
GetItemStr(){
    wxString sOut;
    sOut << mBody ;
    return sOut;
}

// ------------------------------------

ConfItemBOOL_cls::
ConfItemBOOL_cls():ConfItem_cls(_CONF_ITEM_BOOL){
}

ConfItemBOOL_cls::
ConfItemBOOL_cls(wxString strHeader, wxString strBody) : 
ConfItem_cls(_CONF_ITEM_BOOL){
    SetItem(strHeader, strBody);
}

void ConfItemBOOL_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody = strBody;
#ifdef WIN32
    mbBody = ( _wtoi(mstrBody.c_str()) != 0);
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mbBody = ( atoi(wxBody.ToAscii()) != 0);
#endif
}

bool ConfItemBOOL_cls::
ReadStrLine(/*std::wstring*/wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
#ifdef WIN32
    mbBody = ( _wtoi(mstrBody.c_str()) != 0);
#else
    wxString wxBody = wxConvFile.cWC2WX(mstrBody.c_str());
    mbBody = ( atoi(wxBody.ToAscii()) != 0);
#endif
    return TRUE;
}

wxString ConfItemBOOL_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mbBody << ENDL;
    return sOut;
}

wxString ConfItemBOOL_cls::
GetItemStr(){
    wxString sOut;
    sOut << mbBody ;
    return sOut;
}

//-------------------------------------

ConfItemBOOLARRAY_cls::
ConfItemBOOLARRAY_cls() : ConfItem_cls(_CONF_ITEM_DOUBLEARRAY){
}

void ConfItemBOOLARRAY_cls::
SetItem(wxString strHeader, wxString strBody){
    mstrHeader = strHeader;
    if (strBody == wxT("")) return;
    mstrBody = strBody;
    mDimension = CountDimension(strBody);
    wxString strTmp;
    strTmp << mstrHeader.c_str() << strBody.c_str();
    ReadStrLine(strTmp);
}

bool ConfItemBOOLARRAY_cls::
ReadStrLine(wxString strLine){
    if (ReadStrLineBase(strLine) == FALSE) return FALSE;
    std::wstringstream strTmp(wxConvFile.cWX2WC(mstrBody));
    for (unsigned int i = 0; i < mDimension; i++)
	strTmp >> mpbBody[i];
    return TRUE;
}

wxString ConfItemBOOLARRAY_cls::
Serialize(){
    wxString sOut;
    sOut << mstrHeader << mpbBody[0];
    for (unsigned int i = 1; i < mDimension; i++)
	sOut << wxT(" ") << mpbBody[i];
    sOut << ENDL;
    return sOut;
}

wxString ConfItemBOOLARRAY_cls::
GetItemStr(){
    wxString sOut;
    sOut << mpbBody[0];
    for (unsigned int i = 0; i < mDimension; i++)
	sOut << wxT(" ") << mpbBody[i];
    return sOut;
}

// --------------------------------

ConfItemList_cls::
ConfItemList_cls(){
    mbSetWithInitValue = TRUE;
}

void ConfItemList_cls::
Clear(){
    mVpCList.clear();
}

void ConfItemList_cls::
Add(ConfItem_cls* pC, wxString strHeader, wxString strBody){
    if (mbSetWithInitValue == TRUE)
	pC->SetItem(strHeader, strBody);
    else
	pC->SetItem(strHeader, wxT(""));
    mVpCList.push_back(pC);
}

wxString ConfItemList_cls::
GetHeaderStrFromPointer(void *pValue){
    wxString strHeader;
    for (unsigned int i = 0; i < mVpCList.size(); i++){
	switch(mVpCList[i]->mType){
	case _CONF_ITEM_INT:
	    if (pValue == &(((ConfItemINT_cls*)mVpCList[i])->mBody))
		strHeader = mVpCList[i]->mstrHeader;
	    break;
	case _CONF_ITEM_INTARRAY:{
	    ConfItemINTARRAY_cls *pCItem =
		(ConfItemINTARRAY_cls*)mVpCList[i];
	    int dimension = pCItem->mDimension;
	    for (int dim = 0; dim < dimension; dim++)
		if (pValue == &(pCItem->mpBody[dim])){
		    strHeader = mVpCList[i]->mstrHeader;
		    strHeader << wxT("::DIM=") << dim;
		}
	    break; }
	case _CONF_ITEM_UINT:
	    if (pValue == &(((ConfItemUINT_cls*)mVpCList[i])->mBody))
		strHeader = mVpCList[i]->mstrHeader;
	    break;
	case _CONF_ITEM_FLOAT:
	    if (pValue == &(((ConfItemFLOAT_cls*)mVpCList[i])->mBody))
		strHeader = mVpCList[i]->mstrHeader;
	    break;
	case _CONF_ITEM_FLOATARRAY:{
	    ConfItemFLOATARRAY_cls *pCItem =
		(ConfItemFLOATARRAY_cls*)mVpCList[i];
	    int dimension = pCItem->mDimension;
	    for (int dim = 0; dim < dimension; dim++)
		if (pValue == &(pCItem->mpBody[dim])){
		    strHeader = mVpCList[i]->mstrHeader;
		    strHeader << wxT("::DIM=") << dim;
		}
	    break; }
	case _CONF_ITEM_DOUBLE:
	    if (pValue == &(((ConfItemDOUBLE_cls*)mVpCList[i])->mBody))
		strHeader = mVpCList[i]->mstrHeader;
	    break;
	case _CONF_ITEM_DOUBLEARRAY:{
	    ConfItemDOUBLEARRAY_cls *pCItem =
		(ConfItemDOUBLEARRAY_cls*)mVpCList[i];
	    int dimension = pCItem->mDimension;
	    for (int dim = 0; dim < dimension; dim++)
		if (pValue == &(pCItem->mpBody[dim])){
		    strHeader = mVpCList[i]->mstrHeader;
		    strHeader << wxT("::DIM=") << dim;
		}
	    break; }
	}
    }
    return strHeader;
}

ConfItem_cls* ConfItemList_cls::
GetPCItem(wxString strHeader){
    for (unsigned int i = 0; i < mVpCList.size(); i++)
	if (strHeader == mVpCList[i]->mstrHeader)
	    return mVpCList[i];
    return NULL;
}

#define MACRO_COPY(TYPE)\
    case _CONF_ITEM_##TYPE:						\
    *(ConfItem##TYPE##_cls*)(mVpCList[i]) =				\
	*(ConfItem##TYPE##_cls*)(pCFrom->mVpCList[i]);		\
    break;

void ConfItemList_cls::
CopyListComponentFromP(ConfItemList_cls *pCFrom){
    for (unsigned int i = 0; i < mVpCList.size(); i++){
	switch(mVpCList[i]->mType){
	    MACRO_COPY(STR);
	    MACRO_COPY(INT);
	    MACRO_COPY(INTARRAY);
	    MACRO_COPY(FLOAT);
	    MACRO_COPY(FLOATARRAY);
	    MACRO_COPY(DOUBLE);
	    MACRO_COPY(DOUBLEARRAY);
	    MACRO_COPY(BOOL);
	    MACRO_COPY(BOOLARRAY);
	    MACRO_COPY(UINT);
	    MACRO_COPY(UINTARRAY);
	    /*
	case _CONF_ITEM_FLOATARRAY:
	    *(ConfItemFLOATARRAY_cls*)(mVpCList[i]) = 
		*(ConfItemFLOATARRAY_cls*)(pCFrom->mVpCList[i]);
	    break;
	    */
	}
    }
}

wxString ConfItemList_cls::
SerializeAll(){
    wxString sOut;
    for (unsigned int i = 0; i < mVpCList.size(); i++)
	sOut << mVpCList[i]->Serialize();
    return sOut;
}

bool ConfItemList_cls::
ReadStrLineAll(wxString strLine){
    bool bRet = FALSE;
    for (unsigned int i = 0; i < mVpCList.size(); i++)
	if (mVpCList[i]->ReadStrLine(strLine) == TRUE) bRet = TRUE;
    return bRet;
}
