// -*- c++ -*- 
// (setq c-basic-offset 4)

#include "clsFilterBG.h"

#include "../COMMON/MyVector.h"
#include "../COMMON/clsFilename.h"
#include "../GLOBJECTS/clsSphere.h"
extern Sphere_cls *GpCSphere;
#include "../MAINCANVAS/clsFilterPanelPain.h"
#include "../POVRAY/clsPovRay.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "../UNDO/clsUndoItemConstParam.h"
#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsHookCtrl.h"

#include "../WIDGETCTRL/EX/clsFileDialogEX.h"
extern wxFileDialogEX *GpCFileDialog;

#include "../../wxWidgetEnumMenu.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include <wx/file.h>

FilterBGGUI_cls::							
FilterBGGUI_cls(FilterBG_cls *pCParent, 
		FilterPanelPain_cls *pCPanelPain) : FilterGUI_cls(pCParent) {
    mpCParent = pCParent;
    SetMNumPanel((wxBoxSizer*)pCPanelPain->GetSizer(), pCPanelPain, 0);
}

void FilterBGGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, int index){
    
    MultiInputPanel_cls* pCurr = InitializeMInputCommon
	(pCParentWindow, mpCParent->GetPCData()->mCstrCustomName.mstrBody);
    FilterBGData_cls *pCData = mpCParent->GetPCData();

    pCurr->mUpdateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM;
    
    int boxWidth = 2;
    
    pCurr->mTextWidth = 105;
    pCurr->mUndoType = _MN_UNDO_AUTO_TAG;

    pCurr->mItemWidth = 160;
    wxString listType[3] = {_("Star Field"), _("Texture Map"), _("Color Curve")};
    mIndexType = pCurr->CombinePointerList(_ITEM_UINT_LIST, &pCData->mCBGType.mBody, 3, listType, _("BG Type"));

    pCurr->mItemWidth = 80;
    mIndexStarNum = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Num/Size"));
    pCurr->CombinePointer(_ITEM_INTEGER, &pCData->mCStarNumber.mBody);
    pCurr->CombinePointer(_ITEM_FLOAT, &pCData->mCStarScale.mBody);
    
    mIndexBright = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Alpha/Scale"));
    pCurr->CombinePointer(_ITEM_FLOAT, &pCData->mCBrightness.mBody);
    pCurr->CombinePointer(_ITEM_FLOAT, &pCData->mCBGScale.mBody);
    
    //pCurr->mItemWidth = -1;
    pCurr->mItemWidth = 160;
    wxString mapType[2] = {_("Fisheye"), _("Spherical")};
    mIndexMappingType = pCurr->CombinePointerList(_ITEM_UINT_LIST, &pCData->mCMappingType.mBody, 2, mapType, _("Map Type"));

    //boxWidth = 1;
    pCurr->mItemWidth = 160;
    //pCurr->mItemWidth = -1;
    wxString clrType[2] = {_("Odd"), _("Even")};
    mIndexColorType = pCurr->CombinePointerList(_ITEM_UINT_LIST, &pCData->mCColorMapType.mBody, 2, clrType, _("Map Type"));
    mIndexColor = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Color"));
    pCurr->mItemWidth = -1;
    ColorCurveBtn_cls *pCBGColor = 
	new ColorCurveBtn_cls(pCurr, pCData->mVpCClrCurve[0], TRUE, TRUE);
    pCurr->CombinePointer(_ITEM_WXPANEL, pCBGColor);

    boxWidth = 2;
    pCurr->mItemWidth = pCurr->mMinimumWidth - 30 + 10;    
    mIndexTextureA = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->mUndoType = _MN_UNDO_MANUAL;
    pCurr->CombinePointer(_ITEM_STRING, &pCData->mCstrTextureA.mstrBody);
    pCurr->mItemWidth = 30;
    pCurr->CombinePointerButton(wxID_ANY, wxT(".."));
    
    pCurr->mItemWidth = pCurr->mMinimumWidth - 30 + 10;    
    mIndexTextureB = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->mUndoType = _MN_UNDO_MANUAL;
    pCurr->CombinePointer(_ITEM_STRING, &pCData->mCstrTextureB.mstrBody);
    pCurr->mItemWidth = 30;
    pCurr->CombinePointerButton(wxID_ANY, wxT(".."));

    std::vector < bool > VbValid;
    VbValid.push_back(FALSE);
    VbValid.push_back(TRUE);
    VbValid.push_back(FALSE);
    std::vector < int > VIndex = SetMNumBasic(pCurr, pCData, VbValid);
    
    pCurr->FinishSetting(NULL);

    pCurr->SetColorForXYZ(VIndex[0]+1);

    SetHookBasic(pCurr);
    
    int HK = _HOOK_COMBOBOX;
    pCurr->SetFuncHook(FuncHookType, HK, mIndexType, FALSE, this);
    pCurr->SetFuncHook(FuncHookType, HK, mIndexMappingType, FALSE, this);
    pCurr->SetFuncHook(FuncHookType, HK, mIndexColorType, FALSE, this);
    pCurr->SetFuncHook(FuncHookTextureA, _HOOK_BUTTON, 
		       mIndexTextureA+2, FALSE, this);
    pCurr->SetFuncHook(FuncHookTextureB, _HOOK_BUTTON, 
		       mIndexTextureB+2, FALSE, this);

    HideAndShow();
}

void FilterBGGUI_cls::
FuncHookType(void *pUsrDef, void *pArg){
    ((FilterBGGUI_cls*)pUsrDef)->HideAndShow();
}

void FilterBGGUI_cls::
FuncHookTextureA(void *pUsrDef, void *pArg){
    ((FilterBGGUI_cls*)pUsrDef)->PathSelection(0);
}

void FilterBGGUI_cls::
FuncHookTextureB(void *pUsrDef, void *pArg){
    ((FilterBGGUI_cls*)pUsrDef)->PathSelection(1);
}

void FilterBGGUI_cls::
HideAndShow(){
    HideAndShowBegin();
    FilterBGData_cls *pCData = mpCParent->GetPCData();

    switch(pCData->mCBGType.mBody){
    case _BG_STARS:
	mpCMNumPanel->ShowItem(mIndexStarNum, TRUE);
	mpCMNumPanel->ShowItem(mIndexMappingType, FALSE);
	mpCMNumPanel->ShowItem(mIndexTextureA, FALSE);
	mpCMNumPanel->ShowItem(mIndexTextureB, FALSE);
	mpCMNumPanel->ShowItem(mIndexColor, FALSE);
	mpCMNumPanel->ShowItem(mIndexColorType, FALSE);
	break;
    case _BG_TEXTURE_MAP:
	mpCMNumPanel->ShowItem(mIndexStarNum, FALSE);
	mpCMNumPanel->ShowItem(mIndexMappingType, TRUE);
	mpCMNumPanel->ShowItem(mIndexTextureA, TRUE);
	if (pCData->mCMappingType.mBody == _MAPPING_FISHEYE)
	    mpCMNumPanel->ShowItem(mIndexTextureB, TRUE);
	else
	    mpCMNumPanel->ShowItem(mIndexTextureB, FALSE);
	mpCMNumPanel->ShowItem(mIndexColor, FALSE);
	mpCMNumPanel->ShowItem(mIndexColorType, FALSE);
	break;
    case _BG_COLOR_CURVE:
	mpCMNumPanel->ShowItem(mIndexStarNum, FALSE);
	mpCMNumPanel->ShowItem(mIndexMappingType, FALSE);
	mpCMNumPanel->ShowItem(mIndexTextureA, FALSE);
	mpCMNumPanel->ShowItem(mIndexTextureB, FALSE);
	mpCMNumPanel->ShowItem(mIndexColor, TRUE);
	mpCMNumPanel->ShowItem(mIndexColorType, TRUE);
	break;
    }
    
    HideAndShowEnd();
}

void FilterBGGUI_cls::
PathSelection(int index){
    GpCFileDialog->SetMessage(wxT("Select Texture"));
    GpCFileDialog->SetWildcard(wxT("image (*.png;*.jpg)|*.png;*.jpg"));
    if (GpCFileDialog->ShowModal() == wxID_CANCEL) return;

    FilterBGData_cls *pCData = mpCParent->GetPCData();
    int updateBit = _COMMAND_UPDATE_DISPLAY_FILTER_PARAM | _COMMAND_UPDATE_DISPLAY_GL;

    if (index == 0)
	GCUndoList.PushNewCommand
	    (new CommandSetValue_cls(&pCData->mCstrTextureA.mstrBody, GpCFileDialog->GetPath(), updateBit), 
	     new CommandSetValue_cls(&pCData->mCstrTextureA.mstrBody, pCData->mCstrTextureA.mstrBody, updateBit));
    else
	GCUndoList.PushNewCommand
	    (new CommandSetValue_cls(&pCData->mCstrTextureB.mstrBody, GpCFileDialog->GetPath(), updateBit), 
	     new CommandSetValue_cls(&pCData->mCstrTextureB.mstrBody, pCData->mCstrTextureB.mstrBody, updateBit));

    GCUndoList.ExecCurrent();
}

FilterBGData_cls::
FilterBGData_cls() /*: FilterData_cls(FALSE)*/ {
    //wxString listTag[2] = {wxT("deinsity"), wxT("alpha") };
    //int updat<eBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM;
    //wxString strParameter = wxT("Filter"); // {ꉻႾ
    //strParameter << wxT(" Parameter");
    
    mCstrCustomName.mstrBody = wxT("BG");
    mCstrCustomName.mstrBody << wxT("%02d");

    ColorCurve_cls *pCClrBGColor = new ColorCurve_cls
	(wxT("BGColor"), wxT("BGColor"), wxT("0.1 0.3 0.5 1"));
    mVpCClrCurve.push_back(pCClrBGColor);

    mCConfList.Add(&mCBGType, wxT("BGType="), wxT("0"));
    mCConfList.Add(&mCStarNumber, wxT("BGStarNumber="), wxT("4000"));
    mCConfList.Add(&mCBGScale, wxT("BGScale="), wxT("1000"));
    mCConfList.Add(&mCStarScale, wxT("StarScale="), wxT("3"));
    mCConfList.Add(&mCBrightness, wxT("BGBrightness="), wxT("1"));
    mCConfList.Add(&mCMappingType, wxT("BGMappingType="), wxT("0"));
    mCConfList.Add(&mCstrTextureA, wxT("BGTectureA="), wxT("NONE"));
    mCConfList.Add(&mCstrTextureB, wxT("BGTectureB="), wxT("NONE"));
    mCConfList.Add(&mCColorMapType, wxT("BGColorMap="), wxT("1"));
}

FilterBGData_cls::
~FilterBGData_cls(){
}

void FilterBGData_cls::
CopyFilterDataFromP(FilterData_cls *pCFrom,
		    ParamGroup_cls *pCTranslate,
		    ParamGroup_cls *pCRotete   ,
		    ParamGroup_cls *pCScale    ,
		    ParamGroup_cls *pCAdditional){
    /*
    if (pCAdditional == NULL) 
	mpCParameter->CopyFromP
	    (((FilterBGData_cls*)pCFrom)->mpCParameter);
    else
	mpCParameter->CopyFromP(pCAdditional);
    */
    FilterData_cls::CopyFilterDataFromP
	(pCFrom, pCTranslate, pCRotete, pCScale);
}

// ---------------------

FilterBG_cls::
FilterBG_cls(FilterPanelPain_cls *pCPanelPain,
	     FilterBGData_cls *pCInitData) :
    FilterBase_cls(pCPanelPain->mpCFilterCtrlLink)
{
    wxString strInit; strInit << _FILTER_BG;
    mCFilterType.SetItem(wxT("FilterType="), strInit);
    mpCData     = new FilterBGData_cls;
    mpCGUIPanel = new FilterBGGUI_cls(this, pCPanelPain);
    mDLStars = 0;
    mDLDome = 0;
    mDLStarNumberCompiled = -1;
    mBGScaleCompiled = -1;
    mStarScaleCompiled = -1;
    
    mCGLTextureStar.LoadImage(wxT("Texture/Star.png"));

    mpCGLTextureA = new GLTexture_cls();
    mpCGLTextureB = new GLTexture_cls();
    mpCGLTextureClrCurve = new GLTexture_cls();
}

FilterBG_cls::
~FilterBG_cls(){
    delete mpCGLTextureA;
    delete mpCGLTextureB;
    delete mpCGLTextureClrCurve;
}

FilterBGData_cls* FilterBG_cls::
GetPCData(){
    return (FilterBGData_cls*)mpCData;
}

void FilterBG_cls::
DrawCenterfacedSquare(float x, float y, float z, float size){
    GLfloat location[3],target[3], top[3], right[3], matrix[16];
    
    glPushMatrix();
    target[X]   = -x; target[Y]   = -y; target[Z]   = -z;
    location[X] =  x; location[Y] =  y; location[Z] =  z;
    MyVector::Normalize(target, (float)1.0);
    top[X] = 0; top[Y] = 0; top[Z] = 1.0;
    
    glTranslatef(location[X], location[Y], location[Z]);
    
    // 덷C 
    MyVector::Orthogonalize(target, top);
    MyVector::Normalize(top, (float)1.0);
    right[X] = target[Y]*top[Z] - target[Z]*top[Y];
    right[Y] = target[Z]*top[X] - target[X]*top[Z];
    right[Z] = target[X]*top[Y] - target[Y]*top[X];
    
    // IuWFNg]s 
    matrix[0]=right[X];matrix[4]= top[X]; matrix[8] = -target[X]; matrix[12] = 0;
    matrix[1]=right[Y];matrix[5]= top[Y]; matrix[9] = -target[Y]; matrix[13] = 0;
    matrix[2]=right[Z];matrix[6]= top[Z]; matrix[10]= -target[Z]; matrix[14]= 0;
    matrix[3] = 0;     matrix[7]= 0     ; matrix[11] = 0        ; matrix[15] = 1;
    
    glMultMatrixf(matrix);
    
    glScalef(size, size, size);
    
    glBegin(GL_QUADS);
    glNormal3f(0,0,0);glTexCoord2f(0,1);glVertex3f(-0.25,+0.25,0);
    glNormal3f(0,0,0);glTexCoord2f(1,1);glVertex3f(+0.25,+0.25,0);
    glNormal3f(0,0,0);glTexCoord2f(1,0);glVertex3f(+0.25,-0.25,0);
    glNormal3f(0,0,0);glTexCoord2f(0,0);glVertex3f(-0.25,-0.25,0);
    glEnd();
    
    glPopMatrix();
}

void FilterBG_cls::
CompileStars(){
    FilterBGData_cls *pCData = GetPCData();
    
    if (glIsList(mDLStars))
	glDeleteLists(mDLDome, 1);
    else
	mDLStars = glGenLists(1);

    int num         = pCData->mCStarNumber.mBody;
    float distance  = pCData->mCBGScale.mBody;
    float starScale = pCData->mCStarScale.mBody;
    
    srand(1); //_͕̎ςȂƂ
    glNewList(mDLStars, GL_COMPILE);
    for (int i = 0; i < num; i++){
	float x = (float)rand()/RAND_MAX-0.5f;
	float y = (float)rand()/RAND_MAX-0.5f;
	float z =  (float)rand()/RAND_MAX-0.5f;
	float v = (float)i/num;
	float a = 1;
	if (v < 0.5) a /= (1 + (0.5 - v) * 10); 
	float s = (float)sqrt(sqrt(1.0+99.0*v)) * starScale / sqrt(a);
	float starnorm = sqrt(x*x + y*y + z*z) + 0.00001;
	x *= distance / starnorm;
	y *= distance / starnorm;
	z *= distance / starnorm;
	
	switch(i % 4){
	case 0: case 1:
	    glColor4f(1, 1, 1, a); break;
	case 2:
	    glColor4f(1, 0.9, 0.8, a); break;
	case 3:
	    glColor4f(0.8, 0.9, 1, a); break;
	}
	
	DrawCenterfacedSquare(x, y, z, s);
    }

    glEndList();
}

void FilterBG_cls::
CompileDome(){
    if (glIsList(mDLDome))
	glDeleteLists(mDLDome, 1);
    else
	mDLDome = glGenLists(1);

    double pos[90/5+1][360/5+1][3];
    double UV[90/5+1][360/5+1][2];

    glNewList(mDLDome, GL_COMPILE);
    for (int itheta = 0; itheta <= 90/5; itheta++){
	double theta = (double)itheta * 5 * 3.14159 / 180;
	for (int iphi = 0; iphi <= 360/5; iphi++){
	    double phi = (double)iphi * 5 * 3.14159 / 180;
	    pos[itheta][iphi][0] = 1 * cos(theta) * cos(phi);
	    pos[itheta][iphi][1] = 1 * cos(theta) * sin(phi);
	    pos[itheta][iphi][2] = 1 * sin(theta);
	    UV[itheta][iphi][0] =  0.5 + 0.5*(1.0-(double)itheta/90*5)*cos(phi);
	    UV[itheta][iphi][1] =  0.5 - 0.5*(1.0-(double)itheta/90*5)*sin(phi);
	}
    }
	    
    for (int itheta = 0; itheta < 90/5; itheta++){
	double (*tmpUV0)[2] = UV[itheta];
	double (*tmpUV1)[2] = UV[itheta+1];
	double (*tmpPos0)[3] = pos[itheta];
	double (*tmpPos1)[3] = pos[itheta+1];
	glBegin(GL_QUAD_STRIP);
	for (int iphi = 0; iphi <= 360/5; iphi++){
	    glTexCoord2dv(tmpUV0[iphi]);
	    glVertex3dv(tmpPos0[iphi]);
	    glTexCoord2dv(tmpUV1[iphi]);
	    glVertex3dv(tmpPos1[iphi]);
	}
	glEnd();
    }

    glEndList();
}

void FilterBG_cls::
ProcessFisheyeRim(wxImage &CImg){
    int width = CImg.GetWidth();
    int height = CImg.GetHeight();
    if (width != height) return;
    if (CImg.HasAlpha() == FALSE) return;

    int x = width; int y = 0;
    
    double dphi = 0.5/width;
    for (double dr = 1; dr > -0.6;  dr-=0.5){
	for (double phi = 0; phi < 3.14159 * 2; phi += dphi){
	    double factor0 = (double)(width-dr)/(double)width;
	    double factor1 = (double)(width-dr-1.5)/(double)width;
	    int x = (double)width * (cos(phi) * 0.5 * factor0 + 0.5);
	    int y = (double)width * (sin(phi) * 0.5 * factor0 + 0.5);
	    if (TRUE || CImg.GetAlpha(x, y) == 0){
		int x0 = (double)width * (cos(phi) * 0.5 * factor1 + 0.5);
		int y0 = (double)width * (sin(phi) * 0.5 * factor1 + 0.5);
		CImg.SetRGB(wxRect(x, y, 1, 1), 
			    CImg.GetRed(x0, y0),
			    CImg.GetGreen(x0, y0),
			    CImg.GetBlue(x0, y0));
		CImg.SetAlpha(x, y, CImg.GetAlpha(x0, y0));
	    }
	}
    }
}

void FilterBG_cls::
GLDrawStarField(){
    FilterBGData_cls *pCData = GetPCData();
    if (pCData->mCStarNumber.mBody != mDLStarNumberCompiled ||
	pCData->mCBGScale.mBody    != mBGScaleCompiled ||
	pCData->mCStarScale.mBody  != mStarScaleCompiled ){
	CompileStars();
	mDLStarNumberCompiled = pCData->mCStarNumber.mBody;
	mBGScaleCompiled      = pCData->mCBGScale.mBody;
	mStarScaleCompiled    = pCData->mCStarScale.mBody;
    }
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glDisable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    mCGLTextureStar.BindTexture();
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    for (int i = 0; i < pCData->mCBrightness.mBody; i++)
	glCallList(mDLStars);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void FilterBG_cls::
GLDrawDome(int index, float scale){
    if (mDLDome == 0) CompileDome();
    
    glPushMatrix();
    glScaled(scale, scale, scale);
    if (index == 1)
	glRotatef(180, 1, 0, 0);
    glCallList(mDLDome);
    glPopMatrix();
}

void FilterBG_cls::
CheckTextureA(){
    FilterBGData_cls *pCData = GetPCData();
    if (pCData->mCstrTextureA.mstrBody.IsSameAs(mstrTexturALoaded) == FALSE){
	if (pCData->mCstrTextureA.mstrBody.IsSameAs(wxT("NONE")) == TRUE)
	    mpCGLTextureA->ClearImage();
	else {
	    wxImage CImg(pCData->mCstrTextureA.mstrBody);
	    if (CImg.IsOk() == TRUE){
		ProcessFisheyeRim(CImg);
		CImg = CImg.Mirror(FALSE);
		mpCGLTextureA->CopyFromImage(CImg);
	    }
	}
	mstrTexturALoaded = pCData->mCstrTextureA.mstrBody;
    }
}

void FilterBG_cls::
CheckTextureB(){
    FilterBGData_cls *pCData = GetPCData();
    if (pCData->mCstrTextureB.mstrBody.IsSameAs(mstrTexturBLoaded) == FALSE){
	if (pCData->mCstrTextureB.mstrBody.IsSameAs(wxT("NONE")) == TRUE)
	    mpCGLTextureB->ClearImage();
	else{
	    wxImage CImg(pCData->mCstrTextureB.mstrBody);
	    if (CImg.IsOk() == TRUE){
		ProcessFisheyeRim(CImg);
		CImg = CImg.Mirror(FALSE);
		mpCGLTextureB->CopyFromImage(CImg);
	    }
	}
	mstrTexturBLoaded = pCData->mCstrTextureB.mstrBody;
    }
}

bool FilterBG_cls::
IsClrStatusChanged(){
    bool bChanged = FALSE;
    ColorCurve_cls *pCClrCurve = GetPCData()->mVpCClrCurve[0];
    if (pCClrCurve->mCbColorCurve.mbBody != mPrevClrType) bChanged = TRUE;
    mPrevClrType = pCClrCurve->mCbColorCurve.mbBody;
    if (GetPCData()->mCColorMapType.mBody != mPrevMapType) bChanged = TRUE;
    mPrevMapType = GetPCData()->mCColorMapType.mBody;
    for (int i = 0; i < 3; i++){
	if (pCClrCurve->mCConstantColor.mpBody[i] != mPrevClrConst[i])
	    bChanged = TRUE;
	mPrevClrConst[i] = pCClrCurve->mCConstantColor.mpBody[i];
    }
    return bChanged;
}

void FilterBG_cls::
CheckTextureClrCurve(){
    ColorCurve_cls *pCClrCurve = GetPCData()->mVpCClrCurve[0];
    ParamGroup_cls *pCClrParam = GetPCData()->mVpCClrCurve[0]->mpCClrCurve;
    
    if (GetPCData()->mCPGStatusBG.IsParamGroupChanged(pCClrParam) == TRUE ||
	IsClrStatusChanged() == TRUE){
	unsigned char *pClrData = pCClrCurve->GetPNewClrMap(256, GetPCData()->mCColorMapType.mBody);
	mpCGLTextureClrCurve->CopyFromDataRGBA(pClrData, 1, 256);
	delete [] pClrData;
    }
}

void FilterBG_cls::
GLDrawTextureFishEye(){
    FilterBGData_cls *pCData = GetPCData();

    CheckTextureA();
    CheckTextureB();

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    if (mpCGLTextureA->mTextureID != 0){
	mpCGLTextureA->BindTexture();
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	for (int i = 0; i < pCData->mCBrightness.mBody; i++){
	    if (i == 0)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	    else
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    glColor4f(1, 1, 1, pCData->mCBrightness.mBody - i);
	    GLDrawDome(0, pCData->mCBGScale.mBody);
	}
    }
    if (mpCGLTextureB->mTextureID != 0){
	mpCGLTextureB->BindTexture();
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	for (int i = 0; i < pCData->mCBrightness.mBody; i++){
	    if (i == 0)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	    else
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    glColor4f(1, 1, 1, pCData->mCBrightness.mBody - i);
	    GLDrawDome(1, pCData->mCBGScale.mBody);
	}
    }
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
}

void FilterBG_cls::
GLDrawTextureSpherical(int type){
    FilterBGData_cls *pCData = GetPCData();

    GLTexture_cls *pCTexture;
    if (type  == 0){
	pCTexture = mpCGLTextureA;
	CheckTextureA();
    }
    if (type == 1){
	pCTexture = mpCGLTextureClrCurve;
	CheckTextureClrCurve();
    }

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    float scale = pCData->mCBGScale.mBody;
    glPushMatrix();
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glScaled(scale, scale, scale);
    glCullFace(GL_FRONT);
    if (pCTexture->mTextureID != 0){
	pCTexture->BindTexture();
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	if (type == 0){
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	}
	else{
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	}
	    
	for (int i = 0; i < pCData->mCBrightness.mBody; i++){
	    if (i == 0)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	    else
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    glColor4f(1, 1, 1, pCData->mCBrightness.mBody - i);
	    GpCSphere->Draw(3);
	}
    }
    glPopAttrib();
    glPopMatrix();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
}
    
void FilterBG_cls::
GLDraw(double doubleIndex){
    FilterBGData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return;

    glPushMatrix();

    pCData->UpdateCurrentObj(doubleIndex);
    pCData->GLSetMatrix();
    
    switch(pCData->mCBGType.mBody){
    case _BG_STARS: GLDrawStarField(); break;
    case _BG_TEXTURE_MAP: 
	switch(pCData->mCMappingType.mBody){
	case _MAPPING_FISHEYE: GLDrawTextureFishEye(); break;
	case _MAPPING_SPHERICAL: GLDrawTextureSpherical(0); break;
	}
	break;
    case _BG_COLOR_CURVE:
	GLDrawTextureSpherical(1); break;
    }

    glPopMatrix();
}

bool FilterBG_cls::
IsCommonStatusEffective(){
    bool bEffective = TRUE;
    if (GetPCData()->mpCTranslate->IsConstant() == FALSE) bEffective = FALSE;
    if (GetPCData()->mpCRotation->IsConstant() == FALSE) bEffective = FALSE;
    if (GetPCData()->mpCScale->IsConstant() == FALSE) bEffective = FALSE;
    return bEffective;
}

void FilterBG_cls::
GetStrPovStars(wxString &strBG){

    FilterBGData_cls *pCData = GetPCData();
    int num         = pCData->mCStarNumber.mBody;
    float distance  = pCData->mCBGScale.mBody;
    float starScale = pCData->mCStarScale.mBody;

    srand(1); //_͕̎ςȂƂ
    for (int i = 0; i < num; i++){
	float x = (float)rand()/RAND_MAX-0.5f;
	float y = (float)rand()/RAND_MAX-0.5f;
	float z =  (float)rand()/RAND_MAX-0.5f;
	float v = (float)i/num;
	float a = 1;
	if (v < 0.5) a /= (1 + (0.5 - v) * 10); 
	float s = (float)sqrt(sqrt(1.0+99.0*v)) * starScale / sqrt(a);
	float starnorm = sqrt(x*x + y*y + z*z) + 0.00001;
	x *= distance / starnorm;
	y *= distance / starnorm;
	z *= distance / starnorm;

	float pos[3] = {x, y, z};

	float clr[4] = {1, 1, 1, a};
	switch (i % 4){
	case 2:
	    clr[1] = 0.9; clr[2] = 0.8; break;
	case 3:
	    clr[0] = 0.8; clr[1] = 0.9; break;
	}

	strBG << PovOut3f(wxT("SP("), pos, 1) << wxT(",") << PovOutf(s*0.125, 2) 
	      << wxT(",") << PovOut4fRGBTfromRGBA(clr, 2, FALSE)
	      << wxT(")") << ENDL;
    }
}

wxString FilterBG_cls::
GetStrPovImageMap(wxString strFile, int mapType){
    SplitPath_cls CPath(strFile/*GetPCData()->mCstrTextureA.mstrBody*/);
    wxString strImgType;
    wxString strExt = CPath.GetPathStr(_FN_EXT);
    if (strExt.IsSameAs(wxT(".png")) || strExt.IsSameAs(wxT(".PNG")))
	strImgType = wxT("png");
    if (strExt.IsSameAs(wxT(".jpg")) || strExt.IsSameAs(wxT(".JPG")) ||
	strExt.IsSameAs(wxT(".jpeg")) || strExt.IsSameAs(wxT(".JPEG")))
	strImgType = wxT("jpeg");
        
    wxString strImageMap;
    strImageMap << wxT("image_map { ") << strImgType << wxT(" \"") << strFile/*GetPCData()->mCstrTextureA.mstrBody*/ << wxT("\" interpolate 2");
    if (mapType == 1)
	strImageMap << wxT(" map_type 1");
    strImageMap << ENDLCLOSE;
    
    return strImageMap;
}

wxString FilterBG_cls::
GetStrPovFinish(){
    wxString strFinish;
    strFinish << wxT("finish { ambient ") << PovOutf(GetPCData()->mCBrightness.mBody, 2) << wxT(" diffuse 0 specular 0}");
    return strFinish;
}

void FilterBG_cls::
GetStrPovTextureFishEye(wxString &strBG, int type){
    FilterBGData_cls *pCData = GetPCData();
    if (type == 0 &&
	pCData->mCstrTextureA.mstrBody.IsSameAs(wxT("NONE"))) return;
    if (type == 1 &&
	pCData->mCstrTextureB.mstrBody.IsSameAs(wxT("NONE"))) return;

    wxString strDome = wxT("DOMEA");
    if (type == 1) strDome = wxT("DOMEB");

    strBG << GetTab(1) << wxT("intersection {") << ENDL
	  << GetTab(2) << wxT("sphere{ <0,0,0>,1") << ENDL
	  << GetTab(3) << wxT("pigment{") << ENDL
	  << GetTab(4) << wxT("average") << ENDL
	  << GetTab(4) << wxT("pigment_map{") << ENDL
	  << GetTab(5) << wxT("[1 function {") << strDome << wxT("(x*FE_CONVERT(x,y,z),y*FE_CONVERT(x,y,z),z).red} color_map {[0 rgbt 0][1 rgbt <4,0,0,0>]}]") << ENDL
	  << GetTab(5) << wxT("[1 function {") << strDome << wxT("(x*FE_CONVERT(x,y,z),y*FE_CONVERT(x,y,z),z).green} color_map {[0 rgbt 0][1 rgbt <0,4,0,0>]}]") << ENDL
	  << GetTab(5) << wxT("[1 function {") << strDome << wxT("(x*FE_CONVERT(x,y,z),y*FE_CONVERT(x,y,z),z).blue} color_map {[0 rgbt 0][1 rgbt <0,0,4,0>]}]") << ENDL
	  << GetTab(5) << wxT("[1 function {") << strDome << wxT("(x*FE_CONVERT(x,y,z),y*FE_CONVERT(x,y,z),z).transmit} color_map {[0 rgbt 0][1 rgbt <0,0,0,4>]}]") << ENDL
	  << GetTab(4) << ENDLCLOSE
	  << GetTab(3) << ENDLCLOSE
	  << GetTab(3) << GetStrPovFinish() << ENDL
	  << GetTab(2) << ENDLCLOSE
	  << GetTab(2) << wxT("box{ <-1.1,-1.1,0>, <1.1,1.1,1.1> pigment { rgbt <0,0,0,1>}}") << ENDL
	  << GetTab(2) << wxT("scale ") << PovOutf(pCData->mCBGScale.mBody, 2) << ENDL
	  << GetTab(2) << wxT("scale <1,-1,1>") << ENDL;
    if (type == 1) 
	strBG << GetTab(2) << wxT("rotate <180, 0, 0>") << ENDL;
    strBG << GetTab(1) << ENDLCLOSE;
}

void FilterBG_cls::
GetStrPovTextureSpherical(wxString &strBG){
    FilterBGData_cls *pCData = GetPCData();
    if (pCData->mCstrTextureA.mstrBody.IsSameAs(wxT("NONE"))) return;
    
    
    strBG << GetTab(1) << wxT("sphere { <0,0,0>,1") << ENDL
	  << GetTab(2) << wxT("texture { pigment { rgbt <0,0,0,1> } }") << ENDL
	  << GetTab(2) << wxT("interior_texture {") << ENDL
	  << GetTab(3) << wxT("pigment {") << GetStrPovImageMap(pCData->mCstrTextureA.mstrBody, 1) << ENDLCLOSE
	  << GetTab(3) << GetStrPovFinish() << ENDL
	  << GetTab(2) << ENDLCLOSE
	  << GetTab(2) << wxT("scale ") << PovOutf(pCData->mCBGScale.mBody, 2) << ENDL
	  << GetTab(2) << wxT("rotate <90, 0, 0>") << ENDL
	  << GetTab(1) << ENDLCLOSE;
}

void FilterBG_cls::
GetStrPovSkySphere(wxString &strBG, double doubleIndex){
    strBG << wxT("sky_sphere{") << ENDL
	  << GetTab(1) << wxT("pigment{") << ENDL;
    if (GetPCData()->mCColorMapType.mBody == 0)
	strBG << GetTab(2) << wxT("function{asin(z)/3.14159 + 0.5}") << ENDL;
    else
	strBG << GetTab(2) << wxT("function{abs(asin(z)*2/3.14159)}") << ENDL;

    GetPCData()->mVpCClrCurve[0]->GetStrPovColorMap(strBG, GetTab(3));
    strBG << GetTab(1) << ENDLCLOSE;
    strBG << GetTab(1) << wxT("emission rgb <1, 1, 1>") << ENDL;
    GetPCData()->GetStrPovModelView(strBG, doubleIndex);
    strBG << ENDLCLOSE;
}

void FilterBG_cls::
GetStrPovMacroForStars(wxString &strBG){
    strBG << wxT("#macro SP(POS, R, RGBT)") << ENDL;
    strBG << wxT("#local RGBT2 = <RGBT.red, RGBT.green, RGBT.blue, 1 - (1 - RGBT.t) * 0.2>;") << ENDL;
    strBG << wxT("sphere{ POS, R") << ENDL;
    strBG << GetTab(1) << wxT("pigment { aoi color_map{[0 rgbt <0,0,0,1>] [0.8 rgbt RGBT2][1 rgbt RGBT]} }") << GetStrPovFinish() << ENDLCLOSE;
    strBG << wxT("#end") << ENDL << ENDL;
}

void FilterBG_cls::
GetStrPovMacroForFishEye(wxString &strBG){
    strBG << wxT("#declare FE_CONVERT = function {") << ENDL
	      << wxT("\tatan2(sqrt(x*x+y*y),z)*2/pi/sqrt(x*x+y*y)") << ENDL
	      << ENDLCLOSE 
	      << ENDL;
    
    if (GetPCData()->mCstrTextureA.mstrBody.IsSameAs(wxT("NONE")) == FALSE){
	strBG << wxT("#declare DOMEA = function {") << ENDL
	      << GetTab(1) << wxT("pigment{") << ENDL
	      << GetTab(2) << GetStrPovImageMap(GetPCData()->mCstrTextureA.mstrBody) << ENDL
	      << GetTab(2) << wxT("translate <-0.5, -0.5, 0>") << ENDL
	      << GetTab(2) << wxT("scale 2") << ENDL
	      << GetTab(1) << ENDLCLOSE
	      << ENDLCLOSE;
    }
    if (GetPCData()->mCstrTextureB.mstrBody.IsSameAs(wxT("NONE")) == FALSE){
	strBG << wxT("#declare DOMEB = function {") << ENDL
	      << GetTab(1) << wxT("pigment{") << ENDL
	      << GetTab(2) << GetStrPovImageMap(GetPCData()->mCstrTextureB.mstrBody) << ENDL
	      << GetTab(2) << wxT("translate <-0.5, -0.5, 0>") << ENDL
	      << GetTab(2) << wxT("scale 2") << ENDL
	      << GetTab(1) << ENDLCLOSE
	      << ENDLCLOSE;
    }
}

void FilterBG_cls::
GetStrPovBGMain(wxString &strBG, double doubleIndex){
    int type = GetPCData()->mCBGType.mBody;
    
    strBG << wxT("union{") << ENDL;
    switch(type){
    case _BG_STARS: GetStrPovStars(strBG); break;
    case _BG_TEXTURE_MAP: 
	switch(GetPCData()->mCMappingType.mBody){
	case _MAPPING_FISHEYE: 
	    GetStrPovTextureFishEye(strBG, 0);
	    GetStrPovTextureFishEye(strBG, 1); break;
	case _MAPPING_SPHERICAL: GetStrPovTextureSpherical(strBG); break;
	}
	break;
    }
    GetPCData()->GetStrPovModelView(strBG, doubleIndex);
    strBG << GetTab(1) << wxT("no_shadow") << ENDL
	  << ENDLCLOSE;
}

void FilterBG_cls::
GetStrPovBG(wxString &strBG, double doubleIndex, PovRayCtrl_cls *pCPovRayCtrl){

    int type = GetPCData()->mCBGType.mBody;

    // macro
    switch(type){
    case _BG_STARS: GetStrPovMacroForStars(strBG); break;
    case _BG_TEXTURE_MAP:
	switch(GetPCData()->mCMappingType.mBody){
	case _MAPPING_FISHEYE: GetStrPovMacroForFishEye(strBG); break;
	}
	break;
    }

    // J[J[uunion̊O
    if (type == _BG_COLOR_CURVE){
	GetStrPovSkySphere(strBG, doubleIndex); 
	return;
    }

    if (IsCommonStatusEffective() == TRUE){
	if (pCPovRayCtrl->mCommonBGStatus == 0){
	    
	    wxString strBGCommon;
	    GetStrPovBGMain(strBGCommon, doubleIndex);
	    
	    wxString strFile = 
		pCPovRayCtrl->GetWorkingDirWithSep(TRUE);
	    strFile << wxT("Common") << GetPCData()->mCstrCustomName.mstrBody << wxT(".pov");
		
	    wxFile CFile(strFile, wxFile::write);
	    CFile.Write(strBGCommon);
	    CFile.Close();
	}

	strBG << wxT("#include \"Common") << GetPCData()->mCstrCustomName.mstrBody << wxT(".pov\"") << ENDL;
    }
    else
	GetStrPovBGMain(strBG, doubleIndex);

    pCPovRayCtrl->mCommonBGStatus = TRUE;
}
