// -*- c++ -*- 
// (setq c-basic-offset 4)

#include "clsFilterBoolean.h"

#include "../MAINCANVAS/clsFilterPanelPain.h"
#include "../PARTICLEDATA/clsArrayAndIndex.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/clsMultiInputPanel.h"

#include "../../wxWidgetEnum.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

ThreadBooleanCheckBase_cls::
ThreadBooleanCheckBase_cls(wxThreadKind kind) : 
    wxThread(kind) {
}

void ThreadBooleanCheckBase_cls::
SetParameters(bool *pbFinishedFlag,
	      ArrayAndIndex_cls* pIndexAndZ, float (*pX)[3],
	      float *pAlphaFactor,
      	      int   operatorType, float borderFactor,
	      float dX, float dY, float dZ,
	      float scaleX, float scaleY, float scaleZ,
	      int ini, int fin){
    
    mpbFinishedFlagLink = pbFinishedFlag;
    mpIndexAndZ = pIndexAndZ;
    mpX = pX;
    mpAlphaFactor = pAlphaFactor;
    mOperatorType = operatorType;
    mBorderFactor = borderFactor;
    mdX = dX; mdY = dY; mdZ = dZ;
    mScaleX = scaleX; mScaleY = scaleY; mScaleZ = scaleZ;
    mIni = ini; mFin = fin;
    
    if (Create() != wxTHREAD_NO_ERROR) exit (0);
	//MsgBoxExit(NULL, "Error!", "Can't create thread");
    
}

// --

ThreadBooleanCheckBox_cls::
ThreadBooleanCheckBox_cls(wxThreadKind kind) :
    ThreadBooleanCheckBase_cls(kind) {
}

wxThread::ExitCode ThreadBooleanCheckBox_cls::
Entry(){
    ArrayAndIndex_cls *pIndexAndZ = mpIndexAndZ;
    float (*pX)[3] = mpX;
    float *pAlphaFactor = mpAlphaFactor;
    float dX = mdX;
    float dY = mdY;
    float dZ = mdZ;
    float scaleX = mScaleX;
    float scaleY = mScaleY;
    float scaleZ = mScaleZ;
    int   operatorType = mOperatorType;
    float borderFactor = mBorderFactor;

    if (borderFactor == 0){
	float inner, outer;
	switch(operatorType){
	case 0: // intersect
	    inner = 1; outer = 0;
	    break;
	case 1: // intersect
	    inner = 0; outer = 1;
	    break;
	case 2:
	    for (int i = mIni; i < mFin; i++){
		int index = pIndexAndZ[i].index;
		if (pX[index][X]-dX < scaleX/2 && pX[index][X]-dX > -scaleX/2 &&
		    pX[index][Y]-dY < scaleY/2 && pX[index][Y]-dY > -scaleY/2 &&
		    pX[index][Z]-dZ < scaleZ/2 && pX[index][Z]-dZ > -scaleZ/2)
		    pAlphaFactor[i] = 1;
	    }
	    *mpbFinishedFlagLink = TRUE;
	    return 0; 
	    break;
	}
	for (int i = mIni; i < mFin; i++){
	    int index = pIndexAndZ[i].index;
	    if (pX[index][X]-dX < scaleX/2 && pX[index][X]-dX > -scaleX/2 &&
		pX[index][Y]-dY < scaleY/2 && pX[index][Y]-dY > -scaleY/2 &&
		pX[index][Z]-dZ < scaleZ/2 && pX[index][Z]-dZ > -scaleZ/2)
		pAlphaFactor[i] *= inner;
	    else
		pAlphaFactor[i] *= outer;
	}
    }
    else{
	float border = 1.0 + borderFactor;
	float borderFactorX = scaleX/2 * borderFactor;
	float borderFactorY = scaleY/2 * borderFactor;
	float borderFactorZ = scaleZ/2 * borderFactor;
	for (int i = mIni; i < mFin; i++){
	    int index = pIndexAndZ[i].index;
	    float factor = 1;
	    float posX = pX[index][X]-dX;
	    float posY = pX[index][Y]-dY;
	    float posZ = pX[index][Z]-dZ;

	    if      (posX > scaleX/2 * border) factor *= 0;
	    else if (posX > scaleX/2)  factor *= (1.0 - (posX - scaleX/2) / borderFactorX) ;
	    else if (posX > -scaleX/2) factor *= 1.0;
	    else if (posX > -scaleX/2 * border) factor *= (1.0 + (posX + scaleX/2) / borderFactorX);
	    else factor *= 0;

	    if      (posY > scaleY/2 * border) factor *= 0;
	    else if (posY > scaleY/2)  factor *= (1.0 - (posY - scaleY/2) / borderFactorY) ;
	    else if (posY > -scaleY/2) factor *= 1.0;
	    else if (posY > -scaleY/2 * border) factor *= (1.0 + (posY + scaleY/2 ) / borderFactorY);
	    else factor *= 0;

	    if      (posZ > scaleZ/2 * border) factor *= 0;
	    else if (posZ > scaleZ/2)  factor *= (1.0 - (posZ - scaleZ/2) / borderFactorZ) ;
	    else if (posZ > -scaleZ/2) factor *= 1.0;
	    else if (posZ > -scaleZ/2 * border) factor *= (1.0 + (posZ + scaleZ/2 ) / borderFactorZ);
	    else factor *= 0;

	    switch(operatorType){
	    case 0:
		pAlphaFactor[i] *= factor;
		break;
	    case 1:
		pAlphaFactor[i] *= (1.0f - factor);
		break;
	    case 2:
		pAlphaFactor[i] += factor;
		if (pAlphaFactor[i] > 1) pAlphaFactor[i] = 1;
		break;
	    }
	}
    }
    
    *mpbFinishedFlagLink = TRUE;
    return 0;
}

// ---

ThreadBooleanCheckSphere_cls::
ThreadBooleanCheckSphere_cls(wxThreadKind kind) :
    ThreadBooleanCheckBase_cls(kind) {
}

wxThread::ExitCode ThreadBooleanCheckSphere_cls::
Entry(){
    ArrayAndIndex_cls *pIndexAndZ = mpIndexAndZ;
    float (*pX)[3] = mpX;
    float *pAlphaFactor = mpAlphaFactor;
    float dX = mdX;
    float dY = mdY;
    float dZ = mdZ;
    float scaleX = mScaleX;
    float scaleY = mScaleY;
    float scaleZ = mScaleZ;
    int   operatorType = mOperatorType;
    float borderFactor = mBorderFactor;

    if (borderFactor == 0 || TRUE){ // !!_FIX_ME_!!
	float inner, outer;
	switch(operatorType){
	case 0: // intersect
	    inner = 1; outer = 0;
	    break;
	case 1: // intersect
	    inner = 0; outer = 1;
	    break;
	case 2:
	    for (int i = mIni; i < mFin; i++){
		int index = pIndexAndZ[i].index;
		float difference2 = 
		    (pX[index][X]-dX) * (pX[index][X]-dX) / scaleX / scaleX +
		    (pX[index][Y]-dY) * (pX[index][Y]-dY) / scaleY / scaleY +
		    (pX[index][Z]-dZ) * (pX[index][Z]-dZ) / scaleZ / scaleZ;
		if (difference2 < 0.25) pAlphaFactor[i] =1;
	    }
	    *mpbFinishedFlagLink = TRUE;
	    return 0; 
	    break;
	}
	for (int i = mIni; i < mFin; i++){
	    int index = pIndexAndZ[i].index;
	    float difference2 = 
		(pX[index][X]-dX) * (pX[index][X]-dX) / scaleX / scaleX +
		(pX[index][Y]-dY) * (pX[index][Y]-dY) / scaleY / scaleY +
		(pX[index][Z]-dZ) * (pX[index][Z]-dZ) / scaleZ / scaleZ;
	    if (difference2 < 0.25) 
		pAlphaFactor[i] *= inner;
	    else
		pAlphaFactor[i] *= outer;
	}
    }
    
    *mpbFinishedFlagLink = TRUE;
    return 0;
}

// -----------------------

FilterBooleanGUI_cls::							
FilterBooleanGUI_cls(FilterBoolean_cls *pCParent, 
		     FilterPanelPain_cls *pCPanelPain) : FilterGUI_cls(pCParent) {
    mpCParent = pCParent;
    SetMNumPanel((wxBoxSizer*)pCPanelPain->GetSizer(), pCPanelPain, 0);
}

void FilterBooleanGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, int index){
    
    MultiInputPanel_cls* pCurr = InitializeMInputCommon
	(pCParentWindow, mpCParent->GetPCData()->mCstrCustomName.mstrBody);
    FilterBooleanData_cls *pCData = mpCParent->GetPCData();
    
    pCurr->mUpdateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM;
    
    int boxWidth = 2;
    pCurr->mTextWidth = 80;
    pCurr->mItemWidth = -1;
    pCurr->mUndoType = _MN_UNDO_AUTO_TAG;
    pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Operator"));
    wxString listOperator[3] = {wxT("Intersect"), wxT("Difference"), wxT("Add") };
    pCurr->CombinePointerList(_ITEM_UINT_LIST, &pCData->mCOperatorType.mBody, 3, listOperator);
    wxString listRegion[3] = {wxT("Box"), wxT("Plane"), wxT("Sphere")};
    pCurr->CombinePointerList(_ITEM_UINT_LIST, &pCData->mCRegionType.mBody, 3, listRegion);
    
    std::vector < int > VIndex = SetMNumBasic(pCurr, pCData);

    pCurr->mTextWidth = 80;
    pCurr->mItemWidth = -1;
    pCurr->mUndoType = _MN_UNDO_AUTO_TAG;
    pCurr->CombinePointer(_ITEM_FLOAT, &pCData->mCBorderFactor.mBody, _("Border"));
    
    pCurr->FinishSetting(NULL);

    SetHookBasic(pCurr);

    pCurr->SetColorForXYZ(VIndex[0]+1);
    pCurr->SetColorForXYZ(VIndex[1]+1);
    pCurr->SetColorForXYZ(VIndex[2]+1);
}

// ---------------------

FilterBooleanData_cls::
FilterBooleanData_cls() {
    mCstrCustomName.mstrBody = _("Boolean");
    mCstrCustomName.mstrBody << wxT("%02d");
    mCConfList.Add(&mCOperatorType, wxT("OperatorType="), wxT("0"));
    mCConfList.Add(&mCRegionType, wxT("RegionType="), wxT("0"));
    mCConfList.Add(&mCBorderFactor, wxT("BorderFactor="), wxT("0.01"));
    
    //LOOP_XYZ mDummyRotate[xyz] = 0;
}

// ---------------------

FilterBoolean_cls::
FilterBoolean_cls(FilterPanelPain_cls *pCPanelPain,
		  FilterBooleanData_cls *pCInitData) :
    FilterBase_cls(pCPanelPain->mpCFilterCtrlLink)
{
    wxString strInit; strInit << _FILTER_BOOLEAN;
    mCFilterType.SetItem(wxT("FilterType="), strInit);
    mpCData     = new FilterBooleanData_cls;
    mpCGUIPanel = new FilterBooleanGUI_cls(this, pCPanelPain);
}

FilterBoolean_cls::
~FilterBoolean_cls(){
}

FilterBooleanData_cls* FilterBoolean_cls::
GetPCData(){
    return (FilterBooleanData_cls*)mpCData;
}

void FilterBoolean_cls::
CheckBoolean(ArrayAndIndex_cls* pIndexAndZ, 
	     float (*pX)[3], int downsized, float *pAlphaFactor,
	     double doubleIndex){
    
    FilterBooleanData_cls *pCData = GetPCData();
    ParamGroup_cls *pCTranslate = pCData->mpCTranslate;
    ParamGroup_cls *pCScale     = pCData->mpCScale;

    if (pCData->mCbValid.mbBody == FALSE) return;
    
    std::vector < double > VTrans = 
	pCTranslate->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    std::vector < double > VScale = 
	pCScale->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);

    float dX = VTrans[0];
    float dY = VTrans[1];
    float dZ = VTrans[2];
    float scaleX = VScale[0];
    float scaleY = VScale[1];
    float scaleZ = VScale[2];

    int operatorType   = pCData->mCOperatorType.mBody;
    int regionType     = pCData->mCRegionType.mBody;
    float borderFactor = pCData->mCBorderFactor.mBody;

    int threadNumber= 1;
    if (downsized > 10000) threadNumber = 2;
    if (downsized > 100000) threadNumber = 3;
    if (downsized > 1000000) threadNumber = 4;
    ThreadBooleanCheckBase_cls **ppCThreadBooleanCheck = 
	new ThreadBooleanCheckBase_cls* [threadNumber];
    bool *bAFinished = new bool [threadNumber];
    
    for (int core = 0; core < threadNumber; core++){
	bAFinished[core] = FALSE;

	switch(regionType){
	case 0: case 1:
	    ppCThreadBooleanCheck[core] = 
		new ThreadBooleanCheckBox_cls(wxTHREAD_DETACHED);
	    break;
	case 2:
	    ppCThreadBooleanCheck[core] = 
		new ThreadBooleanCheckSphere_cls(wxTHREAD_DETACHED);
	    break;
	}
	    
	int PInit = downsized / threadNumber * core;
	int PFin  = downsized / threadNumber * (core+1);
	if (core == threadNumber - 1) PFin = downsized;

	ppCThreadBooleanCheck[core]->SetParameters
	    (&bAFinished[core], pIndexAndZ, pX,
	     pAlphaFactor, operatorType, borderFactor,
	     dX, dY, dZ, scaleX, scaleY, scaleZ,
	     PInit, PFin);
	ppCThreadBooleanCheck[core]->Run();
    }
	
    bool bFinishedAll = FALSE;
    while (bFinishedAll == FALSE){
	wxMilliSleep(3);
	bFinishedAll = TRUE;
	for (int core = 0; core < threadNumber; core++)
	    bFinishedAll &= bAFinished[core];
    }
	
    delete [] bAFinished;
    delete [] ppCThreadBooleanCheck;
}

void FilterBoolean_cls::
GLDraw(double doubleIndex, double scale, bool bHittest){
    FilterBooleanData_cls *pCData = GetPCData();
    if (pCData->mpCSceneObjParam->mbObjectVisible == FALSE) return;

    glPushMatrix();
    ParamGroup_cls *pCTranslate = pCData->mpCTranslate;
    ParamGroup_cls *pCRotate    = pCData->mpCRotation;
    ParamGroup_cls *pCScale     = pCData->mpCScale;

    if (GetPCData()->mbSelected == FALSE)
	glColor4f(0.3f, 0.8f, 0.3f, 1.0f);
    else
	glColor4f(0.8f, 0.7f, 0.3f, 1.0f);

    std::vector < double > VTrans = 
	pCTranslate->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    glTranslatef(VTrans[0], VTrans[1], VTrans[2]);

    std::vector < double > VRot = 
	pCRotate->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    
    glRotatef(VRot[2], 0, 0, 1);
    glRotatef(VRot[0], 1, 0, 0);
    glRotatef(VRot[1], 0, 1, 0);

    std::vector < double > VScale = 
	pCScale->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    glScalef(VScale[0], VScale[1], VScale[2]);
    
    glBegin(GL_LINES);
    glVertex3d(-0.5, -0.5, -0.5);
    glVertex3d(-0.5, -0.5, +0.5);
    glVertex3d(+0.5, -0.5, -0.5);
    glVertex3d(+0.5, -0.5, +0.5);
    glVertex3d(+0.5, +0.5, -0.5);
    glVertex3d(+0.5, +0.5, +0.5);
    glVertex3d(-0.5, +0.5, -0.5);
    glVertex3d(-0.5, +0.5, +0.5);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3d(-0.5, -0.5, -0.5);
    glVertex3d(+0.5, -0.5, -0.5);
    glVertex3d(+0.5, +0.5, -0.5);
    glVertex3d(-0.5, +0.5, -0.5);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3d(-0.5, -0.5, +0.5);
    glVertex3d(+0.5, -0.5, +0.5);
    glVertex3d(+0.5, +0.5, +0.5);
    glVertex3d(-0.5, +0.5, +0.5);
    glEnd();

    glPopMatrix();
}
