// -*- c++ -*- 
// (setq c-basic-offset 4)

#include "clsFilterCopy.h"
#include "clsFilterCtrl.h"

#include "../MAINCANVAS/clsFilterPanelPain.h"
#include "../PARTICLEDATA/clsPeriodicBoundary.h"
#include "../PARTICLEDATA/clsArrayAndIndex.h"
#include "../POVRAY/clsPovRay.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "../UNDO/clsUndoItemConstParam.h"
#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsHookCtrl.h"

#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

extern wxFrame  *GpCMainFrameCLS;

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

#include <algorithm>

// -----------------------

FilterCopyGUI_cls::							
FilterCopyGUI_cls(FilterCopy_cls *pCParent, 
		  FilterPanelPain_cls *pCPanelPain) : FilterGUI_cls(pCParent) {
    mpCParent = pCParent;
    SetMNumPanel((wxBoxSizer*)pCPanelPain->GetSizer(), pCPanelPain, 0);
}

void FilterCopyGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, int index){
    
    MultiInputPanel_cls* pCurr = InitializeMInputCommon
	(pCParentWindow, mpCParent->GetPCData()->mCstrCustomName.mstrBody);
    FilterCopyData_cls *pCData = mpCParent->GetPCData();

    pCurr->mUpdateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM;

    int boxWidth = 2;
    pCurr->mTextWidth = 80;
    pCurr->mItemWidth = 90;
    pCurr->mUndoType = _MN_UNDO_AUTO_TAG;

    int indexArray0 = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Array 0"));
    pCurr->CombinePointer(_ITEM_INTEGER, &pCData->mCArrayMin[0].mBody);
    pCurr->CombinePointer(_ITEM_INTEGER, &pCData->mCArrayMax[0].mBody);

    int indexArray1 = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Array 1"));
    pCurr->CombinePointer(_ITEM_INTEGER, &pCData->mCArrayMin[1].mBody);
    pCurr->CombinePointer(_ITEM_INTEGER, &pCData->mCArrayMax[1].mBody);

    int indexArray2 = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Array 2"));
    pCurr->CombinePointer(_ITEM_INTEGER, &pCData->mCArrayMin[2].mBody);
    pCurr->CombinePointer(_ITEM_INTEGER, &pCData->mCArrayMax[2].mBody);
    
    pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->mItemWidth = -1;
    int indexBound = pCurr->CombinePointerButton(wxID_ANY, _("Boundary Size"));
    pCurr->GetLastItem().mAlign = _ITEM_ALIGN_RIGHT;
    int indexShear = pCurr->CombinePointerButton(wxID_ANY, _("Shearing Box"));
    pCurr->GetLastItem().mAlign = _ITEM_ALIGN_RIGHT;

    pCurr->FinishSetting(NULL);
    
    SetHookBasic(pCurr);
    pCurr->SetFuncHook(FuncHookPeriodic, _HOOK_BUTTON, indexBound, FALSE, this);
    pCurr->SetFuncHook(FuncHookShear, _HOOK_BUTTON, indexShear, FALSE, this);
}

void FilterCopyGUI_cls::
FuncHookPeriodic(void *pUsrDef, void *pArg){
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuDialogPeriodic);
    GpCMainFrameCLS->AddPendingEvent(com);
}

void FilterCopyGUI_cls::
FuncHookShear(void *pUsrDef, void *pArg){
    ((FilterCopyGUI_cls*)pUsrDef)->mpCParent->OpenShearDialog();
}

// ---------------------

FilterCopyData_cls::
FilterCopyData_cls() : FilterData_cls(FALSE) {
    mCstrCustomName.mstrBody = _("CopyBox");
    mCstrCustomName.mstrBody << wxT("%02d");

    mCConfList.Add(&mCArrayMin[0], wxT("CopyArrayMin0="), wxT("0"));
    mCConfList.Add(&mCArrayMax[0], wxT("CopyArrayMax0="), wxT("1"));
    mCConfList.Add(&mCArrayMin[1], wxT("CopyArrayMin1="), wxT("0"));
    mCConfList.Add(&mCArrayMax[1], wxT("CopyArrayMax1="), wxT("1"));
    mCConfList.Add(&mCArrayMin[2], wxT("CopyArrayMin2="), wxT("0"));
    mCConfList.Add(&mCArrayMax[2], wxT("CopyArrayMax2="), wxT("1"));

    mCConfList.Add(&mCShearArray[0], wxT("ShearArray0="), wxT("0 0 0"));
    mCConfList.Add(&mCShearArray[1], wxT("ShearArray1="), wxT("0 0 0"));
    mCConfList.Add(&mCShearArray[2], wxT("ShearArray2="), wxT("0 0 0"));
    mCConfList.Add(&mCShearPhase, wxT("ShearPhase="), wxT("0"));
}

// ---------------------

FilterCopy_cls::
FilterCopy_cls(FilterPanelPain_cls *pCPanelPain,
	       FilterCopyData_cls *pCInitData) :
    FilterBase_cls(pCPanelPain->mpCFilterCtrlLink)
{
    wxString strInit; strInit << _FILTER_COPY;
    mCFilterType.SetItem(wxT("FilterType="), strInit);
    mpCData     = new FilterCopyData_cls;
    mpCGUIPanel = new FilterCopyGUI_cls(this, pCPanelPain);
}

FilterCopyData_cls* FilterCopy_cls::
GetPCData(){
    return (FilterCopyData_cls*)mpCData;
}

int FilterCopy_cls::
GetNumber(){
    FilterCopyData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return 1;
    int xnum = pCData->mCArrayMax[0].mBody - pCData->mCArrayMin[0].mBody;
    int ynum = pCData->mCArrayMax[1].mBody - pCData->mCArrayMin[1].mBody;
    int znum = pCData->mCArrayMax[2].mBody - pCData->mCArrayMin[2].mBody;
    return xnum * ynum * znum;
}

int FilterCopy_cls::
GetBitBoundary(int index){
    FilterCopyData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return 0;
     
    int bit = 7;
    int xnum = pCData->mCArrayMax[X].mBody - pCData->mCArrayMin[X].mBody;
    int ynum = pCData->mCArrayMax[Y].mBody - pCData->mCArrayMin[Y].mBody;
    int znum = pCData->mCArrayMax[Z].mBody - pCData->mCArrayMin[Z].mBody;
    int ix = index % xnum;
    int iy = ((index - ix) / xnum ) % ynum;
    int iz = ((index - ix) - xnum * iy) / (xnum * ynum);

    if (ix == xnum - 1) bit -= 1;
    if (iy == ynum - 1) bit -= 2;
    if (iz == znum - 1) bit -= 4;
    return bit;
}

std::vector < float > FilterCopy_cls::
GetVTranslate(int index, double time){
    std::vector < float > VRet;
    
    FilterCopyData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return VRet;
    
    int xnum = pCData->mCArrayMax[X].mBody - pCData->mCArrayMin[X].mBody;
    int ynum = pCData->mCArrayMax[Y].mBody - pCData->mCArrayMin[Y].mBody;
    int znum = pCData->mCArrayMax[Z].mBody - pCData->mCArrayMin[Z].mBody;
    int ix = index % xnum;
    int iy = ((index - ix) / xnum ) % ynum;
    int iz = ((index - ix) - xnum * iy) / (xnum * ynum);

    ix += pCData->mCArrayMin[X].mBody;
    iy += pCData->mCArrayMin[Y].mBody;
    iz += pCData->mCArrayMin[Z].mBody;

    float array[3][3];
    LOOP_XYZ mpCParentCtrl->mpCPeriodicLink->GetBoundaryVector(array[xyz], xyz);
    
    float dx = ix * array[0][X] + iy * array[1][X] + iz * array[2][X];
    float dy = ix * array[0][Y] + iy * array[1][Y] + iz * array[2][Y];
    float dz = ix * array[0][Z] + iy * array[1][Z] + iz * array[2][Z];

    float dpos0[3] = {dx, dy, dz};
    float dt = pCData->mCShearPhase.mBody;
    for (int i = 0; i < 3; i++){
	dx += dpos0[i] * pCData->mCShearArray[i].mpBody[X] * (time-dt);
	dy += dpos0[i] * pCData->mCShearArray[i].mpBody[Y] * (time-dt);
	dz += dpos0[i] * pCData->mCShearArray[i].mpBody[Z] * (time-dt);
    }
	
    VRet.push_back(dx);
    VRet.push_back(dy);
    VRet.push_back(dz);
    return VRet;
}

void FilterCopy_cls::
ZSortBegin(double cameraPos[3], double cameraDir[3], double time){
    int num = GetNumber();
    mpCIndexAndZ = new ArrayAndIndex_cls[num];

    for (int i = 0; i < num; i++){
	mpCIndexAndZ[i].index = i;
	std::vector < float > VTrans = GetVTranslate(i, time);

	if (VTrans.size() == 0) continue;

	mpCIndexAndZ[i].z = - 
		VTrans[X] * cameraDir[X] - 
		VTrans[Y] * cameraDir[Y] - 
		VTrans[Z] * cameraDir[Z];
	/*
	mpCIndexAndZ[i].z = -
	    (VTrans[X]-cameraPos[X]) * (VTrans[X]-cameraPos[X]) -
	    (VTrans[Y]-cameraPos[Y]) * (VTrans[Y]-cameraPos[Y]) -
	    (VTrans[Z]-cameraPos[Z]) * (VTrans[Z]-cameraPos[Z]);
	*/
    }

    std::sort(mpCIndexAndZ, mpCIndexAndZ + num);
}

void FilterCopy_cls::
GLTranslate(int index, double time){
    
    std::vector < float > VTrans = GetVTranslate(mpCIndexAndZ[index].index, time);
    if (VTrans.size() == 0) return;

    glTranslatef(VTrans[X], VTrans[Y], VTrans[Z]);
}

void FilterCopy_cls::
ZSortEnd(){
    delete [] mpCIndexAndZ;
}

void FilterCopy_cls::
OpenShearDialog(){
    MultiInputDialog_cls *pCDLG = new
	MultiInputDialog_cls(&pCDLG, mpCGUIPanel->mpCMNumPanel, wxID_ANY, 
			     _("Shearing Box"), wxDP, wxDS);
    FilterCopyData_cls *pCData = GetPCData();

    float array[3][3];
    for (int i = 0; i < 3; i++)
	LOOP_XYZ array[i][xyz] = pCData->mCShearArray[i].mpBody[xyz];
    float phase = pCData->mCShearPhase.mBody;
    
    MultiInputPanel_cls *pCurr = pCDLG->mpCPanel;

    pCurr->mTextWidth = 20;
    int boxWidth = 3;
    int indexX = pCurr->CombinePointer
	    (_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("X"));
    LOOP_XYZ
	pCurr->CombinePointer(_ITEM_FLOAT, &(array[0][xyz]));

    int indexY = pCurr->CombinePointer
	    (_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Y"));
    LOOP_XYZ
	pCurr->CombinePointer(_ITEM_FLOAT, &(array[1][xyz]));
    int indexZ = pCurr->CombinePointer
	    (_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Z"));
    LOOP_XYZ
	pCurr->CombinePointer(_ITEM_FLOAT, &(array[2][xyz]));

    pCurr->mTextWidth = 80;
    pCurr->mItemWidth = -1;
    pCurr->CombinePointer(_ITEM_FLOAT, &phase, _("Shear Offset"));
    
    pCDLG->MakeGeometry();
    pCDLG->UpdateValues();

    pCurr->SetColorForXYZ(indexX+1);
    pCurr->SetColorForXYZ(indexY+1);
    pCurr->SetColorForXYZ(indexZ+1);
    
    if (pCDLG->ShowModal() == FALSE) return;

    for (int i = 0; i < 3; i++)
	LOOP_XYZ pCData->mCShearArray[i].mpBody[xyz] = array[i][xyz];
    pCData->mCShearPhase.mBody = phase;
}

wxString FilterCopy_cls::
GetStrPovTranslate(int index, double time){
    wxString strTranstale;

    std::vector < float > VTrans = GetVTranslate(index, time);
    if (VTrans.size() == 0) return strTranstale;

    strTranstale << PovRayCtrl_cls::PovOut3f(wxT("translate"), VTrans.data());
    
    return strTranstale;
}
