// -*- c++ -*- 
// (setq c-basic-offset 4)

#include "clsFilterFog.h"

#include "../GLOBJECTS/clsCube.h"
#include "../GLOBJECTS/clsSphere.h"
#include "../MAINCANVAS/clsFilterPanelPain.h"
#include "../PARTICLEDATA/clsGLDrawData.h"
#include "../POVRAY/clsPovRay.h"
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsHookCtrl.h"

#include "../../wxWidgetEnum.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

// -----------------------

FilterFogGUI_cls::							
FilterFogGUI_cls(FilterFog_cls *pCParent, 
		 FilterPanelPain_cls *pCPanelPain) : FilterGUI_cls(pCParent) {
    mpCParent = pCParent;
    SetMNumPanel((wxBoxSizer*)pCPanelPain->GetSizer(), pCPanelPain, 0);
}

void FilterFogGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, int index){
    
    MultiInputPanel_cls* pCurr = InitializeMInputCommon
	(pCParentWindow, mpCParent->GetPCData()->mCstrCustomName.mstrBody);
    FilterFogData_cls *pCData = mpCParent->GetPCData();

    pCurr->mUpdateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM;
    
    int boxWidth = 2;
    
    pCurr->mTextWidth = 80;
    pCurr->mUndoType = _MN_UNDO_AUTO_TAG;
    
    //pCurr->mComboWidth = 88;
    //pCurr->mItemWidth = 88;
    wxString listType[3] = {wxT("EXP"), wxT("EXP 2"), wxT("LINEAR") };
    mIndexType = pCurr->CombinePointerList(_ITEM_UINT_LIST, &pCData->mCFogType.mBody, 3, listType, _("Fog Type"));
    
    pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Color/alpha"));
    
    pCurr->mItemWidth = -1;
    ColorCurveBtn_cls *pCBtnFogColor = new ColorCurveBtn_cls
	(pCurr, pCData->mVpCClrCurve[0], TRUE, FALSE);
    pCurr->CombinePointer(_ITEM_WXPANEL, pCBtnFogColor);

    pCurr->mItemWidth = 88;
    pCurr->mUndoType = _MN_UNDO_MANUAL;
    pCurr->CombinePointer(_ITEM_DOUBLE, &(pCData->mpCParameter->mVCParam[1].mInterpolatedValue));
    
    pCurr->mItemWidth = -1;    
    mIndexDensity = pCurr->CombinePointer(_ITEM_DOUBLE, &(pCData->mpCParameter->mVCParam[0].mInterpolatedValue), _("Density"));
    
    pCurr->mItemWidth = 88;
    pCurr->mUndoType = _MN_UNDO_AUTO_TAG;
    mIndexClip = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Near/Far"));
    pCurr->CombinePointer(_ITEM_FLOAT, &pCData->mCFogStart.mBody);
    pCurr->CombinePointer(_ITEM_FLOAT, &pCData->mCFogEnd.mBody);

    pCurr->FinishSetting(NULL);

    SetHookBasic(pCurr);
    
    pCurr->SetFuncHook(FuncHookType, _HOOK_COMBOBOX, mIndexType, FALSE, this);

    HideAndShow();
}

void FilterFogGUI_cls::
FuncHookType(void *pUsrDef, void *pArg){
    ((FilterFogGUI_cls*)pUsrDef)->HideAndShow();
}

void FilterFogGUI_cls::
HideAndShow(){
    HideAndShowBegin();
    FilterFogData_cls *pCData = mpCParent->GetPCData();

    if (pCData->mCFogType.mBody == _FOG_LINEAR){
	mpCMNumPanel->ShowItem(mIndexDensity, FALSE);
	mpCMNumPanel->ShowItem(mIndexClip, TRUE);
    }
    else{
	mpCMNumPanel->ShowItem(mIndexClip, FALSE);
	mpCMNumPanel->ShowItem(mIndexDensity, TRUE);
    }

    HideAndShowEnd();
}

// ---------------------

FilterFogData_cls::
FilterFogData_cls() : FilterData_cls(FALSE) {
    wxString listTag[2] = {wxT("deinsity"), wxT("alpha") };
    int updateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM;
    wxString strParameter = wxT("Filter"); // {ꉻႾ
    strParameter << wxT(" Parameter");
    mpCParameter = new ParamGroup_cls(strParameter, 2, listTag, updateBit, FALSE);
    double pParamInit[2] = {0, 1};
    mpCParameter->SetKeyFrame(pParamInit, TRUE);
    mpCSceneObjParam->mVpCParamGroup.push_back(mpCParameter);
    
    mCstrCustomName.mstrBody = _("Fog");
    mCstrCustomName.mstrBody << wxT("%02d");

    ColorCurve_cls *pCClrFogColor = new ColorCurve_cls
	(wxT("FogColor"), wxT("FogColor"), wxT("0.5 0.5 0.5 1"));
    mVpCClrCurve.push_back(pCClrFogColor);

    mCConfList.Add(&mCFogType,  wxT("FogType="),  wxT("0"));
    mCConfList.Add(&mCFogStart, wxT("FogStart="), wxT("0"));
    mCConfList.Add(&mCFogEnd,   wxT("FogEnd="),   wxT("1"));
}

FilterFogData_cls::
~FilterFogData_cls(){
    delete mpCParameter;
}

void FilterFogData_cls::
CopyFilterDataFromP(FilterData_cls *pCFrom,
		    ParamGroup_cls *pCTranslate,
		    ParamGroup_cls *pCRotete   ,
		    ParamGroup_cls *pCScale    ,
		    ParamGroup_cls *pCAdditional){
    if (pCAdditional == NULL) 
	mpCParameter->CopyFromP
	    (((FilterFogData_cls*)pCFrom)->mpCParameter);
    else
	mpCParameter->CopyFromP(pCAdditional);
    FilterData_cls::CopyFilterDataFromP
	(pCFrom, pCTranslate, pCRotete, pCScale);
}

void FilterFogData_cls::
UpdateCurrentObj(double doubleIndex){
    SceneObj_cls::UpdateCurrentObj(doubleIndex);

    mpCParameter->GetInterpolatedValue(doubleIndex, 0);
}

// ---------------------

FilterFog_cls::
FilterFog_cls(FilterPanelPain_cls *pCPanelPain,
		  FilterFogData_cls *pCInitData) :
    FilterBase_cls(pCPanelPain->mpCFilterCtrlLink)
{
    wxString strInit; strInit << _FILTER_FOG;
    mCFilterType.SetItem(wxT("FilterType="), strInit);
    mpCData     = new FilterFogData_cls;
    mpCGUIPanel = new FilterFogGUI_cls(this, pCPanelPain);
}

FilterFog_cls::
~FilterFog_cls(){
}

FilterFogData_cls* FilterFog_cls::
GetPCData(){
    return (FilterFogData_cls*)mpCData;
}

void FilterFog_cls::
SetGLClearColor(double doubleIndex, float clrBG[3]){
    FilterFogData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return;

    ParamGroup_cls *pCParameter = pCData->mpCParameter;

    float fogColor[4];// = {0.5, 0.5, 0.5, 1};
    pCData->mVpCClrCurve[0]->GetColor4fScaled(fogColor, 0);

    float fogDens = pCParameter->mVCParam[0].GetInterpolatedValue(doubleIndex);
    float alpha = pCParameter->mVCParam[1].GetInterpolatedValue(doubleIndex);

    for (int i = 0; i < 3; i++)
	fogColor[i] = fogColor[i] * alpha + clrBG[0] * (1.0 - alpha);
    
    glClearColor( fogColor[0], fogColor[1], fogColor[2], 1.0f);
}

void FilterFog_cls::
SetGLFog(double doubleIndex){
    FilterFogData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return;
    
    ParamGroup_cls *pCParameter = pCData->mpCParameter;

    float fogColor[4];// = {0.5, 0.5, 0.5, 1};
    pCData->mVpCClrCurve[0]->GetColor4fScaled(fogColor, 0);

    float fogDens = pCParameter->mVCParam[0].GetInterpolatedValue(doubleIndex);
    float alpha = pCParameter->mVCParam[1].GetInterpolatedValue(doubleIndex);
    fogColor[3] = alpha;

    glFogfv(GL_FOG_COLOR, fogColor); //F
    switch(pCData->mCFogType.mBody){
    case _FOG_EXP:
	glFogi(GL_FOG_MODE, GL_EXP); 
	glFogf(GL_FOG_DENSITY, fogDens * alpha);
	break;
    case _FOG_EXP2:
	glFogi(GL_FOG_MODE, GL_EXP2);
	glFogf(GL_FOG_DENSITY, fogDens * alpha);
	break;
    case _FOG_LINEAR:
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogf(GL_FOG_START, pCData->mCFogStart.mBody);
	glFogf(GL_FOG_END, pCData->mCFogEnd.mBody); 
	break;
    }
    
    glEnable(GL_FOG);
}

void FilterFog_cls::
GLDraw(double doubleIndex, double scale, bool bHittest, int mode){
}

void FilterFog_cls::
GLDrawRealFog(double doubleIndex){
}

wxString FilterFog_cls::
GetStrPovFog(double doubleIndex){
    wxString strFog;
    FilterFogData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return strFog;

    ParamGroup_cls *pCParameter = pCData->mpCParameter;

    float fogColor[4];
    pCData->mVpCClrCurve[0]->GetColor4fScaled(fogColor, 0);

    float fogDens = pCParameter->mVCParam[0].GetInterpolatedValue(doubleIndex);
    float alpha = pCParameter->mVCParam[1].GetInterpolatedValue(doubleIndex);

    fogColor[3] = 1.0 - alpha;

    float bgColor[4];
    for (int i = 0; i < 4; i++)	bgColor[i] = fogColor[i];
    if (alpha < 1) bgColor[3] = 1;
    
    float distance;
    switch(pCData->mCFogType.mBody){
    case _FOG_EXP:
    case _FOG_EXP2:
	if (fogDens == 0) return strFog;
	distance = 1.0 / fogDens;
	break;
    case _FOG_LINEAR:
	distance = pCData->mCFogEnd.mBody / 2;
	break;
    }

    strFog << wxT("background {") << PovRayCtrl_cls::PovOut4f(wxT("rgbt "), bgColor, 4) << wxT("}") << ENDL;
    strFog << wxT("fog { ") << ENDL
	   << wxT("\tdistance ") << distance  << ENDL
	   << wxT("\tcolor ") << PovRayCtrl_cls::PovOut4f(wxT("rgbt "), fogColor, 4) << ENDL
	   << wxT("}") << ENDL;

    return strFog;
}
