// -*- c++ -*- 
// (setq c-basic-offset 4)

#include "clsFilterObject.h"

#include "clsFilterCtrl.h"

#include "../GLATTRIB/clsGLAttribute.h"
#include "../GLATTRIB/clsShaderCtrl.h"
#include "../GLOBJECTS/clsCube.h"
#include "../GLOBJECTS/clsSphere.h"
extern Sphere_cls *GpCSphere;
#include "../MAINCANVAS/clsFilterPanelPain.h"
#include "../PARTICLEDATA/clsGLDrawData.h"
#include "../POVRAY/clsPovRay.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "../UNDO/clsUndoItemConstParam.h"
#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsHookCtrl.h"

#include "../WIDGETCTRL/EX/clsFileDialogEX.h"
extern wxFileDialogEX *GpCFileDialog;

#include "../../clsMain.h"
#include "../../wxWidgetEnum.h"

#include "../COMMON/CommonDefinitions.h"
#include "../COMMON/CommonDefinitionsGL.h"

// -----------------------

FilterObjectGUI_cls::							
FilterObjectGUI_cls(FilterObject_cls *pCParent, 
		     FilterPanelPain_cls *pCPanelPain) : FilterGUI_cls(pCParent) {
    mpCParent = pCParent;
    SetMNumPanel((wxBoxSizer*)pCPanelPain->GetSizer(), pCPanelPain, 0);
    HideAndShow();
}

void FilterObjectGUI_cls::
SetMNumPanel(wxBoxSizer *pBoxSizer, wxWindow *pCParentWindow, int index){
    
    MultiInputPanel_cls* pCurr = InitializeMInputCommon
	(pCParentWindow, mpCParent->GetPCData()->mCstrCustomName.mstrBody);
    FilterObjectData_cls *pCData = mpCParent->GetPCData();

    pCurr->mUpdateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM | _COMMAND_UPDATE_DISPLAY_TEXTURE ;

    int boxWidth = 2;
    pCurr->mTextWidth = 80;
    pCurr->mItemWidth = -1;
    pCurr->mUndoType = _MN_UNDO_AUTO_TAG;
    
    mIndexType = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Obj/Shader"));
    
    wxString listObject[4] = {wxT("Box"), wxT("Plane"), wxT("Sphere"), wxT("Import")};
    pCurr->CombinePointerList(_ITEM_UINT_LIST, &pCData->mCObjectType.mBody, 4, listObject);

    wxString strListShader[4] = {_("original"), _("Gouroud"), _("Phong"), _("Toon")};
    pCurr->CombinePointerList(_ITEM_UINT_LIST, &pCData->mCTexture.mCShaderType.mBody, 4, strListShader);

    mIndexPath = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth);
    
    pCurr->mItemWidth = pCurr->mMinimumWidth - 30 + 10;
    pCurr->mUndoType = _MN_UNDO_MANUAL;
    pCurr->CombinePointer(_ITEM_STRING, &pCData->mCstrImportPath.mstrBody);
    pCurr->mItemWidth = 30;
    pCurr->CombinePointerButton(wxID_ANY, wxT(".."));

    pCurr->mItemWidth = 120;
    mIndexArrayVector = pCurr->CombinePointerButton(wxID_ANY, _("Affine map"));
    pCurr->GetLastItem().mAlign = _ITEM_ALIGN_RIGHT;
    
    std::vector < int > VIndex = SetMNumBasic(pCurr, pCData);

    pCurr->mTextWidth = 80;
    pCurr->mItemWidth = -1;
    pCurr->mUndoType = _MN_UNDO_AUTO_TAG;

    boxWidth = 3;
    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Material"));
    ColorCurveBtn_cls *pCBtnAmbient = new ColorCurveBtn_cls
	(pCurr, pCData->mVpCClrCurve[0], TRUE, FALSE);
    ColorCurveBtn_cls *pCBtnDiffuse = new ColorCurveBtn_cls
	(pCurr, pCData->mVpCClrCurve[1], TRUE, FALSE);
    ColorCurveBtn_cls *pCBtnSpecular = new ColorCurveBtn_cls
	(pCurr, pCData->mVpCClrCurve[2], TRUE, FALSE);
    pCurr->CombinePointer(_ITEM_WXPANEL, pCBtnAmbient, wxT("Ambient"));
    pCurr->CombinePointer(_ITEM_WXPANEL, pCBtnDiffuse, wxT("Diffuse"));
    pCurr->CombinePointer(_ITEM_WXPANEL, pCBtnSpecular, wxT("Specular"));
    
    pCurr->FinishSetting(NULL);
    
    pCurr->SetColorForXYZ(VIndex[0]+1);
    pCurr->SetColorForXYZ(VIndex[1]+1);
    pCurr->SetColorForXYZ(VIndex[2]+1);

    SetHookBasic(pCurr);

    pCurr->SetFuncHook(FilterObjectGUI_cls::FuncHookType, _HOOK_COMBOBOX, 
		       mIndexType+1, FALSE, this);
    pCurr->SetFuncHook(FilterObjectGUI_cls::FuncHookButton, _HOOK_BUTTON, 
		       mIndexPath+2, FALSE, this);
    pCurr->SetFuncHook(FilterObjectGUI_cls::FuncHookArrayVector, _HOOK_BUTTON, 
		       mIndexArrayVector, FALSE, this);
}

void FilterObjectGUI_cls::
FuncHookType(void *pUsrDef, void *pArg){
    ((FilterObjectGUI_cls*)pUsrDef)->HideAndShow();
}

void FilterObjectGUI_cls::
FuncHookButton(void *pUsrDef, void *pArg){
    ((FilterObjectGUI_cls*)pUsrDef)->PathSelection();
}

void FilterObjectGUI_cls::
HideAndShow(){
    HideAndShowBegin();
    FilterObjectData_cls *pCData = mpCParent->GetPCData();
    if (pCData->mCObjectType.mBody == _FILTER_OBJECT_IMPORT &&
	mpCMNumPanel->mbPanelOpen == TRUE){
	mpCMNumPanel->ShowItem(mIndexPath, TRUE);
	mpCMNumPanel->ShowItem(mIndexArrayVector, TRUE);
    }
    else{
	mpCMNumPanel->ShowItem(mIndexPath, FALSE);
	mpCMNumPanel->ShowItem(mIndexArrayVector, FALSE);
    }
    HideAndShowEnd();
}

void FilterObjectGUI_cls::
PathSelection(){
    GpCFileDialog->SetMessage(wxT("Select Mesh File"));
    GpCFileDialog->SetWildcard(wxT("PovMesh (*.pov)|*.pov|Wavefront Object (*.obj)|*.obj"));
    if (GpCFileDialog->ShowModal() == wxID_CANCEL) return;
    
    FilterObjectData_cls *pCData = mpCParent->GetPCData();
    int updateBit = _COMMAND_UPDATE_DISPLAY_FILTER_PARAM | _COMMAND_UPDATE_DISPLAY_GL;

    GCUndoList.PushNewCommand
	(new CommandSetValue_cls(&pCData->mCstrImportPath.mstrBody, GpCFileDialog->GetPath(), updateBit), 
	 new CommandSetValue_cls(&pCData->mCstrImportPath.mstrBody, pCData->mCstrImportPath.mstrBody, updateBit));
    GCUndoList.ExecCurrent();

    //mpCMNumPanel->UpdateValues();
}

void FilterObjectGUI_cls::
FuncHookArrayVector(void *pUsrDef, void *pArg){
    FilterObjectGUI_cls* pCThis = (FilterObjectGUI_cls*)pUsrDef;
    
    MultiInputDialog_cls *pCDLG = new
	MultiInputDialog_cls(&pCDLG, pCThis->mpCMNumPanel, 
			     wxID_ANY, wxT("Array Vector"),
			     wxDefaultPosition);
    MultiInputPanel_cls *pCurr = pCDLG->mpCPanel;
    //pCDLG->mpCUserDefinedClueDialog = pCThis->mpCMNumPanel;

    FilterObjectData_cls *pCData = pCThis->mpCParent->GetPCData();
    
    float array0[3], array1[3], array2[3];
    LOOP_XYZ{
	array0[xyz] = pCData->mCArray0Vector.mpBody[xyz];
	array1[xyz] = pCData->mCArray1Vector.mpBody[xyz];
	array2[xyz] = pCData->mCArray2Vector.mpBody[xyz];
    }

    int boxWidth = 3;
	
    pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Array 0"));
    pCurr->CombinePointer(_ITEM_FLOAT, &array0[X]);
    pCurr->CombinePointer(_ITEM_FLOAT, &array0[Y]);
    pCurr->CombinePointer(_ITEM_FLOAT, &array0[Z]);
    pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Array 1"));
    pCurr->CombinePointer(_ITEM_FLOAT, &array1[X]);
    pCurr->CombinePointer(_ITEM_FLOAT, &array1[Y]);
    pCurr->CombinePointer(_ITEM_FLOAT, &array1[Z]);
    pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Array 2"));
    pCurr->CombinePointer(_ITEM_FLOAT, &array2[X]);
    pCurr->CombinePointer(_ITEM_FLOAT, &array2[Y]);
    pCurr->CombinePointer(_ITEM_FLOAT, &array2[Z]);

    pCDLG->MakeGeometry();
    if (pCDLG->ShowModal() == wxID_CANCEL) return;

    bool bDifferent = FALSE;
    LOOP_XYZ{
	bDifferent |= (array0[xyz] != pCData->mCArray0Vector.mpBody[xyz]);
	bDifferent |= (array1[xyz] != pCData->mCArray1Vector.mpBody[xyz]);
	bDifferent |= (array2[xyz] != pCData->mCArray2Vector.mpBody[xyz]);
    }
    if (bDifferent == FALSE) return;

    GCUndoList.OpenBlock();
    
    float *pOrgValueFloat;
    int updateBit = _COMMAND_UPDATE_DISPLAY_EXPOSURE | _COMMAND_UPDATE_DISPLAY_GL;
    LOOP_XYZ{
	pOrgValueFloat = &pCData->mCArray0Vector.mpBody[xyz];
	GCUndoList.PushNewCommand
	    (new CommandSetValue_cls(pOrgValueFloat, array0[xyz], updateBit),
	     new CommandSetValue_cls(pOrgValueFloat, *pOrgValueFloat, updateBit));
	GCUndoList.ExecCurrent();
    }

    LOOP_XYZ{
	pOrgValueFloat = &pCData->mCArray1Vector.mpBody[xyz];
	GCUndoList.PushNewCommand
	    (new CommandSetValue_cls(pOrgValueFloat, array1[xyz], updateBit),
	     new CommandSetValue_cls(pOrgValueFloat, *pOrgValueFloat, updateBit));
	GCUndoList.ExecCurrent();
    }

    LOOP_XYZ{
	pOrgValueFloat = &pCData->mCArray2Vector.mpBody[xyz];
	GCUndoList.PushNewCommand
	    (new CommandSetValue_cls(pOrgValueFloat, array2[xyz], updateBit),
	     new CommandSetValue_cls(pOrgValueFloat, *pOrgValueFloat, updateBit));
	GCUndoList.ExecCurrent();
    }

    GCUndoList.CloseBlock();

}

// ---------------------

FilterObjectData_cls::
FilterObjectData_cls() {
    mCstrCustomName.mstrBody = _("Object");
    mCstrCustomName.mstrBody << wxT("%02d");
    
    mCConfList.Add(&mCObjectType, wxT("ObjectType="), wxT("0"));
    mCConfList.Add(&mCstrImportPath, wxT("ImportPath="), wxT("NONE"));

    mCConfList.Add(&mCArray0Vector, wxT("ArrayVector0="), wxT("1 0 0"));
    mCConfList.Add(&mCArray1Vector, wxT("ArrayVector1="), wxT("0 1 0"));
    mCConfList.Add(&mCArray2Vector, wxT("ArrayVector2="), wxT("0 0 1"));
    
    mCConfList.Add(&mCTexture.mCShaderType, wxT("ShaderType="), wxT("2"));
    mCConfList.Add(&mCTexture.mCShininess, wxT("Shininess="), wxT("10"));
    mCConfList.Add(&mCTexture.mCTransmitType, wxT("TransmitType="), wxT("0"));
    mCConfList.Add(&mCTexture.mCAlpha, wxT("Alpha="), wxT("1"));
    mCConfList.Add(&mCTexture.mCReflectType, wxT("ReflectType="), wxT("0"));
    mCConfList.Add(&mCTexture.mCReflectMin, wxT("ReflectMin="), wxT("0.2"));
    mCConfList.Add(&mCTexture.mCReflectMax, wxT("ReflectMax="), wxT("0.5"));
    mCConfList.Add(&mCTexture.mCIOR, wxT("IOR="), wxT("1"));

    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	wxString strTagPattern = wxT("NormalPattern");
	strTagPattern << i << wxT("=");
	mCConfList.Add(&mCTexture.mCNormalPattern[i], strTagPattern, wxT("0"));
	wxString strTagDegree = wxT("NormalDegree");
	strTagDegree << i << wxT("=");
	mCConfList.Add(&mCTexture.mCNormalDegree[i], strTagDegree, wxT("0.3"));
	wxString strTagScale = wxT("NormalScale");
	strTagScale << i << wxT("=");
	mCConfList.Add(&mCTexture.mCNormalScale[i], strTagScale, wxT("1"));
    }

    ColorCurve_cls *pCClrCurveAmbient = new ColorCurve_cls
	(wxT("Ambient"), wxT("Ambient"), wxT("0.2 0.16 0.04 1"));
    ColorCurve_cls *pCClrCurveDiffuse = new ColorCurve_cls
	(wxT("Diffuse"), wxT("Diffuse"));
    ColorCurve_cls *pCClrCurveSpeculer = new ColorCurve_cls
	(wxT("Specular"), wxT("Specular"));
    
    mVpCClrCurve.push_back(pCClrCurveAmbient);
    mVpCClrCurve.push_back(pCClrCurveDiffuse);
    mVpCClrCurve.push_back(pCClrCurveSpeculer);
}

wxString FilterObjectData_cls::
GetStrPovMatrix(){
    wxString strMatrix;
    float *p0 = mCArray0Vector.mpBody;
    float *p1 = mCArray1Vector.mpBody;
    float *p2 = mCArray2Vector.mpBody;
    strMatrix << wxT("\tmatrix<") 
	      << p0[X] << wxT(",") << p0[Y] << wxT(",") << p0[Z] << wxT(",")
	      << p1[X] << wxT(",") << p1[Y] << wxT(",") << p1[Z] << wxT(",")
	      << p2[X] << wxT(",") << p2[Y] << wxT(",") << p2[Z] << wxT(",")
	      << 0 << wxT(",") << 0 << wxT(",") << 0 << wxT(">");

    return strMatrix;
}

void FilterObjectData_cls::
DrawMesh(int currentIndex){
    glDisable(GL_CULL_FACE);
    wxString strFile = wxString::Format(mCstrImportPath.mstrBody, currentIndex);
    mCMeshObject.ReadMeshFromFile(strFile);
    GLfloat matrix[16];
    matrix[ 0] = mCArray0Vector.mpBody[X];
    matrix[ 1] = mCArray0Vector.mpBody[Y];
    matrix[ 2] = mCArray0Vector.mpBody[Z];
    matrix[ 3] = 0;
    matrix[ 4] = mCArray1Vector.mpBody[X];
    matrix[ 5] = mCArray1Vector.mpBody[Y];
    matrix[ 6] = mCArray1Vector.mpBody[Z];
    matrix[ 7] = 0;
    matrix[ 8] = mCArray2Vector.mpBody[X];
    matrix[ 9] = mCArray2Vector.mpBody[Y];
    matrix[10] = mCArray2Vector.mpBody[Z];
    matrix[11] = 0;
    matrix[12] = 0;
    matrix[13] = 0;
    matrix[14] = 0;
    matrix[15] = 1;

    glMultMatrixf(matrix);
        
    mCMeshObject.GLDraw();
    glEnable(GL_CULL_FACE);
}

// ---------------------

FilterObject_cls::
FilterObject_cls(FilterPanelPain_cls *pCPanelPain,
		  FilterObjectData_cls *pCInitData) :
    FilterBase_cls(pCPanelPain->mpCFilterCtrlLink)
{
    wxString strInit; strInit << _FILTER_OBJECT;
    mCFilterType.SetItem(wxT("FilterType="), strInit);
    mpCData     = new FilterObjectData_cls;
    mpCGUIPanel = new FilterObjectGUI_cls(this, pCPanelPain);
}

FilterObject_cls::
~FilterObject_cls(){
}

FilterObjectData_cls* FilterObject_cls::
GetPCData(){
    return (FilterObjectData_cls*)mpCData;
}

void FilterObject_cls::
GLDraw(double doubleIndex, double scale, bool bHittest){
    FilterObjectData_cls *pCData = GetPCData();
    if (pCData->mpCSceneObjParam->mbObjectVisible == FALSE) return;
    
    glPushMatrix();
    ParamGroup_cls *pCTranslate = pCData->mpCTranslate;
    ParamGroup_cls *pCRotate    = pCData->mpCRotation;
    ParamGroup_cls *pCScale     = pCData->mpCScale;

    if (GetPCData()->mbSelected == FALSE)
	glColor4f(0.3f, 0.8f, 0.3f, 1.0f);
    else
	glColor4f(0.8f, 0.7f, 0.3f, 1.0f);

    std::vector < double > VTrans = 
	pCTranslate->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    glTranslatef(VTrans[0], VTrans[1], VTrans[2]);

    std::vector < double > VRot = 
	pCRotate->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    
    glRotatef(VRot[2], 0, 0, 1);
    glRotatef(VRot[0], 1, 0, 0);
    glRotatef(VRot[1], 0, 1, 0);

    std::vector < double > VScale = 
	pCScale->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    glScalef(VScale[0], VScale[1], VScale[2]);
    
    glBegin(GL_LINES);
    glVertex3d(-0.5, -0.5, -0.5);
    glVertex3d(-0.5, -0.5, +0.5);
    glVertex3d(+0.5, -0.5, -0.5);
    glVertex3d(+0.5, -0.5, +0.5);
    glVertex3d(+0.5, +0.5, -0.5);
    glVertex3d(+0.5, +0.5, +0.5);
    glVertex3d(-0.5, +0.5, -0.5);
    glVertex3d(-0.5, +0.5, +0.5);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3d(-0.5, -0.5, -0.5);
    glVertex3d(+0.5, -0.5, -0.5);
    glVertex3d(+0.5, +0.5, -0.5);
    glVertex3d(-0.5, +0.5, -0.5);
    glEnd();

    glBegin(GL_LINE_LOOP);
    glVertex3d(-0.5, -0.5, +0.5);
    glVertex3d(+0.5, -0.5, +0.5);
    glVertex3d(+0.5, +0.5, +0.5);
    glVertex3d(-0.5, +0.5, +0.5);
    glEnd();

    glPopMatrix();
}

void FilterObject_cls::
GLDrawRealObject(int currentIndex, double doubleIndex, int lightNumber, 
		 bool bCopyInstance){
    
    FilterObjectData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return;

    GLAttribute_cls::mspCShaderCtrl->
	BeginShader(pCData->mCTexture.mCShaderType.mBody, lightNumber);

    glPushMatrix();

    ParamGroup_cls *pCTranslate = pCData->mpCTranslate;
    ParamGroup_cls *pCRotate    = pCData->mpCRotation;
    ParamGroup_cls *pCScale     = pCData->mpCScale;
    
    std::vector < double > VTrans = 
	pCTranslate->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    glTranslatef(VTrans[0], VTrans[1], VTrans[2]);

    std::vector < double > VRot = 
	pCRotate->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    
    float clrMaterial[3][4];
    pCData->mVpCClrCurve[0]->GetColor4fNormal(clrMaterial[0], 0);
    pCData->mVpCClrCurve[1]->GetColor4fNormal(clrMaterial[1], 0);
    pCData->mVpCClrCurve[2]->GetColor4fNormal(clrMaterial[2], 0);

    pCData->mCTexture.GetAlphaInfo(clrMaterial);
    glEnable(GL_BLEND);

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, clrMaterial[0]);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, clrMaterial[1]);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, clrMaterial[2]);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 20);
    
    glRotatef(VRot[2], 0, 0, 1);
    glRotatef(VRot[0], 1, 0, 0);
    glRotatef(VRot[1], 0, 1, 0);

    std::vector < double > VScale = 
	pCScale->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);

    glScalef(VScale[0], VScale[1], VScale[2]);
    
    glEnable(GL_NORMALIZE);
    switch(pCData->mCObjectType.mBody){
    case _FILTER_OBJECT_BOX:
	if (bCopyInstance == TRUE) break;
	glScalef(0.5, 0.5, 0.5);
	Cube_cls::Draw();
	break;
    case _FILTER_OBJECT_PLANE:
	if (bCopyInstance == TRUE) break;
	glScalef(0.5, 0.5, 0.5);
	glBegin(GL_QUADS);
	glNormal3d(0, 0, 1); glVertex3d(-1, -1, 0); 
	glNormal3d(0, 0, 1); glVertex3d(+1, -1, 0); 
	glNormal3d(0, 0, 1); glVertex3d(+1, +1, 0); 
	glNormal3d(0, 0, 1); glVertex3d(-1, +1, 0); 
	glEnd();
	break;
    case _FILTER_OBJECT_SPHERE:
	if (bCopyInstance == TRUE) break;
	glScalef(0.5, 0.5, 0.5);
	GpCSphere->Draw(3);
	break;
    case _FILTER_OBJECT_IMPORT:
	if (bCopyInstance == FALSE) break;
	pCData->DrawMesh(currentIndex);
	break;
    }

    glPopMatrix();

    GLAttribute_cls::mspCShaderCtrl->
	EndShader(pCData->mCTexture.mCShaderType.mBody);
}

void FilterObject_cls::
OnSelected(){
    FilterObjectData_cls *pCData = GetPCData();
    mpCParentCtrl->mspCParentMainLink->ConnectTextureToDialog(&pCData->mCTexture, pCData->mCstrCustomName.mstrBody);
}

wxString FilterObject_cls::
GetStrPovInclude(){
    FilterObjectData_cls *pCData = GetPCData();
    
    wxString strInclude;
    if (pCData->mCObjectType.mBody == _FILTER_OBJECT_IMPORT)
	strInclude << pCData->mCMeshObject.GetStrPovInclude();
    return strInclude;
}

wxString FilterObject_cls::
GetStrPovRealObject(double doubleIndex, wxString strTranslate, 
		    bool bCopyInstance, wxString strTab){
    wxString strObject;

    FilterObjectData_cls *pCData = GetPCData();
    if (pCData->mCbValid.mbBody == FALSE) return strObject;

    switch(pCData->mCObjectType.mBody){
    case _FILTER_OBJECT_BOX:
	if (bCopyInstance == TRUE) return strObject;
	strObject << strTab << wxT("box { <-0.5, -0.5, -0.5>, <0.5, 0.5, 0.5>") << ENDL;
	break;
    case _FILTER_OBJECT_PLANE:
	if (bCopyInstance == TRUE) return strObject;
	strObject << strTab << wxT("polygon { 4, <-0.5, -0.5, 0>, <0.5, -0.5, 0>, <0.5, 0.5, 0>, <-0.5, 0.5, 0>") << ENDL;
	break;
    case _FILTER_OBJECT_SPHERE:
	if (bCopyInstance == TRUE) return strObject;
	strObject << strTab << wxT("sphere { <0, 0, 0>, 0.5") << ENDL;
	break;
    case _FILTER_OBJECT_IMPORT:
	if (bCopyInstance == FALSE) return strObject;
	strObject << strObject << wxT("object {") << ENDL
		  << wxT("\t") << pCData->mCMeshObject.mstrPovLabel << ENDL;
	strObject << pCData->GetStrPovMatrix() << ENDL;
    }

    float clr[3][4];
    pCData->mVpCClrCurve[0]->GetColor4fNormal(clr[0], 0);
    pCData->mVpCClrCurve[1]->GetColor4fNormal(clr[1], 0);
    pCData->mVpCClrCurve[2]->GetColor4fNormal(clr[2], 0);
    
    pCData->mCTexture.GetAlphaInfo(clr);

    wxString strFinishOptions = 
	pCData->mCTexture.GetStrPovFinishOptions();
    
    strObject << PovRayCtrl_cls::GetStrPovTextureColor(strTab + wxT("\t"), clr, 0.156*0.156, strFinishOptions);

    wxString strInteriorOptions = 
	pCData->mCTexture.GetStrPovInteriorOptions();
    if (strInteriorOptions.IsEmpty() == FALSE)
	strObject << strTab << wxT("\tinterior {") << strInteriorOptions << wxT("}") << ENDL;
    

    strObject << pCData->mCTexture.GetStrPovNormal(strTab);
    
    double alpha = 1;
    double reflectMin  = pCData->mCTexture.mCReflectMin.mBody; // ˗
    double reflectMax  = pCData->mCTexture.mCReflectMax.mBody;
    bool   bKillTarget = FALSE; // ̂̂targetɂȂ
    strObject << PovRayCtrl_cls::GetStrPovPhotonSetting(alpha, reflectMin, reflectMax, bKillTarget);
    
    strObject << strTab;
    pCData->GetStrPovModelView(strObject, doubleIndex);
    if (strTranslate.IsEmpty() != TRUE)
	strObject << strTab << wxT("\t") << strTranslate << ENDL;
    strObject << strTab << wxT("}") << ENDL;
    
    return strObject;
}
