// -*- c -*- 
// "clsColorCurve.h"

#include "clsColorCurve.h"
#include "clsColorCurveGUI.h"

#include "../WIDGETCTRL/EX/clsColourDialogEX.h"
extern wxColourDialogEX* GpCClrDialog;

#include "../UNDO/clsUndoItemColor.h"
extern UndoManager_cls GCUndoList;

#include "../../wxWidgetEnumMenu.h"

#include <wx/wx.h>
#include <wx/artprov.h>
#include <wx/dcbuffer.h>
#include <wx/colordlg.h>
#include <wx/clipbrd.h>

#include "../COMMON/CommonDefinitions.h"

#include <sstream>

#define R 0
#define G 1
#define B 2
#define A 3

#define ART(a) wxArtProvider::GetBitmap(wxT(a))

BEGIN_EVENT_TABLE(ColorCurveBtn_cls, wxPanel)    
EVT_LEFT_DOWN(ColorCurveBtn_cls::OnMouse)
EVT_RIGHT_DOWN(ColorCurveBtn_cls::OnBtnRight)
EVT_PAINT    (ColorCurveBtn_cls::OnPaint)
EVT_MENU     (menuCopyColor, ColorCurveBtn_cls::OnCopyColor)
EVT_MENU     (menuPasteColor, ColorCurveBtn_cls::OnPasteColor)
END_EVENT_TABLE()

ColorCurveBtn_cls::
ColorCurveBtn_cls(wxWindow *pCParent, ColorCurve_cls *pCColorLinked,
		  bool bUseButton, bool bUseCurve, bool bWorkAsButton) : 
wxPanel(pCParent, wxID_ANY, wxDefaultPosition, wxSize(-1,20)) {
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    mpCColorLinked = pCColorLinked;
    mbUseButton = bUseButton;
    mbUseCurve = bUseCurve;
    mbWorkAsButton = bWorkAsButton;
    mbRepeat = FALSE;
    mpPopupMenu = new wxMenu;
    mpPopupMenu->Append(menuCopyColor, _("Copy Color"));
    mpPopupMenu->Append(menuPasteColor, _("Paste Color"));
}

ColorCurveBtn_cls::
~ColorCurveBtn_cls(){
    delete mpPopupMenu;
}

void ColorCurveBtn_cls::
LinkColorCurve(ColorCurve_cls *pCColorLinked){
    mpCColorLinked = pCColorLinked;
}

void ColorCurveBtn_cls::
OnBtnRight(wxMouseEvent& event){
    PopupMenu(mpPopupMenu);
}

void ColorCurveBtn_cls::
OnCopyColor(wxCommandEvent& event){
    if (mpCColorLinked == NULL) return;
    wxString strSerialize;

    if (mpCColorLinked->mCbColorCurve.mbBody == TRUE){
	strSerialize = mpCColorLinked->mpCClrCurve->Serialize();
    }
    else{
	wxString strColor = mpCColorLinked->mCConstantColor.Serialize();
	strColor.Replace(mpCColorLinked->mCConstantColor.mstrHeader, wxT(""));
	wxString strTag = wxT("COLOR=");
	strSerialize << strTag << strColor;
    }

    wxTextDataObject *pCData = new wxTextDataObject(strSerialize);
    wxTheClipboard->Open();
    wxTheClipboard->SetData(pCData);
    wxTheClipboard->Close();
}

void ColorCurveBtn_cls::
OnPasteColor(wxCommandEvent& event){
    if (mpCColorLinked == NULL) return;
    if (wxTheClipboard->IsSupported(wxDataFormat(wxDF_TEXT)) == FALSE) return;
    wxTextDataObject CData;
    wxTheClipboard->Open();
    wxTheClipboard->GetData(CData);
    wxTheClipboard->Close();
    wxString strData = CData.GetText();

    if (strData.Matches(wxT("COLOR=*")) == TRUE){
	ConfItemFLOATARRAY_cls CArray(wxT("COLOR="), wxT("0.5 0.5 0.5 1"));
	CArray.ReadStrLine(strData);
	
	GCUndoList.PushNewCommand
	    (new CommandSetColor_cls(&mpCColorLinked->mCConstantColor, CArray.mpBody,
				     _COMMAND_UPDATE_DISPLAY_ATTRIB | _COMMAND_UPDATE_DISPLAY_GL),
	     new CommandSetColor_cls(&mpCColorLinked->mCConstantColor, mpCColorLinked->mCConstantColor.mpBody, 
				     _COMMAND_UPDATE_DISPLAY_ATTRIB | _COMMAND_UPDATE_DISPLAY_GL));
	GCUndoList.ExecCurrent();
    }
    if (strData.Matches(wxT("-ParamGroup=*"))){
	wxString listTag[4] = {wxT("R"), wxT("G"), wxT("B"), wxT("A")};
	int updateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_COLOR | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM | _COMMAND_UPDATE_DISPLAY_ATTRIB | _COMMAND_UPDATE_SNAPSHOT_DATA;
	
	ParamGroup_cls *pCOrg = mpCColorLinked->mpCClrCurve;
	ParamGroup_cls CParam(pCOrg->mstrGroupLabel,
			      pCOrg->mCDimension.mBody,
			      listTag, updateBit, TRUE);
	
	Parameters_cls::msbUndoCommandMake = FALSE;
	CParam.CopyFromP(pCOrg);
	Parameters_cls::msbUndoCommandMake = TRUE;
	
	std::wstring wstrLine;
	std::wistringstream istream((wchar_t*)strData.c_str());
	std::getline(istream, wstrLine);
	wstrLine = wxT("-ParamGroup=ANY");
	CParam.ReadStrLine(istream, wstrLine.c_str());

	pCOrg->CopyFromP(&CParam);
    }
	
    Refresh(FALSE);
}


void ColorCurveBtn_cls::
OnMouse(wxMouseEvent &event){
    if (mbWorkAsButton == TRUE){
	SendButtonEvent(GetParent(), GetId());
	return;
    }

    if (mpCColorLinked == NULL || mbUseButton == FALSE) return;
    int x = event.GetX();
    if (x < 20 && mbUseCurve){
	mpCColorLinked->mCbColorCurve.mbBody = !mpCColorLinked->mCbColorCurve.mbBody;
	Refresh(FALSE);

	SendMouseEvent(GetParent(), event);
	GetParent()->AddPendingEvent(event);
	if (mpCColorLinked->mCbColorCurve.mbBody == FALSE) return;
    }

    if (mpCColorLinked->mCbColorCurve.mbBody == TRUE){
	mspCColorCurveCtrlLinked->LinkColorCurve(mpCColorLinked);
	SendMenuEvent(wxGetTopLevelParent(this), menuDialogColorCurve);
	return; 
    }

    GpCClrDialog->SetColourFloat(mpCColorLinked->mCConstantColor.mpBody);
    if (GpCClrDialog->ShowModal() == wxID_OK){
	float newclr[4] = {0, 0, 0, 1};
	LOOP_RGB newclr[rgb] = GpCClrDialog->GetPColorFloat()[rgb];
	GCUndoList.PushNewCommand
	    (new CommandSetColor_cls(&mpCColorLinked->mCConstantColor, newclr,
				     _COMMAND_UPDATE_DISPLAY_ATTRIB | _COMMAND_UPDATE_DISPLAY_GL),
	     new CommandSetColor_cls(&mpCColorLinked->mCConstantColor, mpCColorLinked->mCConstantColor.mpBody, 
				     _COMMAND_UPDATE_DISPLAY_ATTRIB | _COMMAND_UPDATE_DISPLAY_GL));
	GCUndoList.ExecCurrent();
    }
    
    GetParent()->AddPendingEvent(event);
}

void ColorCurveBtn_cls::
OnPaint(wxPaintEvent &event){
    if (mpCColorLinked != NULL){
	wxAutoBufferedPaintDC dc(this);
	mpCColorLinked->SetXViewRange(mViewMinX, mViewMaxX);
	mpCColorLinked->DrawColorOnPaintDC(dc, this, mbUseButton & mbUseCurve);
	mpCColorLinked->SetXViewRange(0, 1);
    }
    event.Skip();
}

ColorCurveCtrl_cls *ColorCurveBtn_cls::mspCColorCurveCtrlLinked = NULL;

// -----

ColorCurve_cls::
ColorCurve_cls(wxString strTag, wxString strLabel, wxString strInit, int keynum, double (*comp)[5]){
    mstrTagBase = strTag;
    mstrLabel = strLabel;

    mCConfList.Add(&mCbColorCurve, strTag + wxT("UseColorCurve="), wxT("0"));
    mCConfList.Add(&mCConstantColor, strTag + wxT("ConstantColor="), strInit);
    mCConfList.Add(&mCScalingMin, strTag + wxT("ScalingMin="), wxT("0"));
    mCConfList.Add(&mCScalingMax, strTag + wxT("ScalingMax="), wxT("1"));
    mCConfList.Add(&mCWrapType, strTag + wxT("WrapType="), wxT("0"));
    mCConfList.Add(&mCDependType, strTag + wxT("DependType="), wxT("0"));
    
    wxString listTag[4] = {wxT("R"), wxT("G"), wxT("B"), wxT("A")};
    wxString strClrTag = strTag;
    strClrTag << wxT(" ColorCurve");
    int updateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_COLOR | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM | _COMMAND_UPDATE_DISPLAY_ATTRIB | _COMMAND_UPDATE_SNAPSHOT_DATA;
    mpCClrCurve = new ParamGroup_cls(strClrTag, 4, listTag, updateBit, TRUE);
    
    if (keynum == 0){
	double pClrInit[4] = {0, 1, 0, 1};
	mpCClrCurve->SetKeyFrame(pClrInit, TRUE);
	double pClrFin[4] = {1, 0, 0, 1};
	mpCClrCurve->SetKeyFrame(pClrFin, TRUE, 1);
    }
    else
	for (int i = 0; i < keynum; i++)
	    mpCClrCurve->SetKeyFrame(&comp[i][1], TRUE, comp[i][0]);
}

ColorCurve_cls::
~ColorCurve_cls(){
    delete mpCClrCurve;
}

void ColorCurve_cls::
CopyColorFromP(ColorCurve_cls *pCClrCurve, bool bCurveOnly){
    mpCClrCurve->CopyFromP(pCClrCurve->mpCClrCurve);
    if (bCurveOnly == FALSE)
	mCConfList.CopyListComponentFromP(&(pCClrCurve->mCConfList));
}

void ColorCurve_cls::
Flip(){
    mpCClrCurve->Flip();
}

void ColorCurve_cls::
SwapRGB(){
    mpCClrCurve->Swap(3);
}

void ColorCurve_cls::
PreCalcScaleFactor(){
    mMin = mCScalingMin.mBody;
    mScale = mCScalingMax.mBody - mCScalingMin.mBody;
    if (mScale == 0){
	mCScalingMax.mBody = mCScalingMin.mBody + 1;
	mScale = 1;
    }
}

unsigned char *ColorCurve_cls::
GetPNewClrMap(int size, int type){
    unsigned char *pMap = new unsigned char [size * 4];
    if (mCbColorCurve.mbBody == FALSE){
	for (int i = 0; i < size; i++){
	    LOOP_RGB pMap[i*4 + rgb] = mCConstantColor.mpBody[rgb] * 255;
	    pMap[i*4 + 3] = 255;
	}
    }
    else {
	if (type == 0){
	    for (int i = 0; i < size; i++){
		float clr[4];
		GetColor4fNormal(clr, (float)i/255);
		LOOP_RGBA pMap[i*4 + rgba] = clr[rgba] * 255;
	    }
	}
	if (type == 1){
	    for (int i = 0; i < size; i++){
		float clr[4];
		float value = (float)i/128 - 1;
		if (value < 0) value *= -1;
		GetColor4fNormal(clr, value);
		LOOP_RGBA pMap[i*4 + rgba] = clr[rgba] * 255;
	    }
	}
    }
    return pMap;
}

void ColorCurve_cls::
GetColor4fScaled(float *pColor, float value){
    if (mCbColorCurve.mbBody == TRUE){
	value = (value - mMin) / mScale;
	mpCClrCurve->GetInterpolatedValue4Fast(pColor, value, 0, (mCWrapType.mBody == 1));
    }
    else
	memcpy(pColor, mCConstantColor.mpBody, sizeof(float)*4);
}

void ColorCurve_cls::
GetColor4fNormal(float *pColor, float value){
    if (mCbColorCurve.mbBody == TRUE){
	mpCClrCurve->GetInterpolatedValue4Fast(pColor, value, 0, (mCWrapType.mBody == 1));
    }
    else
	memcpy(pColor, mCConstantColor.mpBody, sizeof(float)*4);
}

wxColour ColorCurve_cls::
GetWXColor(wxColour ClrBG, double valueX, bool bIgnoreAlpha){
    if (mCbColorCurve.mbBody == TRUE){
	std::vector < double > VClr = mpCClrCurve->GetInterpolatedValue(valueX, 0, (mCWrapType.mBody == 1));
	if (bIgnoreAlpha == TRUE)
	    return wxColour(VClr[R] * 255, VClr[G] * 255, VClr[B] * 255);
	
	return wxColour
	    (ClrBG.Red()   + (VClr[R] * 255 - ClrBG.Red()) * VClr[A],
	     ClrBG.Green() + (VClr[G] * 255 - ClrBG.Green()) * VClr[A],
	     ClrBG.Blue()  + (VClr[B] * 255 - ClrBG.Blue()) * VClr[A]);
    }


    return wxColour(mCConstantColor.mpBody[R] * 255,
		    mCConstantColor.mpBody[G] * 255,
		    mCConstantColor.mpBody[B] * 255);
}

wxString ColorCurve_cls::
GetStrPov(){
    return PovOut3f(wxT("rgb"), mCConstantColor.mpBody, 3);
}

void ColorCurve_cls::
DrawOnDC(wxDC &dc, bool bGraphButton, wxRect rectToDraw, int offset,
	 bool bIgnoreAlpha){
    int rgt = rectToDraw.GetRight();
    int btm = rectToDraw.GetBottom();
    if (mCbColorCurve.mbBody == TRUE){
	dc.SetPen(*wxBLACK_PEN);
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	dc.DrawRectangle(rectToDraw);
	dc.SetPen(*wxTRANSPARENT_PEN);
	for (int x = offset+1; x < rgt; x++){
	    wxColour ClrBG0, ClrBG1;
	    if (bIgnoreAlpha == TRUE){
		ClrBG0.Set(0, 0, 0);
		ClrBG1.Set(0, 0, 0);
	    }
	    else {if ((x % 20) > 10){
		    ClrBG0.Set(192, 192, 192);
		    ClrBG1.Set(64,  64,   64);
		}
		else{
		    ClrBG0.Set(64,  64,  64);
		    ClrBG1.Set(192, 192, 192);
		}
	    }
	    double valueX = GetMappedXInv((double)(x-offset)/(rgt-offset));

	    dc.SetBrush(wxBrush(GetWXColor(ClrBG0, valueX, bIgnoreAlpha)));
	    wxRect rectToPaint0(x, 1, 1, btm/2);
	    dc.DrawRectangle(rectToPaint0);
	    dc.SetBrush(wxBrush(GetWXColor(ClrBG1, valueX, bIgnoreAlpha)));
	    wxRect rectToPaint1(x, btm/2+1, 1, btm/2);
	    dc.DrawRectangle(rectToPaint1);

	} 
	if (bGraphButton == TRUE) dc.DrawBitmap(ART("wxART_GRAPH_P"), 0, 0);
    }
    else{
	dc.SetPen(*wxBLACK_PEN);
	dc.SetBrush(wxBrush(GetWXColor()));
	dc.DrawRectangle(rectToDraw);
#ifdef _DEBUG
	wxString strColor;
	strColor << mCConstantColor.mpBody[R] << wxT(":") 
		 << mCConstantColor.mpBody[G] << wxT(":") 
		 << mCConstantColor.mpBody[B] << wxT(":") 
		 << mCConstantColor.mpBody[A];
	dc.SetFont(*wxSMALL_FONT);
	dc.DrawText(strColor, 20, 5);
#endif
	if (bGraphButton == TRUE) dc.DrawBitmap(ART("wxART_GRAPH_N"), 0, 0);
    }
}

void ColorCurve_cls::
DrawColorOnPaintDC(wxAutoBufferedPaintDC &dc, wxWindow *pCWindow,
		   bool bGraphButton){
    int offset = 0;
    if (bGraphButton == TRUE) offset = 20;
    wxSize sz = pCWindow->GetClientSize();
    wxRect rectToDraw(offset, 0, sz.x-offset, sz.y);

    // For efficiency, do not draw if not exposed
    if (pCWindow->IsExposed(rectToDraw))
	DrawOnDC(dc, bGraphButton, rectToDraw, offset, FALSE);
}

void ColorCurve_cls::
DrawColorOnMemoryDC(wxMemoryDC &dc, bool bGraphButton, int height,
		    bool bIgnoreAlpha){
    int offset = 0;
    if (bGraphButton == TRUE) offset = 20;
    wxSize sz = dc.GetSize();
    if (height ==0) height = sz.y;
    wxRect rectToDraw(offset, 0, sz.x-offset, height);

    DrawOnDC(dc, bGraphButton, rectToDraw, offset, bIgnoreAlpha);
}

void ColorCurve_cls::
GetStrPovColorMap(wxString &strMap, wxString strTab){
    strMap << strTab << wxT("color_map{") << ENDL;
    int keyNum[4], index[4] = {0, 0, 0, 0};
    LOOP_RGBA keyNum[rgba] = mpCClrCurve->mVCParam[rgba].GetKeyNumber();

    double doubleIndex[4];
    double doubleIndexWritten = -1;
    while (TRUE){
	int    finishedChannel = 0;
	int    minIndex = -1;
	double doubleIndexMin = 99;
	LOOP_RGBA {
	    if (index[rgba] < keyNum[rgba])
		doubleIndex[rgba] = mpCClrCurve->mVCParam[rgba].GetKeyFrameDoubleIndex(index[rgba]);
	    else{
		doubleIndex[rgba] = 99;
		finishedChannel++;
	    }
	}
	if (finishedChannel == 4) break;
	
	LOOP_RGBA if (doubleIndex[rgba] < doubleIndexMin){
	    doubleIndexMin = doubleIndex[rgba];
	    minIndex = rgba;
	}

	if (doubleIndexMin != doubleIndexWritten){
	    float clr[4];
	    GetColor4fNormal(clr, doubleIndexMin);

	    strMap << strTab << wxT("\t[") << doubleIndexMin << wxT(" ")
		   << PovOut4fRGBTfromRGBA(clr, 3) << wxT("]") << ENDL;

	    doubleIndexWritten = doubleIndexMin;
	}

	index[minIndex]++;
    }
    
    strMap << strTab << wxT("}") << ENDL;
}

// ---------------------------------------

wxString ColorCurve_cls::
Serialize(){
    wxString strOut;
    strOut << mCConfList.SerializeAll();
    strOut << mpCClrCurve->Serialize();
    return strOut;
}

void ColorCurve_cls::
ReadStrLine(std::wistream &istream, wxString strLine){
    mCConfList.ReadStrLineAll(strLine);
    mpCClrCurve->ReadStrLine(istream, strLine);
}
