// -*- c -*- 
// "clsColorCurve.cpp"

#ifndef _COLOR_CURVE_H_
#define _COLOR_CURVE_H_

#include "../CONFITEM/clsConfItem.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../PARAMETER/clsParamDrawBase.h"
#include "../POVRAY/clsPovSerialize.h"

#include "../WIDGETCTRL/clsWidgetCommon.h"

class ColorCurve_cls;
class ColorCurveCtrl_cls;

#define _DEPEND_ON_VALUE1 0
#define _DEPEND_ON_RADIUS 1
#define _DEPEND_ON_POSITION 2
#define _DEPEND_ON_POS_X 3
#define _DEPEND_ON_POS_Y 4
#define _DEPEND_ON_POS_Z 5
#define _DEPEND_ON_VELOCITY 6
#define _DEPEND_ON_V_X   7
#define _DEPEND_ON_V_Y   8
#define _DEPEND_ON_V_Z   9
#define _DEPEND_ON_ID    10

class ColorCurveBtn_cls : public wxPanel, public ParamDrawBase_cls, public WidgetCommon_cls {
 public:
    ColorCurveBtn_cls(wxWindow *pCParent, ColorCurve_cls *pCColorLinked,
		      bool bUseButton = TRUE, bool bUseCurve = TRUE,
		      bool bWorkAsButton = FALSE);
    ~ColorCurveBtn_cls();
    void LinkColorCurve(ColorCurve_cls *pCColorLinked);
    
    ColorCurve_cls *mpCColorLinked;
    static ColorCurveCtrl_cls *mspCColorCurveCtrlLinked;
    bool mbUseButton;
    bool mbUseCurve;
    bool mbWorkAsButton;
    bool mbRepeat;
    int mIndex;

    wxMenu *mpPopupMenu;

    void OnBtnRight(wxMouseEvent& event);    
    void OnCopyColor(wxCommandEvent& event);
    void OnPasteColor(wxCommandEvent& event);
    void OnMouse(wxMouseEvent& event);
    void OnPaint(wxPaintEvent& event);


 private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

class ColorCurve_cls : public PovSerialize_cls, public ParamDrawBase_cls {
 public:
    ColorCurve_cls(wxString strTagBase, wxString strLabel,
		   wxString strInit = wxT("1 0.8 0.2 1"),
		   int keynum = 0, double (*comp)[5] = NULL);
    ~ColorCurve_cls();

    wxString mstrTagBase;
    wxString mstrLabel;
    
    ConfItemList_cls mCConfList;
    ConfItemBOOL_cls mCbColorCurve;
    ConfItemFLOATARRAY_cls mCConstantColor; 
    ConfItemFLOAT_cls mCScalingMin;
    ConfItemFLOAT_cls mCScalingMax;
    ConfItemUINT_cls mCWrapType;
    ConfItemUINT_cls mCDependType;
    float mMin, mScale;
    
    ParamGroup_cls *mpCClrCurve;
    
    void CopyColorFromP(ColorCurve_cls *pCClrCurve, bool bCurveOnly = FALSE);
    void Flip();
    void SwapRGB();
    void PreCalcScaleFactor();

    unsigned char *GetPNewClrMap(int size, int type);
    void GetColor4fScaled(float *pColor, float value);
    void GetColor4fNormal(float *pColor, float value);
    wxColour GetWXColor(wxColour clrBG = wxNullColour, double valueX = 0,
			bool bIgnoreAlpha = FALSE);
    wxString GetStrPov();
    
    void DrawOnDC(wxDC &dc, bool bGraphButton, wxRect rectToDraw, int offset,
		  bool bIgnoreAlpha);
    void DrawColorOnPaintDC(wxAutoBufferedPaintDC &dc, wxWindow *pCWindow,
			    bool bGraphButton = TRUE);
    void DrawColorOnMemoryDC(wxMemoryDC &dc, bool bGraphButton = TRUE,
			     int height = 0, bool bIgnoreAlpha = FALSE);

    void GetStrPovColorMap(wxString &strMap, wxString strTab);
    
    wxString Serialize();
    void ReadStrLine(std::wistream &istream, wxString strLine);
};

#endif
