
#include "clsGLAttribute.h"
#include "clsGLAttributeGUI.h"
#include "clsShaderCtrl.h"

#include "../GLATTRIB/clsColorCurve.h"
#include "../PARTICLEDATA/clsParticleDataSequence.h"

#include "../MYAPI/MsgBox.h"

#include "../../GL/glsl.h"

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
extern PFNGLPROGRAMPARAMETERIEXTPROC glProgramParameteriEXT;
#elif defined __WXMAC__
#define GL_GLEXT_PROTOTYPES 
#include <OpenGL/glu.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include "../COMMON/CommonDefinitions.h"

GLAttribute_cls::
GLAttribute_cls(PDataSequence_cls *pCParentSequence, 
		int selfIndex, DrawOption_cls *pCDrawOptionLink) : 
    mpCGUI(NULL)
{
    mpCParentSequence = pCParentSequence;
    mpBufferRGBA = NULL;
    mBufferWidth = 0; mBufferHeight = 0;
    mSelfIndex = selfIndex;
    //mShaderCartoonTexID = 0;
    mpCDrawOptionLink = pCDrawOptionLink;

    wxString strAttribTag;
    strAttribTag << wxT("SEQ") << mpCParentSequence->mSelfIndex
		 << wxT("TYPE") << mSelfIndex;
    wxString strLabel;
    strLabel << wxT("Attrib ") << mSelfIndex;

    mpCClrCurveCol = new ColorCurve_cls(strAttribTag + wxT("Col"),
					strLabel + wxT(" : Color"));
    mpCClrCurveMat[_GL_AMBIENT]  = new ColorCurve_cls
	(strAttribTag + wxT("Ambient"), strLabel + wxT(" : Ambient"), 
	 wxT("0.2 0.16 0.04 1"));
    mpCClrCurveMat[_GL_DIFFUSE]  = new ColorCurve_cls
	(strAttribTag + wxT("Diffuse"), strLabel + wxT(" : Diffuse"));
    mpCClrCurveMat[_GL_SPECULAR] = new ColorCurve_cls
	(strAttribTag + wxT("Specular"), strLabel + wxT(" : Specular"));

    mCConfList.Add(&mCDrawType, strAttribTag + wxT("DrawType="), wxT("0"));
    mCConfList.Add(&mCDrawQuality, strAttribTag + wxT("DrawQuality="), wxT("0"));
    mCConfList.Add(&mCBlendType, strAttribTag + wxT("BlendType="), wxT("0"));
    mCConfList.Add(&mCRadiusType, strAttribTag + wxT("RadiusType="), wxT("0"));
    mCConfList.Add(&mCRadiusFactor, strAttribTag + wxT("RadiusFactor="), wxT("1"));
    mCConfList.Add(&mCRadiusFactorForConst, strAttribTag + wxT("RadiusFactorForConst="), wxT("0.001"));
    mCConfList.Add(&mCPointSize, strAttribTag + wxT("PointSize="), wxT("10"));
    mCConfList.Add(&mCAlphaType, strAttribTag + wxT("AlphaType="), wxT("0"));
    mCConfList.Add(&mCShaderType, strAttribTag + wxT("ShaderType="), wxT("0"));
    mCConfList.Add(&mCAlphaFactor, strAttribTag + wxT("AlphaFactro="), wxT("1"));
    mCConfList.Add(&mCAlphaDenendBase, strAttribTag + wxT("AlphaDependBase="), wxT("1"));
    mCConfList.Add(&mCAlphaMax, strAttribTag + wxT("AlphaMax="), wxT("1"));
    //mCConfList.Add(&mCCartoonBoldness, strAttribTag + wxT("CartoonBoldness="), wxT("0.25"));
    
    mCConfList.Add(&mCTexFilePathBillboard,
		   strAttribTag + wxT("TexFileBillboard="), 
		   wxT("Texture/ParticleA.png"));

    mTextureBillboard = 0;
}

GLAttribute_cls::
~GLAttribute_cls(){
    if (mpCGUI != NULL) delete mpCGUI;
    if (mspCShaderCtrl != NULL){
	delete mspCShaderCtrl;
	mspCShaderCtrl = NULL;
    }
    delete mpCClrCurveCol;
    delete mpCClrCurveMat[_GL_AMBIENT];
    delete mpCClrCurveMat[_GL_DIFFUSE];
    delete mpCClrCurveMat[_GL_SPECULAR];
}

void GLAttribute_cls::
InitializeShader(wxWindow *pCParent, bool bShaderAvailable){
    if (mspCShaderCtrl == NULL){
	mspCShaderCtrl = new ShaderCtrl_cls();
	mspCShaderCtrl->InitializeShader(pCParent, bShaderAvailable);
    }
}

void GLAttribute_cls::
BeginShader(int lightNumber){
    mspCShaderCtrl->BeginShader(mCShaderType.mBody, lightNumber);
}

void GLAttribute_cls::
EndShader(){
    mspCShaderCtrl->EndShader(mCShaderType.mBody);
}

void GLAttribute_cls::
InstantiateGUI(wxWindow *pCParent){
    mpCGUI = new GLAttributeGUI_cls(this, pCParent);
}

void GLAttribute_cls::
RefreshTex(){
    if (mTextureBillboard != 0){
	glDeleteTextures(1, &mTextureBillboard);
    }
    
    glGenTextures(1, &mTextureBillboard);
    mCTexImgBillboard.LoadFile(mCTexFilePathBillboard.mstrBody,
			       wxBITMAP_TYPE_PNG);

    glBindTexture(GL_TEXTURE_2D, mTextureBillboard);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    
    GenerateBufferRGBA(&mCTexImgBillboard);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
		 mBufferWidth, mBufferHeight,
		 0, GL_RGBA, GL_UNSIGNED_BYTE, mpBufferRGBA);
    DeleteBufferRGBA();
}

void GLAttribute_cls::
GenerateBufferRGBA(wxImage *pCImage){
    DeleteBufferRGBA(); 
    mBufferWidth  = pCImage->GetWidth();
    mBufferHeight = pCImage->GetHeight();
    unsigned char *pData  = mCTexImgBillboard.GetData();
    unsigned char *pAlpha = mCTexImgBillboard.GetAlpha();
    mpBufferRGBA = new unsigned char [mBufferWidth * mBufferHeight * 4];
    for (int y = 0; y < mBufferHeight; y++){
	int indexy = mBufferWidth * y;
	for (int x = 0; x <  mBufferWidth; x++){
	    int index = indexy + x;
	    mpBufferRGBA[index * 4 + 0] = pData[index * 3 + 0];
	    mpBufferRGBA[index * 4 + 1] = pData[index * 3 + 1];
	    mpBufferRGBA[index * 4 + 2] = pData[index * 3 + 2];
	    mpBufferRGBA[index * 4 + 3] = pAlpha[index];
	}
    }
}

void GLAttribute_cls::
DeleteBufferRGBA(){
    if (mpBufferRGBA != NULL) delete[] mpBufferRGBA;
    mpBufferRGBA = NULL;
    mBufferWidth = 0;
    mBufferHeight = 0;
}

float GLAttribute_cls::
GetRadiusFactor(){
    if (mCRadiusType.mBody == _GL_RADIUS_CONSTANT)
	return mCRadiusFactorForConst.mBody;
    return mCRadiusFactor.mBody;
}

// ----

wxString GLAttribute_cls::
Serialize(){
    wxString strOut;
    for (int i = 0; i < 3; i++)
	strOut << mpCClrCurveMat[i]->Serialize();
    strOut << mpCClrCurveCol->Serialize();
    strOut << mCConfList.SerializeAll();
    return strOut;
}

void GLAttribute_cls::
ReadStrLine(std::wistream &istream, wxString strLine){
    for (int i = 0; i < 3; i++)
	mpCClrCurveMat[i]->ReadStrLine(istream, strLine);
    mpCClrCurveCol->ReadStrLine(istream, strLine);
    mCConfList.ReadStrLineAll(strLine);
}

ShaderCtrl_cls* GLAttribute_cls::mspCShaderCtrl = NULL;

/*
GLuint GLAttribute_cls::msGeomShader;
GLuint GLAttribute_cls::msVertShaderGouraud;
GLuint GLAttribute_cls::msFragShaderGouraud;
GLuint GLAttribute_cls::msGL2ProgramGouraud;
GLuint GLAttribute_cls::msVertShaderPhong;
GLuint GLAttribute_cls::msFragShaderPhong;
GLuint GLAttribute_cls::msGL2ProgramPhong;
GLuint GLAttribute_cls::msVertShaderCartoon;
GLuint GLAttribute_cls::msFragShaderCartoon;
GLuint GLAttribute_cls::msGL2ProgramCartoon;
GLint  GLAttribute_cls::mParamIDLNumGouraud;
GLint  GLAttribute_cls::mParamIDLNumPhong;
GLint  GLAttribute_cls::mParamIDLNumCartoon;
GLint  GLAttribute_cls::mParamIDShadowModeCartoon;
//GLint  GLAttribute_cls::mParamIDTextureCartoon
GLint  GLAttribute_cls::mParamIDBoldnessCartoon;

ConfItemFLOAT_cls GLAttribute_cls::msCCartoonBoldness;
*/
