// -*- c++ -*-
// "clsGLAttributeGUI.h"

#include "clsColorCurve.h"
#include "clsGLAttributeGUI.h"
#include "clsGLAttribute.h"

#include "../MAINCANVAS/clsTopHalfPanel.h"
#include "../PARTICLEDATA/clsSnapData.h"
extern bool GbSnapDataRefresh;
#include "../UNDO/clsUndoManager.h"
#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsHookCtrl.h"

#include "../../wxWidgetEnumMenu.h"

#include "../COMMON/CommonDefinitions.h"

GLAttributeGUI_cls::
GLAttributeGUI_cls(GLAttribute_cls *pCAttribLink, wxWindow *pCParent){
    mpCAttribLink = pCAttribLink;
    mpCParent = pCParent;
    mpCMInputPanel = NULL;
    mTypePrev = -1;
    mAlphaPrev = -1;
    mRadiusTypePrev = -1;
}

void GLAttributeGUI_cls::
SetPanel(){
    mpCMInputPanel = new MultiInputPanel_cls(&mpCMInputPanel, mpCParent);
    
    MultiInputPanel_cls* pC = mpCMInputPanel;

#ifdef __WXMAC__
    pC->mTextWidth = 90;
#else
    pC->mTextWidth = 80;
#endif
    pC->mItemWidth = -1;
    pC->mUpdateBit = 
	_COMMAND_UPDATE_DISPLAY_ATTRIB |
	_COMMAND_UPDATE_DISPLAY_GL ;
    
    int boxWidth = 2;

    pC->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Type/Quality"));
    wxString listType[5] = {_("Point"), _("Billboard"), _("Sphere"),
			    _("Both"), _("Hide") };
    pC->CombinePointerList
	(_ITEM_UINT_LIST, &mpCAttribLink->mCDrawType.mBody, 5, listType);
    
    wxString listQuality[3] = {_("Low"), _("Middle"), _("High")};
    pC->CombinePointerList
	(_ITEM_UINT_LIST, &mpCAttribLink->mCDrawQuality.mBody, 3, listQuality);

    pC->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Modes"));
    wxString listBlend[2] = {_("AlphaBlend"), _("Additive")};
    pC->CombinePointerList
	(_ITEM_UINT_LIST, &mpCAttribLink->mCBlendType.mBody, 
	 2, listBlend);
    wxString strListShader[4] = {_("original"), _("Gouroud"), _("Phong"), _("Toon")};
    mIndexShaderType = pC->CombinePointerList
	(_ITEM_UINT_LIST, &mpCAttribLink->mCShaderType.mBody,
	 4, strListShader);
        
    wxString listRadisu[2] = {_("Use Data"), _("Constant (unity)")};
    mIndexRadiusType = pC->CombinePointerList
	(_ITEM_UINT_LIST, &mpCAttribLink->mCRadiusType.mBody,
	 2, listRadisu, _("Radius"));

    mIndexRadius = pC->CombinePointer
	(_ITEM_FLOAT, &mpCAttribLink->mCRadiusFactor.mBody, _("Radius Factor"));

    mIndexRadiusForConst = pC->CombinePointer
	(_ITEM_FLOAT, &mpCAttribLink->mCRadiusFactorForConst.mBody, _("Radius Factor"));
    
    pC->mItemWidth = 90;    
    boxWidth = 2;
    mIndexPointSize = pC->SetInputAndSlider
	(_ITEM_INTEGER_SPIN, &mpCAttribLink->mCPointSize.mBody, _("Point Size"));
    pC->SetRange(_ITEM_LASTITEM2, 1, 50);
    pC->SetRange(_ITEM_LASTITEM, 1, 50);
    
    pC->mItemWidth = -1;
    wxString listAlpha[2] = {_("Constant (unity)"), _("Depend on R (power -2)")};

    pC->CombinePointerList
	(_ITEM_UINT_LIST, &mpCAttribLink->mCAlphaType.mBody,
	 2, listAlpha, _("Alpha"));

    pC->mItemWidth = 90;
    mIndexAlphaFactor = pC->SetInputAndSlider
	(_ITEM_FLOAT, &mpCAttribLink->mCAlphaFactor.mBody, _("Alpha Factor"));
    pC->SetRange(_ITEM_LASTITEM, 0, 100, 100);

    mIndexAlphaBase = pC->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Base/Max"));
    pC->CombinePointer
	(_ITEM_FLOAT, &mpCAttribLink->mCAlphaDenendBase.mBody);
    pC->CombinePointer
	(_ITEM_FLOAT, &mpCAttribLink->mCAlphaMax.mBody);
    
    ColorCurveBtn_cls *pCBtnCol = new ColorCurveBtn_cls
	(pC, mpCAttribLink->mpCClrCurveCol);

    ColorCurveBtn_cls *pCBtnAmient = new ColorCurveBtn_cls
	(pC, mpCAttribLink->mpCClrCurveMat[_GL_AMBIENT]);
    ColorCurveBtn_cls *pCBtnDiffuse = new ColorCurveBtn_cls
	(pC, mpCAttribLink->mpCClrCurveMat[_GL_DIFFUSE]);
    ColorCurveBtn_cls *pCBtnSpecular = new ColorCurveBtn_cls
	(pC, mpCAttribLink->mpCClrCurveMat[_GL_SPECULAR]);
    
    pC->mItemWidth = -1;
    mIndexClr = pC->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Color"));
    pC->CombinePointer(_ITEM_WXPANEL, pCBtnCol,  _("Color"));

    boxWidth = 3;
    mIndexMat = pC->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Material"));
    pC->CombinePointer(_ITEM_WXPANEL, pCBtnAmient,  wxT("Ambient"));
    pC->CombinePointer(_ITEM_WXPANEL, pCBtnDiffuse, wxT("Diffuse"));
    pC->CombinePointer(_ITEM_WXPANEL, pCBtnSpecular, wxT("Specular"));

    wxPanel *pCDummy = new wxPanel(pC, wxID_ANY, wxDefaultPosition, wxSize(-1, 4));
    pC->CombinePointer(_ITEM_WXPANEL, pCDummy);
    
    pC->FinishSetting(NULL);
    
    pC->SetFuncHook(FuncHookClick, _HOOK_PANELCLICK, -1);
    pC->SetFuncHook
	(TopHalfPanel_cls::FuncHookAttribHideAndShow, _HOOK_COMBOBOX, -1);

    HideAndShow();
}

void GLAttributeGUI_cls::
HideAndShow(){
    if (mTypePrev == mpCAttribLink->mCDrawType.mBody &&
	mAlphaPrev == mpCAttribLink->mCAlphaType.mBody &&
	mRadiusTypePrev == mpCAttribLink->mCRadiusType.mBody) return;

    mpCMInputPanel->Freeze();

    // wxNoteBook̃TCYLayoutƏlɂǂiƎvĵ
    // \͈ɂB
    if (mpCAttribLink->mCDrawType.mBody == _GL_DRAW_POINT){
	mpCMInputPanel->HideItem(mIndexRadius);
	mpCMInputPanel->HideItem(mIndexRadiusForConst);
	mpCMInputPanel->ShowItem(mIndexPointSize);
	mpCMInputPanel->EnableItem(mIndexRadiusType, FALSE);
    }
    else{
	mpCMInputPanel->HideItem(mIndexPointSize);
	if (mpCAttribLink->mCRadiusType.mBody == _GL_RADIUS_CONSTANT){
	    mpCMInputPanel->ShowItem(mIndexRadiusForConst);
	    mpCMInputPanel->HideItem(mIndexRadius);
	}
	else{
	    mpCMInputPanel->ShowItem(mIndexRadius);
	    mpCMInputPanel->HideItem(mIndexRadiusForConst);
	}
	mpCMInputPanel->EnableItem(mIndexRadiusType, TRUE);
    }
    
    if (mpCAttribLink->mCDrawType.mBody == _GL_DRAW_POINT ||
	mpCAttribLink->mCDrawType.mBody == _GL_DRAW_BILLBOARD){
	mpCMInputPanel->ShowItem(mIndexClr);
	mpCMInputPanel->HideItem(mIndexMat);
    }
    else{
	mpCMInputPanel->HideItem(mIndexClr);
	mpCMInputPanel->ShowItem(mIndexMat);
    }

    if (mpCAttribLink->mCAlphaType.mBody == _GL_ALPHA_INVERSE_SQUARE){
	mpCMInputPanel->HideItem(mIndexAlphaFactor);
	mpCMInputPanel->ShowItem(mIndexAlphaBase);
    }
    else{
	mpCMInputPanel->ShowItem(mIndexAlphaFactor);
	mpCMInputPanel->HideItem(mIndexAlphaBase);
    }

    mpCMInputPanel->Layout();

    mpCMInputPanel->Thaw();

    mTypePrev = mpCAttribLink->mCDrawType.mBody;
    mAlphaPrev = mpCAttribLink->mCAlphaType.mBody;
    mRadiusTypePrev = mpCAttribLink->mCRadiusType.mBody;
}

void GLAttributeGUI_cls::
FuncHookClick(void *pPanel, void *pArg){
    MultiInputPanel_cls *pCurr = (MultiInputPanel_cls*)pPanel;
    pCurr->UpdateValues();

    GbSnapDataRefresh = TRUE;
    
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshGL);
    wxGetTopLevelParent(pCurr)->AddPendingEvent(com);
}
