// -*- c -*- 
// "clsShaderCtrl.h"

#include "clsShaderCtrl.h"

#include "../../GL/glsl.h"
#include "../MYAPI/MsgBox.h"

#ifdef WIN32
PFNGLPROGRAMPARAMETERIEXTPROC glProgramParameteriEXT;
#endif

ShaderCtrl_cls::
ShaderCtrl_cls(){
    mCConfList.Clear();
    mCConfList.Add(&mCCartoonBoldness, wxT("CartoonBoldness="), wxT("0.3"));
}

void ShaderCtrl_cls::
InitializeShader(wxWindow *pCParent, bool bShaderAvailable){
    mbShaderAvailable = bShaderAvailable;
    if (mbShaderAvailable == FALSE) return;
    
#ifdef WIN32
    glProgramParameteriEXT = (PFNGLPROGRAMPARAMETERIEXTPROC)wglGetProcAddress("glProgramParameteriEXT");
#endif
    
    MakeShader(pCParent, mGL2ProgramGouraud, 
	       wxT("GL/Gouraud.vert"), wxT("GL/Gouraud.frag"));
    MakeShader(pCParent, mGL2ProgramPhong,
	       wxT("GL/Phong.vert"), wxT("GL/Phong.frag"));
    MakeShader(pCParent, mGL2ProgramCartoon,
	       wxT("GL/Cartoon.vert"), wxT("GL/Cartoon.frag"));
    
    SetShaderParamIDs();
    SetShaderParamCartoon(0, mCCartoonBoldness.mBody);

    glUseProgram(0);
}

void ShaderCtrl_cls::
BeginShader(int shaderType, int lightNumber){
    if (mbShaderAvailable == FALSE) return;

    glUseProgram(0);
    
    switch(shaderType){
    case 0:
	glUseProgram(0);
	break;
    case 1:
	glUseProgram(mGL2ProgramGouraud);
	SetShaderParamGouraud(lightNumber);
	break;
    case 2:{
	GLboolean bFog;
	int fogMode; 
	glGetBooleanv(GL_FOG, &bFog);
	glGetIntegerv(GL_FOG_MODE, &fogMode);
	if (bFog == GL_FALSE) fogMode = 0;
	else 
	    switch(fogMode){
	    case GL_EXP   : fogMode = 1; break;
	    case GL_EXP2  : fogMode = 2; break;
	    case GL_LINEAR: fogMode = 3; break;
	    }
	glUseProgram(mGL2ProgramPhong);
	SetShaderParamPhong(lightNumber, fogMode);
	break; }
    case 3: {
	glUseProgram(mGL2ProgramCartoon);
	SetShaderParamCartoon(lightNumber, mCCartoonBoldness.mBody);
	float ambient[4]; 
	glGetLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	if (lightNumber == 1 && ambient[0] == 0)
	    SetShaderParamCartoonShadowMode(1);
	else
	    SetShaderParamCartoonShadowMode(0);
	break; }
    }
}

void ShaderCtrl_cls::
EndShader(int shaderType){
    if (mbShaderAvailable == FALSE) return;
    glUseProgram(0);
}

void ShaderCtrl_cls::
SetShaderParamIDs(){
    glUseProgram(mGL2ProgramGouraud);
    mParamIDLNumGouraud = 
	glGetUniformLocation(mGL2ProgramGouraud, "lightNumberGouraud");

    glUseProgram(mGL2ProgramPhong);
    mParamIDLNumPhong   = 
	glGetUniformLocation(mGL2ProgramPhong, "lightNumberPhong");
    mParamIDFogModePhong = 
	glGetUniformLocation(mGL2ProgramPhong, "fogMode");
    
    glUseProgram(mGL2ProgramCartoon);
    mParamIDLNumCartoon = 
	glGetUniformLocation(mGL2ProgramCartoon, "lightNumberCartoon");
    mParamIDShadowModeCartoon = 
	glGetUniformLocation(mGL2ProgramCartoon, "shadowVolumeModeCartoon");
    mParamIDBoldnessCartoon = 
	glGetUniformLocation(mGL2ProgramCartoon, "boldnessCartoon");
}

void ShaderCtrl_cls::
SetShaderParamGouraud(int value){
    glUniform1i(mParamIDLNumGouraud, value);
}

void ShaderCtrl_cls::
SetShaderParamPhong(int LNum, int fogMode){
    glUniform1i(mParamIDLNumPhong, LNum);
    glUniform1i(mParamIDFogModePhong, fogMode);
}

void ShaderCtrl_cls::
SetShaderParamCartoon(int value, float boldness){
    glUniform1i(mParamIDLNumCartoon, value);
    glUniform1f(mParamIDBoldnessCartoon, boldness);
}

void ShaderCtrl_cls::
SetShaderParamCartoonShadowMode(bool bShadowMode){
    if (bShadowMode == 0)
	glUniform1i(mParamIDShadowModeCartoon, 1);
    else
	glUniform1i(mParamIDShadowModeCartoon, 0);
}
