// "clsNavigator.h"

#include "clsNavigators.h"

#include "../clsMainGLCanvas.h"

#include "../../GLOBJECTS/clsCone.h"
#include "../../MYGL/clsGLPerspective.h"

#include "../../COMMON/CommonDefinitions.h"

Navigators_cls::
Navigators_cls(MainGLCanvas_cls *pCMainGLCanvas){
    mpCManinGLCanvas = pCMainGLCanvas;
    mpCNaviFBO = NULL;
}

Navigators_cls::
~Navigators_cls(){
    if (mpCNaviFBO != NULL) delete mpCNaviFBO;
}

void Navigators_cls::
InitializeStaticMembers(){
    msCGLBTNRot.MakeTextureFromImage(wxT("ICONS/Rotation.png"));
    msCGLBTNTrans.MakeTextureFromImage(wxT("ICONS/Translation.png"));
    msCGLBTNScale.MakeTextureFromImage(wxT("ICONS/Scale.png"));
}

void Navigators_cls::
ShowArrow(){
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0);
    glVertex3d(+1, 0, 0);
    glEnd();
    glTranslated(0.6, 0, 0);
    glScaled(0.4, 0.4, 0.4);
    Cone_cls::Draw();
}

void Navigators_cls::
MakeImage(PositionAndDirection_cls CCamPos){
    if (mpCNaviFBO == NULL){
	mpCNaviFBO = new FBO_cls (256, 256);
	mpCNaviFBO->MakeFBO(_FBO_USE_RGBA);
    }

    mpCNaviFBO->Begin(TRUE);
    glEnable(GL_DEPTH_TEST);
    
    GLPerspective_cls CGLPerspective(256, 256, 15);
    //CGLPerspective.SetWindowSize(256, 256);
    //CGLPerspective.SetClipSize(256, 256);
    PositionAndDirection_cls CNaviCameraPos;
    CNaviCameraPos = CCamPos;
    LOOP_XYZ CNaviCameraPos.mPos[xyz] = -6 * CCamPos.mTgt[xyz];
    
    CGLPerspective.SetViewportOnGL();
    CGLPerspective.SetPerspectiveOnGL(&CNaviCameraPos);

    CGLPerspective.SetModelOrProjectionViewOnGL
	(&CNaviCameraPos, GL_MODELVIEW);

    glLineWidth(5);
    glTranslated(-0.25, -0.25, -0.25);

    glColor4f(1.0f, 0.1f, 0.1f, 1.0f);
    glPushMatrix(); // +1
    ShowArrow();
    glPopMatrix(); // +-0

    glColor4f(0.1f, 1.0f, 0.1f, 1.0f);
    glPushMatrix(); // +1
    glRotatef(90, 0, 0, 1);
    ShowArrow();
    glPopMatrix(); // +-0

    glColor4f(0.1f, 0.1f, 1.0f, 1.0f);
    glPushMatrix(); // +1
    glRotatef(-90, 0, 1, 0);
    ShowArrow();
    glPopMatrix(); // +-0
    
    mpCNaviFBO->End();
}

void Navigators_cls::
Show(int width, int height){
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    mpCNaviFBO->DrawOnScreen2D(width, height, 64, 64, 32, 0);
    glPopAttrib();
}

bool Navigators_cls::
CheckGLButton(int canvasWidth, int canvasHeight,
	      int mouseX, int mouseY){
    if (msCGLBTNScale.IsMouseOnButton(canvasWidth, canvasHeight,
				      mouseX, mouseY) == TRUE){
	ScaleMode(); return TRUE; }

    if (msCGLBTNRot.IsMouseOnButton(canvasWidth, canvasHeight,
				    mouseX, mouseY) == TRUE){
	RotationMode(); return TRUE; }

    if (msCGLBTNTrans.IsMouseOnButton(canvasWidth, canvasHeight,
				      mouseX, mouseY) == TRUE){
	TranslationMode(); return TRUE; }
    
    return FALSE;
}

void Navigators_cls::
DrawButton(int width, int height){
    GLPerspective_cls::Push2DMatrix(width, height);
    msCGLBTNScale.DrawButton(width, msManipulatorMode == _MANIP_SCALE);
    msCGLBTNRot  .DrawButton(width, msManipulatorMode == _MANIP_ROTATE);
    msCGLBTNTrans.DrawButton(width, msManipulatorMode == _MANIP_TRANSLATE);
    GLPerspective_cls::Pop2DMatrix();
}

bool Navigators_cls::
UpdateStatus(int width, int height, int mouseX, int mouseY){
    bool bRefresh = FALSE;
    bRefresh |= msCGLBTNScale.UpdateStatus(width, height, mouseX, mouseY);
    bRefresh |= msCGLBTNRot.UpdateStatus(width, height, mouseX, mouseY);
    bRefresh |= msCGLBTNTrans.UpdateStatus(width, height, mouseX, mouseY);
    return bRefresh;
}

void Navigators_cls::
RotationMode(){
    msManipulatorMode = _MANIP_ROTATE;
}

void Navigators_cls::
TranslationMode(){
    msManipulatorMode = _MANIP_TRANSLATE;
}

void Navigators_cls::
ScaleMode(){
    msManipulatorMode = _MANIP_SCALE;
}

void Navigators_cls::
SetDragMode(int dragMode){
    if (msDragMode == dragMode) msDragMode = -1;
    else                        msDragMode = dragMode;
}

GLButton_cls Navigators_cls::msCGLBTNTrans(_BTM_LEFT, 4, 52);
GLButton_cls Navigators_cls::msCGLBTNRot  (_BTM_LEFT, 4, 28);
GLButton_cls Navigators_cls::msCGLBTNScale(_BTM_LEFT, 4, 4);
int    Navigators_cls::msManipulatorMode = _MANIP_TRANSLATE;
int    Navigators_cls::msDragMode = -1; //_DRAG_ALL
