// -*- c -*-
// "clsHitTest.h"

#include "clsHitTest.h"
#include "clsMainGLCanvas.h"
#include "../MYGL/clsFBO.h"

HitTest_cls::
HitTest_cls(){
    mBufferSize = 128;
    mbHitTest = TRUE;
    mpCFBO = NULL;
    mbInitialized = FALSE;
}

HitTest_cls::
~HitTest_cls(){
    if (mpCFBO != NULL) delete mpCFBO;
}

void HitTest_cls::
Initialize(){
    if (mbInitialized == TRUE) return;
    mpCFBO = new FBO_cls(128, 128);
    mpCFBO->MakeFBO();
    mbInitialized = TRUE;
}

void HitTest_cls::
StartHitTest(float mouseNX, float mouseNY, float aspect,
	     MainGLCanvas_cls *pCGLCanvas, PositionAndDirection_cls *pCCamPos,
	     bool bHitTest){

    mbHitTest = bHitTest;
    
    glPushMatrix();

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();

    mpCFBO->Begin();

    glLoadIdentity();

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);

    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    float fviewport[4];
    for (int i = 0; i < 4; i++) fviewport[i] = viewport[i];
    
    if (pCGLCanvas->mCGLPerspective.mbPerspective == TRUE){
	if (aspect >= 1){
	    mPointX = (mouseNX-0.5)*fviewport[2] * aspect + fviewport[2]/2;
	    mPointY = ((1.0f-mouseNY)-0.5)*fviewport[3] + fviewport[3]/2 ;
	}
	else{
	    mPointX = (mouseNX-0.5)*fviewport[2] + fviewport[2]/2;
	    mPointY = ((1.0f-mouseNY)-0.5)*fviewport[3]/aspect + fviewport[3]/2 ;
	}
    }
    else{
	mPointX = (mouseNX-0.5)*fviewport[2] + fviewport[2]/2;
	mPointY = ((1.0f-mouseNY) - 0.5)*fviewport[3]/aspect + fviewport[3]/2;
    }
    
    gluPickMatrix(mPointX, mPointY, (float)viewport[2]/25, (float)viewport[3]/25, viewport);

    GLPerspective_cls CGLPerspective = pCGLCanvas->mCGLPerspective;

    CGLPerspective.SetWindowSize(mBufferSize, mBufferSize);
    CGLPerspective.SetClipSize(mBufferSize, mBufferSize);
    CGLPerspective.SetPerspectiveOnGL(pCCamPos, TRUE);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    pCGLCanvas->mCGLPerspective.SetModelOrProjectionViewOnGL(pCCamPos, GL_MODELVIEW);
    
    glSelectBuffer(100, mBuffer4Selection);
    if (bHitTest == TRUE) glRenderMode(GL_SELECT);
    glInitNames();
}

std::vector < std::vector < int > > HitTest_cls::
GetVVHittedObject(){
    std::vector < std::vector <int> > VVRet;

    int hitNum = glRenderMode(GL_RENDER); // ZNg[hIGqbg𓾂
    
    GLuint *pBuf = mBuffer4Selection;
    GLuint depthMinMax = UINT_MAX;
    for (int i = 0; i < hitNum; i++){
	GLuint nameStack = pBuf[0]; //pBuf[3]ȍ~nameStackł邩
	GLuint depthMin = pBuf[1];
	GLuint depthMax = pBuf[2];
	std::vector < int > VStackedName;
	for (GLuint j = 0; j < nameStack; j++){
	    VStackedName.push_back(pBuf[3 + j]); 
	}
	VVRet.push_back(VStackedName);

	pBuf += (3 + nameStack);
    }
    
    return VVRet;
}

void HitTest_cls::
EndHitTest(){
    mVVHittedObjects = GetVVHittedObject();

    mpCFBO->End();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
}

int HitTest_cls::
GetFirstObject(){
    if (mVVHittedObjects.size() == 0)
	return -1;

    if (mVVHittedObjects[0].size() == 0)
	return -1;
    
    return (mVVHittedObjects[0][0]);
}

int HitTest_cls::
GetFirstObjectSub(){
    if (mVVHittedObjects.size() == 0)
	return -1;
    if (mVVHittedObjects[0].size() < 2)
	return -1;
    return (mVVHittedObjects[0][1]);
}
