// "clsMainGLCanvas.h"
// (setq c-basic-offset 4)

#include "../COMMON/CommonDefinitionsGL.h"

#ifdef WIN32
PFNGLTEXIMAGE3DPROC glTexImage3D;
PFNGLCOMPRESSEDTEXIMAGE3DPROC glTexCompressedImage3D;
PFNGLBLENDCOLORPROC glBlendColor;
PFNGLBLENDEQUATIONEXTPROC glBlendEquationEXT;
extern PFNGLBINDBUFFERPROC glBindBuffer; //!!
#endif 

#include "../../GL/glsl.h"

#include "clsDistortion.h"
#include "clsDolly.h"
#include "clsMoveItem.h"
#include "clsMainGLCanvas.h"
#include "clsMouseCtrl.h"
#include "NAVIGATORS/clsNavigators.h"
#include "clsStereoGram.h"

#include "../FILTER/clsFilterBG.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../FILTER/clsFilterCopy.h"
#include "../FILTER/clsFilterDataBox.h"
#include "../FILTER/clsFilterFog.h"
#include "../FILTER/clsFilterStick.h"
#include "../GLATTRIB/clsColorCurve.h"
#include "../GLATTRIB/clsShaderCtrl.h"
#include "../GLOBJECTS/clsCamera.h"
#include "../GLOBJECTS/clsCircle.h"
#include "../GLOBJECTS/clsCone.h"
#include "../GLOBJECTS/clsCube.h"
#include "../GLOBJECTS/clsCylinder.h"
#include "../GLOBJECTS/clsGrid.h"
#include "../GLOBJECTS/clsSphere.h"
extern Sphere_cls *GpCSphere;
#include "../GRAPHIC/clsStereoGram.h"
#include "../MAINCANVAS/clsOverlay.h"
#include "../MAINCANVAS/clsExposure.h"
#include "../MAINCANVAS/clsDrawOption.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../PANELS/clsDataExporter.h"
#include "../PANELS/clsPreference.h"
#include "../PARTICLEDATA/clsParticleDataSequence.h"
#include "../PARTICLEDATA/clsThreadFileReader.h"
extern ThreadFileReader_cls   *GpCThreadReader;
extern bool                   GbThreadReaderFinished;
extern int                    GThreadReaderPercentage;
#include "../PARTICLEDATA/INTERPOLATE_GET/clsThreadInterpolate.h"
extern ThreadInterpolator_cls *GpCThreadInterpolate;
extern bool                   GbThreadInterpolateFinished;
#include "../PARTICLEDATA/clsZSort.h"

extern int                    GDrawingStage;

#include "../POVRAY/clsPovRay.h"
#include "../SCENEOBJ/clsCameraPath.h"
#include "../SCENEOBJ/clsCameraPathGUI.h"
#include "../SCENEOBJ/clsLightSourceGUI.h"

#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include "../COMMON/MyVector.h"
#include "../GLCANVAS/clsGLCanvasCtrl.h"
#include "../MYAPI/MsgBox.h"
#include "../MYGL/clsFBO.h"
#include "../MYGL/clsFishEye.h"
#include "../MYGL/clsGLString2D.h"
#include "../MYGL/clsGLGuides.h"
#include "../MYGL/clsGLInfo.h"
#include "../MYGL/clsGLPerspective.h"
#include "../MYGL/clsGLTexture.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;

#include <wx/tglbtn.h>

long GTestVariableLong = 0;
int GTestVariableInt = 0;

MainGLCanvas_cls* GpCurrentGLPanel;

#include "../GRAPHIC/clsImageIO.h"
#include "../GRAPHIC/clsImgSequence.h"
extern ImageIO_cls GCImgIO;

#include "../../clsFrameMain.h"
#include "../../clsMain.h"

#include <algorithm>
#include <functional>
#include <sstream>

#include <math.h>

#include "../COMMON/CommonDefinitions.h"

bool GbGLLockPlayWait = FALSE;
bool GbGLLockDrawing  = FALSE;
bool GbGLLockProjectLoading = FALSE;
bool GbGLLockForLoadingData = FALSE;
bool GbGLLockForInterplateData = FALSE;
bool GbGLLockModalDialog = FALSE;

#define R 0
#define G 1
#define B 2

#define LOOP_YPRY for (int ypry = 0; ypry < 4; ypry++)

#define _STENCIL_MODE_0 0
#define _STENCIL_MODE_1 1
#define _STENCIL_MODE_2 2

#define _MAINGL_NO_MIRROR -1

#ifdef WIN32
#define _FLAG_SET_CURRENT
#endif
#ifdef __WXGTK__
#define _FLAG_SET_CURRENT
#endif

// -----------------------------------------

BEGIN_EVENT_TABLE(MainGLCanvas_cls, wxGLCanvas)
EVT_PAINT       (MainGLCanvas_cls::OnPaint)
EVT_MOUSE_EVENTS(MainGLCanvas_cls::OnMouse)
EVT_MOTION(MainGLCanvas_cls::OnMouseMotion)
EVT_MENU (GLMenuMoveToCamPos,   MainGLCanvas_cls::OnMenuMoveToCamPos)
EVT_MENU (GLMenuSetKeyPos,      MainGLCanvas_cls::OnMenuSetKeyPos)
EVT_MENU (GLMenuSetKeyPosRot,   MainGLCanvas_cls::OnMenuSetKeyPosRot)
EVT_MENU (GLMenuSetKeyAnc,      MainGLCanvas_cls::OnMenuSetKeyAnc)
EVT_MENU (GLMenuToggleLink,     MainGLCanvas_cls::OnMenuToggleLink)
EVT_MENU (GLMenuToggleLink2,    MainGLCanvas_cls::OnMenuToggleLink2)
EVT_MENU (GLMenuTogglePerspective, MainGLCanvas_cls::OnMenuTogglePerspective)
EVT_KEY_DOWN  (MainGLCanvas_cls::OnKeyDown)
END_EVENT_TABLE()

MainGLCanvas_cls::
MainGLCanvas_cls(GLCanvasCtrl_cls *pCParentCtrl, FrameMain_cls *pCParentFrame,
		 int* attribList, long style, const wxString& name,
		 unsigned int selfIndex, int displayMode) : 
wxGLCanvas(pCParentCtrl, wxID_ANY, wxDefaultPosition, wxDefaultSize,
	   style, name, attribList)
{
    ConstructorMain(pCParentCtrl, pCParentFrame, selfIndex, displayMode);
}

MainGLCanvas_cls::
MainGLCanvas_cls(GLCanvasCtrl_cls *pCParentCtrl, FrameMain_cls *pCParentFrame,
		 int* attribList, long style, const wxString& name,
		 const wxGLContext *pContextToShare,
		 unsigned int selfIndex, int displayMode) : 
    /*
      wxGLCanvas(pCParentCtrl, id, attribList, pos, size, style|wxFULL_REPAINT_ON_RESIZE, name), //{͂gC邯ǂ܂Ȃ
    */
#ifdef __WXGTK__
    wxGLCanvas(pCParentCtrl, wxID_ANY, wxDefaultPosition, wxDefaultSize, 
	       style | wxFULL_REPAINT_ON_RESIZE, name, attribList)
#else
    wxGLCanvas(pCParentCtrl, pContextToShare, 
	       wxID_ANY, wxDefaultPosition, wxDefaultSize, 
	       style | wxFULL_REPAINT_ON_RESIZE, name, attribList)
#endif
{
    ConstructorMain(pCParentCtrl, pCParentFrame, selfIndex, displayMode, pContextToShare);
}

void MainGLCanvas_cls::
ConstructorMain(GLCanvasCtrl_cls *pCParentCtrl, FrameMain_cls *pCParentFrame,
		unsigned int selfIndex, int displayMode,
		const wxGLContext *pContextToShare)
{
    mpCMouseCtrl  = NULL;
    mpCDolly      = NULL;
    mpCMoveItem   = NULL;

    mpCGLCtrl = pCParentCtrl;
    mpCParentFrameLink = pCParentFrame;
    mpCDrawOptionLink = mpCParentFrameLink->mpCParent->mpCDrawOption;

    mpCNavigators = new Navigators_cls(this);

    mpCGLString2DDirect = new GLStrings2D_cls;    
    mpCGLString2DBuffer = new GLStrings2D_cls;
    mpCTagObjectCtrl = new TagObjectCtrl_cls;

    mpCFBOForFishEyeBuffer = NULL;
    mpCFBOForShadow = NULL;
    mpCFBOForOverlay = NULL;
    mCounter = 0;
    
    SetBackgroundColour(*wxBLACK);
    mSelfIndex = selfIndex;
    mbInitialized    = FALSE;
    mbExportDataQueue = FALSE;
    mbExportPovStringQueue = FALSE;
    mbHitTestItemSelect    = FALSE; 
    mbHitTestGuideSelect   = FALSE; 
    mPrevFrame       = -1;
    mPrevStep        = -1;
    mPrevBlurFactor  = -1;
    mVHitObj.push_back(0); mVHitObj.push_back(0);
    mbSaveImageQueue = FALSE;
    mbSaveImgSeqQueue = FALSE;
    mbRefreshQueue   = FALSE;
    mDoubleIndexPrev = -1;

    SetCameraToDefaultPos();

    wxString ssTag;
    ssTag << wxT("Panel") << selfIndex;

    if (selfIndex == 4)
	mCConfList.Add(&mCbShowMessage, ssTag + L"ShowMessage=", L"0");
    else
	mCConfList.Add(&mCbShowMessage, ssTag + L"ShowMessage=", L"1");

    mCConfList.Add(&mCbShowItems , ssTag + L"ShowItems=", L"1");
    mCConfList.Add(&mCbShowGrid, ssTag + L"ShowGrid=", L"1");
    mCConfList.Add(&mCbShowData, ssTag + L"ShowData=", L"1");
    mCConfList.Add(&mCbShowNavigator, ssTag + L"ShowNavigator=", L"1");

    /*
      mCConfList.Add(&mCbStereoFlip, ssTag + L"StereoFlip=", L"0");
    */
    
    mCConfList.Add(&mCGLPersAngle,   ssTag + L"GLPersAngle=", L"35");
    mCConfList.Add(&mCGLOrthoScale,  ssTag + L"GLOrthoScale=", L"1");
    mCConfList.Add(&mCbShowOtherCam, ssTag + L"ShowOtherCam=", L"0");
    mCConfList.Add(&mCbXOR,          ssTag + L"DrawItemByXOR=", L"0");
    mCConfList.Add(&mCGLGeometry,    ssTag + L"GLGeometry=", L"0");
    mCConfList.Add(&mCDisplayMode,   ssTag + L"DisplayMode=", L"0");
    mCConfList.Add(&mCbPivotMode,    ssTag + L"PivotMode=", L"0");
    mCConfList.Add(&mCPathIndex,     ssTag + L"GLPathIndex=", L"0");
    mCConfList.Add(&mCGridType,      ssTag + L"GridType=", L"0");
    mCConfList.Add(&mCbGridAnti,     ssTag + L"GridAntialias=", L"0");
    mCConfList.Add(&mCbShowGridTag,  ssTag + L"GridShowTag=", L"0");
    mCConfList.Add(&mCbDistortion,   ssTag + L"Distortion", L"0");
    mCConfList.Add(&mCbFisheyeGrid,  ssTag + L"FisheyeGrid", L"0");
    mCConfList.Add(&mCGLDepthMin,    ssTag + L"GLDepthMin=", L"0.03");
    mCConfList.Add(&mCGLDepthMax,    ssTag + L"GLDepthMax=", L"5000");

    SetDisplayMode(displayMode);

    mCGLPersAngle.mBody = mCGLPerspective.mDeg;
        
#ifdef __WXGTK__
    mpCGLContext = new wxGLContext(this, pContextToShare);
#else
    mpCGLContext = GetContext();
#endif
   
    mpCCameraPathLinked   = NULL;
    mpCAnchorPathLinked   = NULL;
    mpCDataExporterLinked = NULL;

    //--- SetSubWindows;

    // Status bar ̂ helpString͕\Ȃ̂ŏȗ
    // CENbNj[
    mpPopupMenu = new wxMenu; 

    mpPopupMenu->Append(GLMenuMoveToCamPos, _("Move to Cam Pos"));
    mpPopupMenu->Append
	(GLMenuSetKeyPos   , 
	 wxString::Format(wxT("%s %s"), _("Sey Key Cam Pos"), wxT("(Pos)")));
    mpPopupMenu->Append
	(GLMenuSetKeyPosRot,
	 wxString::Format(wxT("%s %s"), _("Sey Key Cam Pos"), wxT("(Pos+Rot)")));
    mpPopupMenu->Append(GLMenuSetKeyAnc   , _("Sey Key Anc Pos"));
    mpPopupMenu->Append(GLMenuToggleLink  , _("Link/Unlink to Cam"));
    mpPopupMenu->Append(GLMenuToggleLink2, _("Unlink but view remains here"));
    mpPopupMenu->AppendSeparator();
    mpPopupMenu->Append(GLMenuTogglePerspective, _("Toggle Perspective"));

    /*
      mpMenuItemShowPanel = mpPopupMenu->Append(GLMenuShowPanel   , _("Show Control Panel"));
    */

    mpPopupMenu->UpdateUI(GetEventHandler());
}

MainGLCanvas_cls::
~MainGLCanvas_cls(){
    if (mpCMouseCtrl != NULL) delete mpCMouseCtrl;
    if (mpCDolly != NULL)     delete mpCDolly;
    if (mpCMoveItem != NULL)  delete mpCMoveItem;
    delete mpCNavigators;
    delete mpCGLString2DDirect;
    delete mpCGLString2DBuffer;
    delete mpCTagObjectCtrl;
#ifndef __WXMAC__
    //delete mpCGLContext;
#endif
    delete mpPopupMenu;
    
    if (mpCFBOForFishEyeBuffer != NULL) delete mpCFBOForFishEyeBuffer;
    if (mpCFBOForShadow != NULL) delete mpCFBOForShadow;
    if (mpCFBOForOverlay != NULL) delete mpCFBOForOverlay;

    if (mspCDistortion != NULL){
	delete mspCDistortion; mspCDistortion = NULL; }
    if (mspCMask != NULL){
	delete mspCMask; mspCMask = NULL; }
    if (mspCFisheyeGrid != NULL){
	delete mspCFisheyeGrid; mspCFisheyeGrid = NULL; }
}

// ----------

bool MainGLCanvas_cls::
SetGLContext(){
#ifndef __WXMOTIF__
    if (!GetContext()) return FALSE; 
#endif
    
#ifdef _FLAG_SET_CURRENT
    mpCGLContext->SetCurrent(*(wxGLCanvas*)this);
#else
#ifdef __WXMAC__
    SetCurrent();
#else
    mpCGLContext->SetCurrent();
#endif
#endif
    return TRUE;
}

void MainGLCanvas_cls::
InitGL(){
    glClearColor( 0, 0, 0, 1.0f );
    glClearStencil(0);

    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glDisable(GL_COLOR_MATERIAL);
    glLineStipple(2, 0xAAAA);
}

void MainGLCanvas_cls::
InitGLUT(){
    if (msbGLUTInitialized == TRUE) return;
    int argc = 1;
    char **argv = new char* [1];
    argv[0] = new char [6];
    printf(argv[0], "dummy");
    glutInit(&argc, argv);
    delete argv[0]; 
    delete argv;
    msbGLUTInitialized = TRUE;
}

void MainGLCanvas_cls::
InitializeShader(){

    bool bShaderAvailable = TRUE;
    if (glslInit()) bShaderAvailable = FALSE;
    
    mpCParentFrameLink->mpCParent->
	mpCExposure->Initialize(this, bShaderAvailable);

    GLAttribute_cls::InitializeShader(this, bShaderAvailable);
}

void MainGLCanvas_cls::
MakeFisheyeGridTexture(){
    
    wxMemoryDC *pCMemDC = mspCFisheyeGrid->BeginDC(1024, 1024);
    pCMemDC->SetBackground(wxColour(0, 0, 0));
    pCMemDC->Clear();
    
    pCMemDC->SetPen(wxPen(wxColour(255, 255, 255), 2));
    float scaleX = 512;
    float scaleY = 512;
    std::vector < wxPoint > AVertex;
    for (double r = 1.0; r >= 0; r -= 1.0/90 * 10){
	double r0 = r;
	if (r == 1.0) r0 = 0.997;  
	AVertex.clear();
	for (int i = 0; i <= 360; i+= 5){
	    double theta = ((double)i/180) * 3.14159;
	    double x = (1.0 + r0 * cos(theta)) * scaleX;
	    double y = (1.0 + r0 * sin(theta)) * scaleY;
	    AVertex.push_back(wxPoint(x, y));
	}
	pCMemDC->DrawLines((int)AVertex.size(), AVertex.data());
    }
    for (double i = 0; i <= 360; i += 15){
	double theta = ((double)i/180) * 3.14159;
	double x = (1.0 + cos(theta)* 0.997) * scaleX;
	double y = (1.0 + sin(theta)* 0.997) * scaleY;
	pCMemDC->DrawLine(scaleX, scaleY, x, y);
    }
    mspCFisheyeGrid->EndDC(FALSE);
}

bool MainGLCanvas_cls::
InitializeGLIfNot(){
    if (mbInitialized == TRUE) return FALSE;
    
    InitGLUT();
    InitGL();

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    mpCParentFrameLink->mpCGLInfo->Initialize();
    
    mpCParentFrameLink->GetPCActiveFileSequence()->RefreshGLAttribTex();
    
    Camera_cls::InitializeIfNot();
    Circle_cls::InitializeIfNot();
    Cone_cls::InitializeIfNot();
    Cube_cls::InitializeIfNot();
    Cylinder_cls::InitializeIfNot();
    GpCSphere->InitializeIfNot();

    glPopAttrib();
    
    FBO_cls::Initialize();

    /*
      if (mpCParentFrameLink->mpCGLInfo->mMajorVersion >= 2) 
      RedrawPBO_cls::msbSupported = TRUE;
      if (mpCParentFrameLink->mpCGLInfo->mMajorVersion == 1 &&
      mpCParentFrameLink->mpCGLInfo->mMinorVersion >= 5) 
      RedrawPBO_cls::msbSupported = TRUE;
    */
    
    if (mspCMask == NULL){ // ʕ̏
	InitializeShader();

#ifdef WIN32
	glBlendEquationEXT = (PFNGLBLENDEQUATIONEXTPROC) wglGetProcAddress("glBlendEquationEXT");
	glBlendColor = (PFNGLBLENDCOLORPROC) wglGetProcAddress("glBlendColor");
	
#endif 
	glEnable(GL_TEXTURE_2D);
	mspCDrawTimeIndex->Initialize();
	mspCDrawClrLegend->Initialize();

	mspCMask        = new GLTexture_cls;
	mspCFisheyeGrid = new GLTexture_cls;
	mspCMask->LoadImage(wxT("Resource/DefaultMask.png"), TRUE);
	MakeFisheyeGridTexture();
	
	mpCNavigators->InitializeStaticMembers();

	msCHittest.Initialize();
	
	glDisable(GL_TEXTURE_2D);
    }

    mpCMouseCtrl  = new MouseCtrl_cls(mpCParentFrameLink->mpCPreference);
    mpCDolly      = new Dolly_cls(mpCParentFrameLink->mpCPreference);
    mpCMoveItem   = new MoveItem_cls(mpCParentFrameLink->mpCPreference,
				     mspCGLGuide,
				     mspCGLGuideScale,
				     mspCGLGuideYPR);
    
    glDepthFunc(GL_LEQUAL);
    mbInitialized = TRUE;
    return TRUE;
}

// ----

void MainGLCanvas_cls::
SetCameraToDefaultPos(){
    double defaultPos[3] = {-1.12,-0.8,+0.56};
    double defaultTop[3] = { 0,  0, +1};

    LOOP_XYZ mCCameraPos.mPos[xyz] = 3.5 * defaultPos[xyz];
    LOOP_XYZ mCCameraPos.mTgt[xyz] = - defaultPos[xyz];
    LOOP_XYZ mCCameraPos.mTop[xyz] = defaultTop[xyz];
    mCCameraPos.NormalizeAndOrthogonalize();
    mCCameraPosFree   = mCCameraPos;
    mCCameraPosCamera = mCCameraPos;
}

void MainGLCanvas_cls::
UpdateDrawInfo(int width, int height){
    mCDrawInfoPrev = mCDrawInfo;
    mCDrawInfo.mbForceNewDraw = FALSE;
    mCDrawInfo.mCCameraPos = mCCameraPos;
}

PovRayCtrl_cls* MainGLCanvas_cls::
GetPCPovRayCtrlLinked(){
    return mpCParentFrameLink->mpCParent->mpCPovRayCtrl;
}

PositionAndDirection_cls* MainGLCanvas_cls::
GetPCCameraPosition(){
    if (IsAlongAxis() == TRUE)
	return &mCCameraPosOrtho;
    if (mCDisplayMode.mBody == _GL_CANVAS_LINKED_CAM0)
	return &mCCameraPosCamera;
    return &mCCameraPosFree;
}

Position_cls* MainGLCanvas_cls::
GetPCAnchorPosition(){
    if (IsAlongAxis() == TRUE)
	return &msCAnchorPosOrtho;
    return &msCAnchorPos;
}

TimeCtrl_cls* MainGLCanvas_cls::
GetPCTimeCtrl(){
    return mpCParentFrameLink->GetPCTimeCtrl();
}

void MainGLCanvas_cls::
UpdateCameraPos(){
    if (mPrevFrame == -1 && mPrevStep == -1)
	mpCGLCtrl->SyncDisplayModeToCombobox();

    if (mCbPivotMode.mbBody == _PIVOT_FOLLOW_PATH){
	LOOP_XYZ msCAnchorPos.mPos[xyz] = 
	    GetPLinkedAnchorPath()->mCCurrentObj.mPos[xyz];
    }

    if (mCDisplayMode.mBody == _DIRECTION_FREE){
	mCCameraPosFree.PointAt(msCAnchorPos.mPos);
	mCCameraPos = mCCameraPosFree;
    }

    if (IsAlongAxis() == TRUE){
	ModifyOrthoCamera();
	mCCameraPos = mCCameraPosOrtho;
    }

    if (GetPLinkedCameraPath()->mCbLookAnchor.mbBody == TRUE)
	GetPLinkedCameraPath()->mCCurrentObj.
	    PointAt(GetPLinkedAnchorPath()->mCCurrentObj.mPos);
    
    if (mCDisplayMode.mBody == _GL_CANVAS_LINKED_CAM0){
	if (mPrevFrame != GetPCTimeCtrl()->mFrame ||
	    mPrevStep  != GetPCTimeCtrl()->mInterpStep ||
	    mPrevBlurFactor != GetPCTimeCtrl()->mCBlurCtrl.GetFactor())
	    if (mpCCameraPathLinked != NULL)
		mCCameraPosCamera = mpCCameraPathLinked->mCCurrentObj;
	
	if (mpCCameraPathLinked != NULL &&
	    mpCCameraPathLinked->mCbLookAnchor.mbBody == TRUE)
	    mCCameraPosCamera.PointAt(GetPLinkedAnchorPath()->mCCurrentObj.mPos);
	else
	    mCCameraPosCamera.PointAt(msCAnchorPos.mPos);
	/*
	if (!(mCCameraPosCamera == mpCCameraPathLinked->mCCurrentObj))
	    mCCameraPosCamera.PointAt(msCAnchorPos.mPos);
	*/
	
	mCCameraPos = mCCameraPosCamera;
    }

    GetPCCameraPosition()->UpdateYPR();
    mCCameraPos.UpdateYPR();
}

CameraPath_cls* MainGLCanvas_cls::
GetPLinkedCameraPath(){
    if (mpCCameraPathLinked == NULL) 
	return mpCParentFrameLink->mpCParent->mpCCameraPath;
    return mpCCameraPathLinked;
}

LightSRCCtrl_cls* MainGLCanvas_cls::
GetPLinkLightSourceCtrl(){
    return mpCParentFrameLink->mpCParent->mpCLightSRCCtrl;
}

AnchorPath_cls* MainGLCanvas_cls::
GetPLinkedAnchorPath(){
    if (mpCAnchorPathLinked == NULL)
	return mpCParentFrameLink->mpCParent->mpCAnchorPath;
    return mpCAnchorPathLinked;
}

bool MainGLCanvas_cls::
IsDrawingConditionSatisfied(){
    if (IsShownOnScreen() == FALSE || GbGLLockDrawing == TRUE) return FALSE;
    if (GbGLLockForLoadingData == TRUE)	printf("LockLoading\n");
    if (GbGLLockForLoadingData == TRUE || 
	GbGLLockForInterplateData == TRUE ||
	GbGLLockProjectLoading == TRUE ||
	GbGLLockModalDialog == TRUE) return FALSE;
    return TRUE;
}

bool MainGLCanvas_cls::
IsStereo(){
    if (mspCStereogramLink->mCType.mBody != _STEREO_NONE){
	if (mspCStereogramLink->mCEye.mBody == _STEREO_NONE ||
	    mspCStereogramLink->mCEye.mBody == _STEREO_LFT ||
	    mspCStereogramLink->mCEye.mBody == _STEREO_RGT)
	    return FALSE;
	if (mCDisplayMode.mBody == _DIRECTION_FREE || 
	    mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	    return TRUE;
    }
    return FALSE;
}

bool MainGLCanvas_cls::
IsStereoDualOutput(){
    if (IsStereo() == FALSE) return FALSE;
    if (mspCStereogramLink->mCDualOutput.mBody == 1) return TRUE;
    return FALSE;
}

bool MainGLCanvas_cls::
IsLineByLineFishEye(){
    bool bStereo = IsStereo();
    int  type = mspCStereogramLink->mCEye.mBody;
    bool bFishEye = IsFishEye();
    if (IsStereo() == TRUE &&
	mspCStereogramLink->mCEye.mBody == _STEREO_SCANLINE &&
	IsFishEye() == TRUE)
	return TRUE;
    return FALSE;
}

bool MainGLCanvas_cls::
IsOrthographic(){
    return (mCGLGeometry.mBody == _GEOM_ORTHOGRAPHIC);
}

bool MainGLCanvas_cls::
IsAlongAxis(){
    return (mCDisplayMode.mBody >= _GL_CANVAS_DISPLAY_X_PLUS &&
	    mCDisplayMode.mBody <= _GL_CANVAS_DISPLAY_Z_MINUS);
}

bool MainGLCanvas_cls::
IsFishEye(){
    if (mCDisplayMode.mBody == _DIRECTION_FREE ||
	mCDisplayMode.mBody == _GL_CANVAS_COPY_OF_VIEW_0)
	return (mCGLGeometry.mBody == _GEOM_FISHEYE);

    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	return (GetPLinkedCameraPath()->mCCamGeometry.mBody == _GEOM_FISHEYE);

    return FALSE;
}

bool MainGLCanvas_cls::
IsCylinder(){
    if (mCDisplayMode.mBody == _DIRECTION_FREE ||
	mCDisplayMode.mBody == _GL_CANVAS_COPY_OF_VIEW_0)
	return (mCGLGeometry.mBody == _GEOM_PANORAMIC);
    
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	return (GetPLinkedCameraPath()->mCCamGeometry.mBody == _GEOM_PANORAMIC);

    return FALSE;
}

void MainGLCanvas_cls::
WaitPrepaing(){
    if (mCbShowData.mbBody == TRUE){
	// ̃t[̏OK?
	// ɎԂ邱Ƃ
	int count = 0; // SuF[vBcOȂ炽܂ɔB
	while (mpCParentFrameLink->IsTheFrameReady(_FM_FRAME_CURRENT) == FALSE && 
	       count < 10){
	    mpCParentFrameLink->PrepareTheFrame(_FM_FRAME_CURRENT);	
	    count++;
	}
	
	// ̃t[̓ǂݍ݊Jn
	mpCParentFrameLink->PrepareFramesAhead
	    (GetPCTimeCtrl()->mPlayDirection);
    }
}

int MainGLCanvas_cls::
GetFBOMax(){
    int fboMax = 1;
    if (GetMultiPanelMode() != 0) fboMax = 4;
    return fboMax;
}

int MainGLCanvas_cls::
GetMultiPanelMode(){
    if (mCDisplayMode.mBody == _DIRECTION_FREE ||
	mCDisplayMode.mBody == _GL_CANVAS_COPY_OF_VIEW_0){
	if (mCGLGeometry.mBody == _GEOM_FISHEYE) return _MULTIPANEL_FISHEYE;
	if (mCGLGeometry.mBody == _GEOM_PANORAMIC) return _MULTIPANEL_CYLINDER;
    }
	
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0){
	if (GetPLinkedCameraPath()->mCCamGeometry.mBody == _GEOM_FISHEYE)
	    return _MULTIPANEL_FISHEYE;
	if (GetPLinkedCameraPath()->mCCamGeometry.mBody == _GEOM_PANORAMIC)
	    return _MULTIPANEL_CYLINDER;
    }

    return 0;
}

void MainGLCanvas_cls::
SetAngleIfLinked(){
    if (IsOrthographic()) mCGLPerspective.mbPerspective = FALSE;
    else                 mCGLPerspective.mbPerspective = TRUE;

    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0 &&
	mCDisplayMode.mBody == _GL_CANVAS_LINKED_CAM0){
	mCGLPerspective.mDeg = GetPLinkedCameraPath()->mCVAngle.mBody;
	if (GetPLinkedCameraPath()->mCCamGeometry.mBody == _GEOM_PERSPECTIVE &&
	    mCGLPerspective.mDeg >= 120) {
	    mCGLPerspective.mDeg = 120;
	    mstrWarning << wxT("View angle is limited to 120 deg in perspective view.") << ENDL;
	}
    }
    else{
	mCGLPerspective.mDeg = mCGLPersAngle.mBody;
	if (mCGLGeometry.mBody == _GEOM_PERSPECTIVE &&
	    mCGLPerspective.mDeg >= 120) {
	    mCGLPerspective.mDeg = 120;
	    mstrWarning << wxT("View angle is limited to 120 deg in perspective view.") << ENDL;
	}
    }
}

void MainGLCanvas_cls::
ModifyByDomeTilt(PositionAndDirection_cls *pCCam){
    pCCam->RotateYPR(0, GetPLinkedCameraPath()->mCFishEyeOffset.mBody, 0);
}

void MainGLCanvas_cls::
UpdateAnc(){
    Position_cls CCurrAnc = GetAncPos();

    if (mpCCameraPathLinked != NULL)
	mCCameraPos.SetAncPos(mpCAnchorPathLinked->mCCurrentObj.mPos);
    else
	mCCameraPos.SetAncPos(CCurrAnc.mPos);
}

Position_cls MainGLCanvas_cls::
GetAncPos(){
    if (mCDisplayMode.mBody == _DIRECTION_FREE ||
	mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	return msCAnchorPos;

    // Œ猩Ăꍇ̃AJ[ʒuԂ
    return msCAnchorPosOrtho;
}

void MainGLCanvas_cls::
GetTimeInfo(double &time, double &dT){
    int    currentIndex = GetPCTimeCtrl()->GetRemappedIndex();
    double doubleInterpStep = GetPCTimeCtrl()->GetRemappedInterpStepDouble();
    mpCParentFrameLink->GetPCActiveFileSequence()->GetTimeInfo(time, dT, currentIndex, doubleInterpStep);
}

double* MainGLCanvas_cls::
GetAnchorPosForRender(){
    static double ancPosForRender[3];
    LOOP_XYZ ancPosForRender[xyz] = 
	mCCameraPos.mPos[xyz] + mCCameraPos.mTgt[xyz];
    return ancPosForRender;
}

void MainGLCanvas_cls::
UpdateViewInformation(){
    glGetDoublev(GL_MODELVIEW_MATRIX, mMatrixModelView);
    glGetDoublev(GL_PROJECTION_MATRIX, mMatrixProjection);
    glGetIntegerv(GL_VIEWPORT, mMatrixViewport);
}

void MainGLCanvas_cls::
SetKeyPos(){
    std::vector < double > VPos;
    VPos.push_back(GetPCCameraPosition()->mPos[X]);
    VPos.push_back(GetPCCameraPosition()->mPos[Y]);
    VPos.push_back(GetPCCameraPosition()->mPos[Z]);
    
    GCUndoList.OpenBlock();
    GetPLinkedCameraPath()->mpCTranslate->SetKeyFrame(&VPos[0], TRUE);
    GCUndoList.CloseBlock();
}

void MainGLCanvas_cls::
SetKeyPosRot(){
    std::vector < double > VEuler;
    std::vector < double > VPos;
    VPos.push_back(GetPCCameraPosition()->mPos[X]);
    VPos.push_back(GetPCCameraPosition()->mPos[Y]);
    VPos.push_back(GetPCCameraPosition()->mPos[Z]);
    
    VEuler.push_back(GetPCCameraPosition()->mYPR[X]);
    VEuler.push_back(GetPCCameraPosition()->mYPR[Y]);
    VEuler.push_back(GetPCCameraPosition()->mYPR[Z]);
    
    GCUndoList.OpenBlock();
    GetPLinkedCameraPath()->mpCTranslate->SetKeyFrame(&VPos[0], TRUE);
    GetPLinkedCameraPath()->mpCRotation->SetKeyFrame(&VEuler[0], TRUE);
    GCUndoList.CloseBlock();
}

void MainGLCanvas_cls::
SetKeyAnc(){
    GCUndoList.OpenBlock();
    GetPLinkedAnchorPath()->mpCTranslate->
	SetKeyFrame(msCAnchorPos.mPos, TRUE);
    GCUndoList.CloseBlock();
}

void MainGLCanvas_cls::
SetViewport(){
    mCGLPerspective.SetViewportOnGL();
}

void MainGLCanvas_cls::
SetPerspective(PositionAndDirection_cls *pCCam){
    mCGLPerspective.mbPerspective = (IsOrthographic() == FALSE);
    
    if (mbSaveImageQueue == TRUE &&
	mbSaveImgSeqQueue == TRUE &&
	mspCImgSeqLinked != NULL &&
	GetMultiPanelMode() == 0){
	mCGLPerspective.SetClipSize(mspCImgSeqLinked->mClipWidth, 
				    mspCImgSeqLinked->mClipHeight);
    }
    else
	mCGLPerspective.SetClipSize(mPanelWidth, mPanelHeight);
    
    mCGLPerspective.SetPerspectiveOnGL(pCCam);
}

/*
void MainGLCanvas_cls::
SetModelView(PositionAndDirection_cls *pCCam, bool bUpdateMembers){
    mCGLPerspective.SetModelOrProjectionViewOnGL(pCCam, GL_MODELVIEW);
    if (bUpdateMembers == TRUE){
	mpCMoveItem->GetGLMatrixInfo();
	mspCGLGuideYPR->GetGLMatrixInfo();
    }
}
*/

void MainGLCanvas_cls::
SetLineWidth(FBO_cls *pCFBO){
    if (pCFBO != NULL)
	glLineWidth(msCLineWidth.mBody * pCFBO->GetLineWidthFactor());
    else
	glLineWidth(msCLineWidth.mBody / 3);
}

void MainGLCanvas_cls::
ClearBuffer(double doubleIndex){
    
    float clrBG[3] = {
	mspCClrCurveBG->mCConstantColor.mpBody[R],
	mspCClrCurveBG->mCConstantColor.mpBody[G],
	mspCClrCurveBG->mCConstantColor.mpBody[B]};
    
    glClearColor( clrBG[R], clrBG[G], clrBG[B], 1.0f);
    FilterFog_cls  *pCFog   = (FilterFog_cls*)mspCFilterCtrlWorldLink->GetPCFog();
    if (pCFog != NULL) 
	pCFog->SetGLClearColor(doubleIndex, clrBG);
    
    glClear( GL_COLOR_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void MainGLCanvas_cls::
ClearCanvas(){
    glClearColor(0, 0, 0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void MainGLCanvas_cls::
DrawDataCopyLoop(int fboIndex, int currentIndex, int lightNumber,
		 float pLightPosForSV[4], float pLightPhiTheta[2],
		 bool bShadowVolume){
    double doubleIndex = GetPCTimeCtrl()->GetDoubleIndex();
    double time = 0, dT = 0;
    GetTimeInfo(time, dT);
    
    FilterCopy_cls *pCCopy  = 
	(FilterCopy_cls*)mspCFilterCtrlWorldLink->GetPCCopy();
    
    std::vector < FilterStick_cls* > VpCSticks = 
	mspCFilterCtrlObjectLink->GetVpCSticks();

    PDataSequence_cls *pCSeq = mpCParentFrameLink->GetPCActiveFileSequence();
    int sortType = _SORT_ORTHO;
    if (IsFishEye() == TRUE || IsCylinder() == TRUE)
	sortType = _SORT_RADIAL;

    if (pCCopy != NULL){
	// !! {́AqWn猩J̈ʒułׂ
	pCCopy->ZSortBegin(mCCameraPos.mPos, mCCameraPos.mTgt, time);
	int copyNum = pCCopy->GetNumber();
	for (int i = 0; i < copyNum; i++){
	    glPushMatrix();
	    glPushAttrib(GL_ALL_ATTRIB_BITS);
	    pCCopy->GLTranslate(i, time+dT);
	    pCSeq->GLDrawData
		(fboIndex, currentIndex, dT, lightNumber, pLightPosForSV, 
		 pLightPhiTheta[0], pLightPhiTheta[1], bShadowVolume,
		 mSelfIndex, sortType);
	    glPopAttrib();
	    glPopMatrix();
	}

	if (VpCSticks.size() > 0){
	    glPushAttrib(GL_ALL_ATTRIB_BITS);
	    //glDepthMask(GL_TRUE);
	    //glEnable(GL_DEPTH_TEST);
	    for (int i = 0; i < copyNum; i++){
		glPushMatrix();
		pCCopy->GLTranslate(i, time);
		int bitBoundary = pCCopy->GetBitBoundary(i);
		for (unsigned int j = 0; j < VpCSticks.size(); j++)
		    VpCSticks[j]->GLDrawStick(GLDrawData_cls::mspCSnapShot,
					      mspCPeriodicLink, bitBoundary, 
					      currentIndex, mCCameraPos,
					      lightNumber);
		glPopMatrix();
	    }
	    glPopAttrib();
	}
	pCCopy->ZSortEnd();
    }
    else{
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	pCSeq->GLDrawData
	    (fboIndex, currentIndex, dT, lightNumber, pLightPosForSV, 
	     pLightPhiTheta[0], pLightPhiTheta[1], bShadowVolume, 
	     mSelfIndex, sortType);
	glPopAttrib();
	if (VpCSticks.size() > 0){
	    glPushAttrib(GL_ALL_ATTRIB_BITS);
	    //glDepthMask(GL_TRUE);
	    glEnable(GL_DEPTH_TEST);
	    for (unsigned int j = 0; j < VpCSticks.size(); j++)
		VpCSticks[j]->GLDrawStick(GLDrawData_cls::mspCSnapShot,
					  mspCPeriodicLink, 0,
					  currentIndex, mCCameraPos,
					  lightNumber);
	    glPopAttrib();
	}
    }
}

void MainGLCanvas_cls::
DrawRealObjectCopyLoop(int currentIndex, double doubleIndex, int lightNumber){
    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);

    double time = 0, dT = 0;
    GetTimeInfo(time, dT);
    
    FilterCopy_cls *pCCopy  = 
	(FilterCopy_cls*)mspCFilterCtrlWorldLink->GetPCCopy();
 
    if (pCCopy != NULL){
	int copyNum = pCCopy->GetNumber();
	pCCopy->ZSortBegin(mCCameraPos.mPos, mCCameraPos.mTgt, time);
	// !! ȉ mCCameraPos ́AXeIÕJɂׂ
	for (int i = 0; i < copyNum; i++){
	    glPushMatrix();
	    pCCopy->GLTranslate(i, time);
	    mspCFilterCtrlObjectLink->
		GLDrawRealObject(currentIndex, doubleIndex, lightNumber, TRUE, mCCameraPos);
	    glPopMatrix();
	}
	mspCFilterCtrlObjectLink->
	    GLDrawRealObject(currentIndex, doubleIndex, lightNumber, FALSE, mCCameraPos);
	pCCopy->ZSortEnd();
    }
    else{
	mspCFilterCtrlObjectLink->
	    GLDrawRealObject(currentIndex, doubleIndex, lightNumber, TRUE, mCCameraPos);
	mspCFilterCtrlObjectLink->
	    GLDrawRealObject(currentIndex, doubleIndex, lightNumber, FALSE, mCCameraPos);
    }
}

void MainGLCanvas_cls::
DrawDataLitDiffuse(double doubleIndex, double doubleRemapped,
		   int fboIndex, int lightIndex){
    if (mCbShowData.mbBody == FALSE) return;

    double time = 0, dT = 0;
    GetTimeInfo(time, dT);
    int    currentIndex =
	GetPCTimeCtrl()->GetRemappedIndex();
    double doubleInterpStep = 
	GetPCTimeCtrl()->GetRemappedInterpStepDouble();
    
    mpCParentFrameLink->GetPCActiveFileSequence()->SetCameraInfo(mCCameraPos);

    FilterDataBoxData_cls *pCScene = 
	(FilterDataBoxData_cls*)mspCFilterCtrlWorldLink->GetPCDataBox();
    
    float pLightPosForSV[4];
    float pLightPhiTheta[2];
    GetPLinkLightSourceCtrl()->
	GetLightInfo(doubleIndex, pLightPosForSV, pLightPhiTheta,
		     lightIndex, TRUE, pCScene);
    GetPLinkLightSourceCtrl()->GLSetSingleDiffuse(lightIndex, doubleIndex);
    
    FilterFog_cls     *pCFog   = (FilterFog_cls*)mspCFilterCtrlWorldLink->GetPCFog();

    glDisable(GL_LIGHTING);
    mspCFilterCtrlObjectLink->GLDrawRealGrid
	(doubleIndex, mCCameraPos, &mpCTagObjectCtrl->mVCObject);
    glEnable(GL_LIGHTING);
    
    glPushMatrix(); //+1
    if (pCScene != NULL){
	pCScene->UpdateCurrentObj(doubleIndex);
	pCScene->GLSetMatrix();
    }
    if (pCFog != NULL) pCFog->SetGLFog(doubleIndex);
    else glDisable(GL_FOG);

    // ܂ʏ̕`
    DrawDataCopyLoop(fboIndex, currentIndex, //dT, 
		     1, pLightPosForSV, pLightPhiTheta, 
		     FALSE);
    
    // DataBox Wɍڂ邩ǂ glPopMatriẍʒuŌ܂
    if (pCScene != NULL && pCScene->mCInfluence.mBody == _INFLU_PDATA_ONLY){
	glPopMatrix(); glPushMatrix();
    }
    DrawRealObjectCopyLoop(currentIndex, doubleIndex, 1);

    glPopMatrix();//+-0
    
    glPushMatrix(); //+1
    if (pCScene != NULL){
	pCScene->UpdateCurrentObj(doubleIndex);
	pCScene->GLSetMatrix();
    }
    if (pCFog != NULL) pCFog->SetGLFog(doubleIndex);
    else glDisable(GL_FOG);
        
    glClear(GL_STENCIL_BUFFER_BIT);
    glEnable(GL_STENCIL_TEST);
    glStencilMask(255);
    glDepthMask(GL_FALSE);
    
    glStencilFunc(GL_ALWAYS, 0, 255);
    if (mpCDrawOptionLink->mCShadow.mBody == 1)
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
    glColor4f(1,0.5,0.5,0.2); //?
    glDisable(GL_BLEND);

    // e|S
    GDrawingStage = 1;
    DrawDataCopyLoop(fboIndex, currentIndex, //dT, 
		     0, pLightPosForSV, pLightPhiTheta, 
		     TRUE);

    glColorMask(1,1,1,1);
    glDepthMask(GL_TRUE);
    glStencilMask(0);
    
    glStencilFunc(GL_LESS, 0, 255);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    
    //glClear(GL_DEPTH_BUFFER_BIT);
    glDepthMask(GL_FALSE);
    
    if (mpCParentFrameLink->GetPCActiveFileSequence()
	->mpCGLAttribute[0]->mCShaderType.mBody != 0){
	// eh
	glMatrixMode(GL_PROJECTION);
	glPushMatrix(); // +2
	glLoadIdentity();
	gluOrtho2D(0, 1, 0, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix(); // +3
	glLoadIdentity();

	glUseProgram(0);
	glDisable(GL_LIGHTING);
	glColor4f(0, 0, 0, 0);
	glRectf(0,0,1,1);
	
	glPopMatrix(); // +2
	glMatrixMode(GL_PROJECTION);
	glPopMatrix(); // +1
	glMatrixMode(GL_MODELVIEW);

	glPopMatrix(); // +-0
    }
    else{
	GetPLinkLightSourceCtrl()->GLDisableAllLight();
	GDrawingStage = 2;

	if (pCFog != NULL) pCFog->SetGLFog(doubleIndex);
	else glDisable(GL_FOG);
	
	glPopMatrix(); // +-0
	glColor4f(0.0, 0.0, 0.0, 1.0);
	mspCFilterCtrlObjectLink->GLDrawRealGrid
	    (doubleIndex, mCCameraPos, &mpCTagObjectCtrl->mVCObject);

	glPushMatrix(); // +1
	if (pCScene != NULL){
	    pCScene->UpdateCurrentObj(doubleIndex);
	    pCScene->GLSetMatrix();
	}
	
	// !! FIX_ME !! DataBox Wɍڂ邩ǂ glPopMatriẍʒuŌ܂
	DrawDataCopyLoop(fboIndex, currentIndex, //dT, 
			 0, pLightPosForSV, pLightPhiTheta, 
			 FALSE);
	
	if (pCScene != NULL && pCScene->mCInfluence.mBody == _INFLU_PDATA_ONLY){
	    glPopMatrix(); glPushMatrix();
	}
	DrawRealObjectCopyLoop(currentIndex, doubleIndex, 0);

	glPopMatrix(); // +-0

	GDrawingStage = 0;
    }
}

void MainGLCanvas_cls::
DrawDataMain(double doubleIndex, double doubleRemapped,
	     int fboIndex, int shadowNumber){
    
    int    currentIndex =
	GetPCTimeCtrl()->GetRemappedIndex();
    double doubleInterpStep = 
	GetPCTimeCtrl()->GetRemappedInterpStepDouble();
    double time = 0, dT = 0;
    mpCParentFrameLink->GetPCActiveFileSequence()->GetTimeInfo(time, dT, currentIndex, doubleInterpStep);
    
    int lightNumber =  GetPLinkLightSourceCtrl()->mCLightNum.mBody;
    int firstShadowLight = GetPLinkLightSourceCtrl()->GetFirstShadowedLight();
    
    FilterDataBoxData_cls *pCScene = 
	(FilterDataBoxData_cls*)mspCFilterCtrlWorldLink->GetPCDataBox();    
    
    float pLightPosForSV[4];
    float pLightPhiTheta[2];
    GetPLinkLightSourceCtrl()->
	GetLightInfo(doubleIndex, pLightPosForSV, pLightPhiTheta,
		     firstShadowLight, TRUE, pCScene);
    
    ZSort_cls::SetSceneCoordinateInfo(pCScene, doubleIndex);

    FilterFog_cls         *pCFog   = (FilterFog_cls*)mspCFilterCtrlWorldLink->GetPCFog();
    
    if (pCFog != NULL) pCFog->SetGLFog(doubleIndex);
    else glDisable(GL_FOG);
    
    glEnable(GL_LIGHTING);
    glEnable(GL_DEPTH_TEST);

    PositionAndDirection_cls CCameraPosRelative = mCCameraPos;

    glPushMatrix(); // +1
    if (pCScene != NULL){
	pCScene->UpdateCurrentObj(doubleIndex);
	pCScene->GLSetMatrix();
	float posRelative[3];
	LOOP_XYZ posRelative[xyz] = CCameraPosRelative.mPos[xyz];
	pCScene->CalcRelativePosition(posRelative, doubleIndex, pCScene->mpCAnchor);
	LOOP_XYZ CCameraPosRelative.mPos[xyz] = posRelative[xyz];
    }

    mpCParentFrameLink->GetPCActiveFileSequence()->SetCameraInfo(CCameraPosRelative);
    
    if (mCbShowData.mbBody == TRUE){
	DrawDataCopyLoop(fboIndex, currentIndex, //dT, 
			 lightNumber, pLightPosForSV, pLightPhiTheta,
			 FALSE);
    }
    if (pCScene != NULL && pCScene->mCInfluence.mBody == _INFLU_PDATA_ONLY){
	glPopMatrix(); glPushMatrix();
    }
    DrawRealObjectCopyLoop(currentIndex, doubleIndex, lightNumber);
    
    glPopMatrix(); //+-0;
    
    if (shadowNumber > 0){
	glClear(GL_STENCIL_BUFFER_BIT);
	glEnable(GL_STENCIL_TEST);
	glStencilMask(255);
	glDepthMask(GL_FALSE);

	glStencilFunc(GL_ALWAYS, 0, 255);
	if (mpCDrawOptionLink->mCShadow.mBody == 1)
	    glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDisable(GL_BLEND);
	
	GDrawingStage = 1;

	glPushMatrix(); // +1
	if (pCScene != NULL){
	    pCScene->UpdateCurrentObj(doubleIndex);
	    pCScene->GLSetMatrix();
	}

	if (mCbShowData.mbBody == TRUE)
	    DrawDataCopyLoop(fboIndex, currentIndex, //dT, 
			     lightNumber, pLightPosForSV, pLightPhiTheta, TRUE);

	glColorMask(1,1,1,1);
	glDepthMask(GL_TRUE);
	glStencilMask(0);

	glStencilFunc(GL_LESS, 0, 255);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	
	glClear(GL_DEPTH_BUFFER_BIT);
	
	GDrawingStage = 2;
	glPopMatrix(); // +-0
	mspCFilterCtrlObjectLink->GLDrawRealGrid
	    (doubleIndex, mCCameraPos, &mpCTagObjectCtrl->mVCObject);
	GetPLinkLightSourceCtrl()->SetBasicLight(doubleIndex, 1);
	glEnable(GL_LIGHTING);

	glPushMatrix(); // +1
	if (pCScene != NULL){
	    pCScene->UpdateCurrentObj(doubleIndex);
	    pCScene->GLSetMatrix();
	}

	DrawRealObjectCopyLoop(currentIndex, doubleIndex, lightNumber);
	if (mCbShowData.mbBody == TRUE){
	    DrawDataCopyLoop(fboIndex, currentIndex, //dT, 
			     lightNumber, pLightPosForSV, pLightPhiTheta, 
			     FALSE);
	}
	
	GDrawingStage = 0;
	glPopMatrix(); // +-0
    }
}

void MainGLCanvas_cls::  
DrawGrid(double scale, double scale10){
    if (mCbShowGrid.mbBody == FALSE) return;
    glColor4f(0.3f, 0.3f, 0.3f, 1.f);
    Grid_cls::DrawGridFlat(scale10);
}

void MainGLCanvas_cls::
DrawCameraPath(double doubleIndex, double scale, bool bHittest){
    if (mpCCameraPathLinked != NULL && 
	mCCameraPosCamera == mpCCameraPathLinked->mCCurrentObj) return;
    // FIX_ME!! IĂ\ׂ
    if (mCbShowItems.mbBody == FALSE) return; 

    glPushName(GetPLinkedCameraPath()->mObjIndex);
    GetPLinkedCameraPath()->GLDraw(bHittest);
    mspCStereogramLink->DrawFocus(GetPLinkedCameraPath()->mCCurrentObj,
				  doubleIndex, scale);
    glPopName();
    if (GetPLinkedAnchorPath()->mCbVisible.mbBody == TRUE){
	glPushName(GetPLinkedAnchorPath()->mObjIndex);
	GetPLinkedAnchorPath()->GLDraw(bHittest);
	glPopName();
    }
}

void MainGLCanvas_cls::
DrawLightSource(double doubleIndex, double scale, bool bHittest){
    if (mCbShowItems.mbBody == FALSE) return;
    GetPLinkLightSourceCtrl()->GLDraw(bHittest);
}

void MainGLCanvas_cls::
DrawFilters(double doubleIndex, double scale, bool bHittest){

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glEnable(GL_LIGHTING);
    glPopAttrib();

    if (mCbShowItems.mbBody == FALSE) return;

    mspCFilterCtrlBooleanLink->GLDraw(doubleIndex, scale, bHittest);
    mspCFilterCtrlObjectLink->GLDraw(doubleIndex, scale, bHittest);
    mspCFilterCtrlWorldLink->GLDraw(doubleIndex, scale, bHittest);
}

void MainGLCanvas_cls::
DrawAnchor(double scale){
    if (mCbShowItems.mbBody == FALSE) return;

    glColor4f(0.2f, 1.0f, 1.0f, 1.f);
    glPushMatrix(); // +1
    glTranslated(mCCameraPos.mAncPos[0],
		 mCCameraPos.mAncPos[1],
		 mCCameraPos.mAncPos[2]);
    glScaled(scale*0.02, scale*0.02, scale*0.02);
    glCallList(Camera_cls::mDListCrosshair);
    glPopMatrix(); // +-0
}

#define GL_GPU_MEM_INFO_TOTAL_AVAILABLE_MEM_NVX 0x9048
#define GL_GPU_MEM_INFO_CURRENT_AVAILABLE_MEM_NVX 0x9049

bool MainGLCanvas_cls::
DrawWarning(wxPaintDC& dc){
    if (FBO_cls::msbSupported == FALSE){
	dc.SetTextForeground(wxColour(212, 212, 212));
	wxString strMessage = wxT("Your OpenGL version is ");
	strMessage << mpCParentFrameLink->mpCGLInfo->GetGLVersionStr();
	strMessage << wxT("\nFBO is NOT supported. Plase upgrade your OpenGL environment.\nYou can not use most advanced openGL functions on remote desktop.");
	dc.DrawLabel(strMessage, wxNullBitmap, wxRect(0, 0, mPanelWidth, mPanelHeight), wxALIGN_CENTER | wxALIGN_CENTER_VERTICAL);
	return FALSE;
    }
    return TRUE;
}

void MainGLCanvas_cls::
CheckWarnigForDirectDraw(){
    if (mspCStereogramLink->GetStereoNum() == 2)
	mstrWarning << wxT("Direct Display Mode does not support stereogram.") << ENDL;
    if (IsFishEye() == TRUE ||IsCylinder() == TRUE)
	mstrWarning << wxT("Direct Display Mode does not support fisheye or panorama.") << ENDL;
    if (mpCParentFrameLink->mpCParent->mpCExposure->IsDefault() == FALSE)
	mstrWarning << wxT("Direct Display Mode does not support exposure.") << ENDL;
    if (mpCDrawOptionLink->mCShadow.mBody != 0)
	mstrWarning << wxT("Direct Display Mode does not support shadow volume.") << ENDL;
}

void MainGLCanvas_cls::
DrawMessage(){

    if (GLDrawData_cls::msstrWarning.IsEmpty() == FALSE){
	mstrWarning << GLDrawData_cls::msstrWarning;
	GLDrawData_cls::msstrWarning = wxEmptyString;
    }

    if (mCbShowMessage.mbBody == FALSE){
	mstrWarning.Clear();
	return; 
    }

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    glDisable(GL_DEPTH_TEST);
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    void* font = GLUT_BITMAP_HELVETICA_12;
    int index = GetPCTimeCtrl()->GetRemappedIndex();
        
    GLint total_mem_kb = 10;
    glGetIntegerv(GL_GPU_MEM_INFO_TOTAL_AVAILABLE_MEM_NVX, 
		  &total_mem_kb);
    
    GLint cur_avail_mem_kb = 10;
    glGetIntegerv(GL_GPU_MEM_INFO_CURRENT_AVAILABLE_MEM_NVX, 
		  &cur_avail_mem_kb);
    
    wxString strTime;
    int particleNum = mpCParentFrameLink->GetPCActiveFileSequence()->GetParticleNumber(index);
    int copyNum = 1;
    if (mspCFilterCtrlWorldLink->GetPCCopy() != NULL)
	copyNum = mspCFilterCtrlWorldLink->GetPCCopy()->GetNumber();
    strTime << wxT("Time=");
    strTime << mpCParentFrameLink->GetPCActiveFileSequence()->GetTime(index);
    strTime << wxT(":PNum=") << particleNum
	    << wxT("*") << copyNum;
    if (copyNum != 1)
	strTime << wxT("=") << particleNum * copyNum;
    
    strTime << wxT(":MilliSec") << mCStopWatchSnap.Time()
	    << wxT(":Hitting")  << msCHittest.GetFirstObject()
	    << wxT("+")         << msCHittest.GetFirstObjectSub() 
	    << wxT(":Object")   << (long long)MoveItem_cls::mspCSelectedObject
	    << wxT("pointXY=") << msCHittest.mPointX << wxT(",") << msCHittest.mPointY
	;
    
    mpCGLString2DDirect->Draw(font, strTime.ToAscii(), 5, 10, _STRINGS_TOP_LEFT);

    char pBuf[128]; sprintf(pBuf, "GMem0 = %d GMem1 = %d Count = %d UDBlock = %d",
			    total_mem_kb, cur_avail_mem_kb, mCounter, GCUndoList.mBlockDepth);
    mCounter++;

    mpCGLString2DDirect->Draw(font, pBuf, 5, 28, _STRINGS_TOP_LEFT);

    /*
    wxString strUndo = GCUndoList.GetStrList();
    mpCGLString2D->Draw(font, strUndo.ToAscii(), 5, 46, _STRINGS_TOP_LEFT);
    */

    glColor4f(1.0f, 0.5f, 0.5f, 1.0f);
    mpCGLString2DDirect->Draw(font, mstrWarning.ToAscii(), 100, 10, _STRINGS_BOTTOM_LEFT);
    
    glPopAttrib();
}

void MainGLCanvas_cls::
DrawModeAxis(double scale){
}

void MainGLCanvas_cls::
DrawSelectedItem(double scale){

    if (MoveItem_cls::mspCSelectedObject == NULL) return;

    /*
    if (mpCMoveItem->mbMovingItemNow == FALSE) return;
    glColor4f(1.0, 1.0, 0.2, 1.0);
    bool bSelectKey = FALSE;
    
    LOOP_XYZ MoveItem_cls::mspCSelectedObject->mCCurrentObj.mPos[xyz] = MoveItem_cls::msSelectedItemPos[xyz];
    */

    if (mCbShowNavigator.mbBody == FALSE) return; 

    SetLineWidth();
    glDisable(GL_DEPTH_TEST);
    //MoveItem_cls::mspCSelectedObject->GLDrawGuide();
    //MoveItem_cls::mspCSelectedObject->mbDragging = TRUE;
    
    if (mpCParentFrameLink->mpCPreference->mCUseFBO.mBody == 1)
	mpCMoveItem->DrawMouseGuide(mpCNavigators->msManipulatorMode, 
				    mpCNavigators->msDragMode,
				    mPanelWidth, mPanelHeight,
				    mspCFishEyeCtrl->mBufferSize,
				    mspCFishEyeCtrl->mBufferSize);
    else
	mpCMoveItem->DrawMouseGuide(mpCNavigators->msManipulatorMode, 
				    mpCNavigators->msDragMode,
				    mPanelWidth, mPanelHeight,
				    mPanelWidth, mPanelHeight);
	
    glEnable(GL_DEPTH_TEST);
}


void MainGLCanvas_cls::
RefreshEX(){
    // QuadBufferPendEventRefreshȂB
    if (mspCStereogramLink->mbQuadStereo == TRUE)
	Refresh(FALSE);
    else{
	wxPaintEvent com(GetId());
	AddPendingEvent(com);
    }

    if (mSelfIndex == 0){
	MainGLCanvas_cls *pCExt = mpCParentFrameLink->GetPCGLCanvas(4);
	if (pCExt->IsShownOnScreen() == FALSE) return;
	if (pCExt->mCDisplayMode.mBody != _GL_CANVAS_COPY_OF_VIEW_0) return;
	wxPaintEvent com(pCExt->GetId());
	pCExt->AddPendingEvent(com);
	//pCExt->Refresh(FALSE);
	//SendMenuEvent(wxGetTopLevelParent(this), menuEditRefreshGLExt);
    }
}

bool MainGLCanvas_cls::
IsMouseOnClient(){
    wxRect  CScreenRect = GetScreenRect();
    wxPoint CMousePos   = wxGetMousePosition();
    return CScreenRect.Contains(CMousePos);
}

void MainGLCanvas_cls::
ShowPreviousFramePBO(){
}

void MainGLCanvas_cls::
DrawOtherObjects(double doubleIndex, double scale, double scale10){
    glPointSize(6);
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glDepthMask(GL_TRUE);

    glUseProgram(0);

    DrawGrid(scale, scale10);
    DrawCameraPath(doubleIndex, scale);
    DrawLightSource(doubleIndex, scale);
    DrawFilters(doubleIndex, scale);
    //glDisable(GL_LIGHTING);
    mspCFilterCtrlObjectLink->GLDrawRealGrid
	(doubleIndex, mCCameraPos, &mpCTagObjectCtrl->mVCObject);
    
    //glEnable(GL_LIGHTING);

    if (mbHitTestItemSelect == TRUE || mbHitTestGuideSelect == TRUE){

	msCHittest.StartHitTest((float)mpCMouseCtrl->mMouseX/mPanelWidth,
				(float)mpCMouseCtrl->mMouseY/mPanelHeight,
				(float)mPanelWidth/mPanelHeight,
				this, &mCCameraPos, TRUE);
	glInitNames();
	glDisable(GL_DEPTH_TEST);
	if (msCHittest.mbHitTest == FALSE)
	    DrawGrid(scale, scale10);
	if (mbHitTestItemSelect == TRUE){
	    DrawLightSource(doubleIndex, scale, TRUE);
	    DrawCameraPath(doubleIndex, scale, TRUE);
	    DrawFilters(doubleIndex, scale, TRUE);
	}
	if (mbHitTestGuideSelect == TRUE)
	    DrawSelectedItem(scale);
	msCHittest.EndHitTest();
	
	if (mbHitTestItemSelect == TRUE){
	    int itemName = msCHittest.GetFirstObject();
	    if (itemName != -1){
		if (itemName < 20){ // IuWFNgqbg
		    mpCParentFrameLink->mpCParent->SelectObject(itemName);
		    mpCParentFrameLink->mpCParent->SelectFilter(NULL, FALSE);
		    mbRefreshQueue = TRUE;
		}
		else{
		    mpCParentFrameLink->mpCParent->
			SelectFilterByHittest(itemName, TRUE);
		}
	    }
	}

	if (mbHitTestGuideSelect == TRUE){
	    if (msCHittest.GetFirstObjectSub() != -1)
		mpCNavigators->SetDragMode(msCHittest.GetFirstObjectSub());
	}
	
	mbHitTestItemSelect  = FALSE;
	mbHitTestGuideSelect = FALSE;
    }
    
    glDisable(GL_DEPTH_TEST);
    DrawAnchor(scale);
}

void MainGLCanvas_cls::
RefreshOtherOrthoCanvas(){
    for (unsigned int i = 0; i < mVPOtherCanvas.size(); i++){
	if (mVPOtherCanvas[i]->IsAlongAxis() == TRUE)
	    mVPOtherCanvas[i]->Refresh(FALSE);
    }
}

bool MainGLCanvas_cls::
CheckGLRefresh(int frame, int interpStep){
    if (IsShownOnScreen() == FALSE) return TRUE;
    if (frame != mPrevFrame || interpStep != mPrevStep){
	Refresh(FALSE);
	return FALSE;
    }
    return TRUE;
}

void MainGLCanvas_cls::
DrawOnGL(int stereoEye, double doubleIndex, double doubleRemapped){
    double scale = CalcAutoScale();
    double scale10 = 0.0001;
    while (scale10*10 < scale) scale10 *= 10;

    PositionAndDirection_cls StereoCamera = mCCameraPos;
    mspCStereogramLink->mpCStereoParam->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
    mspCStereogramLink->ModifyByStereogram(&StereoCamera, stereoEye);
    
    if (IsFishEye() == TRUE && mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	ModifyByDomeTilt(&StereoCamera);

    glPushAttrib(GL_ALL_ATTRIB_BITS);

    ClearBuffer(doubleIndex);
    std::vector < FilterBG_cls* > VpCBG = mspCFilterCtrlWorldLink->GetVpCBG();
    for (int i = 0; i < (int)VpCBG.size(); i++) VpCBG[i]->GLDraw(doubleIndex);

    mCGLPerspective.SetWindowSize(mPanelWidth, mPanelHeight);
    mCGLPerspective.SetViewportOnGL();
    mCGLPerspective.SetPerspectiveOnGL(&StereoCamera, FALSE);
    mCGLPerspective.SetModelOrProjectionViewOnGL(&StereoCamera, GL_MODELVIEW);
    mpCMoveItem->GetGLMatrixInfo();
    //mspCGLGuideYPR->GetGLMatrixInfo();

    GetPLinkLightSourceCtrl()->SetNonShadowLight(doubleIndex);

    SetLineWidth();
    DrawDataMain(doubleIndex, doubleRemapped, 0, 0);

    SetLineWidth();
    DrawOtherObjects(doubleIndex, scale, scale10);

    glDisable(GL_LIGHTING);
    glColor4f(1, 1, 1, 1);
    mpCTagObjectCtrl->CalcTagCoordinate();
    mpCTagObjectCtrl->DrawTags(mpCGLString2DBuffer);

    DrawSelectedItem(scale);
    
    glPopAttrib();
}

int MainGLCanvas_cls::
GetFBOBufferSize(int type){
    int bufferSize = msAFBOSize[msCFBOSizeIndex.mBody];
    if (type == 1) // 0: p[c 1: pobt@
	if (GetMultiPanelMode() != 0) bufferSize = 4096;
    return bufferSize;
}

void MainGLCanvas_cls::
DrawOnFBO(int stereoEye, double doubleIndex, double doubleRemapped, 
	  int bufferSize){
    
    double scale = CalcAutoScale();
    double scale10 = 0.0001;
    while (scale10*10 < scale) scale10 *= 10;
    
    int fboMax = GetFBOMax();
    
    int shadowNumber = 
	GetPLinkLightSourceCtrl()->GetNumberLightWithShadow();
    if (mpCDrawOptionLink->mCShadow.mBody == FALSE)
	shadowNumber = 0;
    
    if (shadowNumber > 1)
	mpCFBOForShadow->MakeFBO();
    
    int orgDeg = mCGLPerspective.mDeg;
    for (int fboIdx = 0; fboIdx < fboMax; fboIdx++){
	FBO_cls *pCFBO = mspCFishEyeCtrl->mpCFBOParts[fboIdx];

	PositionAndDirection_cls StereoCamera = mCCameraPos;
	mspCStereogramLink->mpCStereoParam->GetInterpolatedValue(doubleIndex, _PARAM_USE_COMMON);
	mspCStereogramLink->ModifyByStereogram(&StereoCamera, stereoEye);

	if (IsFishEye() == TRUE && mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	    ModifyByDomeTilt(&StereoCamera);

	mspCFishEyeCtrl->MakeFishEyeProcessFH
	    (fboIdx, StereoCamera, mCGLPerspective, GetMultiPanelMode());
	mpCMoveItem->GetGLMatrixInfo();
	//mspCGLGuideYPR->GetGLMatrixInfo();
	
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	ClearBuffer(doubleIndex);
	std::vector < FilterBG_cls* > VpCBG =
	    mspCFilterCtrlWorldLink->GetVpCBG();
	for (int i=0; i < (int)VpCBG.size(); i++) VpCBG[i]->GLDraw(doubleIndex);
	
	if (shadowNumber == 0 || shadowNumber == 1)
	    GetPLinkLightSourceCtrl()->SetNonShadowLight(doubleIndex);
	else
	    GetPLinkLightSourceCtrl()->SetBasicLight
		(doubleIndex, mpCDrawOptionLink->mCShadow.mBody);
	
	glClear(GL_DEPTH_BUFFER_BIT);
	SetLineWidth(pCFBO);
	DrawDataMain(doubleIndex, doubleRemapped, fboIdx, shadowNumber);

	SetLineWidth(pCFBO);
	DrawOtherObjects(doubleIndex, scale, scale10);

	glDisable(GL_LIGHTING);
	glColor4f(1, 1, 1, 1);
	mpCTagObjectCtrl->CalcTagCoordinate();
	mpCTagObjectCtrl->DrawTags(mpCGLString2DBuffer);
	
	DrawSelectedItem(scale);
	glPopAttrib();
	
	mspCFishEyeCtrl->MakeFishEyeProcessLH(fboIdx);
	
	if (shadowNumber > 1) {
	    for (int i = 0; i < GetPLinkLightSourceCtrl()->mCLightNum.mBody; i++){
		if (GetPLinkLightSourceCtrl()->mpCLightSRC[i]->
		    mCbCastShadow.mbBody == FALSE) continue;
		
		mpCFBOForShadow->Begin(TRUE);
		
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		SetLineWidth(pCFBO);
		DrawDataLitDiffuse(doubleIndex, doubleRemapped, fboIdx, i);
		glPopAttrib();
		
		SetLineWidth(pCFBO);
		DrawSelectedItem(scale);
		mpCTagObjectCtrl->DrawTags(mpCGLString2DBuffer);
		mpCFBOForShadow->End();
		
		pCFBO->Begin(FALSE);
		glMatrixMode(GL_PROJECTION);
		glPushMatrix(); // +1
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();	// +2    
		
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glUseProgram(0);
		//glEnable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		mpCFBOForShadow->DrawOnScreen2DFit(bufferSize, bufferSize, _TEX_FIT_VERTICAL, TRUE);
		glPopAttrib();
		
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix(); // +1
		glMatrixMode(GL_PROJECTION);
		glPopMatrix(); // +-0
		
		pCFBO->End();
	    }
	}
    }
}

void MainGLCanvas_cls::
GatherFBOMultiPanel(int width, int height, int stereo){
    
    bool bAdd = FALSE;
    
    glClear(GL_STENCIL_BUFFER_BIT);
    if (GetMultiPanelMode() == _MULTIPANEL_FISHEYE){
	
	int bufferSize = GetFBOBufferSize(1);
	if (mpCFBOForFishEyeBuffer == NULL)
	    mpCFBOForFishEyeBuffer = new FBO_cls(bufferSize, bufferSize);
	else
	    mpCFBOForFishEyeBuffer->SetGeometryBuffer(bufferSize, bufferSize);
	int fboDepth = mpCParentFrameLink->mpCPreference->GetFBODepth();
	mpCFBOForFishEyeBuffer->SetDepth(fboDepth);
	
	mpCFBOForFishEyeBuffer->MakeFBO();
	mpCFBOForFishEyeBuffer->Begin();
	SetGLBGColor(1);
	glEnable(GL_TEXTURE_2D);
	mspCFishEyeCtrl->GatherFishEye
	    (mCGLPerspective.mDeg, bAdd, bufferSize, bufferSize,
	     mspCStereogramLink, stereo);
	mpCFBOForFishEyeBuffer->End();

	glEnable(GL_BLEND);

	if (mCbFisheyeGrid.mbBody == TRUE){
	    mpCFBOForFishEyeBuffer->Begin(FALSE);
	    glBlendColor(0.5, 0.25, 0.25, 1);
	    glBlendFunc(GL_CONSTANT_COLOR, GL_ONE_MINUS_SRC_COLOR);
	    mspCFisheyeGrid->DrawOnScreen2DFit(bufferSize, bufferSize, _TEX_FIT_CORNER, TRUE);
	    glBlendColor(1, 1, 1, 1);
	    mpCFBOForFishEyeBuffer->End();
	}
	
	if (mCbDistortion.mbBody == TRUE){
	    mpCFBOForFishEyeBuffer->Begin(FALSE);
	    glBlendFunc(GL_ZERO, GL_SRC_COLOR);
	    mspCMask->DrawOnScreen2DFit(bufferSize, bufferSize, _TEX_FIT_VERTICAL, TRUE);
	    mpCFBOForFishEyeBuffer->End();
	}
	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	
	mpCParentFrameLink->mpCParent->mpCExposure->Begin();

	if (mCbDistortion.mbBody == TRUE){
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    glColor4f(1, 1, 1, 1);
	    mpCFBOForFishEyeBuffer->BindTexture();
	    mspCDistortion->SetGLSetting(width, height);
	    mspCDistortion->GLDrawDistortedTexture((double)width / height, 1, FALSE, 1);
	    mspCStereogramLink->ColorMaskEnd();
	}
	else{
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    glColor4f(1, 1, 1, 1);
	    mspCStereogramLink->ColorMaskBegin(stereo);
	    mpCFBOForFishEyeBuffer->DrawOnScreen2DFit(width, height, _TEX_FIT_VERTICAL, TRUE);
	    mspCStereogramLink->ColorMaskEnd();
	}

	mpCParentFrameLink->mpCParent->mpCExposure->End();
	
	glBindTexture(GL_TEXTURE_2D, 0);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
    }
    
    if (GetMultiPanelMode() == _MULTIPANEL_CYLINDER){
	mpCParentFrameLink->mpCParent->mpCExposure->Begin();
	mspCFishEyeCtrl->GatherCylinder
	    (mCGLPerspective.mDeg, 
	     bAdd, width, height, mspCStereogramLink, stereo);
	mpCParentFrameLink->mpCParent->mpCExposure->End();
    }
}

void MainGLCanvas_cls::
GatherFBOImages(int width, int height, int stereo){
    
    SetGLBGColor(1);

    if (GetMultiPanelMode() != 0){
	GatherFBOMultiPanel(width, height, stereo);
	return;
    }
    
    // ȉSinglePanel
    FBO_cls *pCFBO = mspCFishEyeCtrl->mpCFBOParts[0];
    
    mpCParentFrameLink->mpCParent->mpCExposure->Begin();
    mspCStereogramLink->ColorMaskBegin(stereo);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    pCFBO->DrawOnScreen2DFit(width, height, TRUE, TRUE);
    mspCStereogramLink->ColorMaskEnd();
    mpCParentFrameLink->mpCParent->mpCExposure->End();
}

void MainGLCanvas_cls::
BakeResultToFBOForImage(int stereo){
    
    double alpha = 1.0;
    
    if (GetPCTimeCtrl()->mCBlurCtrl.mbUseMotionBlur == TRUE) 
	alpha = pow((double)GetPCTimeCtrl()->mCBlurCtrl.GetAlpha(), 0.5);
    
    int finalBufferSize = mspCStereogramLink->mBufferSize;

    mpCParentFrameLink->mpCParent->mpCExposure->mAlpha = alpha;
    mpCParentFrameLink->mpCParent->mpCExposure->mBlendMode = 1;
    mspCStereogramLink->DrawOnFBOBegin(stereo);
    GatherFBOImages(finalBufferSize, finalBufferSize, stereo);
    mspCStereogramLink->DrawOnFBOEnd(stereo);
    mpCParentFrameLink->mpCParent->mpCExposure->mAlpha = 1;
    mpCParentFrameLink->mpCParent->mpCExposure->mBlendMode = 0;
}

void MainGLCanvas_cls::
FBODraw(double doubleIndex, double doubleRemapped, int stereo){

    int bufferSize = GetFBOBufferSize(0);
    int fboDepth = mpCParentFrameLink->mpCPreference->GetFBODepth();
    mpCGLString2DBuffer->SetWindowGeometry(bufferSize, bufferSize);
    mspCFishEyeCtrl->MakeFishEyePrepare
	(mCGLPerspective.mClipWidth, mCGLPerspective.mClipHeight, bufferSize, 
	fboDepth, GetFBOMax());
    
    if (mpCFBOForShadow == NULL)
	mpCFBOForShadow = new FBO_cls(bufferSize, bufferSize);
    else
	mpCFBOForShadow->SetGeometryBuffer(bufferSize, bufferSize);
    mpCFBOForShadow->SetDepth(fboDepth);
    
    int orgDeg = mCGLPerspective.mDeg;
    int stereoEye = mspCStereogramLink->GetStereoEye(stereo);
    DrawOnFBO(stereoEye, doubleIndex, doubleRemapped, bufferSize);
    
    // ʌɖ߂
    mCGLPerspective.SetModelOrProjectionViewOnGL(&mCCameraPos, GL_MODELVIEW);
    //SetModelView(&mCCameraPos, FALSE); 
    mCGLPerspective.mDeg = orgDeg;

    BakeResultToFBOForImage(stereo);
}

void MainGLCanvas_cls::
DirectDraw(double doubleIndex, double doubleRemapped){
    
    int orgDeg = mCGLPerspective.mDeg;
    int stereoEye = mspCStereogramLink->GetStereoEye(0);

    DrawOnGL(stereoEye, doubleIndex, doubleRemapped);
}

void MainGLCanvas_cls::
PrepareBaseObjectSettings(){
    double doubleIndex = GetPCTimeCtrl()->GetDoubleIndex();
    
    // ԂωƁA mCPosCameraȂǂǐ
    if (doubleIndex != mDoubleIndexPrev){
	/*
	mpCSceneObjCtrlLink->GetCameraPathPos(mCPosCamera);
	mpCSceneObjCtrlLink->GetAnchorPathPos(mCPosAnchor);
	*/
	mpCMoveItem->InitializeMoveItem();
	mDoubleIndexPrev = doubleIndex;
    }
}

bool MainGLCanvas_cls::
GLDrawMain(){
    WaitPrepaing();
   
    mDrawingFrame = GetPCTimeCtrl()->mFrame;
    mDrawingStep  = GetPCTimeCtrl()->mInterpStep;
    mDrawingBlurFactor      = GetPCTimeCtrl()->mCBlurCtrl.GetFactor();
    double doubleIndex      = GetPCTimeCtrl()->GetDoubleIndex();
    double doubleRemapped   = GetPCTimeCtrl()->GetRemappedIndexDouble();
    int    currentIndex     = GetPCTimeCtrl()->GetRemappedIndex();
    double doubleInterpStep = GetPCTimeCtrl()->GetRemappedInterpStepDouble();

    double time, dT;
    mpCParentFrameLink->GetPCActiveFileSequence()->
	GetTimeInfo(time, dT, currentIndex, doubleInterpStep);
        
    GLDrawData_cls::msCurrentDoubleIndex = doubleIndex;
    mpCTagObjectCtrl->mVCObject.clear();
    
    if (SetGLContext() == FALSE) return FALSE; 

    InitializeGLIfNot();
    //if (InitializeGLIfNot() == TRUE) return FALSE;

    //SetGLBGColor(1);
    mspCStereogramLink->DrawFunctionOnScreen
	(this, &MainGLCanvas_cls::ClearCanvas);
        
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    if (mCDisplayMode.mBody == _GL_CANVAS_COPY_OF_VIEW_0){
	mCCameraPos = mpCParentFrameLink->GetPCGLCanvas(0)->mCCameraPos;
	mCCameraPosFree = mpCParentFrameLink->GetPCGLCanvas(0)->mCCameraPosFree;
    }

    SetPerspective(&mCCameraPos);
    mCGLPerspective.mDepthMIN = mCGLDepthMin.mBody;
    mCGLPerspective.mDepthMAX = mCGLDepthMax.mBody;

    PrepareBaseObjectSettings();
    
    // J
    GetPLinkLightSourceCtrl()->UpdateCurrentObj(doubleIndex);
    GetPLinkedCameraPath()->UpdateCurrentObj(doubleIndex);
    GetPLinkedAnchorPath()->UpdateCurrentObj(doubleIndex);
    UpdateCameraPos();

    mpCNavigators->MakeImage(mCCameraPos);

    mPrevFrame = mDrawingFrame;
    mPrevStep  = mDrawingStep;
    mPrevBlurFactor = mDrawingBlurFactor;
    
    SetAngleIfLinked();
    
    UpdateDrawInfo(mPanelWidth, mPanelHeight);
    int InfoDiff = mCDrawInfo.GetDifference(mCDrawInfoPrev);

    bool bUseFBO = (mpCParentFrameLink->mpCPreference->mCUseFBO.mBody != _PREF_NO_FBO);
    
    if (FALSE || mbExportDataQueue == TRUE){
	ShowPreviousFramePBO();
    }
    else{
	if (GetPCTimeCtrl()->mCBlurCtrl.IsInitBluring() == TRUE) 
	    mspCStereogramLink->ClearFBO();
	
	for (int stereo = 0; stereo < mspCStereogramLink->GetStereoNum(); stereo++){
	    mspCStereogramLink->mBufferSize = GetFBOBufferSize(1);
	    mspCStereogramLink->PrepareStereogram(stereo);
	    
	    if (bUseFBO == TRUE)
		FBODraw(doubleIndex, doubleRemapped, stereo); 
	    else{
		DirectDraw(doubleIndex, doubleRemapped); 
		CheckWarnigForDirectDraw();
		stereo = mspCStereogramLink->GetStereoNum();
	    }
	}
	
	if (bUseFBO == TRUE){
	    mspCStereogramLink->mbFitFishEye = 
		(IsFishEye() == TRUE) ? TRUE : FALSE;
	    mspCStereogramLink->DrawOnScreen2D(mCbDistortion.mbBody, this);
	}
    }
    return TRUE;
}

void MainGLCanvas_cls::
DrawOverLay(){
    int    currentIndex     = GetPCTimeCtrl()->GetRemappedIndex();
    double doubleInterpStep = GetPCTimeCtrl()->GetRemappedInterpStepDouble();
    PDataSequence_cls *pCPDataSeq = 
	mpCParentFrameLink->GetPCActiveFileSequence();
    double time, dT;
    pCPDataSeq->GetTimeInfo(time, dT, currentIndex, doubleInterpStep);
    
    ColorCurve_cls *pAClrCurve[4];
    for (int i = 0; i < 4; i++){
	if (pCPDataSeq->mpCGLAttribute[i]->mCDrawType.mBody == _GL_DRAW_SPHERE)
	    pAClrCurve[i] = pCPDataSeq->mpCGLAttribute[i]->mpCClrCurveMat[0];
	else
	    pAClrCurve[i] = pCPDataSeq->mpCGLAttribute[i]->mpCClrCurveCol;
    };

    if (mbSaveImageQueue == TRUE){
	if ( mspCImgSeqLinked != NULL)
	    mCGLPerspective.SetClipSize(mspCImgSeqLinked->mClipWidth, 
					mspCImgSeqLinked->mClipHeight);
	
	if (mpCFBOForOverlay == NULL){
	    mpCFBOForOverlay = new FBO_cls(mCGLPerspective.mClipWidth,
					   mCGLPerspective.mClipHeight);
	    mpCFBOForOverlay->SetDepth(8);
	}
	else
	    mpCFBOForOverlay->SetGeometryBuffer(mCGLPerspective.mClipWidth,
						mCGLPerspective.mClipHeight);
	    
	mpCFBOForOverlay->MakeFBO(_FBO_USE_RGBA);
	mpCFBOForOverlay->Begin(TRUE);
	mspCDrawTimeIndex->DrawTimeInfo(time + dT, 
					mCGLPerspective.mClipWidth,
					mCGLPerspective.mClipHeight);
	mspCDrawClrLegend->DrawClrLegend(pAClrCurve, 
					 mCGLPerspective.mClipWidth, 
					 mCGLPerspective.mClipHeight);
	mpCFBOForOverlay->End();
    }
    
    mspCDrawTimeIndex->DrawTimeInfo(time + dT, mPanelWidth, mPanelHeight);
    mspCDrawClrLegend->DrawClrLegend(pAClrCurve, mPanelWidth, mPanelHeight);
}

void MainGLCanvas_cls::
ProcessImageSave(){
    if (mbSaveImageQueue == TRUE){
	if (GetPCTimeCtrl()->mCBlurCtrl.IsOutputConditionOK() == TRUE){
	    if (IsStereoDualOutput() == FALSE){
		mspCStereogramLink->CombineStereoBuffer();
		if (mspCStereogramLink->mCType.mBody == _STEREO_NONE)
		    ProcReadImage(0, mpCFBOForOverlay, _STEREO_NONE);
		else
		    ProcReadImage(0, mpCFBOForOverlay, mspCStereogramLink->mCEye.mBody);
	    }
	    else{
		ProcReadImage(0, mpCFBOForOverlay, _STEREO_LFT);
		ProcReadImage(1, mpCFBOForOverlay, _STEREO_RGT);
	    }
	}
    }
}

void MainGLCanvas_cls::
DrawNavigatorAndButtons(){
    if (mCbShowNavigator.mbBody == FALSE) return;
    mCGLPerspective.SetWindowSize(mPanelWidth, mPanelHeight);
    SetViewport();

    glDisable(GL_LIGHTING);
    glDisable(GL_BLEND);
    
    mpCNavigators->Show(mPanelWidth, mPanelHeight);
    if (GpCurrentGLPanel == this)
	mpCNavigators->DrawButton(mPanelWidth, mPanelHeight);

    if (msCHittest.mbHitTest == FALSE)
	msCHittest.mpCFBO->DrawOnScreen2D
	    (mPanelWidth, mPanelHeight, 64, 64, 96, 16);
}

void MainGLCanvas_cls::
PaintDirectly(wxPaintDC &dc){
    if (DrawWarning(dc) == FALSE) return;

    GbGLLockDrawing = TRUE;
    mCStopWatchSnap.Start();
    
    // ʃTCY֌W
    int width  = GetClientSize().GetWidth();
    int height = GetClientSize().GetHeight();
    mPanelWidth = width;
    mPanelHeight = height;

    mspCStereogramLink->SetInfo(width, height, TRUE);
    mpCGLString2DDirect->SetWindowGeometry(width, height);
    if (mspCImgSeqLinked != NULL)
	mspCImgSeqLinked->SetGeom(width, height, mCGLPerspective.mDeg);

    if (mbExportDataQueue == TRUE)
	mpCParentFrameLink->GetPCActiveFileSequence()->ExportData
	    (mpCParentFrameLink->mpCParent->mpCDataExporter, 
	     GetPCTimeCtrl()->GetRemappedIndex(),
	     mpCDrawOptionLink->mCDownSize.mBody);

    if (GLDrawMain() == TRUE){
	mspCStereogramLink->DrawFunctionOnScreen(this, &MainGLCanvas_cls::DrawOverLay);
	//DrawOverLay();
	ProcessImageSave();
	mspCStereogramLink->DrawFunctionOnScreen(this, &MainGLCanvas_cls::DrawNavigatorAndButtons);
	mspCStereogramLink->DrawFunctionOnScreen(this, &MainGLCanvas_cls::DrawMessage);
	mstrWarning.Clear();
	
	if (GetPCTimeCtrl()->mCBlurCtrl.IsLastBluring() == TRUE){
	    if (mbExportDataQueue != TRUE) {
		SwapBuffers(); 
	    }
	    mspCStereogramLink->ClearFBO();
	}
    }
    
    if (GpCurrentGLPanel == this){
	SendMenuEvent(mpCParentFrameLink, menuEditRefreshCamConf);
	SendMenuEvent(mpCParentFrameLink, menuEditRefreshLight);
	SendMenuEvent(mpCParentFrameLink, menuEditRefreshStereo);
    }

    if (mbRefreshQueue == TRUE &&
	mpCParentFrameLink->mpCParent->mpCTime->mbPlayTex == FALSE){
	mbRefreshQueue = FALSE;
	
	if (mCStopWatchSnap.Time() < 1000)
	    SendMenuEvent(mpCParentFrameLink, menuEditRefreshGL);
    }

    mCStopWatchSnap.Pause();
}

void MainGLCanvas_cls::
OnPaint(wxPaintEvent &event){

    if (IsDrawingConditionSatisfied() == FALSE) return;
    
    wxPaintDC dc(this); 

    PaintDirectly(dc);

    if (mbExportPovStringQueue == TRUE)
	mpCParentFrameLink->mpCParent->ExportPovAnim(0); //stereo

    FinishPaint(event);
}

void MainGLCanvas_cls::
FinishPaint(wxPaintEvent& event){
    GbGLLockDrawing = FALSE;
    GbGLLockPlayWait = FALSE;
    event.Skip();
}

void MainGLCanvas_cls::
ProcReadImage(int fbo, FBO_cls *pCFBOOverlay, 
	      int stereo){
    glReadBuffer( GL_BACK );

    // FBOۑ
    if (mbSaveImgSeqQueue == TRUE && mspCImgSeqLinked != NULL)
	return ProcReadImageSequence(fbo, pCFBOOverlay, stereo);
	
    // GLʂۑ
    mbSaveImageQueue = FALSE;
    if (GCImgIO.SelectFile
	(mpCParentFrameLink->mCLastOpenedFileIMG.mstrBody, 
	 mpCParentFrameLink) == TRUE){
	mpCParentFrameLink->mCLastOpenedFileIMG.mstrBody = GCImgIO.mstrFile;
	// !! FIX_ME !! XeIooɕۑꍇ́H
	if (IsStereo() == TRUE)
	    GCImgIO.SaveImageFromGL(mPanelWidth, mPanelHeight); 
	else
	    GCImgIO.SaveImageFromGL(mPanelWidth, mPanelHeight);
    }
}

// @in plwidthheight
void MainGLCanvas_cls::
ProcReadImageSequence(int fbo,FBO_cls *pCFBOOverlay, int stereo){
    int currentFrame = mDrawingFrame;
    int offset = GetPCTimeCtrl()->GetOffsetForImgSeq();
    
    bool bSideBySide = mspCStereogramLink->IsSideBySideForOverlayRequired();

    FBO_cls *pCFBO = mspCStereogramLink->mpCFBO[fbo];
    if (pCFBO != NULL)
	mspCImgSeqLinked->SaveImageFromFBO
	    (&GCImgIO, pCFBO, pCFBOOverlay, mPanelWidth, mPanelHeight, 
	     offset, stereo, bSideBySide);
    
    if (mspCImgSeqLinked->mbSnapMode == TRUE){
	mspCImgSeqLinked->mbSnapMode = FALSE;
	SendMenuEvent(mpCParentFrameLink, menuPlayStop);
    }
    
    // s̃[`ɗv
    if (mpCParentFrameLink->mpCPreference->mCUseFBO.mBody == FALSE){ 
	mstrWarning << wxT("Direct Display Mode does not support sequential image saving.") << ENDL;
	mbSaveImgSeqQueue = FALSE;
	mbSaveImageQueue = FALSE;
	mspCImgSeqLinked->mpCGUI->AbortRender();
    }
}

void MainGLCanvas_cls::
StartSaveImgSeqProcess(ImgSequence_cls* pCImgSeq){
    mspCImgSeqLinked = pCImgSeq;
    mbSaveImageQueue = TRUE;
    mbSaveImgSeqQueue = TRUE;
    if (mspCImgSeqLinked != NULL)
	mspCImgSeqLinked->mCurrentOffset = 0;
    //mspCStereogramLink->ClearFBO();
}

void MainGLCanvas_cls::
StopSaveImgSeqProcess(){
    mbSaveImageQueue = FALSE;
    mbSaveImgSeqQueue = FALSE;
    if (mspCImgSeqLinked != NULL)
	mspCImgSeqLinked->OutputFinish();
}

void MainGLCanvas_cls::
StartExportDataProcess(DataExporter_cls* pCDataExporter){
    mpCDataExporterLinked = pCDataExporter;
    mbExportDataQueue = TRUE;
    if (mpCDataExporterLinked != NULL){
	mpCDataExporterLinked->mCurrentOffset = 0;
	mpCDataExporterLinked->mCurrentIndex = -1;
    }
}

void MainGLCanvas_cls::
StopExportDataProcess(){
    mbExportDataQueue = FALSE;
    if (mpCDataExporterLinked != NULL)
	mpCDataExporterLinked->ExportFinish();
}

void MainGLCanvas_cls::
StartExportPovStringProcess(){
    mbExportPovStringQueue = TRUE;
}

void MainGLCanvas_cls::
StopExportPovStringProcess(){
    mbExportPovStringQueue = FALSE;
    GetPCPovRayCtrlLinked()->AbortExportPovAnim();
}

void MainGLCanvas_cls::
ModifyOrthoCamera(){
    LOOP_XYZ
	mCCameraPosOrtho.mPos[xyz] = msCAnchorPosOrtho.mPos[xyz];
    switch (mCDisplayMode.mBody){
    case _GL_CANVAS_DISPLAY_X_PLUS:
	mCCameraPosOrtho.mPos[X] -= mCGLOrthoScale.mBody;
	break;
    case _GL_CANVAS_DISPLAY_X_MINUS:
	mCCameraPosOrtho.mPos[X] += mCGLOrthoScale.mBody;
	break;
    case _GL_CANVAS_DISPLAY_Y_PLUS:
	mCCameraPosOrtho.mPos[Y] -= mCGLOrthoScale.mBody;
	break;
    case _GL_CANVAS_DISPLAY_Y_MINUS:
	mCCameraPosOrtho.mPos[Y] += mCGLOrthoScale.mBody;
	break;
    case _GL_CANVAS_DISPLAY_Z_PLUS:
	mCCameraPosOrtho.mPos[Z] -= mCGLOrthoScale.mBody;
	break;
    case _GL_CANVAS_DISPLAY_Z_MINUS:
	mCCameraPosOrtho.mPos[Z] += mCGLOrthoScale.mBody;
	break;
    }
    mCCameraPosOrtho.SetAncPos(msCAnchorPosOrtho.mPos);
    mCCameraPosOrtho.SetDirection(mCDisplayMode.mBody);
}

double MainGLCanvas_cls::
CalcAutoScale(){
    double scale = MyVector::GetNormRelative(mCCameraPos.mPos,
					     msCAnchorPos.mPos);
    return scale;
}

void MainGLCanvas_cls::
ShowMenu(int x, int y){
#ifdef WIN32
    PopupMenu(mpPopupMenu, x, y);
#else
    PopupMenu(mpPopupMenu);
#endif
    if (HasCapture() == TRUE) ReleaseMouse();
    return;
}

bool MainGLCanvas_cls::
SetThisToBeActiveGL(){
    if (mpCParentFrameLink->mpCActiveGLCanvasLink != this) {
	mpCParentFrameLink->mpCActiveGLCanvasLink->Refresh(FALSE); 
	mpCParentFrameLink->mpCActiveGLCanvasLink = this;
	return TRUE;
    }
    return FALSE;
}

bool MainGLCanvas_cls::
SetOtherCanvasToBeActiveGL(){
    for (unsigned int i = 0; i < mVPOtherCanvas.size(); i++){
	if (mVPOtherCanvas[i]->IsMouseOnClient() == TRUE){
	    mVPOtherCanvas[i]->SetThisToBeActiveGL();
	    mVPOtherCanvas[i]->SetFocus();
	    mVPOtherCanvas[i]->Refresh(FALSE);
	    return TRUE;
	}
    }
    return FALSE;
}

void MainGLCanvas_cls::
OnKeyDown(wxKeyEvent& event){

    wxWindow *pCTopLevel = wxGetTopLevelParent(this);
    
    if (mSelfIndex == 4)
	pCTopLevel = pCTopLevel->GetParent();
    
    event.SetId(pCTopLevel->GetId());
    pCTopLevel->AddPendingEvent(event);
}

void MainGLCanvas_cls::
InitializeStaticMembers(Overlay_cls *pCOverlayLink){

    mspCClrCurveBG    = new ColorCurve_cls(wxT("BGColor"), wxT(""), 
					   wxT("0 0 0 0"));
    mspCClrCurveOut   = new ColorCurve_cls(wxT("OutColor"), wxT(""), 
					   wxT("0 0 0 0"));

    TagObjectCtrl_cls::InitializeStaticMembers();
    
    msCConfListStatic.Add(&msCFBOSizeIndex, wxT("FBOBufferSize="), wxT("3"));
    msCConfListStatic.Add(&msCLineWidth, wxT("LineWidth="), wxT("5"));
    
    mspCDrawTimeIndex = new DrawTimeIndex_cls(pCOverlayLink);
    mspCDrawClrLegend = new DrawClrLegend_cls(pCOverlayLink);
    mspCFishEyeCtrl   = new FishEye_cls();

    mspCGLGuide      = new GLGuides_cls();
    mspCGLGuideScale = new GLGuides_cls(_GUIDES_SCALE);
    mspCGLGuideYPR   = new GLGuides_cls(_GUIDES_ROTATE);
    
    SceneObj_cls::mspCGLGuideLink      = mspCGLGuide;
    SceneObj_cls::mspCGLGuideYPRLink   = mspCGLGuideYPR;
    SceneObj_cls::mspCGLGuideScaleLink = mspCGLGuideScale;

    mspCDistortion = new Distortion_cls;
    mspCDistortion->ReadWavefrontObject(wxT("Resource/distortion.obj"));

    GpCSphere = new Sphere_cls;
}

void MainGLCanvas_cls::
DestructStaticMembers(){
    delete mspCClrCurveBG;
    delete mspCClrCurveOut;
    delete TagObjectCtrl_cls::mspCClrCurveTag;
    delete mspCDrawTimeIndex;
    delete mspCDrawClrLegend;
    delete mspCFishEyeCtrl;
    delete mspCGLGuide;
    delete mspCGLGuideScale;
    delete mspCGLGuideYPR;

    delete GpCSphere;
}

bool MainGLCanvas_cls::
IsGLBlendEquationEXOK(){
    return (glBlendEquationEXT != NULL);
}

void MainGLCanvas_cls::
TogglePerspective(){
    if (mCGLGeometry.mBody == _GEOM_PERSPECTIVE)
	mCGLGeometry.mBody = _GEOM_ORTHOGRAPHIC;
    else
	mCGLGeometry.mBody = _GEOM_PERSPECTIVE;
    Refresh(FALSE);
}

void MainGLCanvas_cls::
SetDisplayMode(int displayMode, bool bUpdateCombobox){
    mCDisplayMode.mBody = displayMode;

    if (mCGLGeometry.mBody == _GEOM_ORTHOGRAPHIC ||
	mCGLGeometry.mBody == _GEOM_PERSPECTIVE){ 
	if (IsAlongAxis() == TRUE)
	    mCGLGeometry.mBody = _GEOM_ORTHOGRAPHIC;
	else
	    mCGLGeometry.mBody = _GEOM_PERSPECTIVE;
    }
    
    mpCCameraPathLinked = NULL;
    Refresh(FALSE);
}

void MainGLCanvas_cls::
LinkageConnect(){
    SetDisplayMode(_GL_CANVAS_LINKED_CAM0);
    mpCCameraPathLinked = mpCParentFrameLink->mpCParent->mpCCameraPath;
    mpCAnchorPathLinked = mpCParentFrameLink->mpCParent->mpCAnchorPath;

    double doubleIndex = 
	GetPCTimeCtrl()->GetDoubleIndex();

    GetPLinkedCameraPath()->UpdateCurrentObj(doubleIndex);
    GetPLinkedAnchorPath()->UpdateCurrentObj(doubleIndex);

    mCCameraPosCamera = mpCCameraPathLinked->mCCurrentObj;
    mCCameraPosCamera.UpdateYPR(FALSE);
}

void MainGLCanvas_cls::
OnMenuMoveToCamPos(wxCommandEvent& WXUNUSED(event)){
    LinkageConnect();
    mpCGLCtrl->UpdateWindow();
    Refresh(FALSE);
}

void MainGLCanvas_cls::
OnMenuSetKeyPos(wxCommandEvent& event){
    SendMenuEvent(mpCParentFrameLink, menuSetKeyPos);
}

void MainGLCanvas_cls::
OnMenuSetKeyPosRot(wxCommandEvent& event){
    SendMenuEvent(mpCParentFrameLink, menuSetKeyPosRot);
}

void MainGLCanvas_cls::
OnMenuSetKeyAnc(wxCommandEvent& event){
    SendMenuEvent(mpCParentFrameLink, menuSetKeyAnc);
}

void MainGLCanvas_cls::
OnMenuToggleLink(wxCommandEvent& event){
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0)
	SetDisplayMode(_DIRECTION_FREE);
    else
	LinkageConnect();
    mpCGLCtrl->UpdateWindow();
    Refresh(FALSE);
}

void MainGLCanvas_cls::
OnMenuToggleLink2(wxCommandEvent& event){
    if (mCDisplayMode.mBody >= _GL_CANVAS_LINKED_CAM0){
	mCCameraPosFree = mCCameraPosCamera;
	SetDisplayMode(_DIRECTION_FREE);
    }
    
    mpCGLCtrl->UpdateWindow();
    Refresh(FALSE);
}

void MainGLCanvas_cls::
OnMenuTogglePerspective(wxCommandEvent& event){
    TogglePerspective();
}

void MainGLCanvas_cls::
HookGridComboBox(void *pArg, void *pEvent){
}

void MainGLCanvas_cls::
CopySettingToAll(){
}

void MainGLCanvas_cls::
SetGLBGColor(int index){

    if (index == 0)
	glClearColor(mspCClrCurveBG->mCConstantColor.mpBody[R],
		     mspCClrCurveBG->mCConstantColor.mpBody[G],
		     mspCClrCurveBG->mCConstantColor.mpBody[B], 1.0f);
    if (index == 1)
	glClearColor(mspCClrCurveOut->mCConstantColor.mpBody[R],
		     mspCClrCurveOut->mCConstantColor.mpBody[G],
		     mspCClrCurveOut->mCConstantColor.mpBody[B], 1.0f);
}

void MainGLCanvas_cls::
SelectItem(Parameters_cls *pCParam){
}

// ----

void MainGLCanvas_cls::
ReadStrLine(wxString strLine){
    mCConfList.ReadStrLineAll(strLine);
}

void MainGLCanvas_cls::
ReadStrLineStaticMember(std::wistream &istream, wxString strLine){
    msCConfListStatic.ReadStrLineAll(strLine);
    mspCClrCurveBG->ReadStrLine(istream, strLine);
    mspCClrCurveOut->ReadStrLine(istream, strLine);
    TagObjectCtrl_cls::ReadStrLineStaticMember(istream, strLine);
}

wxString MainGLCanvas_cls::
Serialize(){
    return mCConfList.SerializeAll();
}

wxString MainGLCanvas_cls::
SerializeStaticMember(){
    return 
	msCConfListStatic.SerializeAll() + 
	mspCClrCurveBG->Serialize() + 
	mspCClrCurveOut->Serialize() +
	TagObjectCtrl_cls::SerializeStaticMember();
}

// ----------------------------------------------

ColorCurve_cls* MainGLCanvas_cls::mspCClrCurveBG = NULL;
ColorCurve_cls* MainGLCanvas_cls::mspCClrCurveOut = NULL;
DrawTimeIndex_cls* MainGLCanvas_cls::mspCDrawTimeIndex = NULL;
DrawClrLegend_cls* MainGLCanvas_cls::mspCDrawClrLegend = NULL;
FishEye_cls* MainGLCanvas_cls::mspCFishEyeCtrl = NULL;
Position_cls MainGLCanvas_cls::msCAnchorPos(0, 0, 0);
Position_cls MainGLCanvas_cls::msCAnchorPosOrtho(0, 0, 0);
FilterCtrl_cls* MainGLCanvas_cls::mspCFilterCtrlBooleanLink = NULL;
FilterCtrl_cls* MainGLCanvas_cls::mspCFilterCtrlObjectLink  = NULL;
FilterCtrl_cls* MainGLCanvas_cls::mspCFilterCtrlWorldLink   = NULL;
DF3SeqCtrl_cls* MainGLCanvas_cls::mspCDF3SeqCtrl = NULL;
PeriodicBoundary_cls* MainGLCanvas_cls::mspCPeriodicLink = NULL;
StereoGram_cls* MainGLCanvas_cls::mspCStereogramLink = NULL;
int MainGLCanvas_cls::msAFBOSize[7] = {256, 512, 1024, 2048, 4096, 8192, 16384};
ConfItemList_cls MainGLCanvas_cls::msCConfListStatic;
ConfItemUINT_cls MainGLCanvas_cls::msCFBOSizeIndex;
ConfItemFLOAT_cls  MainGLCanvas_cls::msCLineWidth;
int MainGLCanvas_cls::msGLLightIndex[8] = {
    GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3, 
    GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7
};

bool MainGLCanvas_cls::msbGLUTInitialized = FALSE;

HitTest_cls  MainGLCanvas_cls::msCHittest;

GLGuides_cls* MainGLCanvas_cls::mspCGLGuide      = NULL;
GLGuides_cls* MainGLCanvas_cls::mspCGLGuideScale = NULL;
GLGuides_cls* MainGLCanvas_cls::mspCGLGuideYPR   = NULL;
Distortion_cls* MainGLCanvas_cls::mspCDistortion = NULL;
GLTexture_cls*  MainGLCanvas_cls::mspCMask        = NULL;
GLTexture_cls*  MainGLCanvas_cls::mspCFisheyeGrid = NULL;

ImgSequence_cls* MainGLCanvas_cls::mspCImgSeqLinked = NULL;
