
#ifndef _STEREO_GRAM_H_
#define _STEREO_GRAM_H_

#include "../CONFITEM/clsConfItem.h"
#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsDialogEXKey.h"
#include "../WIDGETCTRL/clsDialogController.h"
#include "../WIDGETCTRL/clsWidgetCommon.h"

class FBO_cls;
class StereoGram_cls; 
class ParamGroup_cls;

#include "../../clsFrameMain.h"
#include "../PARAMETER/clsParameters.h"
#include "../MYGL/clsPositionAndDirection.h"

#define _STEREO_NONE    0
#define _STEREO_OFFAXIS 1
#define _STEREO_TOEIN   2

#define _STEREO_LFT         1
#define _STEREO_RGT         2
#define _STEREO_SCANLINE    3
#define _STEREO_RED_CYAN    4
#define _STEREO_YELLOW_BLUE 5
#define _STEREO_SIDEBYSIDE  6
#define _STEREO_QUADBUFFER  7

//#define _STEREO_DOME 1

class MainGLCanvas_cls;

class StereoGramGUI_cls : public wxDialogEXKey {
 public:
    StereoGramGUI_cls(StereoGram_cls *pCParent,
		      wxWindow *pC,
		      wxPoint pos, wxSize size);
    ~StereoGramGUI_cls();

    unsigned int mIndexEye, mIndexToeIn, mIndexDistance, mIndexDistanceFocus;
    unsigned int mIndexDualOutput, mIndexFlip;
    void HideAndShowItems();

    void OnButtonDefault   (wxCommandEvent &event);
    void OnButtonClose     (wxCommandEvent &event);
    static void FuncHookComboBox(void *pArg, void *pEvent);
    static void FuncHookAssign(void *pArg, void *pEvent);
    
    static StereoGram_cls    *mspCParent;
    static MultiInputPanel_cls* mpCMInputPanel;
    void SetMNumMainParam(wxBoxSizer *pBoxSizer);


 private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

class StereoGram_cls : 
public DialogControl_cls, public WidgetCommon_cls
{
    friend class DF3SeqGUI_cls; 
    friend class CameraPath_cls; 
    friend class StereoGramGUI_cls; 
    
 public:
    StereoGram_cls(FrameMain_cls *pC, wxString strBitmap);
    ~StereoGram_cls();

    FrameMain_cls* mpCParentFrame;
    wxString mstrBitmap;
    StereoGramGUI_cls* mpCGUI;
    wxRect mCScreenRect;

    int  mFlipList[2][2];
    bool mbQuadStereo;
    
    void UpdateItems(bool bForce = FALSE);
    void SetToDefault();
    void InitGUI(wxWindow *pC);
    void UpdateTLShown();

    int mBufferSize;
    FBO_cls *mpCFBO[2];

    bool mbFitFishEye;
    int mPanelWidth, mPanelHeight;
    bool mbCameraStatus;

    ParamGroup_cls *mpCStereoParam;
 
    ConfItemList_cls mCConfList;
    ConfItemDOUBLE_cls mCToeInAngle;
    ConfItemUINT_cls mCType;
    ConfItemUINT_cls mCEye;
    ConfItemUINT_cls mCDualOutput;
    ConfItemBOOL_cls mCbFlip;
    double mDistanceToFocus;
    
    void SetInfo(int panelWidth, int panelHeight, bool bCameraStatus);
    int GetStereoNum();
    float GetStereoFocus();
    void ClearFBO();
    void PrepareStereogram(int stereo);
    void DrawOnFBOBegin(int stereo);
    void DrawOnFBOEnd(int stereo);
    void DrawFunctionOnScreen(MainGLCanvas_cls *pCGLCanvas, void (MainGLCanvas_cls::*func)());
    void DrawOnScreen2D_SideBySide_FitCorner();
    
    void DrawOnScreenSimple(int fit);
    void DrawOnScreenSimpleQuad(int fit);
    void DrawOnScreenScanLine(int fit);
    void DrawOnScreenScanLineQuad(int fit, MainGLCanvas_cls *pCGLCanvas);
    void DrawOnScreenQuadNotQuad(int fit, MainGLCanvas_cls *pCGLCanvas);
    void DrawOnScreenQuadQuad(int fit);
    void DrawOnScreenSideBySide(int fit);
    void DrawOnScreenSideBySideQuad(int fit);
    void DrawOnScreenAnaglyph(int fit);
    void DrawOnScreenAnaglyphQuad(int fit);

    void DrawOnScreen2D(bool bUseDistortion, MainGLCanvas_cls *pCGLCanvas);
    void CombineStereoBuffer();
    
    void ModifyByStereogram(PositionAndDirection_cls *pCCam, int stereoEye);
    void DrawFocus(PositionAndDirection_cls CCamPos, 
		   double doubleIndex, double scale);

    int    GetStereoEye(int stereoIndex);
    bool   IsSideBySideForOverlayRequired();
    double GetDistanceToFocus(double doubleIndex);
    double GetStereogramYawAngle(int eyeMode);
    double GetOffsetRight(int eyeMode);
    void   ColorMaskBegin(int stereo);
    void   StencilMaskBegin(int width, int height);
    void   MakeStripePatternOnStencil(int width, int height, bool bOdd);
    void   ColorMaskEnd();
    void   StencilMaskEnd();
    
    //bool  UpdateValues();
    bool  IsGUIShownOnScreen();

    void ReadStrLine(std::wistream &istream, wxString strLine);
    wxString Serialize();
};

#endif
