// -*- c -*-
// "clsGrid.h"
// (setq c-basic-offset 4)

#include "clsCamera.h"

#include <math.h>

#ifdef WIN32
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined __WXMAC__
#include <OpenGL/glu.h>
#else
#include <GL/glu.h>
#endif

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif


bool Camera_cls::mbInitialized = FALSE;
int  Camera_cls::mDListBasic = -1;
int  Camera_cls::mDListCrosshair = -1;
int  Camera_cls::mDListLightSRC = -1;

int  Camera_cls::
DListDraw(int type, int GLMode){
    int    listName = 0;

    if (GLMode == GL_COMPILE || GLMode == GL_COMPILE_AND_EXECUTE){
	listName = glGenLists(1); // VKXg̔ԍ
	glNewList (listName, GLMode); 
    }
    
    switch(type){
    case _CAMERA_BASIC:
	// LocalX
	glBegin(GL_LINE_LOOP);
	glVertex3d(0.1, -0.1, -0.1);  glVertex3d(0.1, +0.1, -0.1);
	glVertex3d(0.1, +0.1, +0.1);  glVertex3d(0.1, -0.1, +0.1);  
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3d(-0.1, -0.1, -0.1);  glVertex3d(-0.1, +0.1, -0.1);
	glVertex3d(-0.1, +0.1, +0.1);  glVertex3d(-0.1, -0.1, +0.1);  
	glEnd();

	glBegin(GL_LINES);
	glVertex3d(-0.1, -0.1, -0.1); glVertex3d(+0.1, -0.1, -0.1); 
	glVertex3d(-0.1, +0.1, -0.1); glVertex3d(+0.1, +0.1, -0.1); 
	glVertex3d(-0.1, +0.1, +0.1); glVertex3d(+0.1, +0.1, +0.1); 
	glVertex3d(-0.1, -0.1, +0.1); glVertex3d(+0.1, -0.1, +0.1); 
	glEnd();
	/*
	// LocalX
	glBegin(GL_LINE_LOOP);
	glVertex3d(1, -1.3, -1);  glVertex3d(1, +1.3, -1);
	glVertex3d(1, +1.3, +1);  glVertex3d(1, -1.3, +1);  
	glEnd();
	
	glBegin(GL_LINES);
	glVertex3d(0, 0, 0); glVertex3d(1, -1.3, -1);
	glVertex3d(0, 0, 0); glVertex3d(1, +1.3, -1);
	glVertex3d(0, 0, 0); glVertex3d(1, +1.3, +1);
	glVertex3d(0, 0, 0); glVertex3d(1, -1.3, +1);
	glEnd();
	*/
	break;
    case _CAMERA_CROSSHAIR:
	glBegin(GL_LINES);
	glVertex3d(-0.5, 0, 0); glVertex3d(+0.5, 0, 0);
	glVertex3d(0, -0.5, 0); glVertex3d(0, +0.5, 0);
	glVertex3d(0, 0, -0.5); glVertex3d(0, 0, +0.5);
	glEnd();
	break;
    case _OBJECT_LIGHTSRC:
	glBegin(GL_LINES);
	for (int i = 0; i < 360; i+=10){
	    double rad = 3.14159 * i / 180;
	    glVertex3d(cos(rad), sin(rad), 0);
	}
	glEnd();
	glBegin(GL_LINE_STRIP);
	for (int i = 0; i <= 360; i+=10){
	    double rad = 3.14159 * i / 180;
	    glVertex3d(0.7 * cos(rad), 0.7 * sin(rad), 0);
	}
	glEnd();
	break;
    }
    
    if (GLMode == GL_COMPILE || GLMode == GL_COMPILE_AND_EXECUTE)
	glEndList(); 
    
    return listName;
}

void Camera_cls::
InitializeIfNot(){
    if (mbInitialized == TRUE) return;
    mDListBasic     = DListDraw(_CAMERA_BASIC, GL_COMPILE);
    mDListCrosshair = DListDraw(_CAMERA_CROSSHAIR, GL_COMPILE);
    mDListLightSRC  = DListDraw(_OBJECT_LIGHTSRC, GL_COMPILE);
    mbInitialized = TRUE;
}

void Camera_cls::
Draw(int type, double scale, double deg){
    InitializeIfNot();
    glPushMatrix();
    glScaled(scale, scale, scale);

    double rad = deg * 3.14159 / 180 / 2;
    double x = cos(rad);
    double z = sin(rad);
	
    switch(type){
    case _CAMERA_BASIC:{
	glCallList(mDListBasic);
	
	glBegin(GL_LINE_LOOP);
	glVertex3d(x, -z, -z);
	glVertex3d(x, +z, -z);
	glVertex3d(x, +z, +z);
	glVertex3d(x, -z, +z);
	glEnd();

	glBegin(GL_LINES);
	glVertex3d(0, 0, 0); glVertex3d(x, -z, -z);
	glVertex3d(0, 0, 0); glVertex3d(x, +z, -z);
	glVertex3d(0, 0, 0); glVertex3d(x, +z, +z);
	glVertex3d(0, 0, 0); glVertex3d(x, -z, +z);
	glEnd();
	break; }
    case _CAMERA_FISHEYE:{
	glCallList(mDListBasic);

	glBegin(GL_LINE_LOOP);
	for (int phi = 0; phi < 360; phi+=10){
	    double radphi = phi * 3.14159 / 180;
	    glVertex3d(x, z * cos(radphi), z * sin(radphi));
	}
	glEnd();

	glBegin(GL_LINE_STRIP);
	for (int phi = -90; phi <= 90; phi+= 10){
	    double radphi = phi * 3.14159 / 180 * deg / 180;
	    glVertex3d(cos(radphi), sin(radphi), 0); 
	}
	glEnd();

	glBegin(GL_LINE_STRIP);
	for (int phi = -90; phi <= 90; phi+= 10){
	    double radphi = phi * 3.14159 / 180 * deg / 180;
	    glVertex3d(cos(radphi), 0, sin(radphi)); 
	}
	glEnd();

	if (deg < 90){
	    glBegin(GL_LINES);
	    glVertex3d(0, 0, 0); glVertex3d(x, -z, 0);
	    glVertex3d(0, 0, 0); glVertex3d(x, +z, 0);
	    glVertex3d(0, 0, 0); glVertex3d(x, 0, +z);
	    glVertex3d(0, 0, 0); glVertex3d(x, 0, -z);
	    glEnd();
	}
	else{
	    glBegin(GL_LINES);
	    glVertex3d(0, 0, 0); glVertex3d(1, 0, 0);
	    glEnd();
	}

	break; }
    case _CAMERA_CROSSHAIR:
	glCallList(mDListCrosshair);
	break;
    
    }
    glPopMatrix();
}
