// -*- c -*-
// "clsGrid.h"
// (setq c-basic-offset 4)

#include "clsCylinder.h"

#include "../COMMON/MyVector.h"

#include <math.h>

#ifdef WIN32
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined __WXMAC__
#include <OpenGL/glu.h>
#else
#include <GL/glu.h>
#endif

#ifndef TRUE
#define TRUE true
#define FALSE false
#endif

#ifndef M_PI
#define M_PI 3.131593f
#endif

#define X 0
#define Y 1
#define Z 2

bool Cylinder_cls::mbInitialized = FALSE;
int  Cylinder_cls::mDList[2] ={-1, -1};

void Cylinder_cls::
Rotate(double phi, double theta, double vec[3]){
    double org[3] = {vec[0], vec[1], vec[2]};

    vec[X] =  cos(phi)*cos(theta)*org[X] + sin(phi)*org[Y] + cos(phi)*sin(theta)*org[Z];
    vec[Y] = -sin(phi)*cos(theta)*org[X] + cos(phi)*org[Y] - sin(phi)*sin(theta)*org[Z];
    vec[Z] = -sin(theta)*org[X] + cos(theta)*org[Z];
}

void Cylinder_cls::
DrawMain(int type, double phi, double theta, int quality, float length){
    int face; 
    switch(quality){
    case 0:
	face = 8;
	break;
    case 1:
	face = 16;
	break;
    case 2:
	face = 24;
	break;
    }
    double factor = M_PI * 2 / face;
    double rfac = 0.98;
    
    if (type == 0){
	glBegin(GL_QUAD_STRIP);
	double normal[3] = {0, 1, 0};
	double vert0[3] = {0.1, rfac, 0};
	double vert1[3] = {length, rfac, 0};
	Rotate(phi, theta, normal);
	Rotate(phi, theta, vert0);
	Rotate(phi, theta, vert1);
	//glNormal3dv(normal);
	glVertex3dv(vert0);
	glVertex3dv(vert1);
	
	for (int i = face-1; i >= 0 ; i--){
	    //M_PI*i/8.
	    double normal[3] = {0, rfac*cos(factor*i), rfac*sin(factor*i)};
	    double vert0[3] = {0.1, rfac*cos(factor*i), rfac*sin(factor*i)};
	    double vert1[3] = {length, rfac*cos(factor*i), rfac*sin(factor*i)};
	    Rotate(phi, theta, normal);
	    Rotate(phi, theta, vert0);
	    Rotate(phi, theta, vert1);
	    //glNormal3dv(normal);
	    glVertex3dv(vert0);
	    glVertex3dv(vert1);
	}
	glEnd();
	glBegin(GL_POLYGON);
	for (int i = face-1; i>=0; i--){
	    double normal[3] = {-1, 0, 0};
	    double vert[3] = {0.1, rfac*cos(factor*i), rfac*sin(factor*i)};
	    Rotate(phi, theta, normal);
	    Rotate(phi, theta, vert);
	    //glNormal3dv(normal);
	    glVertex3dv(vert);
	}
	glEnd();
	glBegin(GL_POLYGON);
	for (int i = 0; i<=face-1; i++){
	    double normal[3] = {1, 0, 0};
	    double vert[3] = {length, rfac*cos(factor*i), rfac*sin(factor*i)};
	    Rotate(phi, theta, normal);
	    Rotate(phi, theta, vert);
	    //glNormal3dv(normal);
	    glVertex3dv(vert);
	}
	glEnd();
    }
    else{
	glBegin(GL_QUAD_STRIP);

	double normal[3] = {0, -1, 0};
	double vert0[3] = {length, rfac, 0};
	double vert1[3] = {0.1, rfac, 0};
	Rotate(phi, theta, normal);
	Rotate(phi, theta, vert0);
	Rotate(phi, theta, vert1);
	//glNormal3dv(normal);
	glVertex3dv(vert0);
	glVertex3dv(vert1);
	for (int i = face-1; i >= 0 ; i--){
	    double normal[3] = {0, -rfac*cos(factor*i), -rfac*sin(factor*i)};
	    double vert0[3] = {length, rfac*cos(factor*i), rfac*sin(factor*i)};
	    double vert1[3] = {0.1, rfac*cos(factor*i), rfac*sin(factor*i)};
	    Rotate(phi, theta, normal);
	    Rotate(phi, theta, vert0);
	    Rotate(phi, theta, vert1);
	    //glNormal3dv(normal);
	    glVertex3dv(vert0);
	    glVertex3dv(vert1);
	}
	glEnd();
	glBegin(GL_POLYGON);
	for (int i = face-1; i>=0; i--){
	    double normal[3] = {1, 0, 0};
	    double vert[3] = {length, rfac*cos(factor*i), rfac*sin(factor*i)};
	    Rotate(phi, theta, normal);
	    Rotate(phi, theta, vert);
	    //glNormal3dv(normal);
	    glVertex3dv(vert);
	}
	glEnd();
	glBegin(GL_POLYGON);
	for (int i = 0; i<=face-1; i++){
	    double normal[3] = {-1, 0, 0};
	    double vert[3] = {0.1, rfac*cos(factor*i), rfac*sin(factor*i)};
	    Rotate(phi, theta, normal);
	    Rotate(phi, theta, vert);
	    //glNormal3dv(normal);
	    glVertex3dv(vert);
	}
	glEnd();
    }
}

int Cylinder_cls::
DListDraw(int GLMode, int type){
    int    listName = 0;
	
    if (GLMode == GL_COMPILE || GLMode == GL_COMPILE_AND_EXECUTE){
	listName = glGenLists(1); // VKXg̔ԍ
	glNewList (listName, GLMode); 
    }

    DrawMain(type);

    if (GLMode == GL_COMPILE || GLMode == GL_COMPILE_AND_EXECUTE)
	glEndList(); 
    
    return listName;
}

void Cylinder_cls::
InitializeIfNot(){
    if (mbInitialized == TRUE) return;
    mDList[0] = DListDraw(GL_COMPILE, 0);
    mDList[1] = DListDraw(GL_COMPILE, 1);
    mbInitialized = TRUE;
}

void Cylinder_cls::
Draw(int type){
    glCallList(mDList[type]);
}
