// -*- c -*-
// "clsGrid.h"
// (setq c-basic-offset 4)

#include "clsGrid.h"

#include "../GLCANVAS/NAVIGATORS/clsTagObject.h"

#ifdef WIN32
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined __WXMAC__
#include <OpenGL/glu.h>
#else
#include <GL/glu.h>
#endif

#include "../COMMON/CommonDefinitions.h"

bool Grid_cls::mbInitialized = FALSE;
int  Grid_cls::mDListFlatInner = -1;
int  Grid_cls::mDListFlatOuter = -1;
int  Grid_cls::mFineness = 10;

#include "../MYGL/clsPositionAndDirection.h"

int  Grid_cls::
DListFlatDraw(int fineness, int GLMode, int gridMode){
    int    listName = 0;

    if (GLMode == GL_COMPILE || GLMode == GL_COMPILE_AND_EXECUTE){
	listName = glGenLists(1); // VKXg̔ԍ
	glNewList (listName, GLMode); 
    }
    
    if (fineness < 3) fineness = 3;

    glBegin(GL_LINES);
    for (int i= -fineness + (gridMode == _GRID_INNER)
	     ; i <= fineness - (gridMode == _GRID_INNER) ; i++){
	
	if (i == 0) continue;
	
	float value = (float)i / fineness;
	glVertex3f(value, -1, 0); glVertex3f(value, +1, 0);
	glVertex3f(-1, value, 0); glVertex3f(+1, value, 0);
    }

    glVertex3d(1.05, 0.05, 0); glVertex3d(1.1, 0, 0);
    glVertex3d(1.05,-0.05, 0); glVertex3d(1.1, 0, 0);
    
    glVertex3d( 0.05, 1.05, 0); glVertex3d(0, 1.1, 0);
    glVertex3d(-0.05, 1.05, 0); glVertex3d(0, 1.1, 0);
    
    glEnd();
    
    if (gridMode == _GRID_OUTER){
	glBegin(GL_LINES);
	glVertex3f(-1, 0,   0); glVertex3f(+1,  0,  0);
	glVertex3f( 0,-1,   0); glVertex3f( 0, +1,  0);
	glVertex3f( 0,  0, -1); glVertex3f( 0,  0, +1);
	glEnd();
    }
    
    if (GLMode == GL_COMPILE || GLMode == GL_COMPILE_AND_EXECUTE)
	glEndList(); 
    
    return listName;
}

void Grid_cls::
InitializeIfNot(){
    if (mbInitialized == TRUE) return;
    mDListFlatInner = DListFlatDraw(mFineness, GL_COMPILE, _GRID_INNER);
    mDListFlatOuter = DListFlatDraw(mFineness, GL_COMPILE, _GRID_OUTER);
    mbInitialized = TRUE;
}

bool Grid_cls::
IsSameTagExist(std::vector < TagObject_cls > *pVCTagObject,
	       wxString strTag){
    int count = pVCTagObject->size();
    for (int i = 0; i < count; i++)
	if (strTag == (*pVCTagObject)[i].mstrTag) return TRUE;
    return FALSE;
}

void Grid_cls::
DrawCubicGrid(double min, double max, int linenum, 
	      PositionAndDirection_cls CCamPos,
	      std::vector < TagObject_cls > *pVCTagObject){

    if (CCamPos.mTgt[Z] <= 0){
	glPushMatrix();
	glTranslatef(0, 0, min);
	DrawBasicGridZ(min, max, linenum, min, pVCTagObject);
	glPopMatrix();
    }
    if (CCamPos.mTgt[Z] >= 0){
	glPushMatrix();
	glTranslatef(0, 0, max);
	DrawBasicGridZ(min, max, linenum, max, pVCTagObject);
	glPopMatrix();
    }
    
    if (CCamPos.mTgt[X] <= 0){
	glPushMatrix();
	glTranslatef(min, 0, 0);
	DrawBasicGridX(min, max, linenum, min, pVCTagObject);
	glPopMatrix();
    }
    if (CCamPos.mTgt[X] >= 0){
	glPushMatrix();
	glTranslatef(max, 0, 0);
	DrawBasicGridX(min, max, linenum, max, pVCTagObject);
	glPopMatrix();
    }

    if (CCamPos.mTgt[Y] <= 0){
	glPushMatrix();
	glTranslatef(0, min, 0);
	DrawBasicGridY(min, max, linenum, min, pVCTagObject);
	glPopMatrix();
    }
    if (CCamPos.mTgt[Y] >= 0){ 
	glTranslatef(0, max, 0);
	DrawBasicGridY(min, max, linenum, max, pVCTagObject);
	glPopMatrix();
    }

    /*
    for (int iz = 0; iz <= 1; iz++){
	float z = (iz == 1) ? max : min;
	for (int iy = 0; iy <= 1; iy++){
	    float y = (iy == 1) ? max : min;
	    for (int ix = 0; ix <= 1; ix++){
		float x = (ix == 1) ? max : min;
		if (bCorner[iz][iy][ix] == FALSE) continue;
		
		wxString strTag = wxString::Format(wxT("(%g, %g, %g)"), x, y, z);
		if (IsSameTagExist(pVstrTag, strTag) == TRUE) continue;
		pVstrTag->push_back(strTag);
		pVTagX->push_back(x);
		pVTagY->push_back(y);
		pVTagZ->push_back(z);
	    }
	}
    }
    */
}

void Grid_cls::
DrawBasicGridX(double min, double max, int linenum, double offset,
	       std::vector < TagObject_cls > *pVCTagObject){
    double width = max - min;
    if (width == 0) width = 1;
    double d = width / linenum;
    if (linenum < 1) linenum = 1;
    glBegin(GL_LINES);
    for (double z = min; z <= max + 1e-6; z+= d){
	glVertex3d(0, min, z);
	glVertex3d(0, max, z);
    }
    for (double y = min; y <= max + 1e-6; y+= d){
	glVertex3d(0, y, min);
	glVertex3d(0, y, max);
    }
    glEnd();

    if (pVCTagObject == NULL) return;

    float x = offset;
    for (int iz = 0; iz <= 1; iz++){
	float z = (iz == 1) ? max : min;
	for (int iy = 0; iy <= 1; iy++){
	    float y = (iy == 1) ? max : min;
	    wxString strTag = wxString::Format(wxT("(%g, %g, %g)"), x, y, z);
	    if (IsSameTagExist(pVCTagObject, strTag) == TRUE) continue;
	    pVCTagObject->push_back(TagObject_cls(strTag, x, y, z));
	}
    }
}

void Grid_cls::
DrawBasicGridY(double min, double max, int linenum, double offset,
	       std::vector < TagObject_cls > *pVCTagObject){
    double width = max - min;
    if (width == 0) width = 1;
    if (linenum < 1) linenum = 1;
    double d = width / linenum;
    glBegin(GL_LINES);
    for (double x = min; x <= max + 1e-6; x+= d){
	glVertex3d(x, 0, min);
	glVertex3d(x, 0, max);
    }
    for (double z = min; z <= max + 1e-6; z+= d){
	glVertex3d(min, 0, z);
	glVertex3d(max, 0, z);
    }
    glEnd();

    if (pVCTagObject == NULL) return;
    
    float y = offset;
    for (int iz = 0; iz <= 1; iz++){
	float z = (iz == 1) ? max : min;
	for (int ix = 0; ix <= 1; ix++){
	    float x = (ix == 1) ? max : min;
	    
	    wxString strTag = wxString::Format(wxT("(%g, %g, %g)"), x, y, z);
	    if (IsSameTagExist(pVCTagObject, strTag) == TRUE) continue;
	    pVCTagObject->push_back(TagObject_cls(strTag, x, y, z));
	}
    }
}

void Grid_cls::
DrawBasicGridZ(double min, double max, int linenum, double offset,
	       std::vector < TagObject_cls > *pVCTagObject){
    double width = max - min;
    if (width == 0) width = 1;
    if (linenum < 1) linenum = 1;
    double d = width / linenum;
    glBegin(GL_LINES);
    for (double y = min; y <= max + 1e-6; y+= d){
	glVertex3d(min, y, 0);
	glVertex3d(max, y, 0);
    }
    for (double x = min; x <= max + 1e-6; x+= d){
	glVertex3d(x, min, 0);
	glVertex3d(x, max, 0);
    }
    glEnd();

    if (pVCTagObject == NULL) return;
    
    float z=offset;
    for (int iy = 0; iy <= 1; iy++){
	float y = (iy == 1) ? max : min;
	for (int ix = 0; ix <= 1; ix++){
	    float x = (ix == 1) ? max : min;
	    
	    wxString strTag = wxString::Format(wxT("(%g, %g, %g)"), x, y, z);
	    if (IsSameTagExist(pVCTagObject, strTag) == TRUE) continue;
	    pVCTagObject->push_back(TagObject_cls(strTag, x, y, z));
	}
    }
}

void Grid_cls::
DrawGridFlat(double scale, int nest){
    InitializeIfNot();
    glPushMatrix();
    glScaled(scale, scale, scale);
    //glCallList(mDListFlatInner);
    for (int i = 0; i < nest; i++){
	if (i == nest - 1)
	    glCallList(mDListFlatOuter);
	else
	    glCallList(mDListFlatInner);
	glScaled(10, 10, 10);
    }
    glPopMatrix();
}
