// -*- c -*- 
// (setq c-basic-offset 4)
// "clsImgSequence.h"

#include "clsImgSequence.h"
#include "clsImageIO.h"
#include "../MYGL/clsPositionAndDirection.h"
#include "../MYGL/clsFBO.h"
#include "../TIMELINE/clsTimeCtrl.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "../UNDO/clsUndoItemConstParam.h"
#include "../WIDGETCTRL/EX/clsFileDialogEX.h"
extern wxFileDialogEX*   GpCFileDialog;
#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsHookCtrl.h"

#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include "../COMMON/CommonDefinitions.h"

#include <wx/file.h>

#ifdef WIN32
#define DELIM L"\\"
#define DELIMWX wxT("\\")
#else
#define DELIM L"/"
#define DELIMWX wxT("/")
#endif

BEGIN_EVENT_TABLE(ImgSequenceGUI_cls, wxDialog)
    EVT_BUTTON(btnImgSeqRender,    ImgSequenceGUI_cls::OnButtonRender)
    EVT_BUTTON(btnImgSeqSnap,      ImgSequenceGUI_cls::OnButtonSnap)
    EVT_BUTTON(btnImgSeqInitSeq,   ImgSequenceGUI_cls::OnButtonInitSeq)
    EVT_BUTTON(btnImgSeqReference, ImgSequenceGUI_cls::OnButtonReference)
    EVT_BUTTON(wxID_CANCEL, ImgSequenceGUI_cls::OnButtonClose)
END_EVENT_TABLE()

ImgSequenceGUI_cls::
ImgSequenceGUI_cls(ImgSequence_cls *pCParent, wxWindow *pC,
    		   wxPoint pos, wxSize size) :
    wxDialogEXKey(pC, wxID_ANY, _("Save Image Sequence"),
		  pos, size){
    mpCParent = pCParent;
    mspCThis = this;
    SetScreenRectPos(&pCParent->mCScreenRect);

    wxIcon icon(wxT("ICONS/ImgSequence.png"), wxBITMAP_TYPE_PNG);
    SetIcon(icon);

    // TCU[
    wxBoxSizer       *pBoxSizerSrc     = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerParam   = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer       *pBoxSizerButtons = new wxBoxSizer(wxHORIZONTAL);
 
    SetMNumImgSeqParam(pBoxSizerParam);
    
#ifdef WIN32
    int buttonWidth=100;
#else
    int buttonWidth=140;
#endif
    mpCButtonRender =
	new wxButtonEX(this, btnImgSeqRender, wxT("Render Anim"),
		       wxDP, wxSize(buttonWidth, -1));
    mpCButtonSnap = 
	new wxButtonEX(this, btnImgSeqSnap, wxT("Render Snap"),
		       wxDP, wxSize(buttonWidth, -1));
    
    wxButtonEX *pCButtonCancel =
	new wxButtonEX(this, wxID_CANCEL, wxT("Close"),
		       wxDP, wxSize(buttonWidth, -1));

    pBoxSizerButtons->Add(mpCButtonRender, 0, wxEXPAND);
    pBoxSizerButtons->Add(mpCButtonSnap  , 0, wxEXPAND);
    pBoxSizerButtons->Add( -1, -1, 1, wxEXPAND);
    pBoxSizerButtons->Add(pCButtonCancel, 0, wxEXPAND);
    
    pBoxSizerSrc->Add(pBoxSizerParam, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerButtons, 0, wxEXPAND);

    SetSizerAndFit(pBoxSizerSrc);

    mpCButtonRender->SetFocus();
}

void ImgSequenceGUI_cls::
SetMNumImgSeqParam(wxBoxSizer *pBoxSizer){
    mpCMNumImgSeqParam = new MultiInputPanel_cls(&mpCMNumImgSeqParam, this);

    MultiInputPanel_cls *pCurr = mpCMNumImgSeqParam;
    pCurr->mUpdateBit = _COMMAND_UPDATE_DISPLAY_IMGSEQ;
    
    int boxWidth;
    pCurr->mTextWidth = 160/*-1*/;
    pCurr->mItemWidth = 90;

    boxWidth = 3;
    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Preview (W, H, V)"));
    pCurr->CombinePointer(_ITEM_INTEGER, &mpCParent->mClipWidthMax, wxT(""), FALSE);
    pCurr->CombinePointer(_ITEM_INTEGER, &mpCParent->mClipHeightMax, wxT(""), FALSE);
    pCurr->CombinePointer(_ITEM_FLOAT, &mpCParent->mAngleOrg, wxT(""), FALSE);

    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Output Size"));

    wxString listTemplate[_IMG_SIZE_MAX];
    for (int i = 0; i < _IMG_SIZE_MAX; i++) listTemplate[i] = GetStrGeom(i);
    pCurr->CombinePointer(_ITEM_INTEGER_SPIN, &mpCParent->mClipWidth);
    pCurr->SetRange(_ITEM_LASTITEM, 0, 8192);
    pCurr->CombinePointer(_ITEM_INTEGER_SPIN, &mpCParent->mClipHeight);
    pCurr->SetRange(_ITEM_LASTITEM, 0, 8192);
    mIndexGeomTemplate = pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &mGeometryTemplate, _IMG_SIZE_MAX, listTemplate);
    
    boxWidth = 3;

    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Initial sequence"));
    pCurr->CombinePointer(_ITEM_INTEGER, &mpCParent->mCInitNum.mBody);
    pCurr->CombinePointerButton(btnImgSeqInitSeq, _("Calc. Auto"));
    pCurr->CombinePointerButton(btnImgSeqReference, _("Reference"));
    
    pCurr->mItemWidth = -1;
    pCurr->CombinePointer(_ITEM_STRING, &mpCParent->mCFileSequence.mstrFormat);

    pCurr->mItemWidth = 90;
    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth,
			  _("Motion Blur (Length/Quality)"));
    pCurr->CombinePointer(_ITEM_BOOL, &mpCParent->mpCTimeCtrlLink->mCbUseMotionBlur.mbBody, _("Use"));
    pCurr->CombinePointer(_ITEM_FLOAT, &mpCParent->mpCTimeCtrlLink->mCBlurCtrl.mBlurLength);
    pCurr->CombinePointer(_ITEM_INTEGER, &mpCParent->mpCTimeCtrlLink->mCBlurCtrl.mBlurQuality);
    pCurr->GetLastItem().SetRange(2, 16, 1);
    
    pCurr->mMinimumWidth = 400;

    pCurr->FinishSetting(pBoxSizer);

    pCurr->mpCHookCtrl->SetFuncHook(&HookGeometry, _HOOK_COMBOBOX);
}

#include "../MYAPI/MsgBox.h"

void ImgSequenceGUI_cls::
OnButtonInitSeq(wxCommandEvent &event){
    mpCParent->SetAutoInit();
    mpCMNumImgSeqParam->UpdateValues();
}

void ImgSequenceGUI_cls::
OnButtonReference(wxCommandEvent &event){
    GpCFileDialog->SetMessage(_("Select a image file to save"));
    GpCFileDialog->SetWildcard(wxT("image(*.png;*.jpg;*.tif)|*.png;*.jpg;*.tif|png (*.png)|*.png|jpg (*.jpg)|*.jpg|tiff (*.tif)|*.tif"));
    GpCFileDialog->SetStyle(wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
    GpCFileDialog->SetInitPathFull(mpCParent->mCFileSequence.GetPathStr(_FN_FULLPATH));

    if (GpCFileDialog->ShowModal() == wxID_OK){
	GCUndoList.PushNewCommand
	    (new CommandSetValue_cls(&mpCParent->mCFileSequence.mstrFormat, GpCFileDialog->GetPath(), _COMMAND_UPDATE_DISPLAY_IMGSEQ),
	     new CommandSetValue_cls(&mpCParent->mCFileSequence.mstrFormat, mpCParent->mCFileSequence.mstrPath, _COMMAND_UPDATE_DISPLAY_IMGSEQ));
	GCUndoList.ExecCurrent();
	mpCParent->mCFileSequence.Set(mpCParent->mCFileSequence.mstrFormat);
	/*
	if (mpCParent->mCFileSequence.mDigitLength != -1)
	    mpCParent->mCInitNum.mBody = mpCParent->mCFileSequence.mDigit;
	*/
	mpCMNumImgSeqParam->UpdateValues();
    }
}

void ImgSequenceGUI_cls::
OnButtonRender(wxCommandEvent &event){
    if (mpCParent->mbInRendering == TRUE){
	AbortRender(); return; }

    mpCMNumImgSeqParam->AssignBase();
    
    mpCParent->mCFileSequence.Set(mpCParent->mCFileSequence.GetFormat());
    mpCMNumImgSeqParam->UpdateValues();

    if (mpCParent->mCFileSequence.mDigitLength == -1){
	if (MsgBoxYesno(this, _("Sequence Format error"), 
			_("Not Sequence. Append 4-digit tag ?")) == FALSE)
	    return;
	mpCParent->mCFileSequence.AppendDefaultSequenceTag();
	mpCMNumImgSeqParam->UpdateValues();
    }

    if (mpCParent->mCInitNum.mBody !=
	- mpCParent->mpCTimeCtrlLink->GetOffsetForImgSeq(0, 0)){
	wxString strMessage;
	strMessage << _("Initial Number is set to ") << mpCParent->mCInitNum.mBody << ENDL;
	strMessage << _("Working area starts at ") << mpCParent->mpCTimeCtrlLink->mCWAMin.mBody << ENDL;
	strMessage << _("Recommended Initial Number is ") << - mpCParent->mpCTimeCtrlLink->GetOffsetForImgSeq(0, 0) << ENDL;
	strMessage << _("Change Initial Number ?");
	if (MsgBoxYesno(this, _("Confirmation"), strMessage) == TRUE)
	    mpCParent->SetAutoInit();
    }

    // !!_FIX_ME_!! XeÎƂ̃`FbN܂
    if (wxFile::Exists(mpCParent->mCFileSequence.GetModifiedStr(mpCParent->mCInitNum.mBody, 0)) == TRUE){
	wxString strMessage;
	strMessage << mpCParent->mCFileSequence.GetModifiedStr(mpCParent->mCInitNum.mBody, 0) << ENDL;
	strMessage << _("This file already exists. Overwrite it ?");
	if (MsgBoxYesno(this, _("Confirmation"), strMessage) == FALSE)
	    return;
    }
	
    SendMenuEvent(GetParent(), menuItemSaveImgSeqBegin);
    mpCButtonRender->SetLabel(wxT("Abort Anim"));
    mpCButtonRender->SetBackgroundColour(wxColour(wxT("RED")));
    mpCButtonRender->SetForegroundColour(wxColour(wxT("WHITE")));
    mpCParent->mbInRendering = TRUE;
}

void ImgSequenceGUI_cls::
OnButtonSnap(wxCommandEvent &event){
    if (mpCParent->mbInRendering == TRUE) return;
    mpCMNumImgSeqParam->AssignBase();

    mpCParent->mCFileSequence.Set(mpCParent->mCFileSequence.GetFormat());
    mpCMNumImgSeqParam->UpdateValues();
    
    SendMenuEvent(GetParent(), menuItemSaveImgSeqBegin);
    mpCParent->mbInRendering = TRUE;
    mpCParent->mbSnapMode = TRUE;
}

void ImgSequenceGUI_cls::
OnButtonClose(wxCommandEvent &event){
    mpCMNumImgSeqParam->UpdateValues();
    CloseEXKey();
}

void ImgSequenceGUI_cls::
AbortRender(){
    SendMenuEvent(GetParent(), menuPlayStop);
    OutputFinish();
}

void ImgSequenceGUI_cls::
OutputFinish(){
    mpCButtonRender->SetLabel(wxT("Render Anim"));
    mpCButtonRender->SetBackgroundColour(wxNullColour);
    mpCButtonRender->SetForegroundColour(wxNullColour);
    mpCParent->mbInRendering = FALSE;
}

void ImgSequenceGUI_cls::
Geometry(){
    MultiInputPanel_cls* pCurr = mpCMNumImgSeqParam;
    int geometry = pCurr->mVCItemList[mIndexGeomTemplate].GetValueInt();
    
    SetGeometry(geometry, &mpCParent->mClipWidth, &mpCParent->mClipHeight);
    
    if (geometry == 0){
	pCurr->mVCItemList[mIndexGeomTemplate-1].GetPCwxWindow()->Enable();
	pCurr->mVCItemList[mIndexGeomTemplate-2].GetPCwxWindow()->Enable();
    }
    else{
	pCurr->mVCItemList[mIndexGeomTemplate-1].GetPCwxWindow()->Disable();
	pCurr->mVCItemList[mIndexGeomTemplate-2].GetPCwxWindow()->Disable();
    }

    pCurr->UpdateValues();
}

void ImgSequenceGUI_cls::
HookGeometry(void *pPanel, void *pArg){
    mspCThis->Geometry();
}

ImgSequenceGUI_cls* ImgSequenceGUI_cls::mspCThis = NULL;

// ----

ImgSequence_cls::
ImgSequence_cls(wxWindow *pC, TimeCtrl_cls *pCTimeCtrl) :
    mCFileSequence(wxGetCwd()+DELIMWX+wxT("Preview%04d.png"))
{
    mClipWidthMax = 640;
    mClipHeightMax = 480;
    mAngleOrg = 35;
    mClipWidth = 640;
    mClipHeight = 480;
    mOffsetX = 0;
    mOffsetY = 0;
    mbInRendering = FALSE;
    mbSnapMode = FALSE;
    mpCGUI = NULL;
    mpCTimeCtrlLink = pCTimeCtrl;
    SetToDefault();
    InitGUI(pC); // GUĨCX^X
}

void ImgSequence_cls::
SetGeom(int width, int height, float angle, bool bSetTrim){
    mClipWidthMax  = width;
    mClipHeightMax = height;
    mAngleOrg = angle;
    if (bSetTrim == TRUE){
	mClipWidth = width;
	mClipHeight = height;
    }
    if (mpCGUI != NULL && mpCGUI->IsShownOnScreen())
	UpdateItems();
}

void ImgSequence_cls::
CalcOffset(int widthOrg, int heightOrg){
    mOffsetX = (widthOrg - mClipWidth) / 2;
    mOffsetY = (heightOrg - mClipHeight) / 2;
}

void ImgSequence_cls::
SetAutoInit(){
    mCInitNum.mBody = - mpCTimeCtrlLink->GetOffsetForImgSeq(0, 0);
}

wxString ImgSequence_cls::
GetWXStrModifiedFilename(){
    return mCFileSequence.GetModifiedStr(mCurrentOffset + mCInitNum.mBody,
					 mCurrentStereo);
}

void ImgSequence_cls::
SetDefaultPath(wxString strPath){
    if (strPath == wxT("") || strPath == wxT("None")) return;
    SplitPath_cls CPath(strPath);
    if (CPath.GetPathStr(_FN_FNAME).Matches(wxT("snapshot"))){
	CPath.mstrFilename.Replace(wxT("snapshot"), wxT("preview%04d"));
	strPath = CPath.GetPathStr(_FN_FULLPATH);
    }
    mCFileSequence.Set(strPath);
}

bool ImgSequence_cls::
SaveImageFromGL(ImageIO_cls* pCImageIO, int widthOrg, int heightOrg, int offset, int stereo){
    // CHECK_ME!! łQ`݂̖łĂH
    if (mCurrentStereo == stereo && offset < mCurrentOffset)
	return TRUE;

    mCurrentStereo = stereo;
    mCurrentOffset = offset;
    bool retVal = FALSE;
    CalcOffset(widthOrg, heightOrg);
    retVal = pCImageIO->SaveImageFromGL(mClipWidth, mClipHeight,
					GetWXStrModifiedFilename(),
					mOffsetX, mOffsetY);
    return retVal;
}

bool ImgSequence_cls::
SaveImageFromFBO(ImageIO_cls* pCImageIO, FBO_cls *pCFBO,
		 FBO_cls *pCFBOOverlay, int widthOrg, int heightOrg, 
		 int offset, int stereo, bool bSideBySide){
    // CHECK_ME!! łQ`݂̖łĂH
    if (mCurrentStereo == stereo && offset < mCurrentOffset)
	return TRUE;

    mCurrentStereo = stereo;
    mCurrentOffset = offset;
    bool retVal = FALSE;
    CalcOffset(widthOrg, heightOrg);

    retVal = pCImageIO->SaveImageFromFBO
	(pCFBO, pCFBOOverlay, mClipWidth, mClipHeight, 
	 GetWXStrModifiedFilename(), bSideBySide);

    return retVal;
}

void ImgSequence_cls::
SetToDefault(){
    mCInitNum.SetItem(L"ImgSequenceInitNum=", L"0");
}

void ImgSequence_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new ImgSequenceGUI_cls(this, pC,
				    mCScreenRect.GetPosition(),
				    mCScreenRect.GetSize());
    SetPDialog(mpCGUI);
}

void ImgSequence_cls::
UpdateItems(bool bForce){
    mpCGUI->mpCMNumImgSeqParam->UpdateValues(_MN_ALLITEM);
}

void ImgSequence_cls::
RenderFrameBoxInRender(int width, int height, GLPerspective_cls CGLPerspective){
    glPushMatrix();
    PositionAndDirection_cls CPos(0, 0, -1);

    CGLPerspective.mbPerspective = FALSE;
    CGLPerspective.SetPerspectiveOnGL(&CPos);
    double rateX = mClipWidth/width;
    double rateY = mClipHeight/height;
    double aspect = CGLPerspective.GetAspectRatio();
    glBegin(GL_LINE_LOOP);
    glVertex2f(-0.5 * aspect * rateX, -0.5 * rateY);
    glVertex2f(+0.5 * aspect * rateX, -0.5 * rateY);
    glVertex2f(+0.5 * aspect * rateX, +0.5 * rateY);
    glVertex2f(-0.5 * aspect * rateX, +0.5 * rateY);
    glEnd();
    glPopMatrix();
}

void ImgSequence_cls::
OutputFinish(){
    if (mpCGUI != NULL)
	mpCGUI->OutputFinish();
}
