// -*- c -*-
// "clsReadHDF5Data.h"


#include "clsReadHDF5Data.h"
#include "clsGrammerHDF5.h"

#include "clsAttribItem.h"
#include "clsDataSetAndName.h"

#include "../PANELS/clsDataLoader.h"

#include <wx/choicdlg.h>

#include "../COMMON/CommonDefinitions.h"

#ifdef _USE_HDF5

ReadHDF5Data_cls::
ReadHDF5Data_cls(){
    //mbOpened = FALSE;
    mpCFile = NULL;
    mpIndexList = NULL;
    mIndexListSize = -1;
}

ReadHDF5Data_cls::
~ReadHDF5Data_cls(){
    CloseProcess();
    if (mpIndexList != NULL)
	delete [] mpIndexList;
}

void ReadHDF5Data_cls::
CloseProcess(){
    for (int i = 0; i < mVpCDataSetAndName.size(); i++)
	delete mVpCDataSetAndName[i];
    for (int i = 0; i < mVpCAttribItem.size(); i++)
	delete mVpCAttribItem[i];
    mVpCDataSetAndName.clear();
    //if (mbOpened == TRUE) mCFile.close();
    if (mpCFile != NULL) mpCFile->close();
}

DataSetAndName_cls* ReadHDF5Data_cls::
GetPCDataSet(wxString strName, int snap){
    DataSetAndName_cls *pCDataSet = NULL;
    int count = 0;
    int size = mVpCDataSetAndName.size();
    for (int i = 0; i < size; i++){
	wxString strCanditName = wxString::FromAscii(mVpCDataSetAndName[i]->mstdstrName.c_str());
	if (strCanditName.IsSameAs(strName)){
	    if (count == snap){
		pCDataSet = mVpCDataSetAndName[i];
		break;
	    }
	    count++;
	}
    }
    return pCDataSet;
}

void ReadHDF5Data_cls::
ReadAttrib(H5::H5Location &CLocation, std::string stdstrName){
    int attribNum = CLocation.getNumAttrs();

    for (int i = 0; i < attribNum; i++){
	H5::Attribute attrib = CLocation.openAttribute(i);
	mVpCAttribItem.push_back(new AttribItem_cls(attrib, stdstrName));
    }
}

void ReadHDF5Data_cls::
ReadGroup(H5::CommonFG &CGroup){
    
    int objNum = CGroup.getNumObjs();

    for (int i = 0; i < objNum; i++){
	
	std::string stdstrName = CGroup.getObjnameByIdx(i);
	H5G_obj_t objType = CGroup.getObjTypeByIdx(i);
	
	if (objType == H5G_GROUP){
	    mVpCDataSetAndName.push_back
		(new DataSetAndName_cls(CGroup, stdstrName, objType));
	    H5::Group CNewGroup = CGroup.openGroup(stdstrName);
	    ReadAttrib(CNewGroup, stdstrName);
	    ReadGroup(CNewGroup);
	}
	
	if (objType == H5G_DATASET){
	    mVpCDataSetAndName.push_back
		(new DataSetAndName_cls(CGroup, stdstrName, objType));
	}

	if (objType == H5G_TYPE){
	    mVpCDataSetAndName.push_back
		(new DataSetAndName_cls(CGroup, stdstrName, objType));
	}
    }
}

void ReadHDF5Data_cls::
ShowErrorMessage(){
    if (mstrError.IsEmpty() == TRUE) return;
    wxMessageBox(mstrError, wxT("HDF5 Format Warning."), wxICON_ERROR);
    mstrError.clear();
}

bool ReadHDF5Data_cls::
CheckGrammer(GrammerHDF5_cls &CGrammerHDF5){
    //GrammerHDF5_cls CGrammerHDF5;

    // O[vJEg
    int sizeDataSet = (int)mVpCDataSetAndName.size();
    int groupNum = 0;
    for (int i = 0; i < sizeDataSet; i++){
	wxString strGroup = mVpCDataSetAndName[i]->mstrType;
	if (strGroup.IsSameAs(wxT("GROUP"))) groupNum++;
    }

    // eO[v̗vf
    std::vector < std::vector < int > > VGrammer(groupNum);
    groupNum = -1;
    for (int i = 0; i < sizeDataSet; i++){
	wxString strGroup = mVpCDataSetAndName[i]->mstrType;
	if (strGroup.IsSameAs(wxT("GROUP"))) groupNum++;
	if (strGroup.IsSameAs(wxT("DATASET"))){
	    if (groupNum == -1) continue;
	    wxString strName = wxString::FromAscii(mVpCDataSetAndName[i]->mstdstrName.c_str());
	    if (strName.IsSameAs(wxT("index")))
		VGrammer[groupNum].push_back(_GRAMMER_INDEX);
	    if (strName.IsSameAs(wxT("xyz")))
		VGrammer[groupNum].push_back(_GRAMMER_XYZ);
	    if (strName.IsSameAs(wxT("vxvyvz")))
		VGrammer[groupNum].push_back(_GRAMMER_VXVYVZ);
	    if (strName.IsSameAs(wxT("axayaz")))
		VGrammer[groupNum].push_back(_GRAMMER_AXAYAZ);
	    if (strName.IsSameAs(wxT("jxjyjz")))
		VGrammer[groupNum].push_back(_GRAMMER_JXJYJZ);
	    if (strName.IsSameAs(wxT("ID")))
		VGrammer[groupNum].push_back(_GRAMMER_ID);
	    if (strName.IsSameAs(wxT("radius")))
		VGrammer[groupNum].push_back(_GRAMMER_RADIUS);
	    if (strName.IsSameAs(wxT("dradius")))
		VGrammer[groupNum].push_back(_GRAMMER_DRADIUS);
	    if (strName.IsSameAs(wxT("value1")))
		VGrammer[groupNum].push_back(_GRAMMER_VALUE1);
	    if (strName.IsSameAs(wxT("dvalue1")))
		VGrammer[groupNum].push_back(_GRAMMER_DVALUE1);
	}
    }

    groupNum++;

    if (groupNum == 0){
	mstrError = _("HDF5 Error. No group detected.");
	return FALSE;
    }
    
    // ŏ̕@L
    CGrammerHDF5.mCGrammerBody.mDimension = VGrammer[0].size();
    for (int i = 0; i < (int)VGrammer[0].size(); i++)
	CGrammerHDF5.mCGrammerBody.mpBody[i] = VGrammer[0][i];
    
    // O[v̗vfňႢ͖`FbN
    bool bSame = TRUE;
    if (groupNum > 1){
	for (int i = 1; i < groupNum; i++){
	    if (VGrammer[i].size() != VGrammer[0].size())
		bSame = FALSE;
	    else
		for (int j = 0; j < (int)VGrammer[0].size(); j++){
		    if (VGrammer[i][j] != VGrammer[0][j])
			bSame = FALSE;
		}
	}
    }

    // xyz͍Œ邩
    for (int i = 0; i < groupNum; i++){
	if (GetPNum(i, wxT("xyz")) == -1){
	    mstrError = _("HDF5 Error. No xyz data in group");
	    return FALSE;
	}
    }

    wxString strItems[9] = {wxT("index"), wxT("vxvyvz"), wxT("axayaz"), wxT("jxjyjz"), wxT("ID"), wxT("radius"), wxT("dradius"), wxT("value1"), wxT("dvalue1")};

    // O[vŗq̈Ⴄ͂Ȃ
    wxString strDiffTag;
    bool bPNumDiff = FALSE;
    for (int i = 0; i < groupNum; i++){
	int PNum = GetPNum(i, wxT("xyz"));
	for (int j = 0; j < 9; j++){
	    int PNumComp = GetPNum(i, strItems[j]);
	    if (PNumComp == -1) continue;
	    if (PNum != PNumComp) {
		bPNumDiff = TRUE; 
		strDiffTag << i << wxT(":") << strItems[j];
		continue;
	    }
	}
    }

    CGrammerHDF5.mStepNumber = groupNum;
    // Agr[g玞ԂƗq𓾂
    
    CGrammerHDF5.mVTime.clear();
    int sizeAttrib = (int)mVpCAttribItem.size();
    for (int i = 0; i < sizeAttrib; i++){
	if (mVpCAttribItem[i]->mstrName.IsSameAs(wxT("time")))
	    CGrammerHDF5.mVTime.push_back(mVpCAttribItem[i]->mValueFloat);
	if (mVpCAttribItem[i]->mstrName.IsSameAs(wxT("number")))
	    CGrammerHDF5.mVPNum.push_back(mVpCAttribItem[i]->mValueInt);
    }

    if (bPNumDiff == TRUE){
	mstrError = _("HDF5 Error. Particle number are not same in a group.");
	mstrError << ENDL << strDiffTag;
	return FALSE;
    }

    if (bSame == FALSE){
	mstrError = _("HDF5 Error. Dataset are not same between groups.");
	return FALSE;
    }
    return TRUE;
}

bool ReadHDF5Data_cls::
ReadHDF5Data(wxString strPath){
    
#ifdef _USE_HDF5
    if (H5::H5File::isHdf5(strPath.ToAscii()) == FALSE) return FALSE;
#endif 
    //mCFile.openFile( strPath.ToAscii(), H5F_ACC_RDWR) ;

    mpCFile = new H5::H5File;

    //mCFile.openFile( strPath.ToAscii(), H5F_ACC_RDONLY) ;
    mpCFile->openFile(strPath.ToAscii(), H5F_ACC_RDONLY) ;
    //mbOpened = TRUE;
    
    mVpCDataSetAndName.clear();
    
    ReadAttrib(*mpCFile, "File");
    ReadGroup(*mpCFile);

    bool bRetValue = TRUE;
    return bRetValue;
}

void ReadHDF5Data_cls::
MakeIndexList(GrammerHDF5_cls &CGrammerHDF5, int snap){
    DataSetAndName_cls *pCDataSet = GetPCDataSet(wxT("index"), snap);
    if (pCDataSet == NULL){
	if (mpIndexList != NULL) delete [] mpIndexList;
	return;
    }
    int dim = pCDataSet->GetDimension(); // ASSERT(dim = 1);
    int dimSize;
    pCDataSet->GetDimensionSize(1, &dimSize);

    if (mpIndexList != NULL && 
	mIndexListSize != dimSize){
	delete [] mpIndexList;
	mpIndexList = NULL;
    }
    
    if (mpIndexList == NULL) mpIndexList = new int [dimSize];
    mIndexListSize = dimSize;
    pCDataSet->CopyDataTo(mpIndexList, H5::PredType::NATIVE_INT);
}

void ReadHDF5Data_cls::
CheckIndexList(int num){
    if (mpIndexList == NULL) return;

    for (int i = 0; i < num; i++) {
	if (mpIndexList[i] >= num || mpIndexList[i] < 0){
	    mpIndexList[i] = 0;
	    mstrError << _("HDF5 Error. Index overflow is detected");
	}
    }
}

void ReadHDF5Data_cls::
MakeSimpleIndexListIfNeeded(int num){
    if (mpIndexList != NULL) return;

    mpIndexList = new int [num];
    mIndexListSize = num;
    for (int i = 0; i < num; i++) mpIndexList[i] = i;
}

int ReadHDF5Data_cls::
GetPNum(int snap, wxString strTag){
    DataSetAndName_cls *pCDataSet = GetPCDataSet(strTag, snap);
    if (pCDataSet == NULL) return -1;
    int dim = pCDataSet->GetDimension(); // ASSERT(dim = 1 or 2);
    int dimSize[2];
    pCDataSet->GetDimensionSize(dim, dimSize); 
    return dimSize[0];
}

void ReadHDF5Data_cls::
CopyHDF5Data3Dim(float (*pData)[3], int snap, wxString strTag){
    DataSetAndName_cls *pCDataSet = GetPCDataSet(strTag, snap);
    if (pCDataSet == NULL) return;
    if (mpIndexList == NULL){
	pCDataSet->CopyDataTo(pData, H5::PredType::NATIVE_FLOAT);
	return; 
    }
    
    int dim = pCDataSet->GetDimension(); // ASSERT(dim = 2);
    int dimSize[2];
    pCDataSet->GetDimensionSize(2, dimSize); // ASSRT(dimSize[1] = 3);
    int datSize = dimSize[0];
    
    float (*pDataTmp)[3] = new float [datSize][3];

    pCDataSet->CopyDataTo(pDataTmp, H5::PredType::NATIVE_FLOAT);
    for (int i = 0; i < datSize; i++)
	memcpy(pData[i], pDataTmp[mpIndexList[i]], sizeof(float)*3);
    
    delete [] pDataTmp;
}

void ReadHDF5Data_cls::
CopyHDF5Data1DimFloat(float *pData,  int snap, wxString strTag){
    DataSetAndName_cls *pCDataSet = GetPCDataSet(strTag, snap);
    if (pCDataSet == NULL) return;
    if (mpIndexList == NULL){
	pCDataSet->CopyDataTo(pData, H5::PredType::NATIVE_FLOAT);
	return; 
    }

    int dim = pCDataSet->GetDimension(); // ASSERT(dim = 1);
    int dimSize[1];
    pCDataSet->GetDimensionSize(1, dimSize); 
    int datSize = dimSize[0];
    float (*pDataTmp) = new float [datSize];
    
    pCDataSet->CopyDataTo(pDataTmp, H5::PredType::NATIVE_FLOAT);
    for (int i = 0; i < datSize; i++)
	pData[i] = pDataTmp[mpIndexList[i]];
    delete [] pDataTmp;
}

void ReadHDF5Data_cls::
CopyHDF5Data1DimInt(int *pData,  int snap, wxString strTag){
    DataSetAndName_cls *pCDataSet = GetPCDataSet(strTag, snap);
    if (pCDataSet == NULL) return;
    if (mpIndexList == NULL){
	pCDataSet->CopyDataTo(pData, H5::PredType::NATIVE_INT);
	return; 
    }

    int dim = pCDataSet->GetDimension(); // ASSERT(dim = 1);
    int dimSize[1];
    pCDataSet->GetDimensionSize(1, dimSize); 
    int datSize = dimSize[0];
    int (*pDataTmp) = new int [datSize];
    
    pCDataSet->CopyDataTo(pDataTmp, H5::PredType::NATIVE_INT);
    for (int i = 0; i < datSize; i++)
	pData[i] = pDataTmp[mpIndexList[i]];
    delete [] pDataTmp;
}

void ReadHDF5Data_cls::
CopyHDF5Data1DimUC(unsigned char *pData,  int snap, wxString strTag){
    DataSetAndName_cls *pCDataSet = GetPCDataSet(strTag, snap);
    if (pCDataSet == NULL) return;
    if (mpIndexList == NULL){
	pCDataSet->CopyDataTo(pData, H5::PredType::NATIVE_UCHAR);
	return; 
    }

    int dim = pCDataSet->GetDimension(); // ASSERT(dim = 1);
    int dimSize[1];
    pCDataSet->GetDimensionSize(1, dimSize); 
    int datSize = dimSize[0];
    unsigned char (*pDataTmp) = new unsigned char [datSize];
    
    pCDataSet->CopyDataTo(pDataTmp, H5::PredType::NATIVE_UCHAR);
    for (int i = 0; i < datSize; i++)
	pData[i] = pDataTmp[mpIndexList[i]];
    delete [] pDataTmp;
}

#else
ReadHDF5Data_cls::
ReadHDF5Data_cls(){}

ReadHDF5Data_cls::
~ReadHDF5Data_cls(){}

bool ReadHDF5Data_cls::
CheckGrammer(GrammerHDF5_cls& CGrammerHDF5){
    return FALSE;
}

bool ReadHDF5Data_cls::
ReadHDF5Data(wxString strPath){
    return FALSE;
}

void ReadHDF5Data_cls::
MakeIndexList(GrammerHDF5_cls &CGrammerHDF5, int snap){
}

void ReadHDF5Data_cls::
CheckIndexList(int num){
}

void ReadHDF5Data_cls::
MakeSimpleIndexListIfNeeded(int num){
}

void ReadHDF5Data_cls::
CopyHDF5Data3Dim(float (*pX)[3], int snap, wxString strTag){
}

void ReadHDF5Data_cls::
CopyHDF5Data1DimFloat(float (*pX), int snap, wxString strTag){
}

void ReadHDF5Data_cls::
CopyHDF5Data1DimInt(int (*pX), int snap, wxString strTag){
}

void ReadHDF5Data_cls::
CopyHDF5Data1DimUC(unsigned char (*pX), int snap, wxString strTag){
}

void ReadHDF5Data_cls::
ShowErrorMessage(){
}
#endif
