
#include "clsExposure.h"
#include "clsExposureGUI.h"

#include "../../GL/glsl.h"

#include "../COMMON/CommonDefinitions.h"

Exposure_cls::
Exposure_cls() : mpCGUI(NULL){
    mCConfList.Clear();
    mCConfList.Add(&mCLevel, wxT("ExposureLevel="), wxT("0 1"));
    mCConfList.Add(&mCGamma, wxT("ExposureGamma="), wxT("1"));
    mCConfList.Add(&mCHueOffset, wxT("ExposureHueOffset="), wxT("0"));
    mCConfList.Add(&mCSaturation, wxT("ExposureSaturation="), wxT("0"));

    mAlpha = 1;
    mBlendMode = 0;
}

Exposure_cls::
~Exposure_cls(){
    if (mpCGUI != NULL) delete mpCGUI;
}

void Exposure_cls::
InstantiateGUI(wxWindow *pCParent){
    mpCGUI = new ExposureGUI_cls(this, pCParent);
}

void Exposure_cls::
Initialize(wxWindow *pCParent, bool bShaderAvailable){
    mbShaderAvailable = bShaderAvailable;
    if (mbShaderAvailable == FALSE) return;
    
    MakeShader(pCParent, mGL2Program,
	       wxT("GL/exposure.vert"), wxT("GL/exposure.frag"));

    glUseProgram(mGL2Program);
    SetParamIDs(mGL2Program);
    SetParamTexture(0);
    glUseProgram(0);

    if (mpCGUI != NULL)
	mpCGUI->UpdateStatus();
}

void Exposure_cls::
Begin(){
    if (mbShaderAvailable == FALSE) return;
    glUseProgram(mGL2Program);
    SetParamExposure();
}

void Exposure_cls::
End(){
    if (mbShaderAvailable == FALSE) return;
    glUseProgram(0);
};

void Exposure_cls::
SetParamIDs(GLint program){
    mParamIDTexture  = glGetUniformLocation(program, "textureImage");
    mParamIDLevelMin = glGetUniformLocation(program, "levelMin");
    mParamIDLevelMax = glGetUniformLocation(program, "levelMax");
    mParamIDExposure = glGetUniformLocation(program, "exposure");
    mParamIDGamma    = glGetUniformLocation(program, "gamma");
    mParamIDHueOffset = glGetUniformLocation(program, "hueOffset");
    mParamIDSaturation = glGetUniformLocation(program, "saturation");
    mParamIDBlendMode = glGetUniformLocation(program, "blendMode");
    mParamIDAlpha     = glGetUniformLocation(program, "alpha");
}

void Exposure_cls::
SetParamTexture(int value){
    glUniform1i(mParamIDTexture, value);
}

void Exposure_cls::
SetParamExposure(){
    float levelMin = mCLevel.mpBody[0];
    float levelMax = mCLevel.mpBody[1];
    float gamma = mCGamma.mBody;
    float hueOffset = mCHueOffset.mBody;
    float saturation = mCSaturation.mBody;

    glUniform4f(mParamIDLevelMin, levelMin, levelMin, levelMin, levelMin);
    glUniform4f(mParamIDLevelMax, levelMax, levelMax, levelMax, levelMax);
    glUniform4f(mParamIDGamma, gamma, gamma, gamma, gamma);
    glUniform1i(mParamIDHueOffset, hueOffset);
    glUniform1i(mParamIDSaturation, saturation);
    glUniform1i(mParamIDBlendMode, mBlendMode);
    glUniform1f(mParamIDAlpha, mAlpha);
}

bool Exposure_cls::
IsDefault(){
    bool bDefault = TRUE;
    if (mCLevel.mpBody[0] != 0 || mCLevel.mpBody[1] != 1) bDefault = FALSE;
    if (mCGamma.mBody != 1) bDefault = FALSE;
    if (mCHueOffset.mBody != 0) bDefault = FALSE;
    if (mCSaturation.mBody != 0) bDefault = FALSE;
    return bDefault;
}

// ----------------

void Exposure_cls::
ReadStrLine(wxString strLine){
    mCConfList.ReadStrLineAll(strLine);
}

wxString Exposure_cls::
Serialize(){
    return mCConfList.SerializeAll();
}
