// -*- c -*-

#include "clsPositionAndDirection.h"

#include "../COMMON/MyVector.h"
//#include "../GLOBJ/GLOBJECTS.h"

#include <sstream>

#define ENDL wxT("\n")
#define X 0
#define Y 1
#define Z 2

#define LOOP_XYZ for (int xyz = 0; xyz < 3; xyz++)

Direction_cls::
Direction_cls(double tgtX, double tgtY, double tgtZ,
	      double topX, double topY, double topZ,
	      double rgtX, double rgtY, double rgtZ){
    mTgt[X] = tgtX; mTgt[Y] = tgtY; mTgt[Z] = tgtZ;
    mTop[X] = topX; mTop[Y] = topY; mTop[Z] = topZ;
    mRgt[X] = rgtX; mRgt[Y] = rgtY; mRgt[Z] = rgtZ;
    LOOP_XYZ mYPRInt[xyz] = 0;
    LOOP_XYZ mYPRPre[xyz] = 0;
    UpdateYPR();
}

Direction_cls::
Direction_cls(Direction_cls &CDir){
    LOOP_XYZ mTgt[xyz] = CDir.mTgt[xyz];
    LOOP_XYZ mTop[xyz] = CDir.mTop[xyz];
    LOOP_XYZ mRgt[xyz] = CDir.mRgt[xyz];
    LOOP_XYZ mYPRInt[xyz] = CDir.mYPRInt[xyz];
    LOOP_XYZ mYPRPre[xyz] = CDir.mYPRPre[xyz];
    LOOP_XYZ mYPR[xyz] = CDir.mYPR[xyz];
    //UpdateYPR();
}

void Direction_cls::
Setv(double tgt[3], double top[3]){
    mTgt[X] = tgt[X]; mTgt[Y] = tgt[Y]; mTgt[Z] = tgt[Z];
    mTop[X] = top[X]; mTop[Y] = top[Y]; mTop[Z] = top[Z];
    NormalizeAndOrthogonalize();
}

void Direction_cls::
SetvUnitZ(int direction){
    if (direction > 0){
	mTgt[X] =  0; mTgt[Y] = 0; mTgt[Z] = 1;
	mTop[X] =  0; mTop[Y] = 1; mTop[Z] = 0;
	mRgt[X] = -1; mRgt[Y] = 0; mRgt[Z] = 0; }
    else{
	mTgt[X] =  0; mTgt[Y] = 0; mTgt[Z] = -1;
	mTop[X] =  0; mTop[Y] = 1; mTop[Z] =  0;
	mRgt[X] =  1; mRgt[Y] = 0; mRgt[Z] =  0;
    }
}

void Direction_cls::
RejectInvalidDirection(){
    // tgttopsȂtopKȕ
    int rad = 0; 
    if (MyVector::Orthogonalize(mTgt, mTop) == FALSE)
	mTop[X] = 0; mTop[Y] = 1; mTop[Z] = 0;
    if (MyVector::Orthogonalize(mTgt, mTop) == FALSE)
	mTop[X] = 1; mTop[Y] = 0; mTop[Z] = 0;

}

void Direction_cls::
NormalizeAndOrthogonalize(){
    MyVector::Normalize    (mTgt, 1.0);
    if (MyVector::Orthogonalize(mTgt, mTop) == FALSE){
	mTgt[X] = 0; mTgt[Y] = 0; mTgt[Z] = -1;
	mTop[X] = 0; mTop[Y] = 1; mTop[Z] = 0;
    }
    MyVector::Normalize    (mTop, 1.0);
    MyVector::GetCrossedVector(mTgt, mTop, mRgt);
}

void Direction_cls::
SetDirection(unsigned int type){
    switch (type){
    case _DIRECTION_X_PLUS:
	mTgt[X] = 1; mTgt[Y] = 0; mTgt[Z] = 0;
	mTop[X] = 0; mTop[Y] = 0; mTop[Z] = 1;
	break;
    case _DIRECTION_X_MINUS:
	mTgt[X] =-1; mTgt[Y] = 0; mTgt[Z] = 0;
	mTop[X] = 0; mTop[Y] = 0; mTop[Z] = 1;
	break;
    case _DIRECTION_Y_PLUS:
	mTgt[X] = 0; mTgt[Y] = 1; mTgt[Z] = 0;
	mTop[X] = 0; mTop[Y] = 0; mTop[Z] = 1;
	break;
    case _DIRECTION_Y_MINUS:
	mTgt[X] = 0; mTgt[Y] =-1; mTgt[Z] = 0;
	mTop[X] = 0; mTop[Y] = 0; mTop[Z] = 1;
	break;
    case _DIRECTION_Z_PLUS:
	mTgt[X] = 0; mTgt[Y] = 0; mTgt[Z] = 1;
	mTop[X] = 0; mTop[Y] = 1; mTop[Z] = 0;
	break;
    case _DIRECTION_Z_MINUS:
	mTgt[X] = 0; mTgt[Y] = 0; mTgt[Z] =-1;
	mTop[X] = 0; mTop[Y] = 1; mTop[Z] = 0;
	break;
    }
    NormalizeAndOrthogonalize();
}

void Direction_cls::
UpdateYPR(bool bContinuous){
    double rgt[3];
    rgt[Z] = mTgt[X] * mTop[Y] - mTgt[Y] * mTop[X];
    rgt[X] = mTgt[Y] * mTop[Z] - mTgt[Z] * mTop[Y];
    rgt[Y] = mTgt[Z] * mTop[X] - mTgt[X] * mTop[Z];

    double yaw, pitch, roll;
    
    yaw   = atan2(/*mMatrix[6]*/-mTgt[Y], /*mMatrix[2]*/-mTgt[X]) + 3.14159; 
    pitch = asin(/*mMatrix[10]*/-mTgt[Z]);
    roll  = -asin(/*mMatrix[8]*/mRgt[Z]/cos(pitch)); 
    if (/*mMatrix[9]*/mTop[Z] < 0) roll = 3.14159 - roll;
    
    mYPR[X] = 180*roll  / 3.14159;
    mYPR[Y] = 180*pitch / 3.14159;
    mYPR[Z] = 180*yaw   / 3.14159;

    if (bContinuous == TRUE){
	if (mYPR[Z] > 180 && mYPRPre[Z] < 180) mYPRInt[Z] --;
	if (mYPR[Z] < 180 && mYPRPre[Z] > 180) mYPRInt[Z] ++;
    }

    LOOP_XYZ mYPRPre[xyz] = mYPR[xyz];
    
    mYPR[X] += 360 * mYPRInt[X];
    mYPR[Z] += 360 * mYPRInt[Z];
}

void Direction_cls::
RotateYPR(double Yaw, double Pitch, double Roll){
    RotateYaw(Yaw);
    NormalizeAndOrthogonalize();
    RotatePitch(Pitch);
    NormalizeAndOrthogonalize();
    RotateRoll(Roll);
}

void Direction_cls::
RotateYPR3d(double pYPR[3]){
    RotateYPR(pYPR[2], pYPR[1], pYPR[0]);
}

void Direction_cls::
RotateYPR4d(double pYPRY[4], bool bInverse){
    if (bInverse == TRUE){
	RotateYaw(-pYPRY[3]);
	RotateRoll(-pYPRY[2]);
	RotatePitch(-pYPRY[1]);
	RotateYaw(-pYPRY[0]);
    }
    else {
	RotateYaw(pYPRY[0]);
	RotatePitch(pYPRY[1]);
	RotateRoll(pYPRY[2]);
	RotateYaw(pYPRY[3]);
    }
}

void Direction_cls::
RotateYaw(double Yaw){
    double YawRad = Yaw * 3.14159 / 180;
    double cosYaw = cos(YawRad);
    double sinYaw = sin(YawRad);

    double tgtN[3] = {
	cosYaw * mTgt[X] - sinYaw * mRgt[X],
	cosYaw * mTgt[Y] - sinYaw * mRgt[Y],
	cosYaw * mTgt[Z] - sinYaw * mRgt[Z] };
    double rgtN[3] = {
	cosYaw * mRgt[X] + sinYaw * mTgt[X],
	cosYaw * mRgt[Y] + sinYaw * mTgt[Y],
	cosYaw * mRgt[Z] + sinYaw * mTgt[Z] };
    
    LOOP_XYZ mTgt[xyz] = tgtN[xyz];
    LOOP_XYZ mRgt[xyz] = rgtN[xyz];
}

void Direction_cls::
RotatePitch(double Pitch){
    double PitchRad = Pitch * 3.14159 / 180;
    double cosPitch = cos(PitchRad);
    double sinPitch = sin(PitchRad);

    double tgtN[3] = {
	cosPitch * mTgt[X] - sinPitch * mTop[X],
	cosPitch * mTgt[Y] - sinPitch * mTop[Y],
	cosPitch * mTgt[Z] - sinPitch * mTop[Z] };
    double topN[3] = {
	cosPitch * mTop[X] + sinPitch * mTgt[X],
	cosPitch * mTop[Y] + sinPitch * mTgt[Y],
	cosPitch * mTop[Z] + sinPitch * mTgt[Z] };
    
    LOOP_XYZ mTgt[xyz] = tgtN[xyz];
    LOOP_XYZ mTop[xyz] = topN[xyz];
}

void Direction_cls::
RotateRoll(double Roll){
    double RollRad = Roll * 3.14159 / 180;
    double cosRoll = cos(RollRad);
    double sinRoll = sin(RollRad);

    double topN[3] = {
	cosRoll * mTop[X] + sinRoll * mRgt[X],
	cosRoll * mTop[Y] + sinRoll * mRgt[Y],
	cosRoll * mTop[Z] + sinRoll * mRgt[Z] };

    double rgtN[3] = {
	cosRoll * mRgt[X] - sinRoll * mTop[X],
	cosRoll * mRgt[Y] - sinRoll * mTop[Y],
	cosRoll * mRgt[Z] - sinRoll * mTop[Z] };

    LOOP_XYZ mTop[xyz] = topN[xyz];
    LOOP_XYZ mRgt[xyz] = rgtN[xyz];
}


// xNg̍WnŌ̒lɕύX
void Direction_cls::
RotTransformToThisCoordinate(double vec[3]){
    double orgVec[3];
    LOOP_XYZ orgVec[xyz] = vec[xyz];
    vec[X] = MyVector::GetScalarProduct(orgVec, mRgt);
    vec[Y] = MyVector::GetScalarProduct(orgVec, mTop);
    vec[Z] = -MyVector::GetScalarProduct(orgVec, mTgt);
}

// --

bool Direction_cls::
ReadStrLine(std::wstring strLine){
    std::wstring::size_type loc = strLine.find(L"Direction=", 0);
    if (loc != std::string::npos){
	std::wstring strLineBody = strLine.c_str() + loc + wcslen(L"Direction=");
	std::wstringstream strTmp(strLineBody);
	LOOP_XYZ strTmp >> mTgt[xyz];
	LOOP_XYZ strTmp >> mTop[xyz];
	LOOP_XYZ strTmp >> mRgt[xyz];
    }
    else 
	return FALSE;
    return TRUE;
}

wxString Direction_cls::
Serialize(){
    wxString sOut;
    sOut << L"Direction=" << mTgt[X] << wxT(" ") << mTgt[Y] << wxT(" ") << mTgt[Z];
    LOOP_XYZ sOut << wxT(" ") << mTop[xyz];
    LOOP_XYZ sOut << wxT(" ") << mRgt[xyz];
    sOut << wxT("\n");
    return sOut;
}

bool operator==(const Direction_cls& CLeft, const Direction_cls& CRight){
    bool retVal = TRUE;
    LOOP_XYZ{
        if (CLeft.mTgt[xyz] != CRight.mTgt[xyz]) retVal = FALSE;
        if (CLeft.mRgt[xyz] != CRight.mRgt[xyz]) retVal = FALSE;
        if (CLeft.mTop[xyz] != CRight.mTop[xyz]) retVal = FALSE;
    }
    return retVal;
}

// ------

Position_cls::
Position_cls(double posX, double posY, double posZ,
	     double velX, double velY, double velZ){
    mPos[X] = posX; mPos[Y] = posY; mPos[Z] = posZ;
    mVel[X] = velX; mVel[Y] = velY; mVel[Z] = velZ;
}

Position_cls::
Position_cls(const Position_cls &CPos){
    LOOP_XYZ mPos[xyz] = CPos.mPos[xyz];
    LOOP_XYZ mVel[xyz] = CPos.mVel[xyz];
}

void Position_cls::
Set(Position_cls &CPos){
    LOOP_XYZ mPos[xyz] = CPos.mPos[xyz];
    LOOP_XYZ mVel[xyz] = CPos.mVel[xyz];
}

void Position_cls::
SetZero(){
    LOOP_XYZ mPos[xyz] = 0;
    LOOP_XYZ mVel[xyz] = 0;
}

bool Position_cls::
ReadStrLine(std::wstring strLine){
    std::wstring::size_type loc = strLine.find(L"Position=", 0);
    if (loc != std::string::npos){
	std::wstring strLineBody = strLine.c_str() + loc + wcslen(L"Position=");
	std::wstringstream strTmp(strLineBody);
	LOOP_XYZ strTmp >> mPos[xyz];
    }
    else 
	return FALSE;
    return TRUE;
}

wxString Position_cls::
Serialize(){
    wxString sOut;
    sOut << L"Position=" << mPos[X] << wxT(" ") << mPos[Y] << wxT(" ") << mPos[Z] << wxT("\n");
    return sOut;
}

bool operator==(const Position_cls& CLeft, const Position_cls& CRight){
    bool retVal = TRUE;
    LOOP_XYZ{
        if (CLeft.mPos[xyz] != CRight.mPos[xyz]) retVal = FALSE;
        if (CLeft.mVel[xyz] != CRight.mVel[xyz]) retVal = FALSE;
    }
    return retVal;
}

// ------

/*
PositionAndDirection_cls::
PositionAndDirection_cls() :
    Position_cls (0, 0, 0,
		  0, 0, 0), 
    Direction_cls(0, 0, -1, 
		  0, 1, 0,
		  1, 0, 0)
{
    mSensitivity = 1;
    mAspectRatio = 1;
    LOOP_XYZ mATranslateScale[xyz] = 1;
    LOOP_XYZ mAncPos[xyz] = 0;
	
}
*/

PositionAndDirection_cls::
PositionAndDirection_cls(double posX, double posY, double posZ,
			 double velX, double velY, double velZ,
			 double tgtX, double tgtY, double tgtZ,
			 double topX, double topY, double topZ,
			 double rgtX, double rgtY, double rgtZ) : 
    Position_cls (posX, posY, posZ,
		  velX, velY, velZ), 
    Direction_cls(tgtX, tgtY, tgtZ, 
		  topX, topY, topZ,
		  rgtX, rgtY, rgtZ)
{
    mSensitivity = 1;
    mAspectRatio = 1;
    LOOP_XYZ mATranslateScale[xyz] = 1;
    LOOP_XYZ mAncPos[xyz] = 0;
}

PositionAndDirection_cls::
PositionAndDirection_cls(Position_cls &CPos, Direction_cls &CDir) :
    Position_cls (CPos),
    Direction_cls(CDir)
{
    mSensitivity = 1;
    mAspectRatio = 1;
    LOOP_XYZ mATranslateScale[xyz] = 1;
    LOOP_XYZ mAncPos[xyz] = 0;
}

void PositionAndDirection_cls::
SetSensitivity(double setsensitivity){
    mSensitivity = setsensitivity;
}

void PositionAndDirection_cls::
SetAspectRatio(double aspectRatio){
    mAspectRatio = aspectRatio;
}

double PositionAndDirection_cls::
GetDistanceToAnc(){
    return MyVector::GetNormRelative(mPos, mAncPos);
}

double PositionAndDirection_cls::
GetProjectedDistanceToAnc(){
    double rel[3];
    LOOP_XYZ rel[xyz] = mAncPos[xyz] - mPos[xyz];
    double distance = MyVector::GetScalarProduct(mTgt, rel);
    if (distance < 0) distance = -distance;
    return distance;
}

double PositionAndDirection_cls::
GetProjectedDistance(double tgtPos[3]){
    double rel[3];
    LOOP_XYZ rel[xyz] = tgtPos[xyz] - mPos[xyz];
    double distance = MyVector::GetScalarProduct(mTgt, rel);
    if (distance < 0) distance = -distance;
    return distance;
}

void PositionAndDirection_cls::
SetAncPos(double pos[3]){
    LOOP_XYZ mAncPos[xyz] = pos[xyz];
}

void PositionAndDirection_cls::
TrackBall(float prevX, float prevY, float currX, float currY){

    float dx = (currX - prevX) * mSensitivity;
    float dy = (currY - prevY) * mSensitivity;

    if (dx > 0.2) dx = 0.2f;
    if (dx < -0.2) dx = -0.2f;
    if (dy > 0.2) dy = 0.2f;
    if (dy < -0.2) dy = -0.2f;
    
    //PointAt(mAncPos);
    
    LOOP_XYZ mPos[xyz] -= mAncPos[xyz];

    float distance = MyVector::GetNorm(mPos);
    
    float dxMass = 1 - currY * currY; 
    float dyMass = 1 - currX * currX; 
    if (dxMass < 0) dxMass = 0;
    if (dyMass < 0) dyMass = 0;

    LOOP_XYZ{
	mPos[xyz]   -= mRgt[xyz] * dx * dxMass * distance ;
	mTgt[xyz]   += mRgt[xyz] * dx * dxMass; }
    
    LOOP_XYZ{
	mPos[xyz]   -= mTop[xyz] * dy * dyMass * distance;
	mTgt[xyz]   += mTop[xyz] * dy * dyMass; }

    LOOP_XYZ
	mTop[xyz]   += (currX * dy - currY * dx) * mRgt[xyz] ;
    
    MyVector::Normalize(mPos, distance);

    LOOP_XYZ mPos[xyz] += mAncPos[xyz];
    
    NormalizeAndOrthogonalize();
}

void PositionAndDirection_cls::
Zoom(float dx, float dy){

    LOOP_XYZ mPos[xyz] -= mAncPos[xyz];

    float distance = MyVector::GetNorm(mPos);
    distance *= (1.0 + dy - dx * 0.2);
    
    MyVector::Normalize(mPos, distance);

    LOOP_XYZ mPos[xyz] += mAncPos[xyz];

    NormalizeAndOrthogonalize();
}

void PositionAndDirection_cls::
Translate(float dx, float dy, 
	  int mode, 
	  PositionAndDirection_cls *pCParentCoord){

    if (pCParentCoord != NULL){
	float lMatrixMove[5][3];

	lMatrixMove[0][X] = pCParentCoord->mRgt[X];
	lMatrixMove[0][Y] = pCParentCoord->mRgt[Y]; 
	lMatrixMove[0][Z] = pCParentCoord->mRgt[Z];
	LOOP_XYZ if (lMatrixMove[0][xyz] < 0) lMatrixMove[0][xyz] *= -1;
	
	lMatrixMove[1][X] = pCParentCoord->mTop[X]; 
	lMatrixMove[1][Y] = pCParentCoord->mTop[Y];
	lMatrixMove[1][Z] = pCParentCoord->mTop[Z];
	LOOP_XYZ if (lMatrixMove[1][xyz] < 0) lMatrixMove[1][xyz] *= -1;
	
	lMatrixMove[2][X] = pCParentCoord->mTgt[X];
	lMatrixMove[2][Y] = pCParentCoord->mTgt[Y];
	lMatrixMove[2][Z] = pCParentCoord->mTgt[Z]; 
	LOOP_XYZ if (lMatrixMove[2][xyz] < 0) lMatrixMove[2][xyz] *= -1;
	
	LOOP_XYZ lMatrixMove[3][xyz] = 1;
	LOOP_XYZ lMatrixMove[4][xyz] = 1;

	LOOP_XYZ{
	    mPos[xyz]   -= mRgt[xyz] * dx * mSensitivity * mAspectRatio * lMatrixMove[mode][xyz] * mATranslateScale[xyz];
	    mPos[xyz]   -= mTop[xyz] * dy * mSensitivity * lMatrixMove[mode][xyz] * mATranslateScale[xyz];
	}
	
    }
    else
	LOOP_XYZ{
	    mPos[xyz]   -= mRgt[xyz] * dx * mSensitivity * mAspectRatio * msbMatrixMove[mode][xyz] * mATranslateScale[xyz];
	    mPos[xyz]   -= mTop[xyz] * dy * mSensitivity * msbMatrixMove[mode][xyz] * mATranslateScale[xyz];
	}
}

void PositionAndDirection_cls::
PointAt(double point[3]){
    /*
    LOOP_XYZ
	mTgt[xyz] = point[xyz] - mPos[xyz];
    NormalizeAndOrthogonalize();
    */
    LOOP_XYZ mAncPos[xyz] = point[xyz];
    PointAtAnc();
}

void PositionAndDirection_cls::
PointAtAnc(){
    //PointAt(mAncPos);
    LOOP_XYZ
	mTgt[xyz] = mAncPos[xyz] - mPos[xyz];
    
    NormalizeAndOrthogonalize();
}

void PositionAndDirection_cls::
RotateAroundOriginYPR(double Yaw, double Pitch, double Roll){
    RotateAroundOriginYaw(Yaw);
    RotateAroundOriginPitch(Pitch);
    RotateAroundOriginRoll(Roll);
}

void PositionAndDirection_cls::
RotateVectorZ(double pVec[3], double cosYaw, double sinYaw){
    double VecN[3] = {
	cosYaw * pVec[0] - sinYaw * pVec[1],
	sinYaw * pVec[0] + cosYaw * pVec[1],
	pVec[2] };
    LOOP_XYZ pVec[xyz] = VecN[xyz];
}

void PositionAndDirection_cls::
RotateAroundOriginYaw(double Yaw){
    double YawRad = Yaw * 3.14159 / 180;
    double cosYaw = cos(YawRad), sinYaw = sin(YawRad);
    RotateVectorZ(mPos, cosYaw, sinYaw);
    RotateVectorZ(mTgt, cosYaw, sinYaw);
    RotateVectorZ(mTop, cosYaw, sinYaw);
    RotateVectorZ(mRgt, cosYaw, sinYaw);
}

void PositionAndDirection_cls::
RotateVectorY(double pVec[3], double cosPitch, double sinPitch){
    double VecN[3] = {
	cosPitch * pVec[0] + sinPitch * pVec[2],
	pVec[1],
	- sinPitch * pVec[0] + cosPitch * pVec[2]};
    LOOP_XYZ pVec[xyz] = VecN[xyz];
}

void PositionAndDirection_cls::
RotateAroundOriginPitch(double Pitch){
    double PitchRad = Pitch * 3.14159 / 180;
    double cosPitch = cos(PitchRad), sinPitch = sin(PitchRad);
    RotateVectorY(mPos, cosPitch, sinPitch);
    RotateVectorY(mTgt, cosPitch, sinPitch);
    RotateVectorY(mTop, cosPitch, sinPitch);
    RotateVectorY(mRgt, cosPitch, sinPitch);
}

void PositionAndDirection_cls::
RotateVectorX(double pVec[3], double cosRoll, double sinRoll){
    double VecN[3] = {
	pVec[0],
	cosRoll * pVec[1] - sinRoll * pVec[2],
	sinRoll * pVec[1] + cosRoll * pVec[2]};
    LOOP_XYZ pVec[xyz] = VecN[xyz];
}

void PositionAndDirection_cls::
RotateAroundOriginRoll(double Roll){
    double RollRad = Roll * 3.14159 / 180;
    double cosRoll = cos(RollRad), sinRoll = sin(RollRad);
    RotateVectorX(mPos, cosRoll, sinRoll);
    RotateVectorX(mTgt, cosRoll, sinRoll);
    RotateVectorX(mTop, cosRoll, sinRoll);
    RotateVectorX(mRgt, cosRoll, sinRoll);
}

void PositionAndDirection_cls::
GetGLMatrix(double matrix[16]){
    
    mRgt[Z] = -(mTgt[X] * mTop[Y] - mTgt[Y] * mTop[X]);
    mRgt[X] = -(mTgt[Y] * mTop[Z] - mTgt[Z] * mTop[Y]);
    mRgt[Y] = -(mTgt[Z] * mTop[X] - mTgt[X] * mTop[Z]);

    // IuWFNg]s
    matrix[0]=mTgt[X];matrix[4]=mRgt[X];matrix[8] = mTop[X]; 
    matrix[1]=mTgt[Y];matrix[5]=mRgt[Y];matrix[9] = mTop[Y]; 
    matrix[2]=mTgt[Z];matrix[6]=mRgt[Z];matrix[10]= mTop[Z]; 
    matrix[3]=0      ;matrix[7]=0      ;matrix[11]=0;
    
    matrix[12] = mPos[X];
    matrix[13] = mPos[Y];
    matrix[14] = mPos[Z];
    matrix[15] = 1;
}

PositionAndDirection_cls PositionAndDirection_cls::
GetPosAndDirectionInThisCoordinate(PositionAndDirection_cls CPosAndDir){
    PositionAndDirection_cls CNew;
    LOOP_XYZ CNew.mRgt[xyz] = 0;
    LOOP_XYZ CNew.mTop[xyz] = 0;
    LOOP_XYZ CNew.mTgt[xyz] = 0;
    
    LOOP_XYZ CNew.mPos[X] += CPosAndDir.mPos[xyz] * mRgt[xyz];
    LOOP_XYZ CNew.mPos[Y] += CPosAndDir.mPos[xyz] * mTop[xyz];
    LOOP_XYZ CNew.mPos[Z] -= CPosAndDir.mPos[xyz] * mTgt[xyz];

    LOOP_XYZ CNew.mRgt[X] += CPosAndDir.mRgt[xyz] * mRgt[xyz];
    LOOP_XYZ CNew.mRgt[Y] += CPosAndDir.mRgt[xyz] * mTop[xyz];
    LOOP_XYZ CNew.mRgt[Z] -= CPosAndDir.mRgt[xyz] * mTgt[xyz];

    LOOP_XYZ CNew.mTop[X] += CPosAndDir.mTop[xyz] * mRgt[xyz];
    LOOP_XYZ CNew.mTop[Y] += CPosAndDir.mTop[xyz] * mTop[xyz];
    LOOP_XYZ CNew.mTop[Z] -= CPosAndDir.mTop[xyz] * mTgt[xyz];

    LOOP_XYZ CNew.mTgt[X] += CPosAndDir.mTgt[xyz] * mRgt[xyz];
    LOOP_XYZ CNew.mTgt[Y] += CPosAndDir.mTgt[xyz] * mTop[xyz];
    LOOP_XYZ CNew.mTgt[Z] -= CPosAndDir.mTgt[xyz] * mTgt[xyz];

    return CNew;
}

bool operator==(const PositionAndDirection_cls& CLeft,
		const PositionAndDirection_cls& CRight){
    bool retVal = TRUE;
    if (!(*(Direction_cls*)&CLeft == *(Direction_cls*)&CRight)) retVal = FALSE;
    if (!(*(Position_cls*)&CLeft == *(Position_cls*)&CRight)) retVal = FALSE;
    return retVal;
}

// ---

bool PositionAndDirection_cls::
ReadStrLine(std::wistream &istream){
    std::wstring strLine;
    std::getline(istream, strLine);
    if (Position_cls::ReadStrLine(strLine) == FALSE) return FALSE;
    std::getline(istream, strLine);
    if (Direction_cls::ReadStrLine(strLine) == FALSE) return FALSE;
    return TRUE;
}

wxString PositionAndDirection_cls::
Serialize(){
    return Position_cls::Serialize() + Direction_cls::Serialize();
}

bool PositionAndDirection_cls::msbMatrixMove[5][3] = {
    {1, 0, 0},
    {0, 1, 0},
    {0, 0, 1},
    {1, 1, 1},
    {1, 1, 1},
};
