// "clsFBO.h"
// (setq c-basic-offset 4)

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
PFNGLGENFRAMEBUFFERSEXTPROC glGenFramebuffersEXT;
PFNGLDELETEFRAMEBUFFERSEXTPROC glDeleteFramebuffersEXT;
PFNGLBINDFRAMEBUFFEREXTPROC glBindFramebufferEXT;
PFNGLFRAMEBUFFERTEXTURE2DEXTPROC glFramebufferTexture2DEXT;
PFNGLGENRENDERBUFFERSEXTPROC glGenRenderbuffersEXT;
PFNGLDELETERENDERBUFFERSEXTPROC glDeleteRenderbuffersEXT;
PFNGLBINDRENDERBUFFEREXTPROC glBindRenderbufferEXT;
PFNGLRENDERBUFFERSTORAGEEXTPROC glRenderbufferStorageEXT;
PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC glFramebufferRenderbufferEXT;
PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC glCheckFramebufferStatusEXT = NULL;
#elif defined __WXMAC__
#define GL_GLEXT_PROTOTYPES 
#include <OpenGL/glu.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

#include "clsFBO.h"
#include "clsPositionAndDirection.h"

#include "../COMMON/CommonDefinitions.h"

FBO_cls::
FBO_cls(int width, int height){
    if (msbInitialized == FALSE) Initialize();
    mWidthBuffer = width; mHeightBuffer = height;
    mbGenerated = FALSE;
    mWidthGenerated = mHeightGenerated = 0;
    mDepth = 16;
    mDepthGenerated = 0;
    mFBOId = 0;
}

FBO_cls::
~FBO_cls(){
    // 쐬ς FBO 폜 eNX`Baseō폜
    if (mbGenerated == TRUE){
	glDeleteRenderbuffersEXT(1, &mRBId);
	glDeleteFramebuffersEXT(1, &mFBOId);
    }
}

void FBO_cls::
Initialize(){
#ifdef WIN32
    glGenFramebuffersEXT =
	(PFNGLGENFRAMEBUFFERSEXTPROC)wglGetProcAddress("glGenFramebuffersEXT");
    glDeleteFramebuffersEXT =
	(PFNGLDELETEFRAMEBUFFERSEXTPROC)wglGetProcAddress("glDeleteFramebuffersEXT");
    glBindFramebufferEXT =
	(PFNGLBINDFRAMEBUFFEREXTPROC)wglGetProcAddress("glBindFramebufferEXT");
    glFramebufferTexture2DEXT =
	(PFNGLFRAMEBUFFERTEXTURE2DEXTPROC)wglGetProcAddress("glFramebufferTexture2DEXT");
    glGenRenderbuffersEXT =
	(PFNGLGENRENDERBUFFERSEXTPROC)wglGetProcAddress("glGenRenderbuffersEXT");
    glDeleteRenderbuffersEXT =
	(PFNGLDELETERENDERBUFFERSEXTPROC)wglGetProcAddress("glDeleteRenderbuffersEXT");
    
    glBindRenderbufferEXT =
	(PFNGLBINDRENDERBUFFEREXTPROC)wglGetProcAddress("glBindRenderbufferEXT");
    glRenderbufferStorageEXT =
	(PFNGLRENDERBUFFERSTORAGEEXTPROC)wglGetProcAddress("glRenderbufferStorageEXT");
    glFramebufferRenderbufferEXT =
	(PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC)wglGetProcAddress("glFramebufferRenderbufferEXT");
    glCheckFramebufferStatusEXT =
	(PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC)wglGetProcAddress("glCheckFramebufferStatusEXT");
    if (glCheckFramebufferStatusEXT == NULL) msbSupported = FALSE;
#endif
    
    if (msbSupported == TRUE){
	GLenum status;                                            
	status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT); 
	if (status != GL_FRAMEBUFFER_COMPLETE_EXT)
	    msbSupported = FALSE;
    }

    msbInitialized = TRUE;
}

void FBO_cls::
SetGeometryBuffer(int width, int height){
    mWidthBuffer = width;
    mHeightBuffer = height;
}

void FBO_cls::
SetDepth(int depth){
    mDepth = depth;
}

void FBO_cls::
MakeFBO(bool bUseRGBA){
    //if (mFBOId != 0) return;
    if (msbSupported == FALSE) return;

    if (mbGenerated == TRUE){
	if (mWidthGenerated == mWidthBuffer &&
	    mHeightGenerated == mHeightBuffer &&
	    mDepth == mDepthGenerated) return;
	else{
	    glDeleteTextures(1, &mTextureID);
	    glDeleteRenderbuffersEXT(1, &mRBId);
	    glDeleteFramebuffersEXT(1, &mFBOId);
	}
    }

    glGenTextures(1, &mTextureID);
    BindTexture();
    
    switch (mDepth){
    case 16:
	if (bUseRGBA == TRUE)
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA16F_ARB, 
			  mWidthBuffer, mHeightBuffer, 0,
			  GL_RGBA, GL_HALF_FLOAT_ARB, 0);
	else
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB16F_ARB, 
			  mWidthBuffer, mHeightBuffer, 0,
			  GL_RGB, GL_HALF_FLOAT_ARB, 0);
	break;
    case 8:
	if (bUseRGBA == TRUE)
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, 
			  mWidthBuffer, mHeightBuffer, 0,
			  GL_RGBA, GL_UNSIGNED_BYTE, 0);
	else
	    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, 
			  mWidthBuffer, mHeightBuffer, 0,
			  GL_RGB, GL_UNSIGNED_BYTE, 0);
	break;
    }
    
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

    // Depth  Stencil ێ郌_[obt@𐶐č쐬
    glGenRenderbuffersEXT(1, &mRBId); 
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, mRBId);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, //GL_DEPTH_COMPONENT,
			     GL_DEPTH_STENCIL_EXT,
			     mWidthBuffer, mHeightBuffer);
    
    // FBO̍쐬
    glGenFramebuffersEXT(1, &mFBOId);

    mbGenerated = TRUE;
    mDepthGenerated = mDepth;
    mWidthGenerated = mWidthBuffer;
    mHeightGenerated = mHeightBuffer;

    ClearFBO();
}

void FBO_cls::
ClearFBO(){
    Begin(TRUE);
    End();
}

void FBO_cls::
Begin(bool bClear, bool bCalledInFBOStack){
    mbClear = bClear;
    //if (msVpCStack.size() != 0)	glPopAttrib();
    
    msVpCStack.push_back(this);
    if (bCalledInFBOStack == FALSE) glPushAttrib(GL_ALL_ATTRIB_BITS);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, mFBOId);
    
    // FBOɃeNX`IuWFNgƃ_[obt@ 
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
			      GL_TEXTURE_2D, mTextureID, 0);
    
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT,
				 GL_RENDERBUFFER_EXT, mRBId);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_STENCIL_ATTACHMENT_EXT,
				 GL_RENDERBUFFER_EXT, mRBId);
    
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);

    // ̂܂ܕ`FBOɈڂ
    glViewport(0, 0, mWidthBuffer, mHeightBuffer);

    glClearColor(0,0,0,0);
    if (bClear == TRUE)
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    else
	glClear(GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

void FBO_cls::
End(){
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
    glPopAttrib();

    msVpCStack.pop_back();
    if (msVpCStack.size() != 0){
	FBO_cls *pCFBO = msVpCStack[msVpCStack.size() - 1];
	pCFBO->Begin(FALSE, TRUE);
	msVpCStack.pop_back(); // Beginŏ镪̓RRō폜
    }
}

float FBO_cls::
GetLineWidthFactor(){
    switch(mWidthBuffer){
    case 256:
	return 0.125;
    case 512:
	return 0.25;
    case 1024: 
	return 0.5;
    case 2048:
	return 1.0;
    case 4096:
	return 2.0;
    }
    return 1;
}

bool FBO_cls::msbInitialized = FALSE;
bool FBO_cls::msbSupported = TRUE;

std::vector < FBO_cls* > FBO_cls::msVpCStack;
