// "clsGLGuides.h"
// (setq c-basic-offset 4)

#include "clsGLGuides.h"

#include "../COMMON/CommonDefinitions.h"

#include <limits.h>

GLGuides_cls::
GLGuides_cls(int type){
    mArrowDir[0] = mArrowDir[1] = mArrowDir[2] = 1;
    mType = type;
}

void GLGuides_cls::
GetGLMatrixInfo(){
    glGetDoublev(GL_MODELVIEW_MATRIX,  mMatrixModelView);
    glGetDoublev(GL_PROJECTION_MATRIX, mMatrixProjection);
    glGetIntegerv(GL_VIEWPORT, mMatrixViewport);
}

void GLGuides_cls::
SetArrowDir(double xlength, double ylength, double zlength){
    mArrowDir[X] = xlength;
    mArrowDir[Y] = ylength;
    mArrowDir[Z] = zlength;
}

void GLGuides_cls::
DrawArrow(double length, int type){
    if (length == 0) return;
    double arrow;
    if (length > 0) arrow = length - 0.6;
    if (length < 0) arrow = length + 0.6;
    switch (type){
    case _GUIDES_TRANSLATE:
	glBegin(GL_TRIANGLE_FAN);
	glVertex3d(length, 0, 0);
	glVertex3d(arrow,  0,      -0.1);
	glVertex3d(arrow, -0.0707, -0.0707);
	glVertex3d(arrow, -0.1,     0);
	glVertex3d(arrow, -0.0707,  0.0707);
	glVertex3d(arrow,  0,       0.1);
	glVertex3d(arrow,  0.0707,  0.0707);
	glVertex3d(arrow,  0.1,     0);
	glVertex3d(arrow,  0.0707,  -0.0707);
	glVertex3d(arrow,  0,      -0.1);
	glEnd();
	break;
    case _GUIDES_SCALE:
	glBegin(GL_QUAD_STRIP);
	glVertex3d(length - 0.2, -0.2, +0.2);
	glVertex3d(length - 0.2, -0.2, -0.2);
	glVertex3d(length + 0.2, -0.2, +0.2);
	glVertex3d(length + 0.2, -0.2, -0.2);
	glVertex3d(length + 0.2, +0.2, +0.2);
	glVertex3d(length + 0.2, +0.2, -0.2);
	glVertex3d(length - 0.2, +0.2, +0.2);
	glVertex3d(length - 0.2, +0.2, -0.2);
	glVertex3d(length - 0.2, -0.2, +0.2);
	glVertex3d(length - 0.2, -0.2, -0.2);
	glEnd();
	glBegin(GL_QUADS);
	glVertex3d(length - 0.2, -0.2, +0.2);
	glVertex3d(length + 0.2, -0.2, +0.2);
	glVertex3d(length + 0.2, +0.2, +0.2);
	glVertex3d(length - 0.2, +0.2, +0.2);
	
	glVertex3d(length - 0.2, +0.2, -0.2);
	glVertex3d(length + 0.2, +0.2, -0.2);
	glVertex3d(length + 0.2, -0.2, -0.2);
	glVertex3d(length - 0.2, -0.2, -0.2);
	glEnd();
	break;
    }
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0); glVertex3d(length, 0, 0);
    glEnd();
}

void GLGuides_cls::
DrawGuide(PositionAndDirection_cls CObj, double scale,
	  PositionAndDirection_cls CParentObjPos, int dragMode,
	  double *pYPR){

    if (mType == _GUIDES_ROTATE){
	DrawGuideRotation(CObj, scale, pYPR, CParentObjPos, dragMode);
	return; 
    }

    double *pos = CObj.mPos;
    
    glPushMatrix();
    glTranslated(pos[X], pos[Y], pos[Z]);
    if (pYPR != NULL){
	glRotated(pYPR[Z], 0, 0, 1);
	glRotated(pYPR[Y], 0, 1, 0);
	glRotated(pYPR[X], 1, 0, 0);
    }
    glScaled(scale * 0.025, scale * 0.025, scale * 0.025);
    
    glPushName(0);

    glDisable(GL_CULL_FACE);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);

    glPushName(0);
    glColor4d(1, 0, 0, 1);
    DrawArrow(mArrowDir[X]*2, mType);
    glPopName();
    glPushName(1);
    glColor4d(0.2, 1, 0.2, 1);
    glRotated(90, 0, 0, 1); 
    DrawArrow(mArrowDir[Y]*2, mType); 
    glRotated(-90, 0, 0, 1);
    glPopName();
    glPushName(2);
    glColor4d(0.2, 0.2, 1, 1);
    glRotated(-90, 0, 1, 0); 
    DrawArrow(mArrowDir[Z]*2, mType); 
    glRotated(90, 0, 1, 0);
    glPopName();

    glPopName();

    if (dragMode == 0){
	glColor4d(1, 0, 0, 1);
	glBegin(GL_LINES);
	glVertex3d(-10, 0, 0); glVertex3d(+10, 0, 0);
	glEnd();
    }
    if (dragMode == 1){
	glColor4d(0.2, 1, 0.2, 1);
	glBegin(GL_LINES);
	glVertex3d(0, -10, 0); glVertex3d(0, +10, 0);
	glEnd();
    }
    if (dragMode == 2){
	glColor4d(0, 0, 1, 1);
	glBegin(GL_LINES);
	glVertex3d(0, 0, -10); glVertex3d(0, 0, +10);
	glEnd();
    }

    glPopMatrix();
}

void GLGuides_cls::
DrawAxisAndCalcDirection(double X0, double Y0, double Z0,
			 double X1, double Y1, double Z1){

    double posU0, posV0, posW0;
    double posU1, posV1, posW1;

    /*
    glBegin(GL_LINES);
    glVertex3d(X0, Y0, Z0); glVertex3d(X1, Y1, Z1);
    glEnd();
    */
    
    GetGLMatrixInfo();

    gluProject(X0, Y0, Z0,
	       mMatrixModelView, mMatrixProjection, mMatrixViewport,
	       &posU0, &posV0, &posW0);
    gluProject(X1, Y1, Z1,
	       mMatrixModelView, mMatrixProjection, mMatrixViewport,
	       &posU1, &posV1, &posW1);
    mbDirectionInv = (posW0 > posW1) ? TRUE : FALSE;
}

void GLGuides_cls::
DrawRing(double radius, int axis){
    glBegin(GL_LINE_LOOP);
    for (int i = 0; i < 360; i+=15){
	float rcosphi = radius * cos(3.14159 * i / 180);
	float rsinphi = radius * sin(3.14159 * i / 180);
	glVertex3d(rcosphi * msTableU[axis][X] + rsinphi * msTableV[axis][X],
		   rcosphi * msTableU[axis][Y] + rsinphi * msTableV[axis][Y], 
		   rcosphi * msTableU[axis][Z] + rsinphi * msTableV[axis][Z]);
    }
    glEnd();
}

/*
void GLGuides_cls::
SetGLColor(int dragMode, int tgtMode){
    switch(tgtMode){
    case _ROTATE_MOVE_Y:
	if (dragMode == -1)
	    glColor4d(0.1, 0.1, 0.7, 1);
	else if (dragMode == tgtMode)
	    glColor4d(0.2, 0.2, 1.0, 1);
	else
	    glColor4d(0.1, 0.1, 0.4, 1);
	break;
    case _ROTATE_MOVE_P:
	if (dragMode == -1)
	    glColor4d(0.1, 0.7, 0.1, 1);
	else if (dragMode == tgtMode)
	    glColor4d(0.2, 1.0, 0.2, 1);
	else
	    glColor4d(0.1, 0.4, 0.1, 1);
	break;
    case _ROTATE_MOVE_R:
	if (dragMode == -1)
	    glColor4d(0.7, 0.1, 0.1, 1);
	else if (dragMode == tgtMode)
	    glColor4d(1.0, 0.2, 0.2, 1);
	else
	    glColor4d(0.4, 0.1, 0.1, 1);
	break;
    }
}
*/

void GLGuides_cls::
DrawGuideRotation(PositionAndDirection_cls CObj, double scale, 
		  double *pYPR, PositionAndDirection_cls CParentObjPos,
		  int dragMode){
    double *pos = CObj.mPos;
    double *YPR = CObj.mYPR;
    glPushMatrix();
    glTranslated(pos[X], pos[Y], pos[Z]);
    glScaled(scale * 0.05, scale * 0.05, scale * 0.05);

    if (dragMode == -1) dragMode = _ROTATE_MOVE_Y;

    glPushName(0);

    glColor4d(0.2, 0.2, 1, 1);
    if (dragMode == _ROTATE_MOVE_Y){
	glBegin(GL_LINES);
	glVertex3d(0, 0, -10); glVertex3d(0, 0, 10);
	glEnd();
	DrawAxisAndCalcDirection(0, 0, 0, 0, 0, 1); 
    }
    glPushName(2);
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0); glVertex3d(0, 0, 1);
    glEnd();
    glPushMatrix();
    glTranslated(0, 0, 1);
    DrawRing(0.5, Z);
    glPopName();
    
    glPopMatrix();

    glColor4d(0.2, 1, 0.2, 1);
    glRotated(pYPR[Z], 0, 0, 1);
    if (dragMode == _ROTATE_MOVE_P){
	glBegin(GL_LINES);
	glVertex3d(0, -10, 0); glVertex3d(0, 10, 0);
	glEnd();
	DrawAxisAndCalcDirection(0, 0, 0, 0, 1, 0); 
    }
    glPushName(1);
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0); glVertex3d(0, 1, 0);
    glEnd();
    glPushMatrix();
    glTranslated(0, 1, 0);
    DrawRing(0.5, Y);
    glPopName();
    glPopMatrix();

    glColor4d(1, 0, 0, 1);
    glRotated(pYPR[Y], 0, 1, 0);
    if (dragMode == _ROTATE_MOVE_R){
	glBegin(GL_LINES);
	glVertex3d(-10, 0, 0); glVertex3d(10, 0, 0);
	glEnd();
	DrawAxisAndCalcDirection(0, 0, 0, 1, 0, 0); 
    }
    glPushName(0);
    glBegin(GL_LINES);
    glVertex3d(0, 0, 0); glVertex3d(1, 0, 0);
    glEnd();
    glPushMatrix();
    glTranslated(1, 0, 0);
    DrawRing(0.5, X);
    glPopName();
    glPopMatrix();
    

    glPopName();

    glPopMatrix();
};



void GLGuides_cls::
DrawGuideArrowBox(double posBox[3], double scale,
		  double boxScale[3], bool bLimitBox){
    
    /*
    double scaleInv[3], center[3], tmpPos[3];
    LOOP_XYZ{
	if (boxScale[xyz] == 0) boxScale[xyz] = 0.1;
	scaleInv[xyz] = 1.0/boxScale[xyz];
    }

    if (bLimitBox == TRUE){
	LOOP_XYZ if (posBox[xyz] < 0) posBox[xyz] = 0;
	LOOP_XYZ if (posBox[xyz+3] < 0) posBox[xyz+3] = 0;
	LOOP_XYZ if (posBox[xyz] > 1) posBox[xyz] = 1;
	LOOP_XYZ if (posBox[xyz+3] > 1) posBox[xyz+3] = 1;
    }
    
    GLOBJ::DrawWireBoxMinMax(posBox, posBox+3);
    LOOP_XYZ center[xyz] = (posBox[xyz] + posBox[xyz+3]) / 2.0;

    PositionAndDirection_cls CTmp;

    SetArrowDir(1, 1, 1);
    DrawGuide(center, scale*0.4, 1, scaleInv, CTmp);
    
    tmpPos[X] = posBox[X]; tmpPos[Y] = center[Y]; tmpPos[Z] = center[Z];
    SetArrowDir(-1, 0, 0);
    DrawGuide(tmpPos, scale*0.25, 2, scaleInv, CTmp);
    tmpPos[X] = posBox[X+3]; tmpPos[Y] = center[Y]; tmpPos[Z] = center[Z];
    SetArrowDir(1, 0, 0);
    DrawGuide(tmpPos, scale*0.25, 3, scaleInv, CTmp);
    tmpPos[X] = center[X]; tmpPos[Y] = posBox[Y]; tmpPos[Z] = center[Z];
    SetArrowDir(0, -1, 0);
    DrawGuide(tmpPos, scale*0.25, 4, scaleInv, CTmp);
    tmpPos[X] = center[X]; tmpPos[Y] = posBox[Y+3]; tmpPos[Z] = center[Z];
    SetArrowDir(0, 1, 0);
    DrawGuide(tmpPos, scale*0.25, 5, scaleInv, CTmp);
    tmpPos[X] = center[X]; tmpPos[Y] = center[Y]; tmpPos[Z] = posBox[Z];
    SetArrowDir(0, 0, -1);
    DrawGuide(tmpPos, scale*0.25, 6, scaleInv, CTmp);
    tmpPos[X] = center[X]; tmpPos[Y] = center[Y]; tmpPos[Z] = posBox[Z+3];
    SetArrowDir(0, 0, 1);
    DrawGuide(tmpPos, scale*0.25, 7, scaleInv, CTmp);
    */
}

std::vector <int> GLGuides_cls::
GetObjectFromList(){
    int hitObj = 0;
    int hitGuide = 0;
    mHitNum = glRenderMode(GL_RENDER); // ZNg[hIGqbg𓾂
    
    GLuint *pBuf = mBuffer4Selection;
    GLuint depthMinMax = UINT_MAX;
    for (int i = 0; i < mHitNum; i++){
	GLuint objNum = pBuf[0]; //pBuf[3]ȍ~int񂪕ł邩
	GLuint depthMin = pBuf[1];
	GLuint depthMax = pBuf[2];
	GLuint CurObj   = pBuf[3]; 
	GLuint CurGuide = 0;
	if (objNum == 2) CurGuide = pBuf[4];
	if (CurObj > 0 && CurObj < 8){
	    hitObj      = CurObj;
	    hitGuide    = CurGuide;
	    depthMinMax = depthMin;
	}
	pBuf += (3 + objNum);
    }
    std::vector <int> retVec;
    retVec.push_back(hitObj); retVec.push_back(hitGuide);
    return retVec;
}

void GLGuides_cls::
InitializeProjection(){
    glSelectBuffer(100, mBuffer4Selection);
    glRenderMode(GL_SELECT);
    glInitNames();

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
}

std::vector < int > GLGuides_cls::
GetSelection(){

    std::vector < int > retVec = GetObjectFromList();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);

    return retVec;
}

int GLGuides_cls::msTableU[3][3] = {{0,1,0}, {0,0,1}, {1,0,0}};
int GLGuides_cls::msTableV[3][3] = {{0,0,1}, {1,0,0}, {0,1,0}};
