// -*- c++ -*- 
// "clsDataLoader.h"
// (setq c-basic-offset 4)

#include "clsDataLoader.h"
#include "clsDataLoaderGUI.h"

#include "../PARTICLEDATA/clsPeriodicBoundary.h"

#include "../../clsMain.h"

#include "../COMMON/CommonDefinitions.h"

void DataLoaderInfo_cls::
SetInfo(DataLoader_cls *pCParent){
    mDataType     = pCParent->mCDataType    .mBody;
    mInterpType   = pCParent->mCInterpType  .mBody;
    mVariablePNum = pCParent->mCVariablePNum.mBody;
    mPNumLimit    = pCParent->mCPNumLimit   .mBody;
    mSearchLimit  = pCParent->mCSearchLimit .mBody;
}

// -----

DataLoader_cls::
DataLoader_cls(Main_cls *pCMain) :
    GrammerBase_cls(wxT("DataLoad"))
{
    mpCGUI = NULL;
    mpCMainLink = pCMain;
    
    mCConfList.Add(&mCLastOpenedFile, wxT("LastOpenedFile="), wxT("NONE"));
    mCConfList.Add(&mCInterpType, wxT("InterpolateType="), wxT("1"));
    mCConfList.Add(&mCVariablePNum, wxT("VariableNumber="), wxT("0"));
    mCConfList.Add(&mCReadSequence, wxT("ReadAsSequence="), wxT("1"));
    mCConfList.Add(&mCSearchLimit, wxT("SearchLimit="), wxT("0"));
    mCConfList.Add(&mCPNumLimit, wxT("PNumLimit="), wxT("0"));
    mCConfList.Add(&mCAdjustTranslate, wxT("LoaderTranslate="), wxT("0 0 0"));
    mCConfList.Add(&mCAdjustScale, wxT("LoaderScale="), wxT("1"));
    mCConfList.Add(&mCAdjustValue1, wxT("LoaderScaleValue1="), wxT("1"));
    mCConfList.Add(&mCLogarithmValue1, wxT("LoaderLogarithm="), wxT("0"));
}

DataLoader_cls::
~DataLoader_cls(){
}

void DataLoader_cls::
SetToDefault(){
    LOOP_XYZ
	mCAdjustTranslate.mpBody[xyz] = 0;
    mCAdjustScale.mBody  = 1;
    mCAdjustValue1.mBody = 1;
    mCLogarithmValue1.mBody = 0;
}

void DataLoader_cls::
InitGUI(wxWindow *pCParent){
    mpCGUI = new DataLoaderGUI_cls(pCParent, this);
}

void DataLoader_cls::
OpenWindow(wxWindow *pCParent){
    if (mpCGUI == NULL) InitGUI(pCParent);
    if (mpCGUI->IsShownOnScreen() == FALSE){
	mpCGUI->Show(TRUE);
	mpCGUI->SetFocus();
    }
    else
	mpCGUI->SetFocus();

    mpCGUI->Raise();
}

void DataLoader_cls::
UpdateListFromGrammer(){
    if (mpCGUI != NULL)
	mpCGUI->UpdateListFromGrammer();
}

int  DataLoader_cls::
GetHDF5Order(){
    int itemNum = mCGrammerHDF5.mCGrammerBody.mDimension;
    for (int i = 0; i < itemNum; i++)
	if (mCGrammerHDF5.mCGrammerBody.mpBody[i] == _GRAMMER_AXAYAZ ||
	    mCGrammerHDF5.mCGrammerBody.mpBody[i] == _GRAMMER_JXJYJZ)
	    return 3;
    return 1;
}

void DataLoader_cls::
SetLoadInfo(){
    mCLoadInfo.SetInfo(this);
}

bool DataLoader_cls::
OpenDataFileHDF5(wxString strFilePath){
    mCLastOpenedFile.mstrBody = strFilePath;
    if (mpCGUI->CheckDataHDF5(strFilePath) == FALSE)
	return FALSE;
    SetLoadInfo();
    mpCMainLink->OpenDataFileCommand(strFilePath);
    return TRUE;
}

void DataLoader_cls::
OpenDataFile(wxString strFilePath){
    SetLoadInfo();
    mpCMainLink->OpenDataFileCommand(strFilePath);
    mCLastOpenedFile.mstrBody = strFilePath;
}

// ---------

void DataLoader_cls::
ReadStrLine(wxString strLine){
    mCConfList.ReadStrLineAll(strLine);
}

wxString DataLoader_cls::
Serialize(){
    return mCConfList.SerializeAll();
}
