// -*- c -*-
// "clsParameters.cpp"

#ifndef _PARAMETERS_H
#define _PARAMETERS_H

#include "clsKeyFrame.h"
#include "../CONFITEM/clsConfItem.h"
#include "../MYGL/clsPositionAndDirection.h"

#include <vector>
#include <list>
#include <string>

#define _PARAMGL_SELECTED_IN   0
#define _PARAMGL_SELECTED_OUT  1
#define _PARAMGL_NOTSELECT_IN  2
#define _PARAMGL_NOTSELECT_OUT 3

#define _PARAM_USE_COMMON -1

#define _PARAM_CURRENT -1
#define _PARAM_PROCESS_OVRW 2

class GLGuides_cls;

class ParamGroup_cls;
class ParamTL_cls;
class TimeCtrl_cls;
class PDataSeqCtrl_cls;
class Main_cls;

class Parameters_cls {
    friend class TimeLinePanel_cls; 
    friend class ComParamBase_cls; 
 public:
    Parameters_cls(wxString strID,
		   wxString strLabel = wxT("Untitled"),
		   long long updateBit = 0,
		   bool bTimeDependent = FALSE);
    
    // new $B$H(B std::vector $B$NAH$_9g$o$;$O%3%T!<%3%s%9%H%i%/%?I,?\(B
    Parameters_cls(const Parameters_cls &obj); 
    void CopyFromP(const Parameters_cls *pObj);
    void DefineConfList();
    ~Parameters_cls();
    
    std::vector < KeyFrame_cls > mVCKeys;

    ConfItemList_cls mCConfList;
    ConfItemBOOL_cls mCbTimeDependent;
    ConfItemINT_cls  mCPeriodic;
    //ConfItemBOOL_cls mCbShow;

    ParamGroup_cls *mpCParentGroupLink;
    ParamTL_cls    *mpCParamTLLink;
    wxString  mstrID;
    wxString  mstrLabel;
    long long mUpdateBit;
    unsigned int mStamp; // $BA`:n2s?t!#99?7$5$l$?$+$N%A%'%C%/Ey$K;H$&(B
    
    bool mbUpdateCamera; // $B99?7$N:]$K%+%a%i$b%"%C%W%G!<%H(B
    
 protected:
    static int             msFrameNumber;
    static int             msAnimInterp;
    //static unsigned int    msStepQuantum;
 public:
    static double          msCurrentDoubleIndex;
    static double          msResolution; // = 1.0/mStepNumber;

    static TimeCtrl_cls*     mspCTimeCtrlLink;
    static PDataSeqCtrl_cls* mspCPDataSeqCtrlLink;
    static Main_cls*         mspCMainLink;

    static bool            msbUndoCommandMake;
    
 public:
    double mInterpolatedValue;

    void GroupRefresh();
    void RemapKeyFrame(int oldMax, int newMax);
    static void SetFrameNumber(int frame, int animInterp/*, unsigned int quantum*/);
    double Quantize(double doubleIndex, 
		    unsigned int frameNumber = _PARAM_USE_COMMON);

    void SetTimeDependent(bool bTimeDependent);
    void CopyParameterFromP(Parameters_cls *pCParamFrom);
    void Flip();
    void ExecCommandTimeDependent(bool bTimeDependent);
    int GetKeyNumber();
    int GetSelectedKeyNum();
    void SetKeyToValueAt(KeyFrame_cls *pCKey, double doubleIndex);
    
    TimeCtrl_cls*     GetPCTimeCtrl();
    PDataSeqCtrl_cls* GetPCPDataSeqCtrl();
    void CheckSnippingTag(std::vector < bool > *pVbSnipTag);
        
    void SortKey();
    double CalcVelocity(int prev, int next);
    void AdjustAutoControl();

    void InitSet(double value);
    void ExecCommandInitSet(double value, double doubleIndex);
    void InsertKeyQueue(double doubleIndex);
    void InsertKey(double doubleIndex, 
		   double value = 0, double DValue1 = 0, double DValue2 = 0, double resolution = _PARAM_USE_COMMON);
    
    int  CheckOverWrite(double doubleIndex, double Resolution);
    void SetKey(KeyFrame_cls CKey, double resolution = _PARAM_USE_COMMON);
    void ExecCommandSetKey(KeyFrame_cls CKey);
    void ExecCommandSetKey(double doubleIndex, double value, unsigned int type, 
			   double DValue1, double DValue2);
    void ExecCommandOvwtKey(unsigned int index, KeyFrame_cls CKey);
    bool IsValueChanged(double  doubleIndex = _PARAM_CURRENT);
    void SetChangedValueKey(double doubleIndex = _PARAM_CURRENT,
			    double DValue1 = 0, double DValue2 = 0,
			    double resolution = _PARAM_USE_COMMON);
    //bool MoveTimeSelectedKey(double doubleIndexDiff);
    bool MoveSelectedKeyByXYQueue
	(double dX, double dY = 0, double resolution = -1,
	 double minX = 0, double maxX = 0,
	 double minY = 0, double maxY = 0);
    bool MoveSelectedKeyByXY
	(double dX, double dY = 0, double resolution = -1,
	 double minX = 0, double maxX = 0,
	 double minY = 0, double maxY = 0);

    void RemapKeys(double rate);
    int  MoveKey(int indexToMove, double doubleIndexDiff,
		 unsigned int frameNumber = _PARAM_USE_COMMON);

    void ExecCommandMoveKey(double doubleIndexFrom, double doubleIndexTo,
			    double resolution);
    void SetKeyType(int index, int type);
    void ExecCommandSetKeyType(unsigned int index, unsigned int type);
    void SetKeyTypeMulti(unsigned int mode);
    void ExecCommandSetKeyTypeMulti(unsigned int num,
				    unsigned int *pIndex,
				    unsigned int *pType);
    
    bool DeleteSelectedKey();
    bool DeleteKeyQueue(double doubleIndex);
    bool DeleteKey(double doubleIndex, 
		   double resolution = _PARAM_USE_COMMON);
    void ExecCommandDeleteKey(int index);
    bool IsKeyFrame(double doubleIndex, double resolution= _PARAM_USE_COMMON);
    
    double GetNextKeyPosDoubleIndex(double cur, 
				    double resolution = _PARAM_USE_COMMON);
    double GetPrevKeyPosDoubleIndex(double cur, 
				    double resolution = _PARAM_USE_COMMON);

    // flag$B$,(BFALSE$B$GBP1~$9$k(BKey$B$,L5$1$l$P(B -1
    int GetKeyFrameIndexFromDouble(double doubleIndex = _PARAM_CURRENT,
				   double resolution = _PARAM_USE_COMMON,
				   bool bFlagGetKeyEvenNotKeyFrame = FALSE);
    int GetKeyFrameIndexToInsert(double doubleIndex, 
				 double resolution = _PARAM_USE_COMMON);   
    int GetAdjacentKey(double doubleIndex, int direction);
    int GetSelectedKeyIndex();

    double GetKeyFrameDoubleIndex(int index);
    double GetKeyFrameValue  (int index);
    double GetKeyFrameValueD1(int index);
    double GetKeyFrameValueD2(int index);

    unsigned int GetInterpTypeFromDouble(double doubleIndex = _PARAM_CURRENT);

    void    CalcInterpolateStep(double interpolateRate, int index);
    double  GetValue1Periodic(double value0, double value1, double period);
    void    CalcInterpolateLinear(double interpolateRate, int index);
    double  GetHermiteVelocity(int index);
    void    CalcInterpolateHermite(double interpolateRate, int index);
    double  GetInterpolatedValue(double doubleIndex = _PARAM_CURRENT, 
				 double resolution = _PARAM_USE_COMMON,
				 bool bRepeat = false);
    
    wxString GetStrExpression(wxString strExpTag);
    //double  GetPos(double doubleIndex, int axis);
    /*
    void    MoveInterpolatedPos(double dpos[3], double dscale[3], double drot[4], int hitObj);
    void    SetInterpolatedPos(double dpos[3]);
    */

    // ----
    bool    SelectKeyFrameRegionQueue
	(double doubleIndexA, double doubleIndexB, 
	 double valueA, double valueB,
	 bool bSelect = TRUE);
    bool    SelectKeyFrameRegion
	(double doubleIndexA, double doubleIndexB, 
	 double valueA, double valueB,
	 bool bSelect = TRUE);
    bool SelectSelectedKeyByXY(double valueX, double resolutionX,
			       double ValueY, double resolutionY);
    bool    SelectKeyByXY(double valueX, double resolutionX,
			  double ValueY, double resolutionY,
			  bool bSingleSelect = TRUE);
    int     OnWhichKeyFrame(double doubleIndex, double resolutionByPixel);
    void    SelectKeyFrameQueue
	(double doubleIndex, double resolutionByPixel,
	 bool bMultiSelect = FALSE, bool bDeselect = FALSE);
    void    SelectKeyFrame
	(double doubleIndex, double resolutionByPixel,
	 bool bMultiSelect = FALSE, bool bDeselect = FALSE);
    void    DeselectAll();

    void ProcStartDraggingQueue(double origin, bool bCtrl, bool bAlt);
    void ProcDraggingQueue(wxMouseEvent& event);
    void SetRegionOriginQueue(double origin);
    void TurnOffDraggingQueue();

    //--
    void CopyKeyFrameToClipBoard();
    void PasteKeyFrameFromClipBoard();
    
    wxString Serialize();
    void ReadStrLine(std::wistream &istream, wxString strLineTag);
};

#endif
