
#include "clsThreadInterpolate.h"

#include "clsThreadInterpolateLinear.h"
#include "clsThreadInterpolateHermite.h"
#include "../clsPeriodicBoundary.h"
#include "../../MYAPI/MsgBox.h"
#include "../../PANELS/clsDataLoader.h"

#include "../../../wxWidgetEnumMenu.h"

#include <omp.h>

#include "../../COMMON/CommonDefinitions.h"

ThreadInterpolator_cls *GpCThreadInterpolate = NULL;
bool                   GbThreadInterpolateFinished = FALSE;
extern wxFrame         *GpCMainFrameCLS;

BasicDataPointers_cls::
BasicDataPointers_cls(float (*pX)[3], float (*pV)[3],
		      float (*pA)[3], float (*pJ)[3],
		      float *pRadius, float *pDRadius,
		      float *pValue1, float *pDValue1,
		      unsigned char *pType, int *pID){
    mpX = pX; mpV = pV; mpA = pA; mpJ = pJ;
    mpRadius = pRadius; mpDRadius = pDRadius;
    mpValue1 = pValue1; mpDValue1 = pDValue1;
    mpType = pType; mpID = pID;
}

void BasicDataPointers_cls::
CopyPointers(float (**ppX)[3], float (**ppV)[3],
	     float (**ppA)[3], float (**ppJ)[3],
	     float **ppRadius, float **ppDRadius,
	     float **ppValue1, float **ppDValue1,
	     unsigned char **ppType, int **ppID){
    *ppX = mpX; *ppV = mpV; *ppA = mpA; *ppJ = mpJ;
    *ppRadius = mpRadius; *ppDRadius = mpDRadius;
    *ppValue1 = mpValue1; *ppDValue1 = mpDValue1;
    *ppType = mpType; *ppID = mpID;
}

// ------

ThreadInterpolator_cls::
ThreadInterpolator_cls(wxThreadKind kind) :
    wxThread(kind) {
}

void ThreadInterpolator_cls::
SetParameters(bool *pbFinishedFlag,
	      DataLoader_cls *pCDataLoader,
	      int PNum, int PNumNext, float dtime,
	      int interpolateType, int variablePNum,
	      int searchLimit, 
	      BasicDataPointers_cls *pCDataPointers,
	      BasicDataPointers_cls *pCDataPointersNext) {
    mpbFinishedFlagLink = pbFinishedFlag;
    mpCDataLoaderLink = pCDataLoader;
    mpCPeriodicLink = pCDataLoader->mpCPeriodicLink;
    mInterpolateType = interpolateType;
    mVariablePNum = variablePNum;
    mSearchLimit = searchLimit;
    mPNum = PNum;
    mPNumNext = PNumNext;
    mDTime = dtime;

    mpCDataPointers = pCDataPointers;
    mpCDataPointersNext = pCDataPointersNext;

    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

int ThreadInterpolator_cls::
CalcThreadNumber(){
    int threadNumber = 1;
    if (mPNum > 100000) threadNumber  = 2;
    if (mPNum > 300000) threadNumber = 3;
    if (mPNum > 1000000) threadNumber = 4;
    if (mPNum > 3000000) threadNumber = 5;
    if (mPNum > 10000000) threadNumber = 6;
    return threadNumber;
}

void ThreadInterpolator_cls::
CalcPInitPFin(int &PInit, int &PFin, 
	      int PNumLocal, int core, int threadNumber){
    PInit = PInit = PNumLocal / threadNumber * core;
    PFin  = PNumLocal / threadNumber * (core+1);
    if (core == threadNumber - 1) PFin = PNumLocal;
}

wxThread::ExitCode ThreadInterpolator_cls::
Entry(){
    
    int PNumLocal = mPNum;
    if (PNumLocal > mPNumNext) PNumLocal = mPNumNext;

    if (mDTime == 0)
	mInterpolateType = _INTERPOLATE_NO;
    
    int PInit;
    int PFin;
    int threadNumber = CalcThreadNumber();
    if (mInterpolateType == _INTERPOLATE_LINEAR){
	ThreadInterpBlockLinear_cls **ppCThreadInterpolate = 
	    new ThreadInterpBlockLinear_cls* [threadNumber];

	for (int core = 0; core < threadNumber; core++){
	    ppCThreadInterpolate[core] = 
		new ThreadInterpBlockLinear_cls(wxTHREAD_JOINABLE);

	    CalcPInitPFin(PInit, PFin, PNumLocal, core, threadNumber);
	    ppCThreadInterpolate[core]->SetParameters
		(PInit, PFin, mPNumNext, mDTime, 
		 mVariablePNum, mSearchLimit, mpCPeriodicLink,
		 mpCDataPointers, mpCDataPointersNext);
	}

	for (int i = 0; i < threadNumber; i++) ppCThreadInterpolate[i]->Run();
	for (int i = 0; i < threadNumber; i++) ppCThreadInterpolate[i]->Wait();
	for (int i = 0; i < threadNumber; i++) delete ppCThreadInterpolate[i];
	delete[] ppCThreadInterpolate;
    }

    if (mInterpolateType == _INTERPOLATE_HERMITE){
	ThreadInterpBlockHermite_cls **ppCThreadInterpolate = 
	    new ThreadInterpBlockHermite_cls* [threadNumber];

	for (int core = 0; core < threadNumber; core++){
	    ppCThreadInterpolate[core] = 
		new ThreadInterpBlockHermite_cls(wxTHREAD_JOINABLE);

	    CalcPInitPFin(PInit, PFin, PNumLocal, core, threadNumber);
	    ppCThreadInterpolate[core]->SetParameters
		(PInit, PFin, mPNumNext, mDTime, 
		 mVariablePNum, mSearchLimit, mpCPeriodicLink,
		 mpCDataPointers, mpCDataPointersNext);
	}

	for (int i = 0; i < threadNumber; i++) ppCThreadInterpolate[i]->Run();
	for (int i = 0; i < threadNumber; i++) ppCThreadInterpolate[i]->Wait();
	for (int i = 0; i < threadNumber; i++) delete ppCThreadInterpolate[i];
	delete[] ppCThreadInterpolate;
    }
    
    *mpbFinishedFlagLink = TRUE;
    GbThreadInterpolateFinished = TRUE;

    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    GpCMainFrameCLS->AddPendingEvent(com);
    GpCThreadInterpolate = NULL;

    printf("Thread Interpolate Finished\n");
    
    return 0;
}
