
#include "clsThreadInterpolateHermite.h"
#include "clsThreadInterpolate.h"
#include "../clsPeriodicBoundary.h"
#include "../../MYAPI/MsgBox.h"
#include "../../PANELS/clsDataLoader.h"

#include "../../../wxWidgetEnumMenu.h"

#include <omp.h>

#include "../../COMMON/CommonDefinitions.h"

ThreadInterpBlockHermite_cls::
ThreadInterpBlockHermite_cls(wxThreadKind kind) :
    wxThread(kind) {
}

void ThreadInterpBlockHermite_cls::
SetParameters(int PInit, int PFin,
	      int PNumNext, float dtime, int variablePNum,
	      int searchLimit, 
	      PeriodicBoundary_cls *pCPeriodicLink,
	      BasicDataPointers_cls *pCDataPointers,
	      BasicDataPointers_cls *pCDataPointersNext){

    mVariablePNum = variablePNum;
    mSearchLimit = searchLimit;
    mPeriodicType = pCPeriodicLink->mCType.mBody;
    mPeriodicEffect = pCPeriodicLink->mCEffect.mBody;

    float boundaryVector[3][3];
    pCPeriodicLink->GetBoundaryVector(boundaryVector[X], X);
    pCPeriodicLink->GetBoundaryVector(boundaryVector[Y], Y);
    pCPeriodicLink->GetBoundaryVector(boundaryVector[Z], Z);

    // R͂e\
    LOOP_XYZ mPeriodicSize[xyz] = boundaryVector[xyz][xyz];

    mPIni = PInit;
    mPFin = PFin;
    pCDataPointers->CopyPointers(&mpX, &mpV, &mpA, &mpJ,
				 &mpRadius, &mpDRadius, 
				 &mpValue1, &mpDValue1,
				 &mpType, &mpID);

    mPNumNext = PNumNext;
    pCDataPointersNext->CopyPointers(&mpXNext, &mpVNext, &mpANext, &mpJNext,
				     &mpRadiusNext, &mpDRadiusNext,
				     &mpValue1Next, &mpDValue1Next,
				     &mpTypeNext, &mpIDNext);
    mDTime = dtime;
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

void ThreadInterpBlockHermite_cls::
InterpolateDataHermiteCore(int i, int iNext, float dtime){
    //float (*pX)[3] = mpX;
    //float (*pV)[3] = mpV;
    float (*pA)[3] = mpA;
    float (*pJ)[3] = mpJ;
    float (*pXNext)[3] = mpXNext;
    
    pA[i][X] = ((-3*mpX[i][X]+3*mpXNext[iNext][X])/dtime/dtime - (2*mpV[i][X] + mpVNext[iNext][X])/dtime);
    pA[i][Y] = ((-3*mpX[i][Y]+3*mpXNext[iNext][Y])/dtime/dtime - (2*mpV[i][Y] + mpVNext[iNext][Y])/dtime);
    pA[i][Z] = ((-3*mpX[i][Z]+3*mpXNext[iNext][Z])/dtime/dtime - (2*mpV[i][Z] + mpVNext[iNext][Z])/dtime);

    pJ[i][X] = ((2*mpX[i][X]-2*mpXNext[iNext][X])/dtime/dtime/dtime + (mpV[i][X] + mpVNext[iNext][X])/dtime/dtime);
    pJ[i][Y] = ((2*mpX[i][Y]-2*mpXNext[iNext][Y])/dtime/dtime/dtime + (mpV[i][Y] + mpVNext[iNext][Y])/dtime/dtime);
    pJ[i][Z] = ((2*mpX[i][Z]-2*mpXNext[iNext][Z])/dtime/dtime/dtime + (mpV[i][Z] + mpVNext[iNext][Z])/dtime/dtime);
    
    mpDRadius[i] = (mpRadiusNext[iNext] - mpRadius[i]) / dtime;
    mpDValue1[i] = (mpValue1Next[iNext] - mpValue1[i]) / dtime;
}

void ThreadInterpBlockHermite_cls::
InterpolateDataHermite(){
    float dtime = mDTime;

    for (int i = mPIni; i < mPFin; i++){

	mpA[i][X] = ((-3*mpX[i][X]+3*mpXNext[i][X])/dtime/dtime - (2*mpV[i][X] + mpVNext[i][X])/dtime);
	mpA[i][Y] = ((-3*mpX[i][Y]+3*mpXNext[i][Y])/dtime/dtime - (2*mpV[i][Y] + mpVNext[i][Y])/dtime);
	mpA[i][Z] = ((-3*mpX[i][Z]+3*mpXNext[i][Z])/dtime/dtime - (2*mpV[i][Z] + mpVNext[i][Z])/dtime);
	
	mpJ[i][X] = ((2*mpX[i][X]-2*mpXNext[i][X])/dtime/dtime/dtime + (mpV[i][X] + mpVNext[i][X])/dtime/dtime);
	mpJ[i][Y] = ((2*mpX[i][Y]-2*mpXNext[i][Y])/dtime/dtime/dtime + (mpV[i][Y] + mpVNext[i][Y])/dtime/dtime);
	mpJ[i][Z] = ((2*mpX[i][Z]-2*mpXNext[i][Z])/dtime/dtime/dtime + (mpV[i][Z] + mpVNext[i][Z])/dtime/dtime);

	mpDRadius[i] = (mpRadiusNext[i] - mpRadius[i]) / dtime;
	mpDValue1[i] = (mpValue1Next[i] - mpValue1[i]) / dtime;
    }
}

void ThreadInterpBlockHermite_cls::
InterpolateDataHermitePeriodic(){
    float dtime = mDTime;

    /*
    if (mPeriodicType == 1){
	if (mPeriodicEffect == 0)
	    for (int i = mPIni; i < mPFin; i++)
		InterpolateDataHermiteCorePeriodicXYEeffect0(i, i, dtime);
	if (mPeriodicEffect == 1)
	    for (int i = mPIni; i < mPFin; i++)
		InterpolateDataHermiteCorePeriodicXYEeffect1(i, i, dtime); }
    if (mPeriodicType == 2){
	if (mPeriodicEffect == 0)
	    for (int i = mPIni; i < mPFin; i++)
		InterpolateDataHermiteCorePeriodicXYZEeffect0(i, i, dtime);
	if (mPeriodicEffect == 1)
	    for (int i = mPIni; i < mPFin; i++)
		InterpolateDataHermiteCorePeriodicXYZEeffect1(i, i, dtime); }
    */
}

void ThreadInterpBlockHermite_cls::
InterpolateDataHermiteSearchingID(){
    float dtime = mDTime;
    
    void (ThreadInterpBlockHermite_cls::*pFunction)(int, int, float) = NULL;
    if (mPeriodicType == 0)
	pFunction = &ThreadInterpBlockHermite_cls::InterpolateDataHermiteCore;
    /*
    if (mPeriodicType == 1){
	if (mPeriodicEffect == 0) pFunction = &ThreadInterpBlockLinear_cls::InterpolateDataLinearCorePeriodicXYEeffect0;
	if (mPeriodicEffect == 1) pFunction = &ThreadInterpBlockLinear_cls::InterpolateDataLinearCorePeriodicXYEeffect1;
    }
    if (mPeriodicType == 2){
	if (mPeriodicEffect == 0) pFunction = &ThreadInterpBlockLinear_cls::InterpolateDataLinearCorePeriodicXYZEeffect0;
	if (mPeriodicEffect == 1) pFunction = &ThreadInterpBlockLinear_cls::InterpolateDataLinearCorePeriodicXYZEeffect1;
    }
    */

    for (int i = mPIni; i < mPFin; i++){
	int ID = mpID[i];
	int iNext = -1;
	if (i < mPNumNext && ID == mpIDNext[i]) {
	    (this->*pFunction)(i, i, dtime);
	    continue;
	}
	iNext = SearchIDNearBy(i, ID, mSearchLimit);
	/*
	iNext = SearchIDNear(i, ID, 10);
	if (iNext == -1) iNext = SearchIDNear(i, ID, 100);
	if (mSearchLimit < 2){
	    if (iNext == -1) iNext = SearchIDNear(i, ID, 1000);
	    if (iNext == -1) iNext = SearchIDNear(i, ID, 10000);
	    if (mSearchLimit < 1)
		if (iNext == -1) iNext = SearchIDFull(ID);
	}
	*/
	if (iNext == -1){
	    mpV[i][X] = 0; mpV[i][Y] = 0; mpV[i][Z] = 0;
	    mpDRadius[i] = 0; mpDValue1[i] = 0;
	    continue;
	}
	
	(this->*pFunction)(i, iNext, dtime);
    }
}

void ThreadInterpBlockHermite_cls::
InterpolateDataHermiteSearchingIDSorted(){
    float dtime = mDTime;
    
    void (ThreadInterpBlockHermite_cls::*pFunction)(int, int, float) = NULL;
    if (mPeriodicType == 0)
	pFunction = &ThreadInterpBlockHermite_cls::InterpolateDataHermiteCore;
    /*
    if (mPeriodicType == 1){
	if (mPeriodicEffect == 0) pFunction = &ThreadInterpBlockLinear_cls::InterpolateDataLinearCorePeriodicXYEeffect0;
	if (mPeriodicEffect == 1) pFunction = &ThreadInterpBlockLinear_cls::InterpolateDataLinearCorePeriodicXYEeffect1;
    }
    if (mPeriodicType == 2){
	if (mPeriodicEffect == 0) pFunction = &ThreadInterpBlockLinear_cls::InterpolateDataLinearCorePeriodicXYZEeffect0;
	if (mPeriodicEffect == 1) pFunction = &ThreadInterpBlockLinear_cls::InterpolateDataLinearCorePeriodicXYZEeffect1;
    }
    */

    int tgtIndex = 0;
    for (int i = mPIni; i < mPFin; i++){
	int ID = mpID[i];
	int iNext = SearchIDFrom(&tgtIndex, ID);
	if (iNext == -1){
	    mpV[i][X] = 0; mpV[i][Y] = 0; mpV[i][Z] = 0;
	    mpDRadius[i] = 0; mpDValue1[i] = 0;
	    continue;
	}

	(this->*pFunction)(i, iNext, dtime);
    }
}

wxThread::ExitCode ThreadInterpBlockHermite_cls::
Entry(){
    if (mVariablePNum == _INTERPOLATE_PNUM_VARIABLE)
	InterpolateDataHermiteSearchingID();

    /*
    if (mVariablePNum == _INTERPOLATE_PNUM_ID_SORTED)
	InterpolateDataHermiteSearchingIDSorted();
    */
    
    if (mVariablePNum == _INTERPOLATE_PNUM_CONSTANT) {
	if (mPeriodicType == 0)
	    InterpolateDataHermite();
	else
	    InterpolateDataHermitePeriodic();
    }
    return 0;
}
