// -*- c -*- 
// "clsGLDrawBillboard.h"
// (setq c-basic-offset 4)

#include "clsGLDrawBillboard.h"
#include "clsArrayAndIndex.h"

#include "../GLATTRIB/clsGLAttribute.h"

#include "clsSnapData.h"

#include "../../wxWidgetEnumMenu.h"

#ifdef WIN32
#include <windows.h>
#include <GL/glu.h>
#include <GL/glext.h>
#elif defined __WXMAC__
#define GL_GLEXT_PROTOTYPES 
#include <OpenGL/glu.h>
#else
#define GL_GLEXT_PROTOTYPES 
#include <GL/glu.h>
#endif

extern int       GDrawingPercentage;
extern int       GDrawingStage;
extern bool      GbAbort;
extern wxFrame  *GpCMainFrameCLS;

#include "../COMMON/CommonDefinitions.h"

GLDrawBillboard_cls::
GLDrawBillboard_cls(){
    mpVertexArrayBufferA = NULL;
    mpVertexArrayBufferB = NULL;
    mpColorBufferA = NULL;
    mpColorBufferB = NULL;
}

GLDrawBillboard_cls::
~GLDrawBillboard_cls(){
    if (mpVertexArrayBufferA != NULL) delete[] mpVertexArrayBufferA;
    if (mpVertexArrayBufferB != NULL) delete[] mpVertexArrayBufferB;
    if (mpColorBufferA != NULL) delete[] mpColorBufferA;
    if (mpColorBufferB != NULL) delete[] mpColorBufferB;
}

void GLDrawBillboard_cls::
GLDrawBillboardBlockMain(int offset, int pnumber, 
			 float *pRotatedArray, float radiusFactor, 
			 bool bConstantRadius, bool bConstantColor,
			 float steppingAlphaFactor,
			 SnapData_cls *pCDataSnap,
			 GLAttribute_cls *pCGLAttribute, float *pValue1){
    if (radiusFactor == 0) return;

    float (*pX)[3]  = pCDataSnap->mpX;
    float  *pRadius = pCDataSnap->mpRadius;

    int   alphaType       = pCGLAttribute->mCAlphaType.mBody;
    float alphaDependBase = pCGLAttribute->mCAlphaDenendBase.mBody;
    float alphaMax        = pCGLAttribute->mCAlphaMax.mBody;

    //printf("Make Vertex\n");
    
    if (bConstantRadius == TRUE)
	for (int p = 0; p < pnumber; p++){
	    int i = (mspIndexAndZ + offset + p)->index;
	    int poffset = p*12;
	    for (int j = 0; j < 12; j++)
		msVertexArray[poffset + j] = pRotatedArray[j] * radiusFactor;
	    float *pParticleX = pX[i];
	    float *pArray = msVertexArray+poffset;
	    pArray[ 0] += pParticleX[X];
	    pArray[ 1] += pParticleX[Y];
	    pArray[ 2] += pParticleX[Z];
	    pArray[ 3] += pParticleX[X];
	    pArray[ 4] += pParticleX[Y];
	    pArray[ 5] += pParticleX[Z];
	    pArray[ 6] += pParticleX[X];
	    pArray[ 7] += pParticleX[Y];
	    pArray[ 8] += pParticleX[Z];
	    pArray[ 9] += pParticleX[X];
	    pArray[10] += pParticleX[Y];
	    pArray[11] += pParticleX[Z];
	}
    else
	for (int p = 0; p < pnumber; p++){
	    int i = (mspIndexAndZ + offset + p)->index;
	    int poffset = p*12;
	    float radius = radiusFactor * pRadius[i];
	    for (int j = 0; j < 12; j++)
		msVertexArray[poffset + j] = pRotatedArray[j] * radius;
	    float *pParticleX = pX[i];
	    float *pArray = msVertexArray+poffset;
	    pArray[ 0] += pParticleX[X];
	    pArray[ 1] += pParticleX[Y];
	    pArray[ 2] += pParticleX[Z];
	    pArray[ 3] += pParticleX[X];
	    pArray[ 4] += pParticleX[Y];
	    pArray[ 5] += pParticleX[Z];
	    pArray[ 6] += pParticleX[X];
	    pArray[ 7] += pParticleX[Y];
	    pArray[ 8] += pParticleX[Z];
	    pArray[ 9] += pParticleX[X];
	    pArray[10] += pParticleX[Y];
	    pArray[11] += pParticleX[Z];
	}

    if (bConstantColor == TRUE && alphaType == _GL_ALPHA_NORMAL){ 
	//printf("Main %d %d\n", offset, pnumber);
	glTexCoordPointer(2, GL_FLOAT, 0, msTexCoordArray);
	glVertexPointer(3, GL_FLOAT, 0, msVertexArray);
	glDrawElements(GL_QUADS, 4 * pnumber, GL_UNSIGNED_INT, mspIndices);
    }
    else{
	float pColor[4];
	//ColorCurve_cls *pCClrCurve = pCGLAttribute->mpCClrCurveCol;
	
	if (alphaType == _GL_ALPHA_INVERSE_SQUARE){
	    float alphaFactor = steppingAlphaFactor;
	    for (int p = 0; p < pnumber; p++){
		int i = (mspIndexAndZ + offset + p)->index;
		float radius = radiusFactor * pRadius[i];
		memcpy(pColor, pCDataSnap->mpColor[i], sizeof(float)*4);
		pColor[3] /= (radius * radius);
		pColor[3] *= alphaFactor;
		if (pColor[3] > alphaMax)
		    pColor[3] = alphaMax;

		if (pColor[3] < 0.0035) continue;

		glColor4fv(pColor);
		glTexCoordPointer(2, GL_FLOAT, 0, msTexCoordArray);
		glVertexPointer(3, GL_FLOAT, 0, msVertexArray);
		glDrawElements(GL_QUADS, 4, GL_UNSIGNED_INT, mspIndices + 4*p);
	    }
		
	}
	else{
	    float alphaFactor = pCGLAttribute->mCAlphaFactor.mBody * steppingAlphaFactor;
	    for (int p = 0; p < pnumber; p++){
		int i = (mspIndexAndZ + offset + p)->index;
		memcpy(pColor, pCDataSnap->mpColor[i], sizeof(float)*4);
		pColor[3] *= alphaFactor;
		glColor4fv(pColor);
		glTexCoordPointer(2, GL_FLOAT, 0, msTexCoordArray);
		glVertexPointer(3, GL_FLOAT, 0, msVertexArray);
		glDrawElements(GL_QUADS, 4, GL_UNSIGNED_INT, mspIndices + 4*p);
	    }
	}
    }
    
    //printf("Main End\n");
}

void GLDrawBillboard_cls::
GLDrawBillboardBlock(int downsized, 
		     float *pRotatedArray, float radiusFactor, 
		     bool bConstantRadius, bool bConstantColor,
		     float steppingAlphaFactor,
		     SnapData_cls *pCDataSnap,
		     GLAttribute_cls *pCGLAttribute, float *pValue1){
    
    int progressInterval = 20000;
    if (downsized > 2000000)
	progressInterval = downsized / 20;
    progressInterval /= PACKNUM;

    for (int i_ = 0; i_ < downsized; i_+=PACKNUM) {
	int number = PACKNUM;
	if (i_ + number >= downsized) number = downsized - i_;

	if (i_ % (PACKNUM * progressInterval) == 0){
	    GDrawingPercentage = (long long)100 * i_ / downsized;
	    UpdateStatusBar();
	}


	GLDrawBillboardBlockMain(i_ , number, pRotatedArray,
				 radiusFactor, bConstantRadius, bConstantColor,
				 steppingAlphaFactor, pCDataSnap, 
				 pCGLAttribute, pValue1);
	
	if (GbAbort == TRUE){
	    i_ = downsized;
	    GbAbort = FALSE;
	}
    }
}

// ----------------------

ThreadCaleVertex_cls::
ThreadCaleVertex_cls(wxThreadKind kind) : 
    wxThread(kind) {
}

void ThreadCaleVertex_cls::
SetParameters(bool *pbFinishedFlag, int offset, int pnumber,
	      float *pVertexBuffer, float *pColorBuffer,
	      float *pRotatedArray, 
	      GLAttribute_cls   *pCGLAttribute,
	      ArrayAndIndex_cls *pArrayIndexAndZ,
	      SnapData_cls      *pCDataSnap,
	      bool bConstantRadius,
	      float radiusFactor, float steppingAlphaFactor){
    mpbFinishedFlagLink = pbFinishedFlag;
    mOffset        = offset;
    mPNumber       = pnumber;
    mpVertexBuffer = pVertexBuffer;
    mpColorBuffer  = pColorBuffer;
    mpRotatedArray = pRotatedArray;
    mpCGLAttribute = pCGLAttribute;
    mpArrayIndexAndZ = pArrayIndexAndZ;
    mpCDataSnap    = pCDataSnap;
    mbConstantRadius = bConstantRadius;
    mRadiusFactor = radiusFactor;
    mSteppingAlphaFactor = steppingAlphaFactor;

    if ( Create() != wxTHREAD_NO_ERROR) 
	//MsgBoxExit(NULL, "Error!", "Can't create thread");
	exit(0);
}

wxThread::ExitCode ThreadCaleVertex_cls::
Entry(){
    int    offset        = mOffset;
    int    pnumber       = mPNumber;
    float *pVertexBuffer = mpVertexBuffer;
    float *pColorBuffer  = mpColorBuffer;
    float *pRotatedArray = mpRotatedArray;
    ArrayAndIndex_cls *pArrayAndIndexZ = mpArrayIndexAndZ;
    SnapData_cls      *pCDataSnap = mpCDataSnap;
    float (*pX)[3]       = pCDataSnap->mpX;
    float *pRadius       = pCDataSnap->mpRadius;
    float radiusFactor   = mRadiusFactor;
    float steppingAlphaFactor = mSteppingAlphaFactor;

    int   alphaType       = mpCGLAttribute->mCAlphaType.mBody;
    float alphaDependBase = mpCGLAttribute->mCAlphaDenendBase.mBody;
    float alphaMax        = mpCGLAttribute->mCAlphaMax.mBody;
    
    if (mbConstantRadius == TRUE)
	for (int p = 0; p < pnumber; p++){
	    int i = (pArrayAndIndexZ + offset + p)->index;
	    int poffset = p*12;
	    float *pVertex = pVertexBuffer + poffset;
	    for (int j = 0; j < 12; j++)
		pVertex[j] = pRotatedArray[j] * radiusFactor;
	    float *pParticleX = pX[i];
	    pVertex[ 0] += pParticleX[X];
	    pVertex[ 1] += pParticleX[Y];
	    pVertex[ 2] += pParticleX[Z];
	    pVertex[ 3] += pParticleX[X];
	    pVertex[ 4] += pParticleX[Y];
	    pVertex[ 5] += pParticleX[Z];
	    pVertex[ 6] += pParticleX[X];
	    pVertex[ 7] += pParticleX[Y];
	    pVertex[ 8] += pParticleX[Z];
	    pVertex[ 9] += pParticleX[X];
	    pVertex[10] += pParticleX[Y];
	    pVertex[11] += pParticleX[Z];
	}
    else
	for (int p = 0; p < pnumber; p++){
	    int i = (pArrayAndIndexZ + offset + p)->index;
	    int poffset = p*12;
	    float radius = radiusFactor * pRadius[i];
	    float *pVertex = pVertexBuffer + poffset;
	    for (int j = 0; j < 12; j++)
		pVertex[j] = pRotatedArray[j] * radius;
	    float *pParticleX = pX[i];
	    pVertex[ 0] += pParticleX[X];
	    pVertex[ 1] += pParticleX[Y];
	    pVertex[ 2] += pParticleX[Z];
	    pVertex[ 3] += pParticleX[X];
	    pVertex[ 4] += pParticleX[Y];
	    pVertex[ 5] += pParticleX[Z];
	    pVertex[ 6] += pParticleX[X];
	    pVertex[ 7] += pParticleX[Y];
	    pVertex[ 8] += pParticleX[Z];
	    pVertex[ 9] += pParticleX[X];
	    pVertex[10] += pParticleX[Y];
	    pVertex[11] += pParticleX[Z];
	}


    if (alphaType == _GL_ALPHA_INVERSE_SQUARE){
	float alphaFactor = steppingAlphaFactor;
	for (int p = 0; p < pnumber; p++){
	    int i = (pArrayAndIndexZ + offset + p)->index;
	    float radius  = radiusFactor * pRadius[i];
	    float *pColor = pColorBuffer + p*4;
	    memcpy(pColor, pCDataSnap->mpColor[i], sizeof(float)*4);
	    pColor[3] /= (radius * radius);
	    pColor[3] *= alphaFactor;
	    if (pColor[3] > alphaMax)
		pColor[3] = alphaMax;
	}
    }
    else{
	float alphaFactor = mpCGLAttribute->mCAlphaFactor.mBody 
	    * steppingAlphaFactor;
	for (int p = 0; p < pnumber; p++){
	    int i = (pArrayAndIndexZ + offset + p)->index;
	    float *pColor = pColorBuffer + p*4;
	    memcpy(pColor, pCDataSnap->mpColor[i], sizeof(float)*4);
	    pColor[3] *= alphaFactor;
	    
	}
    }

    *mpbFinishedFlagLink = TRUE;
    return 0;
}

void GLDrawBillboard_cls::
GLDrawBillboardMultitheadDraw(float *pVertex, float *pColor, int number){
    for (int i = 0; i < number; i++){
	if (pColor[i*4+3] < 0.0035) continue;
	glColor4fv(pColor + i*4); 
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);glVertex3fv(pVertex + i*12);
	glTexCoord2f(1, 0);glVertex3fv(pVertex + i*12 + 3);
	glTexCoord2f(1, 1);glVertex3fv(pVertex + i*12 + 6);
	glTexCoord2f(0, 1);glVertex3fv(pVertex + i*12 + 9);
	glEnd();
    }
}

void GLDrawBillboard_cls::
GLDrawBillboardMultithead(int downsized, 
			  float *pRotatedArray, float radiusFactor, 
			  bool bConstantRadius, bool bConstantColor,
			  float steppingAlphaFactor,
			  SnapData_cls *pCDataSnap,
			  GLAttribute_cls *pCGLAttribute, float *pValue1){

    int progressInterval = 200000;
    if (downsized > 2000000)
	progressInterval = downsized / 20;
    progressInterval /= _GL_DRAW_BILLBOARD_BUFFER;
    
    if (mpVertexArrayBufferA == NULL) 
	mpVertexArrayBufferA = new float [_GL_DRAW_BILLBOARD_BUFFER * 4 * 3];
    if (mpVertexArrayBufferB == NULL) 
	mpVertexArrayBufferB = new float [_GL_DRAW_BILLBOARD_BUFFER * 4 * 3];
    if (mpColorBufferA == NULL)
	mpColorBufferA = new float [_GL_DRAW_BILLBOARD_BUFFER * 4];
    if (mpColorBufferB == NULL)
	mpColorBufferB = new float [_GL_DRAW_BILLBOARD_BUFFER * 4];

    float *pVertexArrayCalc;
    float *pColorCalc;
    float *pVertexArrayDraw = NULL;
    float *pColorDraw = NULL;

    int calculating = 0;
    int drawing     = -1;
    int number2Draw = -1;
    for (int i_ = 0; i_ < downsized; i_ += _GL_DRAW_BILLBOARD_BUFFER){
	int number = _GL_DRAW_BILLBOARD_BUFFER; 
	if (i_ + number >= downsized) number = downsized - i_;

	if (i_ % (_GL_DRAW_BILLBOARD_BUFFER * progressInterval) == 0){
	    GDrawingPercentage = (long long)100 * i_ / downsized;
	    UpdateStatusBar();
	}

	if (calculating == 0){
	    pVertexArrayCalc = mpVertexArrayBufferA;
	    pColorCalc       = mpColorBufferA;
	}
	else{
	    pVertexArrayCalc = mpVertexArrayBufferB;
	    pColorCalc       = mpColorBufferB;
	}
	
	bool bFinished = FALSE;
	ThreadCaleVertex_cls *pCThread = 
	    new ThreadCaleVertex_cls(wxTHREAD_DETACHED);
	pCThread->SetParameters(&bFinished, i_, number, 
				pVertexArrayCalc, pColorCalc,
				pRotatedArray,
				pCGLAttribute, mspIndexAndZ,
				pCDataSnap, bConstantRadius,
				radiusFactor, steppingAlphaFactor);

	printf ("Thead Calc Begin\n");
	pCThread->Run();

	if (pVertexArrayDraw != NULL){
	    printf("Routine Draw Begin\n");
	    GLDrawBillboardMultitheadDraw(pVertexArrayDraw, pColorDraw, 
					  number2Draw);
	    printf("Routine Draw End\n");
	}
	
	while (bFinished == FALSE){
	    wxMilliSleep(2);
	}
	printf ("Thread Calc End\n");

	pVertexArrayDraw = pVertexArrayCalc;
	pColorDraw       = pColorCalc;
	calculating      = 1 - calculating;
	number2Draw      = number;

	if (GbAbort == TRUE){
	    GbAbort = FALSE;
	    return;
	}
    }
    
    printf("Routine Draw Fin Begin\n");
    GLDrawBillboardMultitheadDraw(pVertexArrayDraw, pColorDraw, 
				  number2Draw);
    printf("Routine Draw Fin End\n");
}

void GLDrawBillboard_cls::
GLDrawBillboardSequential(float *pParticleX, float vertexArray[12],
			  int drawType, int prevDrawType){

    if (drawType != prevDrawType){
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glDisable(GL_NORMALIZE);
    }
    
    vertexArray[0] += pParticleX[X];
    vertexArray[1] += pParticleX[Y];
    vertexArray[2] += pParticleX[Z];
    vertexArray[3] += pParticleX[X];
    vertexArray[4] += pParticleX[Y];
    vertexArray[5] += pParticleX[Z];
    vertexArray[6] += pParticleX[X];
    vertexArray[7] += pParticleX[Y];
    vertexArray[8] += pParticleX[Z];
    vertexArray[9] += pParticleX[X];
    vertexArray[10] += pParticleX[Y];
    vertexArray[11] += pParticleX[Z];
    
    glBegin(GL_QUADS);
    glTexCoord2fv(msTexCoordArray+0);glVertex3fv(vertexArray + 0);
    glTexCoord2fv(msTexCoordArray+2);glVertex3fv(vertexArray + 3);
    glTexCoord2fv(msTexCoordArray+4);glVertex3fv(vertexArray + 6);
    glTexCoord2fv(msTexCoordArray+6);glVertex3fv(vertexArray + 9);
    glEnd();
}

unsigned int GLDrawBillboard_cls::mspIndices[PACKNUM * 4];
float        GLDrawBillboard_cls::msTexCoordArray[PACKNUM * 4 * 2];
float        GLDrawBillboard_cls::msVertexArray[PACKNUM * 4 * 3];

ArrayAndIndex_cls* GLDrawBillboard_cls::mspIndexAndZ = NULL;
int                GLDrawBillboard_cls::msDownsizedAtFBO0 = 0;


