//(setq c-basic-offset 4)

#include "../COMMON/MyVector.h"
#include "../FILTER/clsFilterCtrl.h"
#include "../GLATTRIB/clsColorCurve.h"
#include "../GLATTRIB/clsGLAttribute.h"
#include "../GLOBJECTS/clsCylinder.h"
#include "../GLOBJECTS/clsSphere.h"
extern Sphere_cls *GpCSphere;
#include "../MAINCANVAS/clsDrawOption.h"
#include "../MAINCANVAS/clsInformationGUI.h"
#include "../MAINCANVAS/clsInformation.h"

#include "clsArrayAndIndex.h"
#include "clsGLDrawData.h"
#include "clsParticleDataBlock.h"
#include "clsShadowVolume.h"
#include "clsSnapData.h"
#include "clsZSort.h"

#include "../../wxWidgetEnumMenu.h"

int              GDrawingPercentage = 100;
int              GDrawingStage = 0;
extern bool      GbAbort;
extern wxFrame  *GpCMainFrameCLS;

#include <omp.h>
#include <algorithm>

#include <wx/stopwatch.h>

#define R 0
#define G 1
#define B 2

ArrayAndIndex_cls::
ArrayAndIndex_cls(){
    alphaFactor = 1.0;
}

ThreadA_cls::
ThreadA_cls(wxThreadKind kind) : wxThread(kind){
}

void ThreadA_cls::
SetParameters(bool *pbFinishedFlag, float *pX, int offset, int pnumber, 
	      float *vertexArray, float *pRotatedArray){
    mpbFinishedFlagLink = pbFinishedFlag;
    mpX = pX;
    mOffset = offset;
    mPNumber = pnumber;
    mvertexArray = vertexArray;
    mpRotatedArray = pRotatedArray;
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	;//MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadA_cls::
Entry(){

    for (int p = 0; p < mPNumber; p++){
	int poffset = p*12;
	for (int j = 0; j < 12; j++){
	    mvertexArray[poffset + j] = mpRotatedArray[j] * 0.1;
	}
	float *pParticleX = &mpX[(mOffset + p)*3];
	mvertexArray[poffset + 0] += pParticleX[X];
	mvertexArray[poffset + 1] += pParticleX[Y];
	mvertexArray[poffset + 2] += pParticleX[Z];
	mvertexArray[poffset + 3] += pParticleX[X];
	mvertexArray[poffset + 4] += pParticleX[Y];
	mvertexArray[poffset + 5] += pParticleX[Z];
	mvertexArray[poffset + 6] += pParticleX[X];
	mvertexArray[poffset + 7] += pParticleX[Y];
	mvertexArray[poffset + 8] += pParticleX[Z];
	mvertexArray[poffset + 9] += pParticleX[X];
	mvertexArray[poffset +10] += pParticleX[Y];
	mvertexArray[poffset +11] += pParticleX[Z];

    }

    *mpbFinishedFlagLink = TRUE;
    Delete();
    return 0;
}

GLDrawData_cls::
GLDrawData_cls(){
    if (msbInitialized == FALSE){
	for (int i = 0; i < 4 * PACKNUM; i++)
	    mspIndices[i] = i;
	for (int i = 0; i < PACKNUM; i++){
	    msTexCoordArray[i * 4 * 2 + 0 * 2 + 0] = 0;
	    msTexCoordArray[i * 4 * 2 + 0 * 2 + 1] = 0;
	    msTexCoordArray[i * 4 * 2 + 1 * 2 + 0] = 1;
	    msTexCoordArray[i * 4 * 2 + 1 * 2 + 1] = 0;
	    msTexCoordArray[i * 4 * 2 + 2 * 2 + 0] = 1;
	    msTexCoordArray[i * 4 * 2 + 2 * 2 + 1] = 1;
	    msTexCoordArray[i * 4 * 2 + 3 * 2 + 0] = 0;
	    msTexCoordArray[i * 4 * 2 + 3 * 2 + 1] = 1;
	}
	msbInitialized = TRUE;
    }
    for (int i = 0; i < 3; i++) mpZero4[i] = 0;
    mpZero4[3] = 1;

    mpCShadowVolume = new ShadowVolume_cls(&mspIndexAndZ,
					   &mCParticlePos);
    mpCZSort = new ZSort_cls;
    mSortType = _SORT_ORTHO;
}

GLDrawData_cls::
~GLDrawData_cls(){
    if (mspCSnapShot != NULL){
	delete mspCSnapShot;
	mspCSnapShot = NULL;
    }
    if (mspIndexAndZ != NULL){
	delete [] mspIndexAndZ;
	mspIndexAndZ = NULL;
    }

    delete mpCShadowVolume;
    delete mpCZSort;
}

void GLDrawData_cls::
SetCameraPos(double *pCamPos){
    LOOP_XYZ mCameraPos[xyz] = pCamPos[xyz];
}

void GLDrawData_cls::
SetCameraDir(double *pCamDir){
    LOOP_XYZ mCameraDir[xyz] = pCamDir[xyz];
}

void GLDrawData_cls::
SetLightSourceInfo(int lightNumber, 
		   float *pLightPos, //float *pLightClr, 
		   double lightPhi, double lightTheta){
    mLightNum = lightNumber;
    LOOP_XYZ mLightPos[xyz] = pLightPos[xyz];
    
    double lightTgt[3], lightTop[3];
    lightTop[0] = 0; lightTop[1] = 0; lightTop[2] = 1;
    lightTgt[0] = -mLightPos[0];
    lightTgt[1] = -mLightPos[1];
    lightTgt[2] = -mLightPos[2];
    MyVector::Normalize(lightTgt, 1.0f);
    MyVector::Orthogonalize(lightTgt, lightTop);
    MyVector::Normalize(lightTop, 1.0f);
    
    mCLightPos.Setv(lightTgt, lightTop);
    mCParticlePos = mCLightPos;

    mLightPhi      = lightPhi;
    mLightTheta    = lightTheta;
}

float GLDrawData_cls::
GetSteppingAlphaFactor(int step){
    float factor1 = 0.8 - 0.6*(float)step/100;
    if (factor1 < 0.2) factor1 = 0.2; 
    return 1.0f + ((float)step - 1.0f) * factor1;
}

float GLDrawData_cls::
GetSteppingRadiusFactor(int step){
    float factor1 = 0.2 + 0.6*(float)step/100;
    if (factor1 > 0.8) factor1 = 0.8;
    return sqrt(1.0f + ((float)step - 1.0f) * factor1);
}

void GLDrawData_cls::
GLDraw(int fboIndex, PDataBlock_cls *pCDataBlock,
       GLAttribute_cls **ppCGLAttribute, 
       double dT){

    static unsigned int sAlphaType[_MAX_PARTICLE_TYPE];// = {-1, -1, -1, -1};
    static float        sAlphaFactor[_MAX_PARTICLE_TYPE];// = {-1, -1, -1, -1};
    static float        sAlphaDependBase[_MAX_PARTICLE_TYPE];// = {-1, -1, -1, -1};
    static float        sAlphaMax[_MAX_PARTICLE_TYPE];//        = {-1, -1, -1, -1};

    GbAbort = FALSE;
    GDrawingPercentage = 0;
    int step = ppCGLAttribute[0]->mpCDrawOptionLink->mCDownSize.mBody;

    if (mspCSnapShot == NULL)
	mspCSnapShot = new SnapData_cls(pCDataBlock, ppCGLAttribute, dT, step);
    
    bool bDiff = FALSE;
    if (mspCSnapShot->IsItSameSnap(pCDataBlock, dT, step) != TRUE) bDiff = TRUE;
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	if (sAlphaType[i]       != ppCGLAttribute[i]->mCAlphaType.mBody)
	    bDiff = TRUE;
	if (sAlphaFactor[i]     != ppCGLAttribute[i]->mCAlphaFactor.mBody)
	    bDiff = TRUE;
	if (sAlphaDependBase[i] != ppCGLAttribute[i]->mCAlphaDenendBase.mBody)
	    bDiff = TRUE;
	if (sAlphaMax[i]        != ppCGLAttribute[i]->mCAlphaMax.mBody)
	    bDiff = TRUE;
    }

    // ႢČvZ
    if (bDiff == TRUE){
	delete mspCSnapShot;
	//printf("MakeSnapshot\n");
	mspCSnapShot = new SnapData_cls(pCDataBlock, ppCGLAttribute, dT, step);
	//printf("MakeSnapshotEnd\n");
	if (mspCInfoPanel->mpCCtrlLink->mCType.mBody == 0)
	    mspCInfoPanel->UpdateInfo(mspCSnapShot);
	else
	    mspCInfoPanel->UpdateInfo(pCDataBlock);
    }

    for (int i = 0; i < 4; i++){
	sAlphaType[i]   = ppCGLAttribute[i]->mCAlphaType.mBody;
	sAlphaFactor[i] = ppCGLAttribute[i]->mCAlphaFactor.mBody;
	sAlphaDependBase[i] = ppCGLAttribute[i]->mCAlphaDenendBase.mBody;
	sAlphaMax[i] = ppCGLAttribute[i]->mCAlphaMax.mBody;
    }
    
    mCStopWatch.Start(0);

    if (ppCGLAttribute[0]->mpCDrawOptionLink->mCClassify.mBody == 0){ 
	bool bDependOnColor = 
	    ppCGLAttribute[0]->mpCClrCurveCol->mCbColorCurve.mbBody;
	// q̎ޕ
	switch(ppCGLAttribute[0]->mCDrawType.mBody){
	case _GL_DRAW_POINT:
	    if (bDependOnColor == TRUE)
		GLDrawPointAll(fboIndex, mspCSnapShot, ppCGLAttribute[0]);
	    else
		GLDrawPointArrayElement
		    (fboIndex, mspCSnapShot, ppCGLAttribute[0]);
	    break;
	case _GL_DRAW_SPHERE:
	    GLDrawSphereAll(fboIndex, mspCSnapShot, ppCGLAttribute[0]);
	    break;
	case _GL_DRAW_BILLBOARD:
	    GLDrawBillboardAll(fboIndex, mspCSnapShot, ppCGLAttribute[0]);
	    break;
	}
    }
    else{
	GLDrawSequential(fboIndex, mspCSnapShot, ppCGLAttribute);
    }
    
    mCStopWatch.Pause();
    GDrawingPercentage = 100;
    UpdateStatusBar();
}

void GLDrawData_cls::
GLDrawPointArrayElement(int fboIndex, SnapData_cls *pCDataSnap, 
			GLAttribute_cls *pCGLAttribute){

    int PNum = pCDataSnap->mPNumBlock;
    int downsized = ZSort(fboIndex, pCDataSnap, pCGLAttribute);

    unsigned int *pIndices = new unsigned int [downsized];
    for (int i = 0; i < downsized; i++) pIndices[i] = mspIndexAndZ[i].index;
    
    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);

    glPointSize(pCGLAttribute->mCPointSize.mBody);

    int step = pCGLAttribute->mpCDrawOptionLink->mCDownSize.mBody;
    float steppingAlphaFactor = 1.0f + ((float)step - 1.0f) * (0.8 - 0.6*(float)step/100);
    
    float pColor[4];
    pCGLAttribute->mpCClrCurveCol->GetColor4fScaled(pColor, 0); //!!
    pColor[3] *= pCGLAttribute->mCAlphaFactor.mBody;
    pColor[3] *= steppingAlphaFactor;

    glColor4fv(pColor);

    if (pColor[3] < 1 || pCGLAttribute->mCBlendType.mBody == _GL_BLEND_ONE){
	glEnable(GL_BLEND);
	switch (pCGLAttribute->mCBlendType.mBody){
	case _GL_BLEND_ONE_MINUS_SRC_ALPHA:
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	    break;
	case _GL_BLEND_ONE:
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    glDisable(GL_DEPTH_TEST);
	    break;
	}
    }

    glEnableClientState(GL_VERTEX_ARRAY);    
    int elementMax = 1000000;
    for (int i = 0; i < downsized; i+=elementMax){
	int fin = elementMax;
	if (i + fin > downsized) fin = downsized - i;
	glVertexPointer(3, GL_FLOAT, 0, pCDataSnap->mpX);
	glDrawElements(GL_POINTS, fin, GL_UNSIGNED_INT, pIndices + i);
    }

    //glVertexPointer(3, GL_FLOAT, 0, pCDataSnap->mpX);
    //glDrawElements(GL_POINTS, downsized, GL_UNSIGNED_INT, pIndices);
    glDisableClientState(GL_VERTEX_ARRAY);

    glDisable(GL_BLEND);

    delete[] pIndices;
}

void GLDrawData_cls::
GLDrawPointAll(int fboIndex,
	       SnapData_cls *pCDataSnap,
	       GLAttribute_cls *pCGLAttribute){

    int PNum = pCDataSnap->mPNumBlock;
    int downsized = ZSort(fboIndex, pCDataSnap, pCGLAttribute);

    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);

    glPointSize(pCGLAttribute->mCPointSize.mBody);

    glEnable(GL_BLEND);
    switch (pCGLAttribute->mCBlendType.mBody){
    case _GL_BLEND_ONE_MINUS_SRC_ALPHA:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	break;
    case _GL_BLEND_ONE:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glDisable(GL_DEPTH_TEST);
	break;
    }
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    
    int step = pCGLAttribute->mpCDrawOptionLink->mCDownSize.mBody;
    float steppingAlphaFactor = 1.0f + ((float)step - 1.0f) * (0.8 - 0.6*(float)step/100);
    float totalAlphaFactor = steppingAlphaFactor * pCGLAttribute->mCAlphaFactor.mBody;
    
    int alphaType  = pCGLAttribute->mCAlphaType.mBody;
    float alphaMax = pCGLAttribute->mCAlphaMax.mBody; 
    float radiusFactor = pCGLAttribute->GetRadiusFactor();
    float (*pPos)[3] = pCDataSnap->mpX;
    float *pValue1;
    pCDataSnap->SetPValue1(&pValue1, pCGLAttribute, 0);
    
    int progressInterval = 100000;
    if (downsized > 2000000) 
	progressInterval = downsized / 20;

    float pColor[4];
    for (int i_ = 0; i_ < downsized; i_++){

	if (i_ % progressInterval == 0){
	    GDrawingPercentage = (long long)100 * i_ / downsized;
	    UpdateStatusBar();
	    if (msbUseAbortion == TRUE && mCStopWatch.Time() > msAbortLimitMS){
		i_ = downsized; 
		msstrWarning << wxT("Auto Aborted");
		continue;
	    }
	}

	int i = (mspIndexAndZ + i_)->index;
	
	pCGLAttribute->
	    mpCClrCurveCol->GetColor4fScaled(pColor, pValue1[i]);
	if (alphaType == _GL_ALPHA_INVERSE_SQUARE){
	    pColor[3] /= (radiusFactor * radiusFactor);
	    if (pColor[3] > alphaMax) pColor[3] = alphaMax;
	}
	pColor[3] *= totalAlphaFactor;
	
	glColor4fv(pColor);

	glBegin(GL_POINTS);
	glVertex3fv(pPos[i]);
	glEnd();
	
	if (GbAbort == TRUE){
	    i_ = downsized;
	    GbAbort = FALSE;
	}
    }
}

void GLDrawData_cls::
GetVertexArrayLookAtCamera(float pParticlePos[3], float array[12]){
    static float *spCameraPos = mCameraPos;
    float x0[3] = {1, 0, 0};
    float x[3], y[3];
    float p[3] = {pParticlePos[0] - spCameraPos[0],
		  pParticlePos[1] - spCameraPos[1], 
		  pParticlePos[2] - spCameraPos[2]};
    if (p[1] == 0 && p[2] == 0){
	x0[0] = 0; x0[1] = 1; x0[2] = 0;
    }
    float normp2 = p[0]*p[0] + p[1]*p[1] + p[2]*p[2];
    float normp = sqrt(normp2);
    float iproduct_per_normp2 = (x0[0]*p[0] + x0[1]*p[1] + x0[2]*p[2]) / normp2;

    x[0] = x0[0] - iproduct_per_normp2 * p[0];
    x[1] = x0[1] - iproduct_per_normp2 * p[1];
    x[2] = x0[2] - iproduct_per_normp2 * p[2];
    
    float normx = sqrt(x[0]*x[0] + x[1]*x[1] + x[2]*x[2]);
    x[0] /= normx;
    x[1] /= normx;
    x[2] /= normx;
    
    y[0] = -(p[1]*x[2] - p[2]*x[1]) / normp;
    y[1] = -(p[2]*x[0] - p[0]*x[2]) / normp;
    y[2] = -(p[0]*x[1] - p[1]*x[0]) / normp;
    
    array[ 0] = - x[0] - y[0];
    array[ 1] = - x[1] - y[1];
    array[ 2] = - x[2] - y[2];
    array[ 3] = + x[0] - y[0];
    array[ 4] = + x[1] - y[1];
    array[ 5] = + x[2] - y[2];
    array[ 6] = + x[0] + y[0];
    array[ 7] = + x[1] + y[1];
    array[ 8] = + x[2] + y[2];
    array[ 9] = - x[0] + y[0];
    array[10] = - x[1] + y[1];
    array[11] = - x[2] + y[2];
}

void GLDrawData_cls::
MultiplyMatrixOnVertexArrayOrg(GLfloat matrix[16], float array[3]){
    for (int i = 0; i < 4; i++)
	for (int j = 0; j < 3; j++)
	    array[i*3 + j] = 
		matrix[0+j] * msVertexArrayOrg[i*3 + 0] + 
		matrix[4+j] * msVertexArrayOrg[i*3 + 1] + 
		matrix[8+j] * msVertexArrayOrg[i*3 + 2];
}

void GLDrawData_cls::
GLDrawBillboardAll(int fboIndex, SnapData_cls *pCDataSnap, 
		   GLAttribute_cls *pCGLAttribute){

    int PNum = pCDataSnap->mPNumBlock;
    int downsized = ZSort(fboIndex, pCDataSnap, pCGLAttribute);

    GLfloat mMatrix[16];
    glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, mMatrix);
    float rotatedArray[12];
    MultiplyMatrixOnVertexArrayOrg(mMatrix, rotatedArray);

    glDisable(GL_LIGHTING);
    float pColor[4];
    pCGLAttribute->mpCClrCurveCol->GetColor4fScaled(pColor, 0); //!!
    pColor[3] *= pCGLAttribute->mCAlphaFactor.mBody;
    glColor4fv(pColor);
    
    int step = pCGLAttribute->mpCDrawOptionLink->mCDownSize.mBody;
    float steppingRadiusFactor = sqrt(1.0f + ((float)step - 1.0f) * (0.2 + 0.6*(float)step/100));
    float steppingAlphaFactor = 1.0f + ((float)step - 1.0f) * (0.8 - 0.6*(float)step/100);

    glEnable(GL_TEXTURE_2D);
    switch (pCGLAttribute->mCBlendType.mBody){
    case _GL_BLEND_ONE_MINUS_SRC_ALPHA:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	break;
    case _GL_BLEND_ONE:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	break;
    }
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);
    
    glEnable(GL_BLEND);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glBindTexture(GL_TEXTURE_2D, pCGLAttribute->mTextureBillboard);
    
    glPushMatrix();

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    int   alphaType    = pCGLAttribute->mCAlphaType.mBody;
    float radiusFactor = pCGLAttribute->GetRadiusFactor() *
	steppingRadiusFactor;
    bool  bConstantRadius = 
	(pCGLAttribute->mCRadiusType.mBody == _GL_RADIUS_CONSTANT);
    bool  bConstantColor = 
	(pCGLAttribute->mpCClrCurveCol->mCbColorCurve.mbBody == FALSE);
    float *pValue1;
    pCDataSnap->SetPValue1(&pValue1, pCGLAttribute, 0);

    bool bUseMultiThread = FALSE;
    if (downsized > 600000 && 
	(bConstantColor != TRUE || alphaType != _GL_ALPHA_NORMAL))
	bUseMultiThread = TRUE;
    
    //printf("DrawBillboard\n");
    //bUseMultiThread = FALSE;
    if (bUseMultiThread == TRUE)
	GLDrawBillboardMultithead(downsized, rotatedArray,
				  radiusFactor, bConstantRadius, bConstantColor,
				  steppingAlphaFactor, pCDataSnap, 
				  pCGLAttribute, pValue1);
    if (bUseMultiThread == FALSE)
	GLDrawBillboardBlock(downsized, rotatedArray,
			     radiusFactor, bConstantRadius, bConstantColor,
			     steppingAlphaFactor, pCDataSnap, 
			     pCGLAttribute, pValue1);
    printf("DrawBillboardEnd\n");
    
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);

    glPopMatrix();
    
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
}

void GLDrawData_cls::
GLDrawSphereConstRadiusConstantColor(int downsized, float (*pPos)[3], 
				     float radiusFactor, float quality){
    for (int ibase = 0; ibase < downsized; ibase+=PACKNUM) {
	int number = PACKNUM;
	if (ibase + number >= downsized) number = downsized - ibase;
	
	if (ibase % (PACKNUM * 500) == 0){
	    GDrawingPercentage = (long long)100 * ibase / downsized;
	    UpdateStatusBar();
	    //glFlush();
	    if (msbUseAbortion == TRUE && mCStopWatch.Time() > msAbortLimitMS){
		ibase = downsized; 
		msstrWarning << wxT("Auto Aborted");
		continue;
	    }
	}
	
	glPushMatrix();
	int i = (mspIndexAndZ + ibase)->index;
	glTranslatef(pPos[i][X]/radiusFactor, pPos[i][Y]/radiusFactor, pPos[i][Z]/radiusFactor);
	GpCSphere->Draw(1 + quality);
	int prev = i;

	if (/*USETEXTURE=*/FALSE){
	    for (int i_ = ibase+1; i_ < ibase + number; i_ ++){
		int i = (mspIndexAndZ + i_)->index;
		glTranslatef((pPos[i][X] - pPos[prev][X])/radiusFactor,
			     (pPos[i][Y] - pPos[prev][Y])/radiusFactor,
			     (pPos[i][Z] - pPos[prev][Z])/radiusFactor);
		GpCSphere->Draw(1 + quality);
		prev = i;
	    }
	}
	else{
	    for (int i_ = ibase+1; i_ < ibase + number; i_ ++){
		int i = (mspIndexAndZ + i_)->index;
		glTranslatef((pPos[i][X] - pPos[prev][X])/radiusFactor,
			     (pPos[i][Y] - pPos[prev][Y])/radiusFactor,
			     (pPos[i][Z] - pPos[prev][Z])/radiusFactor);
		GpCSphere->DrawNoTex(1 + quality);
		prev = i;
	    }
	}
	
	glPopMatrix();
	
	if (GbAbort == TRUE){
	    ibase = downsized;
	    GbAbort = FALSE;
	}
	
    }
}

void GLDrawData_cls::
GLDrawSphereConstRadiusVariableColor
(int downsized, float (*pPos)[3], float *pValue1,
 float radiusFactor, float quality, float steppingAlphaFactor,
 GLAttribute_cls *pCGLAttribute){
    float pClrMat[3][4];
    ColorCurve_cls *pCClrCurveMat[3];
    for (int mat = 0; mat < 3; mat++){
	pCClrCurveMat[mat] = pCGLAttribute->mpCClrCurveMat[mat];
	pCClrCurveMat[mat]->PreCalcScaleFactor();
    }
    
    for (int ibase = 0; ibase < downsized; ibase+=PACKNUM) {
	int number = PACKNUM;
	if (ibase + number >= downsized) number = downsized - ibase;
	
	if (ibase % (PACKNUM * 500) == 0){
	    GDrawingPercentage = (long long)100 * ibase / downsized;
	    UpdateStatusBar();
	    //glFlush();
	    if (msbUseAbortion == TRUE && mCStopWatch.Time() > msAbortLimitMS){
		ibase = downsized; 
		msstrWarning << wxT("Auto Aborted");
		continue;
	    }
	}
	
	glPushMatrix();
	int i = (mspIndexAndZ + ibase)->index;
	float filterAlhplaFactor = ((mspIndexAndZ + ibase))->alphaFactor;
	glTranslatef(pPos[i][X]/radiusFactor, pPos[i][Y]/radiusFactor, pPos[i][Z]/radiusFactor);
	
	for (int mat = 0; mat < 3; mat++){
	    pCClrCurveMat[mat]->GetColor4fScaled(pClrMat[mat], pValue1[i]);
	    pClrMat[mat][3] *= pCGLAttribute->mCAlphaFactor.mBody;
	    pClrMat[mat][3] *= steppingAlphaFactor;
	    pClrMat[mat][3] *= filterAlhplaFactor;
	}
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,  pClrMat[0]);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  pClrMat[1]);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, pClrMat[2]);

	GpCSphere->Draw(1 + quality);
	int prev = i;
	for (int i_ = ibase+1; i_ < ibase + number; i_ ++){
	    int i = (mspIndexAndZ + i_)->index;
	    float filterAlhplaFactor = ((mspIndexAndZ + i_))->alphaFactor;
	    glTranslatef((pPos[i][X] - pPos[prev][X])/radiusFactor,
			 (pPos[i][Y] - pPos[prev][Y])/radiusFactor,
			 (pPos[i][Z] - pPos[prev][Z])/radiusFactor);

	    for (int mat = 0; mat < 3; mat++){
		pCClrCurveMat[mat]->GetColor4fScaled(pClrMat[mat], pValue1[i]);
		pClrMat[mat][3] *= pCGLAttribute->mCAlphaFactor.mBody;
		pClrMat[mat][3] *= steppingAlphaFactor;
		pClrMat[mat][3] *= filterAlhplaFactor;
	    }
	    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,  pClrMat[0]);
	    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  pClrMat[1]);
	    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, pClrMat[2]);

	    GpCSphere->Draw(1 + quality);
	    prev = i;
	}
	glPopMatrix();
	
	if (GbAbort == TRUE){
	    ibase = downsized;
	    GbAbort = FALSE;
	}
	
    }
}

void GLDrawData_cls::
GLDrawSphereVariableRadiusConstColor
(int downsized, float (*pPos)[3], float *pRadius,
 float radiusFactor, float quality){
    for (int i_ = 0; i_ < downsized; i_++){
	int i = (mspIndexAndZ + i_)->index;
	DrawSphere(pPos[i], pRadius[i] * radiusFactor, quality);
	if (GbAbort == TRUE){
	    i_ = downsized;
	    GbAbort = FALSE;
	}
    }
}

void GLDrawData_cls::
GLDrawSphereVariableRadiusVariableColor
(int downsized, float (*pPos)[3], float *pRadius, float *pValue1,
 float radiusFactor, float quality, float steppingAlphaFactor,
 GLAttribute_cls *pCGLAttribute){
    float pClrMat[3][4];
    ColorCurve_cls *pCClrCurveMat[3];
    for (int mat = 0; mat < 3; mat++){
	pCClrCurveMat[mat] = pCGLAttribute->mpCClrCurveMat[mat];
	pCClrCurveMat[mat]->PreCalcScaleFactor();
    }

    for (int i_ = 0; i_ < downsized; i_++){
	int i = (mspIndexAndZ + i_)->index;

	for (int mat = 0; mat < 3; mat++){
	    pCClrCurveMat[mat]->GetColor4fScaled(pClrMat[mat], pValue1[i]);
	    pClrMat[mat][3] *= pCGLAttribute->mCAlphaFactor.mBody;
	    pClrMat[mat][3] *= steppingAlphaFactor;
	}
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,  pClrMat[0]);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  pClrMat[1]);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, pClrMat[2]);
	
	DrawSphere(pPos[i], pRadius[i] * radiusFactor, quality);
	if (GbAbort == TRUE){
	    i_ = downsized;
	    GbAbort = FALSE;
	}
    }
}

void GLDrawData_cls::
GLDrawCylinderVariableRadius(int downsized, float (*pPos)[3], float *pRadius,
			     float radiusFactor, float quality){
    for (int i_ = 0; i_ < downsized; i_++){
	int i = (mspIndexAndZ + i_)->index;
	DrawCylinder(pPos[i], pRadius[i] * radiusFactor, quality);
	if (GbAbort == TRUE){
	    i_ = downsized;
	    GbAbort = FALSE;
	}
    }
}

void GLDrawData_cls::
DrawCylinder(float pos[3], float scale, int quality){
    // Look At LightSourceInf
    mCParticlePos.mPos[0] = pos[0];
    mCParticlePos.mPos[1] = pos[1];
    mCParticlePos.mPos[2] = pos[2];
    
    double matrix[16];
    mCParticlePos.GetGLMatrix(matrix);
    glPushMatrix();
    glMultMatrixd(matrix); 
    glScaled(scale, scale, scale);

    glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
    Cylinder_cls::Draw(1);

    glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
    Cylinder_cls::Draw(0);
    
    glPopMatrix();
}

void GLDrawData_cls::
GLDrawSphereAll(int fboIndex, SnapData_cls *pCDataSnap, 
		GLAttribute_cls *pCGLAttribute){

    int PNum = pCDataSnap->mPNumBlock;
    int downsized = ZSort(fboIndex, pCDataSnap, pCGLAttribute);

    glEnable(GL_LIGHTING);

    int step = pCGLAttribute->mpCDrawOptionLink->mCDownSize.mBody;
    float steppingAlphaFactor = GetSteppingAlphaFactor(step);
    float steppingRadiusFactor = GetSteppingRadiusFactor(step);

    float pClrMat[3][4];
    for (int mat = 0; mat < 3; mat++){
	pCGLAttribute->
	    mpCClrCurveMat[mat]->GetColor4fScaled(pClrMat[mat], 0); //!!
	pClrMat[mat][3] *= pCGLAttribute->mCAlphaFactor.mBody;
	pClrMat[mat][3] *= steppingAlphaFactor;
    }

    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,  pClrMat[0]);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  pClrMat[1]);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, pClrMat[2]);

    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 20);

    glDisable(GL_TEXTURE_2D);
    switch (pCGLAttribute->mCBlendType.mBody){
    case _GL_BLEND_ONE_MINUS_SRC_ALPHA:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	break;
    case _GL_BLEND_ONE:
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	//glDisable(GL_DEPTH_TEST);
	break;
    }

    if (mbShadowVolume == FALSE)
	glDepthMask(GL_TRUE);
    else
	glDepthMask(GL_FALSE);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glEnable(GL_NORMALIZE);

    float radiusFactor = pCGLAttribute->GetRadiusFactor() *
	steppingRadiusFactor;
    bool  bConstantRadius = 
	(pCGLAttribute->mCRadiusType.mBody == _GL_RADIUS_CONSTANT);
    bool  bConstantColor = 
	(pCGLAttribute->mpCClrCurveMat[0]->mCbColorCurve.mbBody == FALSE) &&
	(pCGLAttribute->mpCClrCurveMat[1]->mCbColorCurve.mbBody == FALSE) &&
	(pCGLAttribute->mpCClrCurveMat[2]->mCbColorCurve.mbBody == FALSE);

    if (mspCFilterCtrlBooleanLink->IsBooleanChangeConstantColor() == TRUE)
	bConstantColor = FALSE; 
    
    int   quality = pCGLAttribute->mCDrawQuality.mBody;
    
    float (*pPos)[3] = pCDataSnap->mpX;
    float *pRadius   = pCDataSnap->mpRadius;
    float *pValue1;
    // !! _FIX_ME !!AmbientŔ    
    pCDataSnap->SetPValue1(&pValue1, pCGLAttribute, 1);

    if (mbShadowVolume == FALSE){
	pCGLAttribute->BeginShader(mLightNum);
	glPushMatrix();
	if (bConstantRadius == TRUE){
	    glScalef(radiusFactor, radiusFactor, radiusFactor);
	    if (bConstantColor == TRUE)
		GLDrawSphereConstRadiusConstantColor
		    (downsized, pPos, radiusFactor, quality);
	    else
		GLDrawSphereConstRadiusVariableColor
		    (downsized, pPos, pValue1, radiusFactor, quality, 
		     steppingAlphaFactor, pCGLAttribute);
	}
	else{
	    if (bConstantColor == TRUE)
		GLDrawSphereVariableRadiusConstColor
		    (downsized, pPos, pRadius, radiusFactor, quality);
	    else
		GLDrawSphereVariableRadiusVariableColor
		    (downsized, pPos, pRadius, pValue1, radiusFactor, quality,
		     steppingAlphaFactor, pCGLAttribute);
	}
	glPopMatrix();
	pCGLAttribute->EndShader();
    }
    else{
	// ShadowVolume[hȂ
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);
	float lightDir[3] = {mLightPos[0], mLightPos[1], mLightPos[1]};
	
	mpCShadowVolume->CheckShadow
	    (downsized, pPos, pRadius, radiusFactor, mCameraPos, lightDir,
	     mLightPhi, mLightTheta);
	
	glPushMatrix();
	if (bConstantRadius == TRUE)
	    mpCShadowVolume->GLDrawCylinderConstantRadius
		(downsized, pPos, pRadius, radiusFactor, quality);
	else
	    mpCShadowVolume->GLDrawCylinderVariableRadius
		(downsized, pPos, pRadius, radiusFactor, quality);
	glPopMatrix();
    }
    
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
}

// ----

void GLDrawData_cls::
DrawSphere(float pos[3], float scale, int quality){
    glPushMatrix();
    glTranslatef(pos[X], pos[Y], pos[Z]);
    glScalef(scale, scale, scale);
    GpCSphere->Draw(1 + quality);
    glPopMatrix();
}

/*
int GLDrawData_cls::
ZSortOrthA(int PNum, int step, float (*pX)[3], int sortType){
    int downsized = 0;
    // ̒xłOpenMPőȂȂc
    for (int i = 0; i < PNum; i+=step){ 
	mspIndexAndZ[downsized].index = i;
	mspIndexAndZ[downsized].z = - //\[gtɂ邽
	    (pX[i][X]-mCameraPos[X]) * mCameraDir[X] - 
	    (pX[i][Y]-mCameraPos[Y]) * mCameraDir[Y] - 
	    (pX[i][Z]-mCameraPos[Z]) * mCameraDir[Z];
	downsized++;
    }
    return downsized;
}

int GLDrawData_cls::
ZSortSphericalA(int PNum, int step, float (*pX)[3], int sortType){
    int downsized = 0;
    // ̒xłOpenMPőȂȂc
    for (int i = 0; i < PNum; i+=step){ 
	mspIndexAndZ[downsized].index = i;
	mspIndexAndZ[downsized].z = - //\[gtɂ邽
	    (pX[i][X]-mCameraPos[X]) * (pX[i][X]-mCameraPos[X]) - 
	    (pX[i][Y]-mCameraPos[Y]) * (pX[i][Y]-mCameraPos[Y]) - 
	    (pX[i][Z]-mCameraPos[Z]) * (pX[i][Z]-mCameraPos[Z]);
	downsized++;
    }
    return downsized;
    
}
*/

void GLDrawData_cls::
ZSortAtFBO0(int PNum, int step, float (*pX)[3], int sortOrder/*, int sortType*/){

    if (mspIndexAndZ != NULL) delete [] mspIndexAndZ;
    mspIndexAndZ =  new ArrayAndIndex_cls [PNum/step + 1];
    
    int downsized = 0;
    switch(sortOrder){
    case 0:
	for (int i = 0; i < PNum; i+=step){
	    mspIndexAndZ[downsized].index = i;
	    downsized++;
	}
	
	downsized = mspCFilterCtrlBooleanLink->
	    CheckBoolean(mspIndexAndZ, pX, downsized, msCurrentDoubleIndex);
	
	break;
    case 1: 
	if (mSortType == _SORT_ORTHO)
	    downsized = mpCZSort->
		ZSortPrepare(mspIndexAndZ, PNum, step, pX, 
			     mCameraPos, mCameraDir, 0);
	else
	    downsized = mpCZSort->
		ZSortPrepare(mspIndexAndZ, PNum, step, pX, 
			     mCameraPos, mCameraDir, 1);

	// !!_FIX_ME_!! ̏Ԃtɂł΁A
	// SortԂ񏭂ł
	downsized = mspCFilterCtrlBooleanLink->
	    CheckBoolean(mspIndexAndZ, pX, downsized, msCurrentDoubleIndex);
	
	if (downsized > 100000)
	    mpCZSort->ZSortParallel(mspIndexAndZ, downsized);
	else
	    std::sort(mspIndexAndZ, mspIndexAndZ + downsized);

	break;
    case 2:
	for (int i = 0; i < PNum; i+=step){ // ̒xłOpenMPőȂȂc
	    mspIndexAndZ[downsized].index = i;
	    mspIndexAndZ[downsized].z = //߂`
		(pX[i][X]-mCameraPos[X]) * (pX[i][X]-mCameraPos[X]) +
		(pX[i][Y]-mCameraPos[Y]) * (pX[i][Y]-mCameraPos[Y]) + 
		(pX[i][Z]-mCameraPos[Z]) * (pX[i][Z]-mCameraPos[Z]);
	    downsized++;
	}

	if (downsized > 100000)
	    mpCZSort->ZSortParallel(mspIndexAndZ, downsized);
	else
	    std::sort(mspIndexAndZ, mspIndexAndZ + downsized);
	break;
    }
    msDownsizedAtFBO0 = downsized;
}

int  GLDrawData_cls::
ZSort(int fboIndex, 
      SnapData_cls* pCDataSnap, GLAttribute_cls *pCGLAttribute){
    int PNum = pCDataSnap->mPNumBlock;
    int step = pCGLAttribute->mpCDrawOptionLink->mCDownSize.mBody;
    int sortOrder = pCGLAttribute->mpCDrawOptionLink->mCSortOrder.mBody;
    //int sortType = _SORT_ORTHO;
    if (pCGLAttribute->mCDrawType.mBody == _GL_DRAW_BILLBOARD &&
	pCGLAttribute->mCDrawQuality.mBody >= 1)
	mSortType = _SORT_RADIAL;
    if (fboIndex == 0) ZSortAtFBO0(PNum, step, pCDataSnap->mpX, 
				   sortOrder/*, sortType*/);
    return msDownsizedAtFBO0;
}

void GLDrawData_cls::
GLDrawSequential(int fboIndex, SnapData_cls *pCDataSnap, 
		 GLAttribute_cls **ppCGLAttribute){

    wxStopWatch CStopWatch;
    CStopWatch.Start(0);
    int dispatchTime = 0;

    for (int i = 0; i < 4; i++)
	for (int mat = 0; mat < 3; mat++){
	    ppCGLAttribute[i]->mpCClrCurveMat[mat]->PreCalcScaleFactor();
	}
    
    GLfloat mMatrix[16];
    glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, mMatrix);
    float rotatedArray[12];
    float vertexArray[12];
    MultiplyMatrixOnVertexArrayOrg(mMatrix, rotatedArray);

    int PNum = pCDataSnap->mPNumBlock;

    int step = ppCGLAttribute[0]->mpCDrawOptionLink->mCDownSize.mBody;
    float steppingAlphaFactor = GetSteppingAlphaFactor(step);
    float steppingRadiusFactor = GetSteppingRadiusFactor(step);
    
    int sortOrder = ppCGLAttribute[0]->mpCDrawOptionLink->mCSortOrder.mBody;
    int downsized = ZSort(fboIndex, pCDataSnap, ppCGLAttribute[0]);
    
    int prevParticleType = -1;
    int prevBlendMode = -1;
    int prevDrawType = -1;
    
    bool bConstantColorInBillboard[_MAX_PARTICLE_TYPE];
    bool bConstantColorInSphere[_MAX_PARTICLE_TYPE];
    int   alphaType[_MAX_PARTICLE_TYPE];
    float alphaDependBase[_MAX_PARTICLE_TYPE];
    float alphaMax[_MAX_PARTICLE_TYPE];
    float alphaFactor[_MAX_PARTICLE_TYPE];
    int   qualityList[_MAX_PARTICLE_TYPE];
    int   blendModeList[_MAX_PARTICLE_TYPE];
    int   drawTypeList[_MAX_PARTICLE_TYPE];
    int   radiusTypeList[_MAX_PARTICLE_TYPE];
    float radiusFactorList[_MAX_PARTICLE_TYPE];
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	bConstantColorInBillboard[i] = (ppCGLAttribute[i]->mpCClrCurveCol->mCbColorCurve.mbBody != TRUE);
	bConstantColorInSphere[i] = 
	    (ppCGLAttribute[i]->mpCClrCurveMat[0]->mCbColorCurve.mbBody!=TRUE)&
	    (ppCGLAttribute[i]->mpCClrCurveMat[1]->mCbColorCurve.mbBody!=TRUE)&
	    (ppCGLAttribute[i]->mpCClrCurveMat[2]->mCbColorCurve.mbBody!=TRUE);
	if (mspCFilterCtrlBooleanLink->
	    IsBooleanChangeConstantColor() == TRUE)
	    bConstantColorInSphere[i] = FALSE;
	alphaType[i]        = ppCGLAttribute[i]->mCAlphaType.mBody; 
	alphaDependBase[i]  = ppCGLAttribute[i]->mCAlphaDenendBase.mBody;
	alphaMax[i]         = ppCGLAttribute[i]->mCAlphaMax.mBody;
	alphaFactor[i]      = ppCGLAttribute[i]->mCAlphaFactor.mBody;
	qualityList[i]      = ppCGLAttribute[i]->mCDrawQuality.mBody;
	blendModeList[i]    = ppCGLAttribute[i]->mCBlendType.mBody;
	drawTypeList[i]     = ppCGLAttribute[i]->mCDrawType.mBody;
	radiusTypeList[i]   = ppCGLAttribute[i]->mCRadiusType.mBody;
	radiusFactorList[i] = ppCGLAttribute[i]->GetRadiusFactor() * 
	    steppingRadiusFactor;
    }

    glEnable(GL_BLEND);
    glBindTexture(GL_TEXTURE_2D, ppCGLAttribute[0]->mTextureBillboard);
    
    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);
    glDisable(GL_LIGHTING);
    //glDisable(GL_NORMALIZE);

    float pColor[4];
    float pClrMat[3][4];

    unsigned char *pType     = pCDataSnap->mpTypeLink;
    float (*pPos)[3] = pCDataSnap->mpX;
    float *pRadius = pCDataSnap->mpRadius;
    float *pValue1Mat[_MAX_PARTICLE_TYPE];    
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	pCDataSnap->SetPValue1(&pValue1Mat[i], ppCGLAttribute[i], 1);
    
    if (mbShadowVolume == TRUE){
	float lightDir[3] = {mLightPos[0], mLightPos[1], mLightPos[2]};
	mpCShadowVolume->CheckShadowSequential
	    (downsized, drawTypeList, radiusTypeList, 
	     radiusFactorList, pType, pPos, pRadius,
	     mCameraPos, lightDir,
	     mLightPhi, mLightTheta);
	
	GLDrawShadowVolumeSequential
	    (pCDataSnap, downsized, drawTypeList, radiusTypeList, 
	     radiusFactorList, pType, pPos, pRadius);
	return;
    }
    
    glEnable(GL_TEXTURE_2D);

    int progressInterval = 100000;
    if (downsized > 2000000) 
	progressInterval = downsized / 20;
    
    for (int i_ = 0; i_ < downsized; i_++){
	
	if (i_ % progressInterval == 0 && CStopWatch.Time()>dispatchTime+200){
	    dispatchTime = CStopWatch.Time();
	    GDrawingPercentage = (long long)100 * i_ / downsized;
	    UpdateStatusBar();
	    //glFlush();
	    if (msbUseAbortion == TRUE && mCStopWatch.Time() > msAbortLimitMS){
		i_ = downsized; 
		msstrWarning << wxT("Auto Aborted");
		continue;
	    }
	}
	
	int i = (mspIndexAndZ+i_)->index;
	float filterAlphaFactor = ((mspIndexAndZ + i_))->alphaFactor;
	
	int particleType = pType[i];/*pCDataSnap->mpTypeLink[i];*/
	
	GLAttribute_cls *pCGLAttribute =  ppCGLAttribute[particleType];
	int blendMode = blendModeList[particleType];
	int drawType  = drawTypeList[particleType];
	if (blendMode != prevBlendMode)
	    switch(blendMode){
	    case _GL_BLEND_ONE_MINUS_SRC_ALPHA:
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		break;
	    case _GL_BLEND_ONE:
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		break;
	    }

	if (drawType == _GL_DRAW_POINT){
	    if (particleType != prevParticleType ||
		bConstantColorInBillboard[particleType] == FALSE){
		pCGLAttribute->EndShader();
		memcpy(pColor, pCDataSnap->mpColor[i], sizeof(float)*4);
		pColor[3] *= (steppingAlphaFactor*filterAlphaFactor);//*alphaFactor[particleType]);
		glColor4fv(pColor);
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);
		glPointSize(ppCGLAttribute[particleType]->mCPointSize.mBody);
	    }
	    glBegin(GL_POINTS);
	    glVertex3fv(pPos[i]);
	    glEnd();

	    prevParticleType = particleType;
	    prevBlendMode = blendMode;
	    prevDrawType = drawType;
	    continue;
	}

	float radiusFactor = radiusFactorList[particleType];
	if (radiusTypeList[particleType] != _GL_RADIUS_CONSTANT)
	    radiusFactor *= pRadius[i];
	
	switch(drawType){
	case _GL_DRAW_SPHERE: {

	    if (particleType != prevParticleType ||
		bConstantColorInSphere[particleType] == FALSE ||
		alphaType[particleType] != _GL_ALPHA_NORMAL){
		for (int mat = 0; mat < 3; mat++){
		    ppCGLAttribute[particleType]->
			mpCClrCurveMat[mat]->GetColor4fScaled
			(pClrMat[mat], pValue1Mat[particleType][i]); 
		    if (alphaType[particleType] == _GL_ALPHA_INVERSE_SQUARE){
			pClrMat[mat][3] *= alphaDependBase[particleType] * alphaDependBase[particleType] / (radiusFactor * radiusFactor);
			if (pClrMat[mat][3] > alphaMax[particleType])
			    pClrMat[mat][3] = alphaMax[particleType];
		    }
		    else
			pClrMat[mat][3] *= alphaFactor[particleType];
		    pClrMat[mat][3] *= (steppingAlphaFactor*filterAlphaFactor);
		    if (pClrMat[mat][3] > 1) pClrMat[mat][3] = 1;
		}
		
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,  pClrMat[0]);
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  pClrMat[1]);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, pClrMat[2]);

		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 20);
		// ͋ʉ΃[v̊Oɏo͂
		
		pCGLAttribute->BeginShader(mLightNum);
	    }
	    
	    if (drawType != prevDrawType){
		glDisable(GL_TEXTURE_2D);
		glEnable(GL_LIGHTING);
		glEnable(GL_NORMALIZE);
		glEnable(GL_DEPTH_TEST);
	    }

	    /*
	    glDisable(GL_TEXTURE_2D);
	    glEnable(GL_LIGHTING);
	    glEnable(GL_NORMALIZE);
	    glEnable(GL_DEPTH_TEST);
	    */

	    glDepthMask(GL_TRUE);
	    DrawSphere(pPos[i], radiusFactor, qualityList[particleType]);
	    glDepthMask(GL_FALSE);
	    break; }
	case _GL_DRAW_BILLBOARD:{
	    
	    // F̔aɂωdownsizingˑ̂
	    // ŌvZ
	    if (particleType != prevParticleType ||
		bConstantColorInBillboard[particleType] == FALSE ||
		alphaType[particleType] != _GL_ALPHA_NORMAL){

		// prevParticleTypeɂ炸textureoffȂ̂
		glEnable(GL_TEXTURE_2D);
		glDisable(GL_LIGHTING);
		
		memcpy(pColor, pCDataSnap->mpColor[i], sizeof(float)*4);
		if (alphaType[particleType] == _GL_ALPHA_INVERSE_SQUARE){
		    pColor[3] /= (radiusFactor * radiusFactor);
		    if (pColor[3] > alphaMax[particleType])
			pColor[3] = alphaMax[particleType];
		}
		pColor[3] *= (steppingAlphaFactor*filterAlphaFactor);
		glColor4fv(pColor);
		pCGLAttribute->EndShader();
	    }
	    
	    if (pColor[3] < 0.0035){ 
		prevParticleType = particleType;
		prevBlendMode = blendMode;
		prevDrawType = drawType;
		continue;
	    }

	    if (qualityList[particleType] == 0){
		vertexArray[ 0] = rotatedArray[ 0] * radiusFactor;
		vertexArray[ 1] = rotatedArray[ 1] * radiusFactor;
		vertexArray[ 2] = rotatedArray[ 2] * radiusFactor;
		vertexArray[ 3] = rotatedArray[ 3] * radiusFactor;
		vertexArray[ 4] = rotatedArray[ 4] * radiusFactor;
		vertexArray[ 5] = rotatedArray[ 5] * radiusFactor;
		vertexArray[ 6] = rotatedArray[ 6] * radiusFactor;
		vertexArray[ 7] = rotatedArray[ 7] * radiusFactor;
		vertexArray[ 8] = rotatedArray[ 8] * radiusFactor;
		vertexArray[ 9] = rotatedArray[ 9] * radiusFactor;
		vertexArray[10] = rotatedArray[10] * radiusFactor;
		vertexArray[11] = rotatedArray[11] * radiusFactor;
	    }
	    else{
		GetVertexArrayLookAtCamera(pPos[i], vertexArray);
		for (int j = 0; j < 12; j++)
		    vertexArray[j] *= radiusFactor;
	    }
	    
	    GLDrawBillboardSequential(pPos[i], vertexArray,
				      drawType, prevDrawType);
	    
	    break;}
	}

	prevParticleType = particleType;
	prevBlendMode = blendMode;
	prevDrawType = drawType;

	if (GbAbort == TRUE){
	    i_ = downsized;
	    GbAbort = FALSE;
	}
    }

    glEnable(GL_DEPTH_TEST);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
}

void GLDrawData_cls::
GLDrawShadowVolumeSequential
(SnapData_cls *pCDataSnap, int downsized, 
 int drawTypeList[_MAX_PARTICLE_TYPE],
 int radiusTypeList[_MAX_PARTICLE_TYPE],
 float radiusFactorList[_MAX_PARTICLE_TYPE],
 unsigned char *pType, float (*pPos)[3], float *pRadius){
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);

    int progressInterval = 100000;
    if (downsized > 2000000)
	progressInterval = downsized / 20;

    for (int i_ = 0; i_ < downsized; i_++){
	
	if (i_ % progressInterval == 0){
	    GDrawingPercentage = (long long)100 * i_ / downsized;
	    UpdateStatusBar();
	    //glFlush();
	    if (msbUseAbortion == TRUE && mCStopWatch.Time() > msAbortLimitMS){
		i_ = downsized; 
		msstrWarning << wxT("Auto Aborted");
		continue;
	    }
	}
	
	int i = (mspIndexAndZ+i_)->index;
	int particleType = pType[i];
	int drawType = drawTypeList[particleType];
	
	if (drawType != _GL_DRAW_SPHERE) continue;
	float radiusFactor = radiusFactorList[particleType];
	if (radiusTypeList[particleType] != _GL_RADIUS_CONSTANT)
	    radiusFactor *= pRadius[i];

	mpCShadowVolume->DrawCylinder(pPos[i], radiusFactor, i, i_);
    }

    glEnable(GL_DEPTH_TEST);
}

bool         GLDrawData_cls::msbInitialized = FALSE;
float        GLDrawData_cls::msVertexArrayOrg[12] = {
    -1, -1, 0,
    +1, -1, 0,
    +1, +1, 0,
    -1, +1, 0
};
 
SnapData_cls* GLDrawData_cls::mspCSnapShot = NULL;
InformationGUI_cls* GLDrawData_cls::mspCInfoPanel = NULL;
FilterCtrl_cls* GLDrawData_cls::mspCFilterCtrlBooleanLink = NULL;
double GLDrawData_cls::msCurrentDoubleIndex = 0;
int    GLDrawData_cls::msAbortLimitMS = 1000;
bool   GLDrawData_cls::msbUseAbortion = FALSE;
wxString GLDrawData_cls::msstrWarning;
