
#include "../HDF5/clsReadHDF5Data.h"

#include "clsParticleDataBlock.h"
#include "clsParticleDataSnap.h"

#include "../COMMON/CommonDefinitionsForH.h"
#include "clsThreadFileReader.h"
extern ThreadFileReader_cls   *GApCThreadReader[_MAX_READ_THREAD];
extern bool                    GAbThreadReaderFinished[_MAX_READ_THREAD];
extern int                     GAThreadReaderIndex[_MAX_READ_THREAD];
#include "INTERPOLATE_GET/clsThreadInterpolate.h"
extern ThreadInterpolator_cls *GpCThreadInterpolate;
extern bool                    GbThreadInterpolateFinished;

extern wxFrame       *GpCMainFrameCLS;

#include "../PANELS/clsDataLoader.h"

#include "../COMMON/CommonDefinitions.h"

#include "stdio.h"
#include <fstream>

#include <omp.h>

PDataBlock_cls::
PDataBlock_cls(PDataSnap_cls *pCParentSnap){
    mpCParentSnap = pCParentSnap;
    mPNumBlock = _PNUM_UNDEFINED;
    mbAllocated = FALSE;
    mInterpolateType = _INTERPOLATE_HERMITE;
    mOrder           = _DATABLOCK_ORDER_3;
    mVariablePNum    = _INTERPOLATE_PNUM_CONSTANT;
    mbUseID = TRUE;

    mpA = NULL;
    mpJ = NULL;
    mpID = NULL;
}

PDataBlock_cls::
~PDataBlock_cls(){
    ReleaseData();
}

void PDataBlock_cls::
Allocate(int pNum){
    if (mbAllocated == TRUE) ReleaseData();

    mPNumBlock = pNum;
    printf("Allocate Block Num=%d Index=%d\n", mPNumBlock, mpCParentSnap->mSelfIndex);
    
    if (pNum == 0) return;

    mpX = new float [mPNumBlock][3];
    mpV = new float [mPNumBlock][3];

    if (mOrder == _DATABLOCK_ORDER_3){
	printf("new AJ\n");
	mpA = new float [mPNumBlock][3];
	mpJ = new float [mPNumBlock][3];
    }

    mpRadius  = new float [mPNumBlock];
    mpDRadius = new float [mPNumBlock];
    mpType    = new unsigned char [mPNumBlock];
    mpValue1  = new float [mPNumBlock];
    mpDValue1 = new float [mPNumBlock];

    if (mbUseID == TRUE){
	mpID = new int [mPNumBlock];
    }

    mpCDataPointers = new BasicDataPointers_cls
	(mpX, mpV, mpA, mpJ, mpRadius, mpDRadius, 
	 mpValue1, mpDValue1, mpType, mpID);
    mbAllocated = TRUE;
    printf("Allocation End\n");
}

void PDataBlock_cls::
ReleaseData(){
    if (mbAllocated == FALSE){
	//printf("Release failed\n");
	return;
    }
    else
	printf("Release Block\n");
    
    delete[] mpX;
    delete[] mpV;

    if (mpA != NULL){
	delete[] mpA; mpA = NULL; printf("delete A\n"); }
    if (mpJ != NULL){
	delete[] mpJ; mpJ = NULL; printf("delete J\n"); }

    delete[] mpRadius;
    delete[] mpDRadius;
    delete[] mpType;
    delete[] mpValue1;
    delete[] mpDValue1;

    if (mpID != NULL){
	delete[] mpID; mpID = NULL; }
    
    delete mpCDataPointers;
    mbAllocated = FALSE;
    mPNumBlock = _PNUM_UNDEFINED;
}

int PDataBlock_cls::
GetUsedMemoryKB(){
    if (mPNumBlock == _PNUM_UNDEFINED) return 0;
    long long memKB = (long long)mPNumBlock;

    int factor = 12 + 12 + 4 + 4 + 4 + 4 + 4;
    if (mpA != NULL) factor += 4;
    if (mpJ != NULL) factor += 4;
    if (mpID != NULL) factor += 4;

    memKB *= factor;
    
    memKB /= 1024;
    return (int)memKB;
}

PDataBlock_cls* PDataBlock_cls::
GetPCLinkedBlock(int snap){
    PDataBlock_cls *pCDataBlock = mpCParentSnap->mVpCSnapsInFile[snap]->mpCPDataBlock[0];
    return pCDataBlock;
    //return mpCParentSnap->mVpCSnapsInFile[snap]->mpCPDataBlock[0];
}

void PDataBlock_cls::
LoadData(wxString strFilename){
    if (mPNumBlock == _PNUM_UNDEFINED){
	if (mspCDataLoaderLink->IsItemExistHead(_GRAMMER_TIME) == FALSE)
	    mpCParentSnap->mTime = mpCParentSnap->mSelfIndex;

	int dataType = mspCDataLoaderLink->mCLoadInfo.mDataType;
	//mCDataType.mBody;
	switch(dataType){
	case _DATA_SIMPLEBINARY:
	    LoadHeaderBinary(strFilename);
	    break;
	case _DATA_SIMPLEASCII:
	    LoadHeaderAscii(strFilename);
	    break;
	case _DATA_HDF5_SIMPLE:
	case _DATA_HDF5_TOKI:
	    LoadHeaderHDF5(strFilename);
	    break;
	}

	int interpolateType = mspCDataLoaderLink->mCLoadInfo.mInterpType;//mCInterpType.mBody;
	if (dataType == _DATA_HDF5_TOKI)
	    interpolateType = _INTERPOLATE_NO;
	
	int snapNum = mpCParentSnap->mVpCSnapsInFile.size();
	for (int snap = 0; snap < snapNum; snap++){
	    PDataBlock_cls *pC = GetPCLinkedBlock(snap);
	    pC->mInterpolateType = interpolateType;
	    pC->mVariablePNum    = mspCDataLoaderLink->mCLoadInfo.mVariablePNum;//mCVariablePNum.mBody;
	    pC->mSearchLimit     = mspCDataLoaderLink->mCLoadInfo.mSearchLimit;//mCSearchLimit.mBody;
	    pC->mOrder = _DATABLOCK_ORDER_1;
	    if (pC->mInterpolateType == _INTERPOLATE_HERMITE ||
		dataType == _DATA_HDF5_TOKI)
		pC->mOrder = _DATABLOCK_ORDER_3;
	    if (pC->mInterpolateType == _INTERPOLATE_NO &&
		dataType == _DATA_HDF5_SIMPLE)
		if (mspCDataLoaderLink->GetHDF5Order() == 3)
		    pC->mOrder = _DATABLOCK_ORDER_3;
	}
    }

    printf("Start Loading Data And Allocate %d \n", mpCParentSnap->mSelfIndex);
    
    // _HERE_ ReleaseMemoryIfNessesary!!
    //if (mbAllocated == FALSE) Allocate(mPNumBlock);
    
    mpCParentSnap->AllocateSnapsInFile();
    StartLoadingData(strFilename);
}

void PDataBlock_cls::
LoadHeaderBinary(wxString strFilename){
    std::fstream ifstream(wxFNCONV(strFilename),
			  std::ios::in | std::ios::binary);

    if (ifstream.fail()){
	printf("Read Header Fails!!\n");
	exit(0);
    }
    
    float time;
    int number;
    float dummy[10];

    int itemNumber = mspCDataLoaderLink->mCGrammerHead.mDimension;
    for (int i = 0; i < itemNumber; i++){
	switch(mspCDataLoaderLink->mCGrammerHead.mpBody[i]){
	case _GRAMMER_TIME:
	    ifstream.read((char*)&time, sizeof(float));
	    mpCParentSnap->mTime = time;
	    printf("time=%f\n", time);
	    break;
	case _GRAMMER_NUMBER:
	    ifstream.read((char*)&number, sizeof(int));
	    mPNumBlock = number;
	    printf("number=%d\n", number);
	    break;
	case _GRAMMER_DUMMY1:
	    ifstream.read((char*)dummy, sizeof(float));
	    break;
	case _GRAMMER_DUMMY3:
	    ifstream.read((char*)dummy, sizeof(float) * 3);
	    break;
	case _GRAMMER_DUMMY10:
	    ifstream.read((char*)dummy, sizeof(float) * 10);
	    break;
	}
    }
    
    int numLimit = mspCDataLoaderLink->mCLoadInfo.mPNumLimit;//mCPNumLimit.mBody;
    if (numLimit != 0 && numLimit < number)
	mPNumBlock = numLimit;
}

#define MAXLINE 255

void PDataBlock_cls::
LoadHeaderAscii(wxString strFilename){
    FILE *fp;
    
    fp = fopen(strFilename.mb_str(), "rt");
    
    if (fp == NULL){
	printf("Load Header Ascii Error\n");
	exit(0);
    }
    
    char buf[MAXLINE+1];
    int itemNumber = mspCDataLoaderLink->mCGrammerHead.mDimension;
    for (int i = 0; i < itemNumber; i++){
	if (fgets(buf, MAXLINE, fp) == NULL){
	    //ErrorProcess!! ;
	}
	switch(mspCDataLoaderLink->mCGrammerHead.mpBody[i]){
	case _GRAMMER_TIME:
	    mpCParentSnap->mTime = atof(buf);
	    break;
	case _GRAMMER_NUMBER:
	    mPNumBlock = atoi(buf);
	    break;
	case _GRAMMER_DUMMY1:
	    break;
	case _GRAMMER_DUMMY3:
	    break;
	case _GRAMMER_DUMMY10:
	    break;
	}
    }
    
    fclose(fp);
    
    int number = mPNumBlock;
    int numLimit = mspCDataLoaderLink->mCLoadInfo.mPNumLimit;//mCPNumLimit.mBody;
    if (numLimit != 0 && numLimit < number)
	mPNumBlock = numLimit;
}

void PDataBlock_cls::
LoadHeaderHDF5(wxString strFilename){
    ReadHDF5Data_cls CRead;
    GrammerHDF5_cls CGrammerHDF5;
    CRead.ReadHDF5Data(strFilename);
    if (CRead.CheckGrammer(CGrammerHDF5) == FALSE){
	mPNumBlock = 0;
	mpCParentSnap->mTime = 0;
	CRead.ShowErrorMessage();
	return;
    }

    int snapNum = mpCParentSnap->mVpCSnapsInFile.size();
    for (int snap = 0; snap < snapNum; snap++){
	GetPCLinkedBlock(snap)->mPNumBlock = CGrammerHDF5.mVPNum[snap];
	GetPCLinkedBlock(snap)->mpCParentSnap->mTime = CGrammerHDF5.mVTime[snap];
    }
}

void PDataBlock_cls::
StartLoadingData(wxString strFilename){
    if (ThreadFileReader_cls::IsReadThreadIdleAny() == FALSE) return;

    int threadIndex = ThreadFileReader_cls::GetIdleThreadIndex();
    
    GAThreadReaderIndex[threadIndex] = mpCParentSnap->mSelfIndex;
    GAbThreadReaderFinished[threadIndex] = FALSE;
    GApCThreadReader[threadIndex] = new ThreadFileReader_cls(wxTHREAD_DETACHED);
    GApCThreadReader[threadIndex]->
	SetParameters(threadIndex, //&mpCParentSnap->mbLoadFinished,
		      mspCDataLoaderLink, this,
		      strFilename, mPNumBlock, mbUseID);
    GApCThreadReader[threadIndex]->Run();
}

void PDataBlock_cls::
InterpolateData(PDataBlock_cls *pCNext){
    if (GpCThreadInterpolate != NULL) return;

    float dtime = pCNext->mpCParentSnap->mTime - mpCParentSnap->mTime;
    if (dtime == 0) dtime = 1;
    
    GbThreadInterpolateFinished = FALSE;
    GpCThreadInterpolate = new ThreadInterpolator_cls(wxTHREAD_DETACHED);
    GpCThreadInterpolate->SetParameters
	(&mpCParentSnap->mbInterpolateFinishd, 
	 mspCDataLoaderLink,
	 mPNumBlock, pCNext->mPNumBlock, dtime,
	 mInterpolateType, mVariablePNum, mSearchLimit,
	 mpCDataPointers,
	 pCNext->mpCDataPointers);
    printf("Start Interpolate between %d and %d InterpolateType = %d\n", mpCParentSnap->mSelfIndex, pCNext->mpCParentSnap->mSelfIndex, mInterpolateType);
    GpCThreadInterpolate->Run();
}

void PDataBlock_cls::
Special_Copy(PDataBlock_cls *pCFrom){
    for (int i = 0; i < mPNumBlock; i++){
	mpRadius[i] = pCFrom->mpRadius[i];
	mpType[i]   = pCFrom->mpType[i];
	mpValue1[i] = pCFrom->mpValue1[i];
    }
}

// ----

DataLoader_cls *PDataBlock_cls::mspCDataLoaderLink = NULL;
