
#include "../HDF5/clsGrammerHDF5.h"
#include "../HDF5/clsReadHDF5Data.h"

#include "clsParticleDataSequence.h"
#include "clsParticleDataSnap.h"
#include "clsSnapData.h"

#include "../COMMON/clsFilename.h"


PDataSequence_cls::
PDataSequence_cls(int index, DrawOption_cls *pCDrawOptionLink){
    mSelfIndex = index;
    mSnapNumInFile = 1;
    
    for (int i = 0; i < _MAX_PDATA_INDEX; i++){
	mpCPDataSnap[i] = new PDataSnap_cls(this, i);
    }
    
    for (int i = 0; i < _MAX_PDATA_INDEX - 1; i++){
	mpCPDataSnap[i]->SetPNextSnap(mpCPDataSnap[i+1]);
	mpCPDataSnap[i+1]->SetPPrevSnap(mpCPDataSnap[i]);
    }

    mpCGLDrawData = new GLDrawData_cls;
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	mpCGLAttribute[i] = new GLAttribute_cls(this, i, pCDrawOptionLink);
    
    wxString strTag = wxT("DataSequence"); strTag << mSelfIndex << wxT("=");
    wxString strSelfIndex; strSelfIndex << mSelfIndex;
    mCConfSelfIndex.SetItem(strTag, strSelfIndex);
}

PDataSequence_cls::
~PDataSequence_cls(){
    for (int i = 0; i < _MAX_PDATA_INDEX; i++)
	delete mpCPDataSnap[i];
    delete mpCGLDrawData;
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	delete mpCGLAttribute[i];
}

void PDataSequence_cls::
InstantiateGLAttribGUI(wxWindow *pCParent){
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	mpCGLAttribute[i]->InstantiateGUI(pCParent);
}

void PDataSequence_cls::
RefreshGLAttribTex(){
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	mpCGLAttribute[i]->RefreshTex();
}

void PDataSequence_cls::
Clear(int index, bool bClearFileInfo){
    if (index == _SEQUENCE_ALL_INDEX){
	if (bClearFileInfo == TRUE)
	    mCFileSequence.Set(wxT(""));
	for (int i = 0; i < _MAX_PDATA_INDEX; i++)
	    mpCPDataSnap[i]->Clear(bClearFileInfo);
    }
    else
	mpCPDataSnap[index]->Clear(bClearFileInfo);
}

void PDataSequence_cls::
ReleaseMemoryOutofStep(int animStep){
    // FIX_ME!!
    // TimeRemap邽߂ɁA
    // {́AanimStepԊuFrameŕKvIndex𓾂
    // sKvIndexReleaseȂƂȂ
    for (int index = 0; index < mCFileSequence.mSeqLength; index++){
	if (index % animStep != 0 && index % animStep != 1 &&
	    mpCPDataSnap[index]->mbLoadStarted == TRUE &&
	    mpCPDataSnap[index]->mbLoadFinished == TRUE)
	    mpCPDataSnap[index]->ReleaseData();
    }
}

int  PDataSequence_cls::
SearchLowNeedStepLarge(int currentIndex, bool bSnipped){
    int candidateL = -1;
    for (int index = mCFileSequence.mSeqLength - 1; index > currentIndex; index--)
	if (mpCPDataSnap[index]->mbLoadStarted == TRUE &&
	    mpCPDataSnap[index]->mbLoadFinished == TRUE &&
	    mpCPDataSnap[index]->mbSnipping == bSnipped){
	    candidateL = index;
	    break;
	}
    return candidateL;
}

int  PDataSequence_cls::
SearchLowNeedStepSmall(int currentIndex, bool bSnipped){
    int candidateS = -1;
    for (int index = 0; index < currentIndex; index++)
	if (mpCPDataSnap[index]->mbLoadStarted == TRUE &&
	    mpCPDataSnap[index]->mbLoadFinished == TRUE &&
	    mpCPDataSnap[index]->mbSnipping == bSnipped){
	    candidateS = index;
	    break;
	}
    return candidateS;
}

bool PDataSequence_cls::
ReleaseMemoryAStep(int currentIndex, int animStep){
    int diffL0 = 0; // L[̂Ȃ폜
    int candidateL0 = SearchLowNeedStepLarge(currentIndex, FALSE);
    if (candidateL0 != -1)
	diffL0 = candidateL0 - currentIndex;

    int diffS0 = 0; // L[̂Ȃ폜
    int candidateS0 = SearchLowNeedStepSmall(currentIndex, FALSE);
    if (candidateS0 != -1)
	diffS0 = currentIndex - candidateS0;

    int diffL1 = 0; // L[̂폜
    int candidateL1 = SearchLowNeedStepLarge(currentIndex, TRUE);
    if (candidateL1 != -1)
	diffL1 = candidateL1 - currentIndex;

    int diffS1 = 0; // L[̂폜
    int candidateS1 = SearchLowNeedStepSmall(currentIndex, TRUE);
    if (candidateS1 != -1)
	diffS1 = currentIndex - candidateS1;

    bool bRelease = FALSE;
    if (diffL0 > diffS0 * 2 && diffL0 > 3 * animStep && diffL0 * 2 > diffL1){
	mpCPDataSnap[candidateL0]->ReleaseData(); bRelease = TRUE;
    }
    
    if (bRelease == FALSE && candidateS0 != -1 && diffS0 * 2 > diffS1){
	mpCPDataSnap[candidateS0]->ReleaseData(); bRelease = TRUE;
    }

    if (bRelease == FALSE && diffL1 > diffS1 * 2 && diffL1 > 3 * animStep){
	mpCPDataSnap[candidateL1]->ReleaseData(); bRelease = TRUE;
    }

    if (bRelease == FALSE && candidateS1 != -1){
	mpCPDataSnap[candidateS1]->ReleaseData(); bRelease = TRUE;
    }

    return bRelease;
}

void PDataSequence_cls::
SetIndividualFilenames(){
    int index = 0;
    for (int offset = 0; offset < mCFileSequence.mSeqLength; offset++){
	mpCPDataSnap[index]->SetFilePath(mCFileSequence.GetModifiedStr(offset));
	mpCPDataSnap[index]->mVpCSnapsInFile.clear();
	mpCPDataSnap[index]->mVpCSnapsInFile.push_back(mpCPDataSnap[index]);
	index++;
    }
    mSnapNumInFile = 1;
}

void PDataSequence_cls::
SetIndividualFilenamesHDF5(){
    int index = 0;

    wxString strFirstFile = mCFileSequence.GetModifiedStr(0);

    ReadHDF5Data_cls CRead;
    GrammerHDF5_cls CGrammerHDF5;
    CRead.ReadHDF5Data(strFirstFile);
    if (CRead.CheckGrammer(CGrammerHDF5) == FALSE){
	return;
    }
    mSnapNumInFile = (int)CGrammerHDF5.mVTime.size();
    
    for (int offset = 0; offset < mCFileSequence.mSeqLength; offset++){
	for (int snap = 0; snap < mSnapNumInFile; snap++){
	    mpCPDataSnap[index]->SetFilePath(mCFileSequence.GetModifiedStr(offset));
	    mpCPDataSnap[index]->mVpCSnapsInFile.clear();
	    for (int j = 0; j < mSnapNumInFile; j++)
		mpCPDataSnap[index]->mVpCSnapsInFile.push_back(mpCPDataSnap[offset*mSnapNumInFile + j]);
	    index++;
	}
    }
}

int PDataSequence_cls::
OpenDataFile(wxString strFilePathInit, bool bSequential){
    mCFileSequence.Set(strFilePathInit);
    if (bSequential == TRUE)
	mCFileSequence.CalcMaxOffset();
    else
	mCFileSequence.mSeqLength = 1;

    SplitPath_cls CSplit(strFilePathInit);
    if (CSplit.IsHDF5File() == TRUE)
	SetIndividualFilenamesHDF5();
    else
	SetIndividualFilenames();
    return mSnapNumInFile;
}

int PDataSequence_cls::
ClampIndexWithinRange(int index){
    int max = mCFileSequence.mSeqLength * mSnapNumInFile;
    if (max > 0 && index >= max) index = max - 1;
    return index;
}

wxString PDataSequence_cls::
GetFilename(int index){
    index = ClampIndexWithinRange(index);
    /*
    return mCFileSequence.GetModifiedStr(index);
    */
    return mpCPDataSnap[index]->mstrFilePath;
}

int      PDataSequence_cls::
GetFilenameStatus(int index){
    index = ClampIndexWithinRange(index);
    return mpCPDataSnap[index]->mFilenameStatus;
}

int      PDataSequence_cls::
GetParticleNumber(int index){
    index = ClampIndexWithinRange(index);
    return mpCPDataSnap[index]->GetParticleNumber();
}

double   PDataSequence_cls::
GetTime(int index){
    index = ClampIndexWithinRange(index);
    return mpCPDataSnap[index]->mTime;
}

void PDataSequence_cls::
GetTimeInfo(double &time, double &dT, int index, double interpStepDouble){
    time = GetTime(index);
    dT = GetTime(index+1) - GetTime(index);
    dT *= interpStepDouble;
}

/*
double   PDataSequence_cls::
GetDT(int index, double interpStepDouble){
    double dT = GetTime(index+1) - GetTime(index);
    return dT * interpStepDouble;
}
*/

int PDataSequence_cls::
GetSeqLength(){
    return mCFileSequence.mSeqLength;
}

void PDataSequence_cls::
ClearSnipping(){
    for (int index = 0; index < mCFileSequence.mSeqLength; index++)
	mpCPDataSnap[index]->mbSnipping = FALSE;
}

void PDataSequence_cls::
SetSnipping(int index){
    index = ClampIndexWithinRange(index);
    mpCPDataSnap[index]->mbSnipping = TRUE;
}

void PDataSequence_cls::
PrepareIndex(int index, double indexmod, bool bUseBlur, bool bWaitFinish){
    index = ClampIndexWithinRange(index);
    bool bUseInterpolate = (indexmod != 0);
    bUseInterpolate |= bUseBlur;
    bUseInterpolate |= SnapData_cls::IsInterpolateRequired(mpCGLAttribute);
    mpCPDataSnap[index]->PrepareData(bUseInterpolate, bWaitFinish);
}

int PDataSequence_cls::
GetUsedTotalMemoryMB(){
    int totalMemoryKB = 0;
    for (int index = 0; index < mCFileSequence.mSeqLength; index++)
	totalMemoryKB += mpCPDataSnap[index]->GetUsedMemoryKB();
    return totalMemoryKB/1024;
}

int PDataSequence_cls::
GetLoadedFrameNumber(){
    int totalFrameNumber = 0;
    for (int index = 0; index < mCFileSequence.mSeqLength; index++)
	if (mpCPDataSnap[index]->mbLoadFinished == TRUE)
	    totalFrameNumber++;
    return totalFrameNumber;
}

bool PDataSequence_cls::
IsTheIndexReady(int index, double indexmod, bool bUseBlur){
    index = ClampIndexWithinRange(index);
    bool bUseInterpolate = (indexmod != 0);
    bUseInterpolate |= bUseBlur;
    bUseInterpolate |= SnapData_cls::IsInterpolateRequired(mpCGLAttribute);
    return mpCPDataSnap[index]->IsItReady(bUseInterpolate);
}

bool PDataSequence_cls::
IsTheIndexReading(int index){
    index = ClampIndexWithinRange(index);
    return mpCPDataSnap[index]->IsItReading();
}

bool PDataSequence_cls::
IsTheIndexLoaded(int index){
    index = ClampIndexWithinRange(index);
    return mpCPDataSnap[index]->IsItLoaded();
}

void PDataSequence_cls::
Special_Copy(int indexA, int indexB){
    if (IsTheIndexLoaded(indexA) == FALSE) return;
    if (IsTheIndexLoaded(indexB) == FALSE) return;
    int numA = mpCPDataSnap[indexA]->GetParticleNumber();
    int numB = mpCPDataSnap[indexB]->GetParticleNumber();
    if (numA != numB) return;

    mpCPDataSnap[indexA]->Special_Copy(mpCPDataSnap[indexB]);
}


// -----

void PDataSequence_cls::
SetCameraInfo(PositionAndDirection_cls CCameraPos){
    mpCGLDrawData->SetCameraPos(CCameraPos.mPos);
    mpCGLDrawData->SetCameraDir(CCameraPos.mTgt);
}

void PDataSequence_cls::
GLDrawData(int fboIndex, int index, double dT, 
	   int lightNumber, float *pLightPos, 
	   double lightPhi, double lightTheta,
	   bool bShadowVolume, int mSelfIndex, int sortType){
    index = ClampIndexWithinRange(index);
    mpCGLDrawData->mbShadowVolume = bShadowVolume;

    mpCGLDrawData->mbMainFrame = TRUE;
    if (mSelfIndex == 4) mpCGLDrawData->mbMainFrame = FALSE;
    
    mpCGLDrawData->SetLightSourceInfo(lightNumber, pLightPos, lightPhi, lightTheta);
    mpCGLDrawData->mSortType = sortType;
    mpCPDataSnap[index]->GLDrawData
	(fboIndex, mpCGLDrawData, mpCGLAttribute, dT); 
}

void PDataSequence_cls::
ExportData(DataExporter_cls *pCDataExporter, int index, int downsize){
    index = ClampIndexWithinRange(index);
    mpCPDataSnap[index]->ExportData(pCDataExporter, downsize);
}

bool PDataSequence_cls::
GetPovDataRef(wxString &strPData, PovRayCtrl_cls *pCPovRayCtrl,
	      int index, double dT){
    index = ClampIndexWithinRange(index);
    return mpCPDataSnap[index]->GetPovDataRef(strPData, pCPovRayCtrl, mpCGLAttribute, dT);
}

wxString PDataSequence_cls::
SerializeFile(){
    wxString strOut;
    strOut << mCConfSelfIndex.Serialize();
    strOut << mCFileSequence.mstrPath << ENDL;
    return strOut;
}

wxString PDataSequence_cls::
Serialize(){
    wxString strOut;

    strOut << SerializeFile();

    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	strOut << mpCGLAttribute[i]->Serialize();
    }
    
    return strOut;
}

void PDataSequence_cls::
ReadStrLine(std::wistream &istream, wxString strLine, bool bSequential){
    if (mCConfSelfIndex.ReadStrLine(strLine) == TRUE){
	std::wstring wstr;
	std::getline (istream, wstr);
	strLine = wstr.c_str();
	if (strLine != wxT("")) OpenDataFile(strLine, bSequential);
    }
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	mpCGLAttribute[i]->ReadStrLine(istream, strLine);
    }
}
 
