// (setq c-basic-offset 4)

#include "clsGLDrawData.h"
#include "clsParticleDataSnap.h"
#include "clsParticleDataBlock.h"

#include "../COMMON/CommonDefinitionsForH.h"

#include "clsThreadFileReader.h"
extern ThreadFileReader_cls   *GApCThreadReader[_MAX_READ_THREAD];
extern bool                   GAbThreadReaderFinished[_MAX_READ_THREAD];
extern int                    GAThreadReaderIndex[_MAX_READ_THREAD];
#include "INTERPOLATE_GET/clsThreadInterpolate.h"
extern ThreadInterpolator_cls *GpCThreadInterpolate;
extern bool                   GbThreadInterpolateFinished;

#include "../PANELS/clsDataExporter.h"
#include "../POVRAY/clsPovRay.h"

#include "../../wxWidgetEnumMenu.h"
#include "../../clsFrameMain.h"

extern bool GbGLLockForLoadingData;
extern bool GbGLLockForInterplateData;
extern bool GbQuitApplication;

extern wxFrame  *GpCMainFrameCLS;

#include "fstream"

PDataSnap_cls::
PDataSnap_cls(PDataSequence_cls *pCParentSequence, int index){
    mpCParentSequence = pCParentSequence;
    mSelfIndex = index;
    mTime = -9999;
    mbSnipping = FALSE;
    for (int i = 0; i < _MAX_PDATA_BLOCK; i++)
	mpCPDataBlock[i] = new PDataBlock_cls(this);
    mFilenameStatus = _FN_STATUS_UNDEF;
    mbLoadStarted  = FALSE;
    mbLoadFinished = FALSE;
    mbInterpolateStarted = FALSE;
    mbInterpolateFinishd = FALSE;
    mpCNextPDSnapLink = NULL;
    mpCPrevPDSnapLink = NULL;
}

PDataSnap_cls::
~PDataSnap_cls(){
    for (int i = 0; i < _MAX_PDATA_BLOCK; i++)
	delete mpCPDataBlock[i];
}

void PDataSnap_cls::
SetPNextSnap(PDataSnap_cls *pCPDataSnap){
    mpCNextPDSnapLink = pCPDataSnap;
}

void PDataSnap_cls::
SetPPrevSnap(PDataSnap_cls *pCPDataSnap){
    mpCPrevPDSnapLink = pCPDataSnap;
}

void PDataSnap_cls::
Clear(bool bClearFileInfo){
    ReleaseData();
    if (bClearFileInfo == FALSE) return;
    mTime = -9999;
    mstrFilePath = wxT("");
    mFilenameStatus = _FN_STATUS_UNDEF;
}

void PDataSnap_cls::
ReleaseData(){
    WaitThreadReading(this, TRUE); // ̃t[̓ǂݍݒҋ@
    for (int i = 0; i < _MAX_PDATA_BLOCK; i++)
	mpCPDataBlock[i]->ReleaseData();
    mbLoadStarted  = FALSE;
    mbLoadFinished = FALSE;
    mbInterpolateStarted = FALSE;
    mbInterpolateFinishd = FALSE;
}

void PDataSnap_cls::
SetFilePath(wxString strFilePath){
    mstrFilePath = strFilePath;
    mFilenameStatus = _FN_STATUS_UNDEF;
}

int  PDataSnap_cls::
GetParticleNumber(){
    int number = 0;
    for (int i = 0; i < _MAX_PDATA_BLOCK; i++)
	if (mpCPDataBlock[i]->mPNumBlock != _PNUM_UNDEFINED)
	    number += mpCPDataBlock[i]->mPNumBlock;
    return number;
}

bool PDataSnap_cls::
IsFileValid(){
    if (mstrFilePath == wxT("")) return FALSE;
    if (mFilenameStatus == _FN_STATUS_INVALID) return FALSE;
    if (mFilenameStatus == _FN_STATUS_VALID) return TRUE;

    std::fstream ifstream(wxFNCONV(mstrFilePath),
			  std::ios::in | std::ios::binary);
    if (ifstream.fail() == TRUE){
	mFilenameStatus = _FN_STATUS_INVALID; return FALSE;
    }
    mFilenameStatus = _FN_STATUS_VALID; return TRUE;
}

void PDataSnap_cls::
WaitThreadReading(PDataSnap_cls *pCSnap, bool bPreparing){
    
    if (ThreadFileReader_cls::IsReadThreadIdleAll() == FALSE &&
	pCSnap->IsPreparing(FALSE) == bPreparing){
	if (GbGLLockForLoadingData == TRUE){
	    printf("waitnig??\n");
	    return;
	}
	
	int threadIndex = 
	    ThreadFileReader_cls::GetThreadIndexReading(mSelfIndex);
	if (threadIndex == -1){
	    printf("not reading %d??\n", mSelfIndex);
	    return;
	}

	GbGLLockForLoadingData = TRUE;
	printf("Lock!!\n");
	int currentIndex = -1;
	int waitLoop = 0;

	// !! ʂ悤Ɍ
	if (GAbThreadReaderFinished[threadIndex] == FALSE)
	    currentIndex = GAThreadReaderIndex[threadIndex];
	
	while (GAbThreadReaderFinished[threadIndex] == FALSE && 
	       GAThreadReaderIndex[threadIndex] == currentIndex){
	    wxMilliSleep(5);
	    waitLoop++;
	    if (waitLoop % 100 == 0){
		//printf("%d Yield\n", waitLoop);
		wxString strLoop;
		strLoop << waitLoop;
		((FrameMain_cls*)GpCMainFrameCLS)->UpdateStatusText();
	    }
	}
	GApCThreadReader[threadIndex] = NULL;
	GbGLLockForLoadingData = FALSE;
	printf("Set::GbGLLockForLoadingData = FALSE\n");
    }
}
 
void PDataSnap_cls::
WaitThreadInterpolate(PDataSnap_cls *pCSnap, bool bPreparing){
    if (GpCThreadInterpolate != NULL &&
	pCSnap->IsPreparing(TRUE) == bPreparing){
	GbGLLockForInterplateData = TRUE;
	while (GbThreadInterpolateFinished == FALSE)
	    wxMilliSleep(3);
	GpCThreadInterpolate = NULL;
	GbGLLockForInterplateData = FALSE;
    }
}

void PDataSnap_cls::
PrepareData(bool bUseInterpolate, bool bWaitFinish){
    if (mbLoadFinished == TRUE && mbInterpolateFinishd == TRUE) return;
    if (mbLoadFinished == FALSE && mbLoadStarted == FALSE){
	printf("Prepare Data -> LoadData %d\n", mSelfIndex);
	LoadData(bWaitFinish);
    }
    
    if (bWaitFinish == TRUE)
	WaitThreadReading(this, TRUE);	// ̃t[̓ǂݍݒҋ@

    if (bUseInterpolate == FALSE) return;

    if (mbInterpolateStarted == FALSE && mbInterpolateFinishd == FALSE){
	InterpolateData(bWaitFinish);
    }

    if (bWaitFinish == TRUE)
	WaitThreadInterpolate(this, TRUE); // ̃t[̕Ԓҋ@
}

void PDataSnap_cls::
AllocateSnapsInFile(){
    int snapNum = mVpCSnapsInFile.size();
    for (int i = 0; i < snapNum; i++){
	PDataBlock_cls *pCPDataBlock = mVpCSnapsInFile[i]->mpCPDataBlock[0];
	pCPDataBlock->Allocate(pCPDataBlock->mPNumBlock);
    }
}

void PDataSnap_cls::
MarkLoadingStateStart(){
    int snapNum = mVpCSnapsInFile.size();
    for (int i = 0; i < snapNum; i++)
	mVpCSnapsInFile[i]->mbLoadStarted = TRUE;
}

void PDataSnap_cls::
MarkLoadingStateFinished(){
    int snapNum = mVpCSnapsInFile.size();
    for (int i = 0; i < snapNum; i++)
	mVpCSnapsInFile[i]->mbLoadFinished = TRUE;
}

void PDataSnap_cls::
LoadData(bool bWaitFinish){

    // I͂Ȃ
    if (GbQuitApplication == TRUE) return;

    /*
    WaitThreadReading(this, FALSE); // ʃt[̓ǂݍݒҋ@
    */
    /*
    while (ThreadFileReader_cls::IsReadThreadIdleAny() == FALSE){
	wxMilliSleep(5);
	printf("Waiting Idle Thread\n");
    }
    */
    
    if (IsFileValid() == TRUE){
	if (bWaitFinish == FALSE &&
	    ThreadFileReader_cls::IsReadThreadIdleAny() == FALSE) return;
	
	while (ThreadFileReader_cls::IsReadThreadIdleAny() == FALSE){
	    wxMilliSleep(5);
	    printf("Waiting Idle Thread\n");
	}
	
	MarkLoadingStateStart();
	printf("Begin Load Data %d\n", mSelfIndex);
	mpCPDataBlock[0]->LoadData(mstrFilePath);
    }
}

void PDataSnap_cls::
InterpolateData(bool bWaitFinish){
    if (mbLoadFinished == FALSE) return; // ̃pXł͖߂Ȃ͂

    if (mpCNextPDSnapLink->mbLoadFinished == FALSE) {
	if (mpCNextPDSnapLink->mbLoadStarted == FALSE){
	    printf("NextStep->LoadData\n");
	    mpCNextPDSnapLink->LoadData(TRUE);
	}
	else{
	    printf("NextStep->LoadData Skip\n");
	}

	if (bWaitFinish == FALSE) return;

	// t[ǂݍݒҋ@
	while(mpCNextPDSnapLink->IsItLoaded() == FALSE)
	    wxMilliSleep(5);
	//WaitThreadReading(mpCNextPDSnapLink, TRUE); 
    }

    WaitThreadInterpolate(this, FALSE); //ʃt[̕Ԓҋ@
    
    // ̃t@C_Ȃ̂͏ĨXebv̂͂
    if (mpCNextPDSnapLink->IsFileValid() == FALSE){
	mbInterpolateStarted = TRUE;
	mpCPDataBlock[0]->InterpolateData(mpCPDataBlock[0]);
	return;
    }

    mbInterpolateStarted = TRUE;
    mpCPDataBlock[0]->InterpolateData(mpCNextPDSnapLink->mpCPDataBlock[0]);

    //WaitThreadInterpolate(this,TRUE); //ʃt[̕Ԓҋ@
}

int PDataSnap_cls::
GetUsedMemoryKB(){
    int usedMemoryKB = 0;
    for (int i = 0; i < _MAX_PDATA_BLOCK; i++){
	usedMemoryKB += mpCPDataBlock[i]->GetUsedMemoryKB();
    }
    return usedMemoryKB;
}

bool PDataSnap_cls::
IsPreparing(bool bUseInterpolate){
    if (bUseInterpolate == FALSE){
	if (mbLoadFinished == TRUE) return FALSE;
	if (mbLoadStarted == TRUE) return TRUE;
	return FALSE;
    }
    if (mbLoadFinished == TRUE && mbInterpolateFinishd == TRUE) return FALSE;
    if (mbLoadFinished == TRUE && mbInterpolateFinishd == FALSE) return TRUE;
    if (mbLoadStarted == TRUE) return TRUE;
    if (mbInterpolateStarted == TRUE) return TRUE;
    return FALSE;
}

bool PDataSnap_cls::
IsItReady(bool bUseInterpolate){
    // `ĂȂΕ\O.K.
    if (IsFileValid() == FALSE) return TRUE;
    if (bUseInterpolate == FALSE)
	return mbLoadStarted & mbLoadFinished;
    return mbLoadStarted & mbLoadFinished & mbInterpolateStarted & mbInterpolateFinishd;
}

bool PDataSnap_cls::
IsItReading(){
    if (mbLoadStarted == TRUE && mbLoadFinished == FALSE) return TRUE;
    return FALSE;
}

bool PDataSnap_cls::
IsItLoaded(){
    return mbLoadStarted & mbLoadFinished;
}

void PDataSnap_cls::
Special_Copy(PDataSnap_cls *pCFrom){
    mpCPDataBlock[0]->Special_Copy(pCFrom->mpCPDataBlock[0]);
}

// -----

void PDataSnap_cls::
GLDrawData(int fboIndex, GLDrawData_cls *pCDrawData, 
	   GLAttribute_cls **ppCGLAttribute,
	   double dT){
    for (int i = 0; i < _MAX_PDATA_BLOCK; i++){
	if (mpCPDataBlock[i]->mPNumBlock == 0 || 
	    mpCPDataBlock[i]->mPNumBlock == _PNUM_UNDEFINED) continue;
	pCDrawData->GLDraw(fboIndex, mpCPDataBlock[i], ppCGLAttribute, dT);
    }
}

void PDataSnap_cls::
ExportData(DataExporter_cls *pCDataExporter, int downsize){
    pCDataExporter->ExportData(mpCPDataBlock[0], downsize, mSelfIndex); //!! _AD_HOC_
}

bool PDataSnap_cls::
GetPovDataRef(wxString &strPData, PovRayCtrl_cls *pCPovRayCtrl, 
	      GLAttribute_cls **ppCGLAttribute, double dT){
    bool bExist = FALSE;
    for (int i = 0; i < _MAX_PDATA_BLOCK; i++){
	if (mpCPDataBlock[i]->mPNumBlock == 0 || 
	    mpCPDataBlock[i]->mPNumBlock == _PNUM_UNDEFINED) continue;
	pCPovRayCtrl->GetPovStringParticle
	    (strPData, mpCPDataBlock[i], 
	     ppCGLAttribute, dT);
	bExist = TRUE;
    }
    return bExist;
}
