
#include "clsSnapData.h"
#include "clsParticleDataBlock.h"
#include "clsParticleDataSnap.h"
#include "clsParticleDataSequence.h"

#include "INTERPOLATE_GET/clsThreadInterpolate.h"

#include "../GLATTRIB/clsColorCurve.h"
#include "../GLATTRIB/clsGLAttribute.h"
#include "../MYAPI/MsgBox.h"

bool GbSnapDataRefresh = FALSE;

#include <omp.h>

ThreadSnapInterpHermite_cls::
ThreadSnapInterpHermite_cls(wxThreadKind kind) :
    wxThread(kind) {
}

void ThreadSnapInterpHermite_cls::
SetParameters(bool *pbFinishedFlag,
	      int PInit, int PFin,
	      int downsizeStep, float DTime,
	      float (*pX)[3], float (*pV)[3],
	      float (*pA)[3], float (*pJ)[3], 
	      float *pRadius, float *pDRadius,
	      float *pValue1, float *pDValue1,
	      float (*pXMem)[3], //float (*pVMem)[3],
	      float *pRadiusMem, 
	      float *pValue1Mem,
	      unsigned char   *pTypeLinkMem,
	      float (*pColorMem)[4], 
	      float *pValue1ForColor[_MAX_PARTICLE_TYPE],
	      ColorCurve_cls* pColorCurveCol[_MAX_PARTICLE_TYPE],
	      int alphaType[_MAX_PARTICLE_TYPE], 
	      float alphaDependBase[_MAX_PARTICLE_TYPE], 
	      float alphaFactor[_MAX_PARTICLE_TYPE]){
    
    mpbFinishedFlagLink = pbFinishedFlag;
    mPIni = PInit;
    mPFin = PFin;
    mDownsizeStep = downsizeStep;
    mDTime = DTime;
    mpX = pX;
    mpV = pV;
    mpA = pA;
    mpJ = pJ;
    mpRadius = pRadius;
    mpDRadius = pDRadius;
    mpValue1 = pValue1;
    mpDValue1 = pDValue1;
    mpTypeLinkMem = pTypeLinkMem;

    mpXMem = pXMem;
    //mpVMem = pVMem;
    mpRadiusMem = pRadiusMem;
    mpValue1Mem = pValue1Mem;

    mpColorMem = pColorMem;
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	mpValue1ForColor[i] = pValue1ForColor[i];
	mpCClrCurveCol[i] = pColorCurveCol[i];
	mAlphaType[i] = alphaType[i];
	mAlphaDependBase[i] = alphaDependBase[i];
	mAlphaFactor[i] = alphaFactor[i];
    }
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

void ThreadSnapInterpHermite_cls::
InterpolateDataHermite(){
    float dT = mDTime;
    
    float (*pX)[3]  = mpX;
    float (*pV)[3]  = mpV;
    float (*pA)[3]  = mpA;
    float (*pJ)[3]  = mpJ;
    float *pRadius  = mpRadius;
    float *pDRadius = mpDRadius;
    float *pValue1  = mpValue1;
    float *pDValue1 = mpDValue1;
    
    float (*pXMem)[3] = mpXMem;
    float *pRadiusMem = mpRadiusMem;
    float *pValue1Mem = mpValue1Mem;
    unsigned char  *pTypeLinkMem = mpTypeLinkMem;
    float (*pColorMem)[4] = mpColorMem;
    float *pValue1ForColor[_MAX_PARTICLE_TYPE];
    ColorCurve_cls *pCClrCurveCol[_MAX_PARTICLE_TYPE];
    int alphaType[_MAX_PARTICLE_TYPE];
    float alphaDependBase[_MAX_PARTICLE_TYPE];
    float alphaFactor[_MAX_PARTICLE_TYPE];
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	pValue1ForColor[i] = mpValue1ForColor[i];
	pCClrCurveCol[i] = mpCClrCurveCol[i];
	alphaType[i] = mAlphaType[i];
	alphaDependBase[i] = mAlphaDependBase[i];
	alphaFactor[i] = mAlphaFactor[i];
    }
    int downsizeStep = mDownsizeStep;
    int PFin = mPFin;

    if (downsizeStep == 1){
	memcpy(pXMem[mPIni], pX[mPIni], 4*3*(mPFin - mPIni)); 
	memcpy(&pRadiusMem[mPIni], &pRadius[mPIni], 4*(mPFin - mPIni));
	memcpy(&pValue1Mem[mPIni], &pValue1[mPIni], 4*(mPFin - mPIni)); 
	for (int i = mPIni; i < PFin; i++){
	    pXMem[i][0] += pV[i][0]*dT + pA[i][0]*dT*dT + pJ[i][0]*dT*dT*dT;
	    pXMem[i][1] += pV[i][1]*dT + pA[i][1]*dT*dT + pJ[i][1]*dT*dT*dT;
	    pXMem[i][2] += pV[i][2]*dT + pA[i][2]*dT*dT + pJ[i][2]*dT*dT*dT;
	    pRadiusMem[i] += pDRadius[i] * dT;
	    pValue1Mem[i] += pDValue1[i] * dT;
	    
	    int particleType = pTypeLinkMem[i];
	    
	    pCClrCurveCol[particleType]->GetColor4fScaled(pColorMem[i], pValue1ForColor[particleType][i]);
	    
	    if (alphaType[particleType] == _GL_ALPHA_INVERSE_SQUARE)
		pColorMem[i][3] *= alphaDependBase[particleType] * alphaDependBase[particleType];
	    else
		pColorMem[i][3] *= alphaFactor[particleType];
	}
	return;
    }
    else{
	int pInit = mPIni / downsizeStep;
	pInit *= downsizeStep;
	if (pInit != mPIni) pInit += downsizeStep;
	for (int i = pInit; i < PFin; i += downsizeStep){
	    pXMem[i][0] = pX[i][0] + pV[i][0] * dT;
	    pXMem[i][1] = pX[i][1] + pV[i][1] * dT;
	    pXMem[i][2] = pX[i][2] + pV[i][2] * dT;
	    pRadiusMem[i] = pRadius[i] + pDRadius[i] * dT;
	    pValue1Mem[i] = pValue1[i] + pDValue1[i] * dT;
	    
	    int particleType = pTypeLinkMem[i];
	    
	    pCClrCurveCol[particleType]->GetColor4fScaled(pColorMem[i], pValue1ForColor[particleType][i]);
	    
	    if (alphaType[particleType] == _GL_ALPHA_INVERSE_SQUARE)
		pColorMem[i][3] *= alphaDependBase[particleType] * alphaDependBase[particleType];
	    else
		pColorMem[i][3] *= alphaFactor[particleType];
	}
	return;
    }
    
}

wxThread::ExitCode ThreadSnapInterpHermite_cls::
Entry(){
    InterpolateDataHermite();
    *mpbFinishedFlagLink = TRUE;
    return 0;
}

// ---------

ThreadSnapInterpLinear_cls::
ThreadSnapInterpLinear_cls(wxThreadKind kind) :
    wxThread(kind) {
}

void ThreadSnapInterpLinear_cls::
SetParameters(bool *pbFinishedFlag,
	      int PInit, int PFin,
	      int downsizeStep, float DTime,
	      float (*pX)[3], float (*pV)[3],
	      //float (*pA)[3], float (*pJ)[3], 
	      float *pRadius, float *pDRadius,
	      float *pValue1, float *pDValue1,
	      float (*pXMem)[3], //float (*pVMem)[3],
	      float *pRadiusMem, 
	      float *pValue1Mem,
	      unsigned char   *pTypeLinkMem,
	      float (*pColorMem)[4], 
	      float *pValue1ForColor[_MAX_PARTICLE_TYPE],
	      ColorCurve_cls* pColorCurveCol[_MAX_PARTICLE_TYPE],
	      int alphaType[_MAX_PARTICLE_TYPE], 
	      float alphaDependBase[_MAX_PARTICLE_TYPE], 
	      float alphaFactor[_MAX_PARTICLE_TYPE]){
    
    mpbFinishedFlagLink = pbFinishedFlag;
    mPIni = PInit;
    mPFin = PFin;
    mDownsizeStep = downsizeStep;
    mDTime = DTime;
    mpX = pX;
    mpV = pV;
    //mpA = pA;
    //mpJ = pJ;
    mpRadius = pRadius;
    mpDRadius = pDRadius;
    mpValue1 = pValue1;
    mpDValue1 = pDValue1;
    mpTypeLinkMem = pTypeLinkMem;

    mpXMem = pXMem;
    //mpVMem = pVMem;
    mpRadiusMem = pRadiusMem;
    mpValue1Mem = pValue1Mem;

    mpColorMem = pColorMem;
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	mpValue1ForColor[i] = pValue1ForColor[i];
	mpCClrCurveCol[i] = pColorCurveCol[i];
	mAlphaType[i] = alphaType[i];
	mAlphaDependBase[i] = alphaDependBase[i];
	mAlphaFactor[i] = alphaFactor[i];
    }
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

void ThreadSnapInterpLinear_cls::
InterpolateDataLinear(){
    float dT = mDTime;
    
    float (*pX)[3]  = mpX;
    float (*pV)[3]  = mpV;
    float *pRadius  = mpRadius;
    float *pDRadius = mpDRadius;
    float *pValue1  = mpValue1;
    float *pDValue1 = mpDValue1;
    
    float (*pXMem)[3] = mpXMem;
    float *pRadiusMem = mpRadiusMem;
    float *pValue1Mem = mpValue1Mem;
    unsigned char  *pTypeLinkMem = mpTypeLinkMem;
    float (*pColorMem)[4] = mpColorMem;
    float *pValue1ForColor[_MAX_PARTICLE_TYPE];
    ColorCurve_cls *pCClrCurveCol[_MAX_PARTICLE_TYPE];
    int alphaType[_MAX_PARTICLE_TYPE];
    float alphaDependBase[_MAX_PARTICLE_TYPE];
    float alphaFactor[_MAX_PARTICLE_TYPE];
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	pValue1ForColor[i] = mpValue1ForColor[i];
	pCClrCurveCol[i] = mpCClrCurveCol[i];
	alphaType[i] = mAlphaType[i];
	alphaDependBase[i] = mAlphaDependBase[i];
	alphaFactor[i] = mAlphaFactor[i];
    }
    int downsizeStep = mDownsizeStep;
    int PFin = mPFin;

    if (downsizeStep == 1){
	memcpy(pXMem[mPIni], pX[mPIni], 4*3*(mPFin - mPIni)); 
	memcpy(&pRadiusMem[mPIni], &pRadius[mPIni], 4*(mPFin - mPIni));
	memcpy(&pValue1Mem[mPIni], &pValue1[mPIni], 4*(mPFin - mPIni)); 
	for (int i = mPIni; i < PFin; i++){
	    pXMem[i][0] += pV[i][0] * dT;
	    pXMem[i][1] += pV[i][1] * dT;
	    pXMem[i][2] += pV[i][2] * dT;
	    pRadiusMem[i] += pDRadius[i] * dT;
	    pValue1Mem[i] += pDValue1[i] * dT;
	    
	    int particleType = pTypeLinkMem[i];
	    
	    pCClrCurveCol[particleType]->GetColor4fScaled(pColorMem[i], pValue1ForColor[particleType][i]);
	    
	    if (alphaType[particleType] == _GL_ALPHA_INVERSE_SQUARE)
		pColorMem[i][3] *= alphaDependBase[particleType] * alphaDependBase[particleType];
	    else
		pColorMem[i][3] *= alphaFactor[particleType];
	}
	return;
    }
    else{
	int pInit = mPIni / downsizeStep;
	pInit *= downsizeStep;
	if (pInit != mPIni) pInit += downsizeStep;
	for (int i = pInit; i < PFin; i += downsizeStep){
	    pXMem[i][0] = pX[i][0] + pV[i][0] * dT;
	    pXMem[i][1] = pX[i][1] + pV[i][1] * dT;
	    pXMem[i][2] = pX[i][2] + pV[i][2] * dT;
	    pRadiusMem[i] = pRadius[i] + pDRadius[i] * dT;
	    pValue1Mem[i] = pValue1[i] + pDValue1[i] * dT;
	    
	    int particleType = pTypeLinkMem[i];
	    
	    pCClrCurveCol[particleType]->GetColor4fScaled(pColorMem[i], pValue1ForColor[particleType][i]);
	    
	    if (alphaType[particleType] == _GL_ALPHA_INVERSE_SQUARE)
		pColorMem[i][3] *= alphaDependBase[particleType] * alphaDependBase[particleType];
	    else
		pColorMem[i][3] *= alphaFactor[particleType];
	}
	return;
    }
    
}

wxThread::ExitCode ThreadSnapInterpLinear_cls::
Entry(){
    InterpolateDataLinear();
    *mpbFinishedFlagLink = TRUE;
    return 0;
}

// -----------------------------------

SnapData_cls::
SnapData_cls(PDataBlock_cls *pCDataBlockLink,
	     GLAttribute_cls **ppCGLAttribute,
	     double dT, int downsizeStep){
    mpCDataBlockLink = pCDataBlockLink;
    mbAllocated  = FALSE;
    mbValue1dash = FALSE;
    mTime = pCDataBlockLink->mpCParentSnap->mTime + dT;
    mDownsizeStep = downsizeStep;
    Allocate(pCDataBlockLink->mPNumBlock);
    //mInterpolateType = pCDataBlockLink->mInterpolateType;
    mOrder = pCDataBlockLink->mOrder;
    Interpolate(ppCGLAttribute, dT);
}

SnapData_cls::
~SnapData_cls(){
    ReleaseData();
}

void SnapData_cls::
Allocate(int pNum){
    mPNumBlock = pNum;
    if (pNum == 0) return;

    //printf("Allocate Snap %p\n", this);

    mpX = new float [mPNumBlock][3];
    mpRadius = new float [mPNumBlock];
    mpValue1 = new float [mPNumBlock];
    mpColor = new float [mPNumBlock][4];
    
    mbAllocated = TRUE;
}

void SnapData_cls::
ReleaseData(){
    if (mbAllocated == FALSE){
	printf("Release Snap failed\n");
	return;
    }
    //printf("Release Snap %p\n", this);
    
    delete[] mpX;
    delete[] mpRadius;
    delete[] mpValue1;
    delete[] mpColor;

    if (mbValue1dash == TRUE){
	delete [] mpValue1dash;
	mbValue1dash = FALSE;
	printf ("delete Value1 dash\n");
    }

    mbAllocated = FALSE;
}

bool SnapData_cls::
IsItSameSnap(PDataBlock_cls *pCDataBlockLink, double dT, int step){
    if (GbSnapDataRefresh == TRUE){
	GbSnapDataRefresh = FALSE;
	printf("GbSnapDataRefresh!!\n");
	return FALSE;
    }
    
    if (mTime != pCDataBlockLink->mpCParentSnap->mTime + dT) return FALSE;
    if (mDownsizeStep != step && mDownsizeStep != 1) return FALSE;

    return TRUE;
}

void SnapData_cls::
SetPValue1(float **ppValue1, GLAttribute_cls *pCGLAttribute, int type){
    int dependType;
    if (type == 0) 
	dependType = pCGLAttribute->mpCClrCurveCol->mCDependType.mBody;
    else
	dependType = pCGLAttribute->mpCClrCurveMat[0]->mCDependType.mBody;
	
    switch(dependType){
    case _DEPEND_ON_VALUE1:
	*ppValue1 = /*pCDataBlockLink->*/mpValue1; break;
    case _DEPEND_ON_RADIUS:
	*ppValue1 = /*pCDataBlockLink->*/mpRadius; break;
    case _DEPEND_ON_VELOCITY:
    case _DEPEND_ON_POSITION:
    case _DEPEND_ON_POS_X: case _DEPEND_ON_POS_Y: case _DEPEND_ON_POS_Z:
    case _DEPEND_ON_V_X:   case _DEPEND_ON_V_Y:   case _DEPEND_ON_V_Z:
    case _DEPEND_ON_ID:
	SetValue1Dash(dependType);
	*ppValue1 = mpValue1dash; break; 
    }
}

void SnapData_cls::
InterpolateNot(PDataBlock_cls *pCDataBlockLink, 
	       GLAttribute_cls **ppCGLAttribute){
    mpTypeLink = pCDataBlockLink->mpType;

    int   alphaType[_MAX_PARTICLE_TYPE];
    float alphaDependBase[_MAX_PARTICLE_TYPE];
    float alphaMax[_MAX_PARTICLE_TYPE];
    float alphaFactor[_MAX_PARTICLE_TYPE];
    ColorCurve_cls *pCClrCurveCol[_MAX_PARTICLE_TYPE];
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	alphaType[i]        = ppCGLAttribute[i]->mCAlphaType.mBody; 
	alphaDependBase[i]  = ppCGLAttribute[i]->mCAlphaDenendBase.mBody;
	alphaMax[i]         = ppCGLAttribute[i]->mCAlphaMax.mBody;
	alphaFactor[i]      = ppCGLAttribute[i]->mCAlphaFactor.mBody;
	ppCGLAttribute[i]->mpCClrCurveCol->PreCalcScaleFactor();
	pCClrCurveCol[i] = ppCGLAttribute[i]->mpCClrCurveCol;
    }

    memcpy(mpX, pCDataBlockLink->mpX, sizeof(float)*3*mPNumBlock);
    memcpy(mpRadius, pCDataBlockLink->mpRadius, sizeof(float)*mPNumBlock);
    memcpy(mpValue1, pCDataBlockLink->mpValue1, sizeof(float)*mPNumBlock);

    float *pValue1[_MAX_PARTICLE_TYPE];
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	SetPValue1(&pValue1[i], ppCGLAttribute[i], 0);

    float (*pColorMem)[4] = mpColor;
    unsigned char *pTypeLinkMem = mpTypeLink;
    
    for (int i = 0; i < mPNumBlock; i++){

	int particleType = pTypeLinkMem[i];

	if (particleType >= _MAX_PARTICLE_TYPE || particleType < 0){
	    particleType = 0;
	}
	
	// F̊{͌vZĂ
	pCClrCurveCol[particleType]->GetColor4fScaled(pColorMem[i], pValue1[particleType][i]);
	if (alphaType[particleType] == _GL_ALPHA_INVERSE_SQUARE){
	    pColorMem[i][3] *= alphaDependBase[particleType] * alphaDependBase[particleType];
	}
	else
	    pColorMem[i][3] *= alphaFactor[particleType];
    }
    
}

bool SnapData_cls::
IsInterpolateRequired(GLAttribute_cls **ppCGLAttribute){
    bool bRequired = FALSE;
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	switch(ppCGLAttribute[i]->mCDrawType.mBody){
	case _GL_DRAW_POINT:
	case _GL_DRAW_BILLBOARD:
	case _GL_DRAW_BOTH:
	    if (ppCGLAttribute[i]->mpCClrCurveCol->mCDependType.mBody >= 
		_DEPEND_ON_POSITION)
		bRequired = TRUE;
	    break;
	case _GL_DRAW_SPHERE:
	    for (int j = 0; j < 3; j++)
		if (ppCGLAttribute[i]->mpCClrCurveMat[j]->mCDependType.mBody >= 
		    _DEPEND_ON_POSITION)
		    bRequired = TRUE;
	    break;
	}
    return bRequired;
}

void SnapData_cls::
Interpolate(GLAttribute_cls **ppCGLAttribute, 
	    double dT){

    if (dT == 0 && IsInterpolateRequired(ppCGLAttribute) == FALSE){
	// !! FIX_ME !! xˑƂdT=0łԂKv
	InterpolateNot(mpCDataBlockLink, ppCGLAttribute);
	return;
    }

    mpTypeLink = mpCDataBlockLink->mpType;

    int   alphaType[_MAX_PARTICLE_TYPE];
    float alphaDependBase[_MAX_PARTICLE_TYPE];
    float alphaFactor[_MAX_PARTICLE_TYPE];
    ColorCurve_cls *pCClrCurveCol[_MAX_PARTICLE_TYPE];
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++){
	alphaType[i]        = ppCGLAttribute[i]->mCAlphaType.mBody; 
	alphaDependBase[i]  = ppCGLAttribute[i]->mCAlphaDenendBase.mBody;
	alphaFactor[i]      = ppCGLAttribute[i]->mCAlphaFactor.mBody;
	ppCGLAttribute[i]->mpCClrCurveCol->PreCalcScaleFactor();
	pCClrCurveCol[i] = ppCGLAttribute[i]->mpCClrCurveCol;
    }

    float (*pX)[3]  = mpCDataBlockLink->mpX;
    float (*pV)[3]  = mpCDataBlockLink->mpV;
    float (*pA)[3]  = mpCDataBlockLink->mpA;
    float (*pJ)[3]  = mpCDataBlockLink->mpJ;
    float *pRadius  = mpCDataBlockLink->mpRadius;
    float *pDRadius = mpCDataBlockLink->mpDRadius;
    float *pValue1  = mpCDataBlockLink->mpValue1;
    float *pDValue1 = mpCDataBlockLink->mpDValue1;
    
    float (*pXMem)[3] = mpX;
    float *pRadiusMem = mpRadius;
    float *pValue1Mem = mpValue1;
    float (*pColorMem)[4] = mpColor;
    unsigned char *pTypeLinkMem = mpTypeLink;

    float *pValue1ForColor[_MAX_PARTICLE_TYPE];
    for (int i = 0; i < _MAX_PARTICLE_TYPE; i++)
	SetPValue1(&pValue1ForColor[i], ppCGLAttribute[i], 0);

    /*
#ifdef _OPENMP
    // 1%xȂۂB
    omp_set_dynamic( 0 );
    omp_set_num_threads( 2 );
#pragma omp parallel for
#endif
    */

    int threadNumber = 1;
    if (mPNumBlock > 100000) threadNumber = 2;
    if (mPNumBlock > 1000000) threadNumber = 3;
    if (mPNumBlock > 10000000) threadNumber = 4;

    PDataSnap_cls *pCSnap = mpCDataBlockLink->mpCParentSnap;
    printf("Snap Interpolate Index=%d Order=%d InterpolateFinishd = %d\n", pCSnap->mSelfIndex, mOrder, pCSnap->mbInterpolateFinishd);
    
    bool *bAFinished = new bool [threadNumber];
    for (int core = 0; core < threadNumber; core++)
	bAFinished[core] = FALSE;

    if (//mInterpolateType == _INTERPOLATE_LINEAR ||
	//mInterpolateType == _INTERPOLATE_NO
	mOrder == _DATABLOCK_ORDER_1){
	ThreadSnapInterpLinear_cls **ppCThreadSnapInterp = 
	    new ThreadSnapInterpLinear_cls* [threadNumber];
	
	for (int core = 0; core < threadNumber; core++){
	    ppCThreadSnapInterp[core] = 
		new ThreadSnapInterpLinear_cls(wxTHREAD_DETACHED);
	    
	    int PInit = mPNumBlock / threadNumber * core;
	    int PFin  = mPNumBlock / threadNumber * (core+1);
	    if (core == threadNumber - 1) PFin = mPNumBlock;
	    
	    ppCThreadSnapInterp[core]->SetParameters
		(&bAFinished[core], PInit, PFin, mDownsizeStep, dT,
		 pX, pV, pRadius, pDRadius, pValue1, pDValue1,
		 pXMem, pRadiusMem, pValue1Mem, pTypeLinkMem,
		 pColorMem, pValue1ForColor, pCClrCurveCol,
		 alphaType, alphaDependBase, alphaFactor);
	    ppCThreadSnapInterp[core]->Run();
	}
	
	bool bFinishedAll = FALSE;
	while (bFinishedAll == FALSE){
	    wxMilliSleep(3);
	    bFinishedAll = TRUE;
	    for (int core = 0; core < threadNumber; core++)
		bFinishedAll &= bAFinished[core];
	}
	
	delete [] bAFinished;
	delete [] ppCThreadSnapInterp;
    }

    if (//mInterpolateType == _INTERPOLATE_HERMITE
	mOrder == _DATABLOCK_ORDER_3){
	ThreadSnapInterpHermite_cls **ppCThreadSnapInterp = 
	    new ThreadSnapInterpHermite_cls* [threadNumber];
	
	for (int core = 0; core < threadNumber; core++){
	    ppCThreadSnapInterp[core] = 
		new ThreadSnapInterpHermite_cls(wxTHREAD_DETACHED);
	    
	    int PInit = mPNumBlock / threadNumber * core;
	    int PFin  = mPNumBlock / threadNumber * (core+1);
	    if (core == threadNumber - 1) PFin = mPNumBlock;
	    
	    ppCThreadSnapInterp[core]->SetParameters
		(&bAFinished[core], PInit, PFin, mDownsizeStep, dT,
		 pX, pV, pA, pJ, pRadius, pDRadius, pValue1, pDValue1,
		 pXMem, pRadiusMem, pValue1Mem, pTypeLinkMem,
		 pColorMem, pValue1ForColor, pCClrCurveCol,
		 alphaType, alphaDependBase, alphaFactor);
	    ppCThreadSnapInterp[core]->Run();
	}
	
	bool bFinishedAll = FALSE;
	while (bFinishedAll == FALSE){
	    wxMilliSleep(3);
	    bFinishedAll = TRUE;
	    for (int core = 0; core < threadNumber; core++)
		bFinishedAll &= bAFinished[core];
	}
	
	delete [] bAFinished;
	delete [] ppCThreadSnapInterp;
    }
}

void SnapData_cls::
AllocateValue1Dash(){
    mbValue1dash = TRUE;
    if (mbValue1dash == TRUE)
	printf("mpValue1 dash Dual Allocate");
    printf("mpValue1 dash Allocate");
    mpValue1dash = new float [mPNumBlock];
}

void SnapData_cls::
SetValue1Dash(int type){
    if (mbValue1dash == TRUE) return;
    AllocateValue1Dash();
    switch(type){
    case _DEPEND_ON_VELOCITY:{
	float (*pV)[3]  = mpCDataBlockLink->mpV;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep){
	    mpValue1dash[i] = sqrt(pV[i][0] * pV[i][0] + pV[i][1] * pV[i][1] + pV[i][2] * pV[i][2]);
	}
	break; }
    case _DEPEND_ON_POSITION:{
	float (*pX)[3]  = mpCDataBlockLink->mpX;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep){
	    mpValue1dash[i] = sqrt(pX[i][0] * pX[i][0] + pX[i][1] * pX[i][1] + pX[i][2] * pX[i][2]);
	}
	break; }
    case _DEPEND_ON_POS_X: {
	float (*pX)[3]  = mpCDataBlockLink->mpX;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep)
	    mpValue1dash[i] = pX[i][0];
	break; }
    case _DEPEND_ON_POS_Y: {
	float (*pX)[3]  = mpCDataBlockLink->mpX;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep)
	    mpValue1dash[i] = pX[i][1];
	break; }
    case _DEPEND_ON_POS_Z: {
	float (*pX)[3]  = mpCDataBlockLink->mpX;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep)
	    mpValue1dash[i] = pX[i][2];
	break; }
    case _DEPEND_ON_V_X: {
	float (*pV)[3]  = mpCDataBlockLink->mpV;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep)
	    mpValue1dash[i] = pV[i][0];
	break; }
    case _DEPEND_ON_V_Y: {
	float (*pV)[3]  = mpCDataBlockLink->mpV;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep)
	    mpValue1dash[i] = pV[i][1];
	break; }
    case _DEPEND_ON_V_Z: {
	float (*pV)[3]  = mpCDataBlockLink->mpV;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep)
	    mpValue1dash[i] = pV[i][2];
	break; }
    case _DEPEND_ON_ID: {
	int *pID  = mpCDataBlockLink->mpID;
	for (int i = 0; i < mPNumBlock; i+=mDownsizeStep)
	    mpValue1dash[i] = (float)pID[i];
	break; }
    }

}
