// -*- c-mode -*
// "clsThreadFileReaderAscii.h"

#include "clsThreadFileReaderAscii.h"
#include "../MYAPI/MsgBox.h"
#include "../PANELS/clsDataLoader.h"

#include <fstream>
#include <sstream>
#include <string.h>

#include "../COMMON/CommonDefinitionsForH.h"

extern int           GAThreadReaderPercentage[_MAX_READ_THREAD];
extern wxFrame       *GpCMainFrameCLS;
//extern int           GThreadNumberReadAscii;

#include "../COMMON/CommonDefinitions.h"

ThreadAsciiReader_cls::
ThreadAsciiReader_cls(wxThreadKind kind) :
    wxThread(kind) {
}

void ThreadAsciiReader_cls::
SetParameters(int selfThreadIndex,
	      bool *pbFinishedFlag,
	      std::string *pstrBuf, int pNum, int headerLine,
	      int threadnum, int modnumber,
	      std::vector < int > VGrammerBody, bool bUseDefaultFormat,
	      float (*pX)[3], float (*pV)[3],
	      float (*pA)[3], float (*pJ)[3], 
	      float *pRadius, unsigned char  *pType, int *pId,
	      float *pValue1){

    mSelfThreadIndex = selfThreadIndex;
    mpbFinishedFlagLink = pbFinishedFlag;
    mpStrBuf = pstrBuf;

    mpX = pX;
    mpV = pV;
    mpA = pA;
    mpJ = pJ;
    mpRadius = pRadius;
    mpType = pType;
    mTypeCache = 0;
    mpID = pId;
    mpValue1 = pValue1;

    mPNum = pNum;
    mHeaderLine = headerLine;
    mThreadNum = threadnum;
    mModNumber = modnumber;
    mVGrammerBody = VGrammerBody;
    mbUseDefaultFormat = bUseDefaultFormat;
    
    if ( Create() != wxTHREAD_NO_ERROR) 
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

void ThreadAsciiReader_cls::
ReadDeafultFormat(std::istringstream &sstream, int progressInterval){
    int type = 0;
    std::string strLine;
    
    for (int i = 0; i < mPNum; i++){
	std::getline(sstream, strLine);

	if (i % mThreadNum != mModNumber) continue;

	if (mModNumber == 0)
	    if (i % progressInterval == 0)
		GAThreadReaderPercentage[mSelfThreadIndex] = i * 100 / mPNum;

	sscanf(strLine.data(), "%f %f %f %f %f %f %f %d %d %f", 
	       &mpX[i][X], &mpX[i][Y], &mpX[i][Z], 
	       &mpV[i][X], &mpV[i][Y], &mpV[i][Z], 
	       &mpRadius[i], &mpID[i], &type, &mpValue1[i]);
	mpType[i] = type;
	
	if (mpType[i] < 0) mpType[i] = 0;
	if (mpType[i] > _MAX_PARTICLE_TYPE) mpType[i] = _MAX_PARTICLE_TYPE;
    }
}

void ThreadAsciiReader_cls::
Read3Format(std::istringstream &sstream, int progressInterval,
	    std::string strFormat, std::vector < char* > VPointer,
	    std::vector < int > VInc){
    //int type = 0;
    std::string strLine;

    char formatBuff[255];
    sprintf(formatBuff, "%s", strFormat.c_str());
    char* p0 = VPointer[0]; int inc0 = VInc[0];
    char* p1 = VPointer[1]; int inc1 = VInc[1];
    char* p2 = VPointer[2]; int inc2 = VInc[2];
    for (int i = 0; i < mPNum; i++){
	std::getline(sstream, strLine);

	if (i % mThreadNum != mModNumber) continue;
	
	if (mModNumber == 0)
	    if (i % progressInterval == 0)
		GAThreadReaderPercentage[mSelfThreadIndex] = i * 100 / mPNum;

	sscanf(strLine.data(), formatBuff, 
	       p0 + inc0*i, p1 + inc1*i, p2 + inc2*i);

	mpType[i] = mTypeCache;
	
	if (mpType[i] < 0) mpType[i] = 0;
	if (mpType[i] > _MAX_PARTICLE_TYPE) mpType[i] = _MAX_PARTICLE_TYPE;
    }
}

void ThreadAsciiReader_cls::
Read6Format(std::istringstream &sstream, int progressInterval, 
	    std::string strFormat, std::vector < char* > VPointer,
	    std::vector < int > VInc){
    //int type = 0;
    std::string strLine;

    char formatBuff[255];
    sprintf(formatBuff, "%s", strFormat.c_str());
    char* p0 = VPointer[0]; int inc0 = VInc[0];
    char* p1 = VPointer[1]; int inc1 = VInc[1];
    char* p2 = VPointer[2]; int inc2 = VInc[2];
    char* p3 = VPointer[3]; int inc3 = VInc[3];
    char* p4 = VPointer[4]; int inc4 = VInc[4];
    char* p5 = VPointer[5]; int inc5 = VInc[5];
    for (int i = 0; i < mPNum; i++){
	std::getline(sstream, strLine);

	if (i % mThreadNum != mModNumber) continue;
	
	if (mModNumber == 0)
	    if (i % progressInterval == 0)
		GAThreadReaderPercentage[mSelfThreadIndex] = i * 100 / mPNum;

	sscanf(strLine.data(), formatBuff, 
	       p0 + inc0*i, p1 + inc1*i, p2 + inc2*i,
	       p3 + inc3*i, p4 + inc4*i, p5 + inc5*i);
	mpType[i] = mTypeCache;
	
	if (mpType[i] < 0) mpType[i] = 0;
	if (mpType[i] > _MAX_PARTICLE_TYPE) mpType[i] = _MAX_PARTICLE_TYPE;
    }
}

void ThreadAsciiReader_cls::
Read12Format(std::istringstream &sstream, int progressInterval,
	     std::string strFormat, std::vector < char* > VPointer,
	     std::vector < int > VInc){
    //int type = 0;
    std::string strLine;

    char formatBuff[255];
    sprintf(formatBuff, "%s", strFormat.c_str());
    char* p0  = VPointer[0];  int inc0  = VInc[0];
    char* p1  = VPointer[1];  int inc1  = VInc[1];
    char* p2  = VPointer[2];  int inc2  = VInc[2];
    char* p3  = VPointer[3];  int inc3  = VInc[3];
    char* p4  = VPointer[4];  int inc4  = VInc[4];
    char* p5  = VPointer[5];  int inc5  = VInc[5];
    char* p6  = VPointer[6];  int inc6  = VInc[6];
    char* p7  = VPointer[7];  int inc7  = VInc[7];
    char* p8  = VPointer[8];  int inc8  = VInc[8];
    char* p9  = VPointer[9];  int inc9  = VInc[9];
    char* p10 = VPointer[10]; int inc10 = VInc[10];
    char* p11 = VPointer[11]; int inc11 = VInc[11];
    for (int i = 0; i < mPNum; i++){
	std::getline(sstream, strLine);

	if (i % mThreadNum != mModNumber) continue;
	
	if (mModNumber == 0)
	    if (i % progressInterval == 0)
		GAThreadReaderPercentage[mSelfThreadIndex] = i * 100 / mPNum;

	sscanf(strLine.data(), formatBuff, 
	       p0 + inc0*i,  p1 + inc1*i,  p2 + inc2*i,
	       p3 + inc3*i,  p4 + inc4*i,  p5 + inc5*i,
	       p6 + inc6*i,  p7 + inc7*i,  p8 + inc8*i,
	       p9 + inc9*i, p10 + inc10*i, p11 + inc11*i
	       );
	mpType[i] = mTypeCache;
	
	if (mpType[i] < 0) mpType[i] = 0;
	if (mpType[i] > _MAX_PARTICLE_TYPE) mpType[i] = _MAX_PARTICLE_TYPE;
    }
}

void ThreadAsciiReader_cls::
Read18Format(std::istringstream &sstream, int progressInterval,
	     std::string strFormat, std::vector < char* > VPointer,
	     std::vector < int > VInc){
    //int type = 0;
    std::string strLine;

    char formatBuff[255];
    sprintf(formatBuff, "%s", strFormat.c_str());
    char* p0  = VPointer[0];  int inc0  = VInc[0];
    char* p1  = VPointer[1];  int inc1  = VInc[1];
    char* p2  = VPointer[2];  int inc2  = VInc[2];
    char* p3  = VPointer[3];  int inc3  = VInc[3];
    char* p4  = VPointer[4];  int inc4  = VInc[4];
    char* p5  = VPointer[5];  int inc5  = VInc[5];
    char* p6  = VPointer[6];  int inc6  = VInc[6];
    char* p7  = VPointer[7];  int inc7  = VInc[7];
    char* p8  = VPointer[8];  int inc8  = VInc[8];
    char* p9  = VPointer[9];  int inc9  = VInc[9];
    char* p10 = VPointer[10]; int inc10 = VInc[10];
    char* p11 = VPointer[11]; int inc11 = VInc[11];
    char* p12 = VPointer[12]; int inc12 = VInc[12];
    char* p13 = VPointer[13]; int inc13 = VInc[13];
    char* p14 = VPointer[14]; int inc14 = VInc[14];
    char* p15 = VPointer[15]; int inc15 = VInc[15];
    char* p16 = VPointer[16]; int inc16 = VInc[16];
    char* p17 = VPointer[17]; int inc17 = VInc[17];
    for (int i = 0; i < mPNum; i++){
	std::getline(sstream, strLine);

	if (i % mThreadNum != mModNumber) continue;
	
	if (mModNumber == 0)
	    if (i % progressInterval == 0)
		GAThreadReaderPercentage[mSelfThreadIndex] = i * 100 / mPNum;

	sscanf(strLine.data(), formatBuff, 
	       p0  + inc0*i,  p1 + inc1*i,   p2 + inc2*i,
	       p3  + inc3*i,  p4 + inc4*i,   p5 + inc5*i,
	       p6  + inc6*i,  p7 + inc7*i,   p8 + inc8*i,
	       p9  + inc9*i,  p10 + inc10*i, p11 + inc11*i,
	       p12 + inc12*i, p13 + inc13*i, p14 + inc14*i,
	       p15 + inc15*i, p16 + inc16*i, p17 + inc17*i
	       );
	mpType[i] = mTypeCache;
	
	if (mpType[i] < 0) mpType[i] = 0;
	if (mpType[i] > _MAX_PARTICLE_TYPE) mpType[i] = _MAX_PARTICLE_TYPE;
    }
}

void ThreadAsciiReader_cls::
FillByDummy(std::string &strFormat, std::vector < char* > &VPointer,
	    std::vector < int > &VInc, int itemNum, char *pDummy){
    int i = VPointer.size();
    for (; i < itemNum; i++){
	strFormat += " %f";
	VPointer.push_back(pDummy);
	VInc.push_back(0);
    }
}

wxThread::ExitCode ThreadAsciiReader_cls::
Entry(){

    std::ios::sync_with_stdio(false);

    std::istringstream sstream(*mpStrBuf);
    std::string strLine;
    for (int i = 0; i < mHeaderLine; i++)
	std::getline(sstream, strLine);

    int progressInterval = 50000;
    if (mPNum > 2000000) progressInterval = 100000;
    progressInterval /= mThreadNum;
    progressInterval *= mThreadNum;

    if (mbUseDefaultFormat == TRUE){
	ReadDeafultFormat(sstream, progressInterval);
	*mpbFinishedFlagLink = TRUE;
	return 0;
    }

    //int type = 0;
    float dummy[10];

    std::string strFormat;
    std::vector < char* > VPointer;
    std::vector < int > VInc;
    for (int i = 0; i < mVGrammerBody.size(); i++){
	std::string strElement;
	switch (mVGrammerBody[i]){
	case _GRAMMER_XYZ:
	    strElement = "%f %f %f";
	    VPointer.push_back((char*)&mpX[0][X]);
	    VInc.push_back(3*4);
	    VPointer.push_back((char*)&mpX[0][Y]);
	    VInc.push_back(3*4);
	    VPointer.push_back((char*)&mpX[0][Z]);
	    VInc.push_back(3*4);
	    break;
	case _GRAMMER_VXVYVZ:
	    strElement = "%f %f %f";
	    VPointer.push_back((char*)&mpV[0][X]);
	    VInc.push_back(3*4);
	    VPointer.push_back((char*)&mpV[0][Y]);
	    VInc.push_back(3*4);
	    VPointer.push_back((char*)&mpV[0][Z]);
	    VInc.push_back(3*4);
	    break;
	case _GRAMMER_RADIUS:
	    strElement = "%f";
	    VPointer.push_back((char*)&mpRadius[0]);
	    VInc.push_back(4);
	    break;
	case _GRAMMER_TYPE:
	    strElement = "%d";
	    VPointer.push_back((char*)&mTypeCache);
	    VInc.push_back(0);
	    break;
	case _GRAMMER_ID:
	    strElement = "%d";
	    VPointer.push_back((char*)&mpID[0]);
	    VInc.push_back(4);
	    break;
	case _GRAMMER_VALUE1:
	    strElement = "%f";
	    VPointer.push_back((char*)&mpValue1[0]);
	    VInc.push_back(4);
	    break;
	case _GRAMMER_DUMMY1:
	    strElement = "%f";
	    VPointer.push_back((char*)dummy);
	    VInc.push_back(0);
	    break;
	case _GRAMMER_DUMMY3:
	    strElement = "%f %f %f";
	    for (int j = 0; j < 3; j++){
		VPointer.push_back((char*)dummy);
		VInc.push_back(0);
	    }
	    break;
	case _GRAMMER_DUMMY10:
	    strElement = "%f %f %f %f %f %f %f %f %f %f";
	    for (int j = 0; j < 10; j++){
		VPointer.push_back((char*)dummy);
		VInc.push_back(0);
	    }
	    break;
	}
	strFormat += strElement;
	if (i != mVGrammerBody.size() - 1)
	    strFormat += " ";
    }

    // 3ȉ̓f[^ƂĕsS
    if (VPointer.size() == 3){
	Read3Format(sstream, progressInterval, strFormat, VPointer, VInc);
	*mpbFinishedFlagLink = TRUE;
	return 0;
    }

    if (VPointer.size() <= 6){
	FillByDummy(strFormat, VPointer, VInc, 6, (char*)dummy);
	Read6Format(sstream, progressInterval, strFormat, VPointer, VInc);
	*mpbFinishedFlagLink = TRUE;
	return 0;
    }

    if (VPointer.size() <= 12){
	FillByDummy(strFormat, VPointer, VInc, 12, (char*)&dummy);
	Read12Format(sstream, progressInterval, strFormat, VPointer, VInc);
	*mpbFinishedFlagLink = TRUE;
	return 0;
    }
    
    FillByDummy(strFormat, VPointer, VInc, 18, (char*)&dummy);
    Read18Format(sstream, progressInterval, strFormat, VPointer, VInc);
    *mpbFinishedFlagLink = TRUE;
    return 0;
}
