// "clsPovObjectMesh.h"
// (setq c-basic-offset 4)

#include "clsPovObjectMesh.h"

#include "../clsPovRay.h"
#include "../clsPovSerialize.h"

#include "../../MYAPI/MsgBox.h"
#include "../../../wxWidgetString.h"

#include "../../COMMON/MyVector.h"
#include "../../COMMON/clsFilename.h"

#include "../../COMMON/CommonDefinitions.h"
#include "../../COMMON/CommonDefinitionsGL.h"

#define ENDL wxT("\n")

#include <wx/mstream.h>
#include <wx/txtstrm.h>

ThreadReadMesh2Sub::
ThreadReadMesh2Sub(char *pBuffer, int fileSize, int posStart,
		   int itemNum, int skipLine, std::vector < float > *pVData,
		   bool *pbFinishedFlagLink) :
    wxThread(wxTHREAD_DETACHED)
{
    mpbFinishedFlagLink = pbFinishedFlagLink;
    mpBuffer = pBuffer;
    mFileSize = fileSize;
    mPosStart = posStart;
    mItemNum = itemNum;
    mSkipLine = skipLine;
    mpVData = pVData;
    
    if ( Create() != wxTHREAD_NO_ERROR)
	MsgBoxExit(NULL, "Error!", "Can't create thread");
}

wxThread::ExitCode ThreadReadMesh2Sub::
Entry(){
    char *ptr = mpBuffer + mPosStart;
    
    int itemNum = mItemNum;
    float vectorX = 0, vectorY = 0, vectorZ = 0;
    
    std::vector <float> *pVData = mpVData;
    int item = 0;
    while (TRUE){
	while (*ptr != '<'){
	    ptr++;
	    if (ptr >= mpBuffer + mFileSize){
		ptr = NULL;
		break;
	    }
	}
	//ptr = strchr(ptr, '<');
	if (ptr == NULL) break;
	
	ptr++;
	vectorX = ReadFloat(&ptr, ',');
	ptr++;
	vectorY = ReadFloat(&ptr, ',');
	ptr++;
	vectorZ = ReadFloat(&ptr, '>');
	
	pVData->push_back(vectorX);
	pVData->push_back(vectorY);
	pVData->push_back(vectorZ);
	item++;

	if (item >= itemNum) break;
    }

    *mpbFinishedFlagLink = TRUE;
    return 0;
}

// ----

void PovObjectMesh_cls::
ClearMesh(){
    // ̊mۗ̈悲Clear@(TCY0̈ꎞxNgƌ)
    std::vector<float>().swap(mVVertices);
    //std::vector<float>().swap(mVUV1D);
    std::vector<float>().swap(mVNormals);
    std::vector<unsigned int>().swap(mVIndex);
    std::vector<unsigned int>().swap(mVIndexNormal);
}

char* PovObjectMesh_cls::
GetMemoryBufferFromFile(wxString strFile){
    FILE *fp = fopen(strFile.ToAscii(), "rb"); //!!
    if (fp == NULL) {
	mstrLoaded = strFile; return NULL; }

    fseek( fp, 0, SEEK_END );
    mFileSize = ftell( fp );
    fseek( fp, 0, SEEK_SET);
    
    char *pBuffer = new char [mFileSize];
    fread( pBuffer, 1, mFileSize, fp);
    fclose(fp);
    return pBuffer;
}

void PovObjectMesh_cls::
ConstructNormalVectors(){
    mVNormals.resize(mVVertices.size());
    
    int normalSize = (int)mVNormals.size();
    int normalNum = normalSize / 3;
    for (int i = 0; i < normalSize; i++)
	mVNormals[i] = 0;
    std::vector < unsigned char > mVOverlap(normalNum);
    
    int faceSize = (int)mVIndex.size();
    int faceNum = faceSize / 3;
    for (int i = 0; i < faceNum; i++){
	int index0 = mVIndex[i*3 + 0];
	int index1 = mVIndex[i*3 + 1];
	int index2 = mVIndex[i*3 + 2];
	float *p0 = &mVVertices[index0*3];
	float *p1 = &mVVertices[index1*3];
	float *p2 = &mVVertices[index2*3];

	double normal[3];
	normal[X] = (p1[Y]-p0[Y])*(p2[Z]-p1[Z]) - (p1[Z]-p0[Z])*(p2[Y]-p1[Y]);
	normal[Y] = (p1[Z]-p0[Z])*(p2[X]-p1[X]) - (p1[X]-p0[X])*(p2[Z]-p1[Z]);
	normal[Z] = (p1[X]-p0[X])*(p2[Y]-p1[Y]) - (p1[Y]-p0[Y])*(p2[X]-p1[X]);
	
	MyVector::Normalize(normal, 1);

	int overlap;

	mVOverlap[index0]++;
	overlap = mVOverlap[index0];
	LOOP_XYZ mVNormals[index0*3 + xyz] =
	    (mVNormals[index0*3 + xyz] * (overlap - 1) + normal[xyz]) / overlap;
	
	mVOverlap[index1]++;
	overlap = mVOverlap[index1];
	LOOP_XYZ mVNormals[index1*3 + xyz] =
	    (mVNormals[index1*3 + xyz] * (overlap - 1) + normal[xyz]) / overlap;

	mVOverlap[index2]++;
	overlap = mVOverlap[index2];
	LOOP_XYZ mVNormals[index2*3 + xyz] =
	    (mVNormals[index2*3 + xyz] * (overlap - 1) + normal[xyz]) / overlap;

    }

    for (int i = 0; i < normalNum; i++){
	if (MyVector::GetNorm(&mVNormals[i*3]) == 0){
	    mVNormals[i*3+X] = 0;
	    mVNormals[i*3+Y] = 0;
	    mVNormals[i*3+Z] = 1;
	}
	MyVector::Normalize(&mVNormals[i*3], 1);
    }
}


void PovObjectMesh_cls::
ReadObjFile(wxString strFile){
    
    if (mstrLoaded == strFile) return;

    char *pBuffer = GetMemoryBufferFromFile(strFile);
    
    if (pBuffer == NULL){
	ClearMesh();
	return;
    }
    
    char *ptr = pBuffer;
    char *pEnd = pBuffer + mFileSize;
    
    bool bSmooth;
 
    ClearMesh();
    mVIndex.reserve(mFileSize / 3 / 7);
    mVIndexNormal.reserve(mFileSize / 3 / 7);
    mVVertices.reserve(mFileSize / 3 / 10);
    //mVUV1D.reserve(mFileSize / 3 / 10);
    mVNormals.reserve(mFileSize / 3 / 10);

    bool bSliceChecked = FALSE;
    bool bIsSliceExist = FALSE;
    
    //objectname
    //if (SearchChar(&ptr, 'o', pEnd) == FALSE) return;
    //if (SearchChar(&ptr, '\n', pEnd) == FALSE) return;
    while (TRUE){
	ptr++;
	if (ptr >= pEnd) break;
	if (*ptr == 'o'){
	    if (SearchChar(&ptr, '\n', pEnd) == FALSE) return;
	}
	if (*ptr == 'v' && *(ptr+1) == 'n'){
	    ptr++; ptr++; ptr++;
	    mVNormals.push_back(ReadFloat(&ptr, ' '));
	    ptr++;
	    mVNormals.push_back(ReadFloat(&ptr, ' '));
	    ptr++;
	    mVNormals.push_back(ReadFloat(&ptr, '\n'));
	}
	if (*ptr == 'v' && *(ptr+1) == ' '){
	    ptr++; ptr++;
	    mVVertices.push_back(ReadFloat(&ptr, ' '));
	    ptr++;
	    mVVertices.push_back(ReadFloat(&ptr, ' '));
	    ptr++;
	    mVVertices.push_back(ReadFloat(&ptr, '\n'));
	}
	if (*ptr == 's'){
	    ptr++; ptr++;
	    bSmooth = ReadBool(&ptr, '\n');
	}
	if (*ptr == 'f'){
	    ptr++;
	    if (IsCharExist(ptr, ' ', '/') == TRUE){
		ptr++;
		mVIndex.push_back(ReadInt(&ptr, '/') - 1);
		ptr++; ptr++ ; // index//index 
		mVIndexNormal.push_back(ReadInt(&ptr, ' ') - 1);

		ptr++;
		mVIndex.push_back(ReadInt(&ptr, '/') - 1);
		ptr++; ptr++ ; // index//index 
		mVIndexNormal.push_back(ReadInt(&ptr, ' ') - 1);

		ptr++;
		mVIndex.push_back(ReadInt(&ptr, '/') - 1);
		ptr++; ptr++ ; // index//index 
		mVIndexNormal.push_back(ReadInt(&ptr, '\n') - 1);
	    }
	    else{
		ptr++;
		mVIndex.push_back(ReadInt(&ptr, ' ') - 1);
		ptr++;
		mVIndex.push_back(ReadInt(&ptr, ' ') - 1);
		ptr++;
		mVIndex.push_back(ReadInt(&ptr, '\n') - 1);
	    }
	}
    }

    delete pBuffer;
    mstrLoaded = strFile;

    // @xNg
    if (mVNormals.size() == 0)
	ConstructNormalVectors();
}

void PovObjectMesh_cls::
ReadMesh2File(wxString strFile){
    
    if (mstrLoaded == strFile) return;
    
    char *pBuffer = GetMemoryBufferFromFile(strFile);
    if (pBuffer == NULL){
	ClearMesh();
	mstrLoaded.Clear();
	return;
    }
    
    // ɈU߂قfgcB
    wxMemoryInputStream CMStream(pBuffer, mFileSize);
	
    wxString sep = wxT(" \t\n=,{}");
    wxTextInputStream CTStream(CMStream, sep);

    ClearMesh();

    mVIndex.reserve(mFileSize / 3 / 7);
    mVIndexNormal.reserve(mFileSize / 3 / 7);
    mVVertices.reserve(mFileSize / 3 / 10);
    //mVUV1D.reserve(mFileSize / 3 / 10);
    mVNormals.reserve(mFileSize / 3 / 10);

    int posNormal = -1, posIndex = -1;

    wxString strWord;
    strWord = CTStream.ReadWord();
    while (!strWord.Matches(L"#declare")){
	if (strWord.Matches(L"//normal_start_at"))
	    CTStream >> posNormal;
	if (strWord.Matches(L"//index_start_at"))
	    CTStream >> posIndex;
	strWord = CTStream.ReadWord();
    }
    mstrPovLabel = CTStream.ReadWord();
    while (!strWord.Matches(L"vertex_vectors")){
	strWord = CTStream.ReadWord();
    }
    int vertexNum;
    CTStream >> vertexNum;
    mVVertices.reserve(vertexNum * 3);

    // Verticex ǂݍ݂ʃXbhŊJn
    bool bFinishedVertex = FALSE;
    ThreadReadMesh2Sub *pCThreadVertices = new ThreadReadMesh2Sub
	(pBuffer, mFileSize, CMStream.TellI(), vertexNum,
	 1, &mVVertices, &bFinishedVertex);
    pCThreadVertices->Run();
    
    // m[}ʒum肵ĂȂΉsłǂĂ
    if (posNormal == -1){
	CTStream.ReadLine(); //dummy
	for (int i = 0; i < vertexNum; i++)
	    wxString strLine = CTStream.ReadLine();
    }
    else
	CMStream.SeekI(posNormal);
    
    int normalNum = -1;
    strWord = CTStream.ReadWord();
    printf ("%ls\n", strWord.c_str());
    if (strWord.Matches(L"normal_vectors")){
	CTStream >> normalNum;
	mVNormals.reserve(normalNum * 3);
    }

    bool bFinishedNormal = FALSE;
    ThreadReadMesh2Sub *pCThreadNormal;
    if (normalNum != -1){
	pCThreadNormal= new ThreadReadMesh2Sub
	    (pBuffer, mFileSize, CMStream.TellI(), normalNum,
	     1, &mVNormals, &bFinishedNormal);
	pCThreadNormal->Run();
    }
    
    // Indexʒum肵ĂȂΉsłǂĂ
    if (normalNum != -1){
	if (posIndex == -1){
	    CTStream.ReadLine(); //dummy
	    for (int i = 0; i < vertexNum; i++)
		wxString strLine = CTStream.ReadLine();
	}
	else
	    CMStream.SeekI(posIndex);
	
	// normal΁AstrWord͐Vɓǂ܂Ɉێ
	strWord = CTStream.ReadWord();
    }
    
    if (strWord.Matches(L"face_indices")){
	int faceNum;
	CTStream >> faceNum;
	mVIndex.reserve(faceNum * 3);

	int indexX, indexY, indexZ;
	int item = 0;
	char *ptr = pBuffer + CMStream.TellI();
	while(TRUE){
	    while (*ptr != '<'){
		ptr++;
		if (ptr >= pBuffer + mFileSize){
		    ptr = NULL;
		    break;
		}
	    }
	    if (ptr == NULL) break;
	    
	    ptr++;
	    indexX = ReadInt(&ptr, ',');
	    ptr++;
	    indexY = ReadInt(&ptr, ',');
	    ptr++;
	    indexZ = ReadInt(&ptr, '>');

	    mVIndex.push_back(indexX);
	    mVIndex.push_back(indexY);
	    mVIndex.push_back(indexZ);
	    item++;
	    
	    if (item >= faceNum) break;
	}
    }

    while (TRUE){
	if (bFinishedVertex == TRUE) break;
	wxMilliSleep(3);
    }

    if (normalNum != -1) {
	while (TRUE){
	    if (bFinishedNormal == TRUE) break;
	    wxMilliSleep(3);
	}
    }

    delete [] pBuffer;
    mstrLoaded = strFile;
}

void PovObjectMesh_cls::
ReadMeshFromFile(wxString strFile){
    SplitPath_cls CSplit(strFile);
    wxString strExt = CSplit.mstrExt;
    if (strExt.Matches(wxT(".pov")) || strExt.Matches(wxT(".POV")))
	ReadMesh2File(strFile);
    if (strExt.Matches(wxT(".obj")) || strExt.Matches(wxT(".OBJ")))
	ReadObjFile(strFile);
}

wxString PovObjectMesh_cls::
GetStrPovInclude(){

    wxString strInclude;

    bool bPov = FALSE;
    SplitPath_cls CSplit(mstrLoaded);
    wxString strExt = CSplit.mstrExt;
    if (strExt.Matches(wxT(".pov")) || strExt.Matches(wxT(".POV")))
	bPov = TRUE;

    if (bPov == TRUE){ // .povȂ
	if (mstrLoaded.IsEmpty() == FALSE){
	    strInclude << wxT("#include \"") 
		       << mspCPovRayLink->GetStrRelativePath(mstrLoaded)
		       << wxT("\"") << ENDL;
	}
    }
    else if (mVVertices.size() == 0){
	mstrPovLabel = wxT("MESH_OBJECT");
	strInclude << wxT("#declare MESH_OBJECT=union{};") << ENDL;
	
    }
    else {
	mstrPovLabel = wxT("MESH_OBJECT");
	strInclude << wxT("#declare MESH_OBJECT=mesh2{") << ENDL;
	
	strInclude << wxT("\tvertex_vectors{")  
		   << wxT("\t\t") << mVVertices.size() / 3 << wxT(",") << ENDL;
	for (unsigned int i = 0; i < mVVertices.size(); i += 3){
	    strInclude << PovSerialize_cls::PovOut3f(wxT(""), mVVertices.data() + i);
	    strInclude << wxT(",") << ENDL;
	}
	strInclude << wxT("\t}") << ENDL;

	strInclude << wxT("\tnormal_vectors{")  
		   << wxT("\t\t") << mVNormals.size() / 3 << wxT(",") << ENDL;
	for (unsigned int i = 0; i < mVNormals.size(); i += 3){
	    strInclude << PovSerialize_cls::PovOut3f(wxT(""), mVNormals.data() + i, 2);
	    strInclude << wxT(",") << ENDL;
	}
	strInclude << wxT("\t}") << ENDL;

	strInclude << wxT("\tface_indices{")  
		   << wxT("\t\t") << mVIndex.size() / 3 << wxT(",") << ENDL;
	for (unsigned int i = 0; i < mVIndex.size(); i += 3){
	    strInclude << PovSerialize_cls::PovOut3ui(wxT(""), mVIndex.data() + i);
	    strInclude << wxT(",") << ENDL;
	}
	strInclude << wxT("\t}") << ENDL;

	/*
	strInclude << wxT("\tnormal_indices{")  
		   << wxT("\t\t") << mVIndex.size() / 3 << wxT(",") << ENDL;
	for (int i = 0; i < mVIndex.size(); i += 3){
	    strInclude << PovSerialize_cls::PovOut3ui(wxT(""), mVIndex.data() + i);
	    strInclude << wxT(",") << ENDL;
	}
	strInclude << wxT("\t}") << ENDL;
	*/
	
	strInclude << wxT("};") << ENDL;
    }


    return strInclude;
}

void PovObjectMesh_cls::
GetMesh2FormatRef(wxString &strMesh2, wxString strTag,
		  bool bNormal, 
		  std::vector < float > *pVVertices,
		  std::vector < float > *pVNormals,
		  std::vector < unsigned int > *pVIndices){
    
    strMesh2 << wxT("//Mesh2file generated by ") << _FRAME_MAIN_APPNAME << ENDL;
    strMesh2 << wxT("//normal_start_at=");
    int posNormalTag = (int)strMesh2.Len();
    strMesh2 << wxT("##########") << ENDL;

    strMesh2 << wxT("//index_start_at=");
    int posIndexTag = (int)strMesh2.Len();
    strMesh2 << wxT("##########") << ENDL;

    int verticesSize = (int)pVVertices->size();
    int verticesNum = verticesSize / 3;

    strMesh2 << wxT("#declare ") << wxT("MESH2OBJ_") << strTag << wxT("=mesh2{") << ENDL
	     << wxT("\tvertex_vectors{") << ENDL
	     << wxT("\t\t") << verticesNum << wxT (",") << ENDL;
    float *pVertex = &((*pVVertices)[0]);
    for (int v = 0; v < verticesNum; v++){
	char buf [64];
	if (v % 3 == 0 && v == verticesNum - 1)
	    sprintf(buf, "\t\t<%0.5f,%0.5f,%0.5f>,\n",
		    *(pVertex + v*3 + X),
		    *(pVertex + v*3 + Y),
		    *(pVertex + v*3 + Z));
	else if (v % 3 == 2)
	    sprintf(buf, "<%0.5f,%0.5f,%0.5f>,\n",
		    *(pVertex + v*3 + X),
		    *(pVertex + v*3 + Y),
		    *(pVertex + v*3 + Z));
	else if (v % 3 == 0)
	    sprintf(buf, "\t\t<%0.5f,%0.5f,%0.5f>,",
		    *(pVertex + v*3 + X),
		    *(pVertex + v*3 + Y),
		    *(pVertex + v*3 + Z));
	else
	    sprintf(buf, "<%0.5f,%0.5f,%0.5f>,",
		    *(pVertex + v*3 + X),
		    *(pVertex + v*3 + Y),
		    *(pVertex + v*3 + Z));
	/*
	strMesh2 << 
	    wxString::Format(wxT("\t\t<%0.5f,%0.5f,%0.5f>,"),
			     *(pVertex + v + X),
			     *(pVertex + v + Y),
			     *(pVertex + v + Z)) << ENDL;
			     */
	strMesh2 << wxString::From8BitData(buf);
    }
    strMesh2 << wxT("\t}") << ENDL;

    if (bNormal == TRUE){
	int normalSize = (int)pVNormals->size();
	int normalNum = normalSize / 3;
	strMesh2 << wxT("\tnormal_vectors{") << ENDL
		 << wxT("\t\t") << normalNum << wxT (",") 
		 << ENDL;
	float *pNormal = &((*pVNormals)[0]);
	for (int n = 0; n < normalNum; n++){
	    char buf[64];
	    if (n % 3 == 0 && n == normalNum - 1)
		sprintf(buf, "\t\t<%0.3f,%0.3f,%0.3f>,\n",
			*(pNormal + n*3 + X),
			*(pNormal + n*3 + Y),
			*(pNormal + n*3 + Z));
	    else if (n % 3 == 2)
		sprintf(buf, "<%0.3f,%0.3f,%0.3f>,\n",
			*(pNormal + n*3 + X),
			*(pNormal + n*3 + Y),
			*(pNormal + n*3 + Z));
	    else if (n % 3 == 0)
		sprintf(buf, "\t\t<%0.3f,%0.3f,%0.3f>,",
			*(pNormal + n*3 + X),
			*(pNormal + n*3 + Y),
			*(pNormal + n*3 + Z));
	    else
		sprintf(buf, "<%0.3f,%0.3f,%0.3f>,",
			*(pNormal + n*3 + X),
			*(pNormal + n*3 + Y),
			*(pNormal + n*3 + Z));
	    /*
	    strMesh2 << 
		wxString::Format(wxT("\t\t<%0.3f,%0.3f,%0.3f>,"),
				 *(pNormal + n + X),
				 *(pNormal + n + Y),
				 *(pNormal + n + Z)) << ENDL;
				 */
	    strMesh2 << wxString::From8BitData(buf);
	}
	strMesh2 << wxT("\t}") << ENDL;
    }

    int faceSize = (int)pVIndices->size();
    int faceNum = faceSize / 3;
    strMesh2 << wxT("\tface_indices{") << ENDL
	     << wxT("\t\t") << faceSize / 3 << wxT (",") 
	     << ENDL;
    unsigned *pFace = &((*pVIndices)[0]);
    for (int i = 0; i < faceNum; i++){
	if (i % 4 == 0) strMesh2 << wxT("\t\t");
	strMesh2 << wxT("<")
		 << *(pFace + i*3 + X) << wxT(",")
		 << *(pFace + i*3 + Y) << wxT(",")
		 << *(pFace + i*3 + Z) << wxT(">")
		 << wxT(",");
	if (i % 4 == 3 || i == faceNum-1) strMesh2 << ENDL;
    }
    strMesh2 << wxT("\t}") << ENDL
	     << wxT("}") << ENDL;
}

void PovObjectMesh_cls::
GLDraw(){
    if (mVVertices.size() == 0) return;

    if (mVVertices.size() == mVNormals.size()) {
	//You can find out maximum number of vertices to be prefetched and the maximum number of indices to be referenced by using glGetIntegerv() with GL_MAX_ELEMENTS_VERTICES and GL_MAX_ELEMENTS_INDICES.
	//S邱ƁI

	if (FALSE) { //glVertex̂Ƃ 
	    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	    glEnableClientState(GL_VERTEX_ARRAY);
	    glEnableClientState(GL_NORMAL_ARRAY);
	    /*
	      if (DoesUseMapping() == TRUE)
	      glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	    */
	    
	    glVertexPointer(3, GL_FLOAT, 0, &mVVertices[0]);
	    glNormalPointer(GL_FLOAT, 0, &mVNormals[0]);
	    /*
	      if (DoesUseMapping() == TRUE)
	      glTexCoordPointer(1, GL_FLOAT, 0, &mVUV1D[0]);
	    */
	    
	    glDrawElements(GL_TRIANGLES, (GLsizei)mVIndex.size() , 
			   GL_UNSIGNED_INT, mVIndex.data());
	    
	    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	    glDisableClientState(GL_VERTEX_ARRAY);
	    glDisableClientState(GL_NORMAL_ARRAY);
	}
	else{
	    int totalSize = (int)mVIndex.size() / 3;
	    for (int block = 0; block < totalSize; block += 50){
		int ini = block;
		int fin = ini + 50;
		if (fin > totalSize) fin = totalSize;
		glBegin(GL_TRIANGLES);
		for (int i = ini; i < fin; i++){
		    int index0 = mVIndex[3*i];
		    int index1 = mVIndex[3*i+1];
		    int index2 = mVIndex[3*i+2];
		    glNormal3fv(&mVNormals[index0*3]);
		    glVertex3fv(&mVVertices[index0*3]);
		    glNormal3fv(&mVNormals[index1*3]);
		    glVertex3fv(&mVVertices[index1*3]);
		    glNormal3fv(&mVNormals[index2*3]);
		    glVertex3fv(&mVVertices[index2*3]);
		}
		glEnd();
	    }
	}
    }
    else{
	glBegin(GL_TRIANGLES);
	int indexSize = (int)mVIndex.size();
	for (int i = 0; i < indexSize; i+=3){
	    for (int j = 0; j < 3; j++){
		int normalOffset = mVIndexNormal[i+j] * 3;
		glNormal3d(mVNormals[normalOffset + X],
			   mVNormals[normalOffset + Y],
			   mVNormals[normalOffset + Z]);
		int vertexOffset = mVIndex[i+j] * 3;
		glVertex3d(mVVertices[vertexOffset + X],
			   mVVertices[vertexOffset + Y],
			   mVVertices[vertexOffset + Z]);
	    }
	}
	glEnd();
    }
}

PovRayCtrl_cls* PovObjectMesh_cls::mspCPovRayLink = NULL;
