// -*- c -*- 
// (setq c-basic-offset 4)
// "clsPovFocalBlur.h"

#include "clsPovFocalBlur.h"

#include "../clsPovRay.h"
#include "../clsPovRayGUI.h"
#include "../../../wxWidgetEnum.h"
#include "../../WIDGETCTRL/clsAtoSTextCtrl.h"
#include "../../WIDGETCTRL/clsDialogEXKey.h"
#include "../../WIDGETCTRL/clsHookCtrl.h"
#include "../../WIDGETCTRL/clsMultiInputPanel.h"

#include "../../COMMON/CommonDefinitions.h"

PovFocalBlur_cls::
PovFocalBlur_cls(){
    mspCThis = this;
    SetToDefault();
}

void PovFocalBlur_cls::
SetToDefault(){
    mCConfList.Clear();
    
    mCConfList.Add(&mCType, wxT("PovFocalBlur="), L"0");
    mCConfList.Add(&mCFocalBlurScale, wxT("PovFocalBlurScale="), L"0");
    mCConfList.Add(&mCFocalBlurQuality, wxT("PovFocalBlurQuality="), L"1");
    mCConfList.Add(&mCFocusLocationType, wxT("PovFocusLocationType="), L"0");
    mCConfList.Add(&mCFocusDistance, wxT("PovFocusDistance="), L"1");
    mCConfList.Add(&mCFocusLocation, wxT("PovFocusLocation="), L"0 0 0");
}

void PovFocalBlur_cls::
SetMNumFocalBlur(wxWindow* pCParentPanel){
    MultiInputPanel_cls* pCurr = 
	MultiInputPanel_cls::Initialize(&mpCMNum, pCParentPanel);

    pCurr->mTextWidth  = -1;
    pCurr->mItemWidth = 180;

    wxString focalBlurType[2] =
	{_("No Focal Blur"), _("Use Focal Blur") };
    mIndexType = pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &mCType.mBody, 2, focalBlurType,_("Focal Blur"));

    pCurr->mTextWidth = 120;
    pCurr->mItemWidth = -1;
    int boxWidth = 2;
    mIndexScale = pCurr->SetInputAndSlider
	(_ITEM_FLOAT, &mCFocalBlurScale.mBody, _("scale"));
    pCurr->SetRange(_ITEM_LASTITEMx2, 0, 1000, 1000);

    mIndexQuality = pCurr->SetInputAndSlider
	(_ITEM_INTEGER_SPIN, &mCFocalBlurQuality.mBody,_("Quality")) ;
    pCurr->SetRange(_ITEM_LASTITEMx2, 0, 5);

    wxString focusLocationType[5] =
	{_("Origin"), _("Anchor Pos"), _("Stereo Focus"), _("Distance"),
	 _("Numeric") };
    mIndexFocusLocationType = pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &mCFocusLocationType.mBody,
	 5, focusLocationType, _("Focus location"));
    
    mIndexFocusDistance = pCurr->SetInputAndSlider
	(_ITEM_FLOAT, &mCFocusDistance.mBody, _("Distance"));
    pCurr->SetRange(_ITEM_LASTITEM, 0, 1000, 100);
    
    boxWidth = 3;
    mIndexFocusLocation = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("location"));
    pCurr->CombinePointer(_ITEM_FLOAT, &mCFocusLocation.mpBody[0]);
    pCurr->CombinePointer(_ITEM_FLOAT, &mCFocusLocation.mpBody[1]);
    pCurr->CombinePointer(_ITEM_FLOAT, &mCFocusLocation.mpBody[2]);
    
#ifdef WIN32
    pCurr->mMinimumWidth = 320;
#else
    pCurr->mMinimumWidth = 340;
#endif
    pCurr->FinishSetting();

    pCurr->SetColorForXYZ(mIndexFocusLocation+1);

    pCurr->mpCHookCtrl->SetFuncHook
	(PovFocalBlur_cls::HookHideAndShow, _HOOK_COMBOBOX);
    
    HideAndShow();
}

void PovFocalBlur_cls::
HideAndShow(){
    HideAndShowBegin();
    MultiInputPanel_cls *pCurr = mpCMNum;
    bool bShow = (mCType.mBody > _POV_FOCALBLUR_NOUSE);
    
    pCurr->ShowItem(mIndexScale, bShow);
    pCurr->ShowItem(mIndexQuality, bShow);
    pCurr->ShowItem(mIndexFocusLocationType, bShow);

    bool bDistance = (mCFocusLocationType.mBody == _POV_FOCUSPOS_DISTANCE);
    pCurr->ShowItem(mIndexFocusDistance, bShow & bDistance);
    
    bool bLocation = (mCFocusLocationType.mBody == _POV_FOCUSPOS_NUMERIC);
    pCurr->ShowItem(mIndexFocusLocation, bShow & bLocation);
    HideAndShowEnd();
}

void PovFocalBlur_cls::
HookHideAndShow(void *pUsrDef, void *pEvent){
    MultiInputPanel_cls* pCPanel = (MultiInputPanel_cls*)pUsrDef;
    mspCThis->HideAndShow();
    RecalcSize(pCPanel);
}

wxString PovFocalBlur_cls::
GetPovStrFocalBlur(PositionAndDirection_cls CCameraPos, Position_cls CAnchorPos,
		   double apatureFactor, double stereoFocus){
    if (mCType.mBody == _POV_FOCALBLUR_NOUSE) return wxT("");
    wxString strFocalBlur;

    strFocalBlur << wxT("\taperture ") << mCFocalBlurScale.mBody * apatureFactor << ENDL;
    
    // FIX_ME!! Ad_Hoc
    int quality = mCFocalBlurQuality.mBody;
    int blurSamples = 8 * quality * quality;
    //float confidence = 0.86 + quality * 0.03;
    int varianceInv = 128 + 256 * quality * (quality - 1);
    if (quality == 0) varianceInv = 64;
    if (blurSamples == 0) blurSamples = 3;
    //if (confidence >= 0.99) confidence = 0.99;
    float confidence = 1.0f - 1.0f/varianceInv;
    strFocalBlur << wxT("\tblur_samples ") << blurSamples << ENDL;
    strFocalBlur << wxT("\tconfidence ") << confidence << ENDL;
    strFocalBlur << wxT("\tvariance 1/") << varianceInv << ENDL;

    double anchorPos[3] = {0, 0, 0};
    switch(mCFocusLocationType.mBody){
    case _POV_FOCUSPOS_ORIGIN:
	break;
    case _POV_FOCUSPOS_ANCHORPOS: 
	LOOP_XYZ anchorPos[xyz] = CAnchorPos.mPos[xyz];
	break;
    case _POV_FOCUSPOS_STEREOFOCUS:
	LOOP_XYZ anchorPos[xyz] = CCameraPos.mPos[xyz] + stereoFocus * CCameraPos.mTgt[xyz];
	break;
    case _POV_FOCUSPOS_DISTANCE: 
	LOOP_XYZ anchorPos[xyz] = CCameraPos.mPos[xyz] + mCFocusDistance.mBody * CCameraPos.mTgt[xyz];
	break; 
    case _POV_FOCUSPOS_NUMERIC:
	LOOP_XYZ anchorPos[xyz] = mCFocusLocation.mpBody[xyz];
	break;
    }

    strFocalBlur << PovSerialize_cls::PovOut3d(wxT("\tfocal_point "), anchorPos, 4) << ENDL;
    
    return strFocalBlur;
}

PovFocalBlur_cls *PovFocalBlur_cls::mspCThis = NULL;
