#ifndef _POV_RAY_H_
#define _POV_RAY_H_

#include "clsPovSerialize.h"

#include "../MYGL/clsPositionAndDirection.h"
#include "../WIDGETCTRL/clsDialogController.h"

#include "../COMMON/CommonDefinitions.h"

#include "../../clsFrameMain.h"

#define _POV_SNAP 0
#define _POV_ANIM 1
#define _POV_COMMON 2

#define _POV_PERSPECTIVE  0
#define _POV_ORTHOGRAPHIC 1
#define _POV_FISHEYE      2
#define _POV_CYLINDER     3

#define _POV_RENDER_WORKAREA  0
#define _POV_RENDER_FULLSCENE 1
#define _POV_RENDER_CUSTOM    2

#define _POV_FOLLOW_CAMERA     0
#define _POV_SAME_AS_PREVIEW0  1
#define _POV_SAME_AS_PREVIEW1  2
#define _POV_SAME_AS_PREVIEW2  3
#define _POV_SAME_AS_PREVIEW3  4

#define _POVRAY_36 0
#define _POVRAY_37 1

#define _POVRAY_MAX_THREAD 16

#define _POV_EXPORT 0
#define _POV_RENDER 1

#include <wx/thread.h>

class ArrayAndIndex_cls;
class FilterBase_cls;
class GLAttribute_cls;
class PDataBlock_cls;
class PovAdvanced_cls;
class PovFocalBlur_cls;
class PovFormat_cls;
class PovIsosurfaceData_cls;
class PovRayGUI_cls;
class PovRadiosity_cls;
class PovRender_cls;
class PovPhotons_cls;
class StereoGram_cls;
class ThreadPovRay_cls;
class TimeCtrl_cls;

class wxNotebook;

class ThreadSPString_cls : public wxThread {
public:
    ThreadSPString_cls(bool *pbFinishedFlagLink,
		       char *pBuf, int init, int fin, int step,
		       unsigned char *pType, 
		       int   drawTypeList[_MAX_PARTICLE_TYPE],
		       int   radiusTypeList[_MAX_PARTICLE_TYPE],
		       float radiusFactorList[_MAX_PARTICLE_TYPE],
		       bool  bConstantColorInSphere[_MAX_PARTICLE_TYPE],
		       float (*pPos)[3], float *pRadius);

    virtual ExitCode Entry();
    
    bool *mpbFinishedFlagLink;
    char *mpBuf;
    int mInit, mFin, mStep;
    unsigned char *mType;
    int *mDrawTypeList;
    int *mRadiusTypeList;
    float *mRadiusFactorList;
    bool  *mbConstantColorInSphere;
    float (*mpPos)[3];
    float *mpRadius;
		       
	
};

class PovRayCtrl_cls : public DialogControl_cls, public PovSerialize_cls
{ 
 public:
    PovRayCtrl_cls(FrameMain_cls *pCParentFrameLink, wxString strBitmap);
    ~PovRayCtrl_cls();
    
    static PovRayCtrl_cls *mspCThis;
    wxString mstrBitmap;

    PovRayGUI_cls*  mpCGUI;
    FrameMain_cls*  mpCParentFrameLink;
    TimeCtrl_cls*   GetPCTimeCtrl();

    bool              mbPovRayThreadFinished;
    ThreadPovRay_cls *mpCThreadPovRay;

    std::vector < wxString > mVstrAnimCommand;

    void InitGUI(wxWindow *pC);
    void SetToDefault();
    void UpdateItems(bool bForce = FALSE);
    
    int GetRegionType();
    int GetRegionInit();
    int GetRegionFin();
    void IncrementAnim();
    wxString GetFileStr(unsigned int type, unsigned int stereoId,
			int index);
    wxString GetWorkingDirWithSep(bool bForceAbs = FALSE);

    static wxString GetStrPovTextureFinish
	(int type, float roughness, wxString strOptions);
    static wxString GetStrPovTextureFastColor
	(wxString strTab, float pClrMat[3][4], float roughness,
	 wxString strFinishOptions, bool bFilterAlpha);
    static wxString GetStrPovTextureGLColor
	(wxString strTab, float pClrMat[3][4], float roughness,
	 wxString strFinishOptions, bool bFilterAlpha);
    static wxString GetStrPovTextureColor
	(wxString strTab, float pClrMat[3][4], float roughness, 
	 wxString strFinishOptions, bool bFilterAlpha = FALSE);
    static wxString GetStrPovPhotonSetting
	(double alpha, double reflectMin, double reflectMax, bool bKillTarget);
    wxString GetPovMacroSphereVariousColor
	(int index, float pClrMat[3][4], bool bConstantRadius, 
	 float radius, float roughness);
    wxString GetPovMacroSphereConstColor
	(int index, float pClrMat[3][4], bool bConstantRadius, 
	 float radius, float roughness);
    
    void GetPovStringParticle(wxString &strPData,
			      PDataBlock_cls *pCDataBlock, 
			      GLAttribute_cls **ppCGLAttribute,
			      double dT);
    wxString GetStrPovParticleObject(wxString strTag);
    wxString GetStrPovGlobal();
    wxString GetStrPovCameraAngle(float angle, bool bFishEye);
    wxString GetStrPovCamera(PositionAndDirection_cls CCameraPos, 
			     Position_cls CAnchorPos,
			     float angle, bool bFishEye, bool bOrthographic,
			     StereoGram_cls *pCStereogram);
    wxString GetPovRayCommand(wxString strFile);
    //wxString GetStrPovIni(wxString strFile, int jpegCompress);
    wxString GetStrRelativePath(wxString strPath);
    
    wxString GetStrPovIni(wxString& strIni, wxString strFile, int jpegCompress);
    bool  OutputPovRayIni(wxString strFile);
    bool  RenderByPovSnap(unsigned int stereo);

    bool  StartExportPovAnim(int type);
    void  AbortExportPovAnim();

    void  PushBackCommands(unsigned int stereo, unsigned int frame);
    void  StartRenderPovAnim();
    void  ProcessRenderPovAnim();
    void  AbortRenderPovAnim();
    
    int  mTabDepth;
    //bool mbInstanceStatus;
    wxString mstrInstanceTranslate;
    
    /*
    bool mbRenderingTwoSnap;
    */
    bool mbExportingAnim, mbRenderingAnim;
    
    /*
    int  mCurrentAnim;
    ThreadPovRay *mpCThreadPovRayAnim;
    std::vector < wxString > mVstrAnimCommand;
    std::vector < wxString > mVstrAnimPostSceneCommand;
    */

    void HideAndShow();
    static void HookRadiocity(void *pArg, void *pEvent);
    static void HookPhotons(void *pArg, void *pEvent);
    static void HookFocalBlur(void *pArg, void *pEvent);
    static void HookBG(void *pArg, void *pEvent);
    static void HookPovrayVersion(void *pArg, void *pEvent);
    static void EnableDisablePovrayVersion(void *pCDLG, void *pDummy);
    
    //--

    ConfItemList_cls mCConfListPovPath;
    ConfItemSTR_cls mCConfPathPovray;
    ConfItemSTR_cls mCConfWorkingDir;
    ConfItemSTR_cls mCConfSnapFile;
    ConfItemSTR_cls mCConfAnimFile;
    
    ConfItemBOOL_cls  mCbAbsolutePath;

    ConfItemList_cls mCConfListPovParam;
    //ConfItemUINT_cls  mCRegionType;
    //ConfItemUINT_cls  mCCameraType;
    ConfItemUINT_cls   mCShowTweaks;
    ConfItemBOOL_cls   mCbUseCommonSceneFile;
    
    int mSnapNumber;
    int mAnimNumber;//, mAnimNumberInit;
    int mCommonPDataStatus;
    int mCommonBGStatus;

    PovAdvanced_cls  *mpCAdvanced;
    PovFocalBlur_cls *mpCFocalBlur;
    PovFormat_cls    *mpCFormat;
    PovRadiosity_cls *mpCRadiosity;
    PovRender_cls    *mpCRender;
    PovPhotons_cls   *mpCPhotons;
    
    void ReadStrLine(wxString& strLine);
    void Serialize(wxString &strConf);
    wxString SerializeForConfigure();

    /*
    wxString mstrPostSceneCommand;
    void CheckIllegalFileName();
        
    
    AtoiSpinCtrl_cls *mpCSnapNumber;


    PovFocalBlur_cls *mpCFocalBlur;
    PovBG_cls        *mpCBG;

    bool mbPovOutForRender;

    int   GetStereoId(int index);
    int   GetStereoTmpFileIOMode(int index);
    
    int   GetCameraGeometry();
    int   GetCameraIndex();
    double GetVAngle();
    void CoupleToStereogram();
    
    wxString InsertTab(wxString strLines);

    wxString GetPovStrCommon();
			
    wxString SerializeCameraGeometryType();
    wxString SerializeRight();
    wxString GetPovStrCamInfo(PositionAndDirection_cls CCamPos,
			      wxString strCamPos, double doubleIndex);

    wxString GetPovStrIniFile(std::wstring strPovFile = L"povout",
			      int jpegCompress = 85);
    static wxString GetPovStrPhotonSetting
	(double alpha, double reflectMin, double reflectMax,
	 bool bKillTarget = FALSE);

    wxString GetPovRayCommand(wxString filename);
    void  PushBackCommands(unsigned int stereo, unsigned int frame);

    void  WaitingForExport(bool bFlag);

    float GetRadianForVisCheck();
    std::wstring msAFormat[_POV_FORMAT_NUMBER];

    double mSplit;
    float mTreeDepthParam[5];
        
    ConfItemList_cls      mCConfListDisp;
    ConfItemUINT_cls      mCPovRayVersion;
    ConfItemFLOAT_cls     mCFileGamma, mCDisplayGamma;
    ConfItemFLOAT_cls     mCAssumedGamma;
    ConfItemUINT_cls      mCWorkThread;
    ConfItemUINT_cls      mCRenderBlockSize;
    ConfItemUINT_cls      mCBoundingMethod;
    ConfItemUINT_cls      mCBufferSize;
    ConfItemINT_cls       mCSSSQuality1, mCSSSQuality2;
    
    ConfItemINT_cls       mCStereoEye;
    ConfItemUINT_cls      mCRemoveFile;
    */

    /*

    ConfItemBOOL_cls      mCbSubsetRow;
    ConfItemINT_cls       mCSubsetRowStart, mCSubsetRowEnd;
    
    ConfItemList_cls  mCConfListFog;
    ConfItemBOOL_cls  mCbWinePov;
*/

    
};

#endif
