// "clsCameraPathGUI.h"
// (setq c-basic-offset 4)

#include "clsCameraPathGUI.h"
#include "clsCameraPath.h"

#include "../GLATTRIB/clsColorCurve.h"
#include "../GLCANVAS/clsMainGLCanvas.h"
extern MainGLCanvas_cls* GpCurrentGLPanel;
#include "../GLCANVAS/NAVIGATORS/clsTagObject.h"
#include "../UNDO/clsUndoManager.h"

#include "../../clsFrameMain.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsHookCtrl.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#include "../COMMON/CommonDefinitions.h"

BEGIN_EVENT_TABLE(CameraPathGUI_cls, wxDialog)
EVT_NOTEBOOK_PAGE_CHANGED(ID_BOOK_VIEW, CameraPathGUI_cls::OnNotebookChanged)
EVT_BUTTON    (btnViewDefault, CameraPathGUI_cls::OnButtonViewDefault)
END_EVENT_TABLE()

CameraPathGUI_cls::
CameraPathGUI_cls(CameraPathCtrl_cls* pCCtrl,
		  wxWindow *pC, wxPoint pos, wxSize size) :
wxDialogEXKey(pC, wxID_ANY, _("View and Camera Config"), 
	      pos, size, ART(pCCtrl->mstrBitmap)) {
    mspCThis = this;
    mpCCtrl = pCCtrl;
    
    wxBoxSizer *pBoxSizerSrc     = new wxBoxSizer(wxVERTICAL);
    //wxBoxSizer *pBoxSizerView    = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerAdv     = new wxBoxSizer(wxVERTICAL);
    wxBoxSizer *pBoxSizerPath    = new wxBoxSizer(wxVERTICAL);

    mpCBook1 = new wxNotebook (this, ID_BOOK_VIEW);
    
    for (int i = 0; i < 5; i++) SetMNumView(i, mpCBook1);
    SetMNumAdv (pBoxSizerAdv);
    SetMNumPath(pBoxSizerPath);

    mpCBook1->AddPage(mpCMNumView[0], wxT("View0"));
    mpCBook1->AddPage(mpCMNumView[1], wxT("View1"));
    mpCBook1->AddPage(mpCMNumView[2], wxT("View2"));
    mpCBook1->AddPage(mpCMNumView[3], wxT("View3"));
    mpCBook1->AddPage(mpCMNumView[4], wxT("Extra"));

    pBoxSizerSrc->Add(mpCBook1, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerAdv, 0, wxEXPAND);
    pBoxSizerSrc->Add(pBoxSizerPath, 0, wxEXPAND);
    
    SetSizerAndFit(pBoxSizerSrc);
    
    int minimumWidth = GetClientSize().GetWidth() - 8;
    //minimumWidth = mpCMNumView->GetPaddingWidth(minimumWidth);
    //minimumWidth = mpCMNumAdv->GetPaddingWidth(minimumWidth);
    //minimumWidth = mpCMNumPath->GetPaddingWidth(minimumWidth);
    //mpCMNumView->SetPaddingWidth(minimumWidth);
    mpCMNumAdv->SetPaddingWidth(minimumWidth);
    mpCMNumPath->SetPaddingWidth(minimumWidth);

    mpCMNumAdv->TogglePanelOpenStatus(1);
    mpCMNumPath->TogglePanelOpenStatus(1);

    ResizeDialog();
}

void CameraPathGUI_cls::
SetMNumView(int i, wxWindow *pCWindow){
    MultiInputPanel_cls* pCurr = 
	MultiInputPanel_cls::Initialize(&mpCMNumView[i], pCWindow);

    MainGLCanvas_cls *pCGLCanvasLink = 
	mpCCtrl->mpCParentFrame->GetPCGLCanvas(i);

    pCurr->mUpdateBit = 
	_COMMAND_UPDATE_DISPLAY_CAMCONF | _COMMAND_UPDATE_DISPLAY_GL |
	_COMMAND_UPDATE_DISPLAY_IMGSEQ;
    
    int boxWidth = 2;
#ifdef WIN32
    pCurr->mTextWidth = 90;
#else
    pCurr->mTextWidth = 100;
#endif
    pCurr->mItemWidth = 80;
    pCurr->mComboWidth = 150;
    
    pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Geometry"));

    wxString strGeometry[4] = {
	_("Perspective"), wxT("Orthographic"), _("Fisheye"), _("Panoramic")};
    int typeSelect = 4;
    pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &pCGLCanvasLink->mCGLGeometry.mBody,
	 typeSelect, strGeometry);
    
    pCurr->CombinePointerButton(btnViewDefault, wxT("Default"));
    pCurr->GetLastItem().mAlign = _ITEM_ALIGN_RIGHT;
	
    boxWidth = 3;
    int indexTranslate = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Translate"));
    pCurr->CombinePointer(_ITEM_DOUBLE, &pCGLCanvasLink->mCCameraPos.mPos[X]);
    pCurr->CombinePointer(_ITEM_DOUBLE, &pCGLCanvasLink->mCCameraPos.mPos[Y]);
    pCurr->CombinePointer(_ITEM_DOUBLE, &pCGLCanvasLink->mCCameraPos.mPos[Z]);

    int indexRotation = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Rotate"));
    pCurr->CombinePointer(_ITEM_DOUBLE, &pCGLCanvasLink->mCCameraPos.mYPR[X]);
    pCurr->CombinePointer(_ITEM_DOUBLE, &pCGLCanvasLink->mCCameraPos.mYPR[Y]);
    pCurr->CombinePointer(_ITEM_DOUBLE, &pCGLCanvasLink->mCCameraPos.mYPR[Z]);

    int indexAnchor = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Pivot"));
    pCurr->CombinePointer(_ITEM_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[X]);
    pCurr->CombinePointer(_ITEM_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[Y]);
    pCurr->CombinePointer(_ITEM_DOUBLE, &MainGLCanvas_cls::msCAnchorPos.mPos[Z]);
    
    /*
    boxWidth = 2;
    mIndexPitchAngle = pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
					     _("Pitch Angle"));
    pCurr->CombinePointer
	(_GP_DOUBLE_SLIDER, &mspCParent->GetActivePath()->mpCStereoGramLinked->mCPitchAngle.mBody);
    pCurr->SetRange(-180, 180, _MN_LASTITEM, 1);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mpCStereoGramLinked->mCPitchAngle.mBody);
    */
    
    pCurr->mItemWidth = -1;
    boxWidth = 2;
    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("View Angle"));
    pCurr->CombinePointer
	(_ITEM_FLOAT_SLIDER, &pCGLCanvasLink->mCGLPersAngle.mBody);
    pCurr->SetRange(_ITEM_LASTITEM, 1, 180);
    pCurr->CombinePointer(_ITEM_FLOAT, &pCGLCanvasLink->mCGLPersAngle.mBody);
    
    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("DepthMIN/MAX"));
    pCurr->CombinePointer
	(_ITEM_FLOAT, &pCGLCanvasLink->mCGLDepthMin.mBody);
    pCurr->CombinePointer
	(_ITEM_FLOAT, &pCGLCanvasLink->mCGLDepthMax.mBody);

    mIndexDistortion = 
	pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth);
    pCurr->CombinePointer
	(_ITEM_BOOL, &pCGLCanvasLink->mCbDistortion.mbBody, _("Distortion"));
    pCurr->CombinePointer
	(_ITEM_BOOL, &pCGLCanvasLink->mCbFisheyeGrid.mbBody, _("Fisheye Grid"));

    pCurr->FinishSetting();

    pCurr->SetColorForXYZ(indexTranslate+1);
    pCurr->SetColorForXYZ(indexRotation+1);
    pCurr->SetColorForXYZ(indexAnchor+1);

    pCurr->mpCHookCtrl->SetFuncHook(FuncHookShow, _HOOK_SHOW);
    pCurr->mpCHookCtrl->SetFuncHook(FuncHookAssign, _HOOK_ASSIGN);
    pCurr->mpCHookCtrl->SetFuncHook(HookHideAndShowView, _HOOK_COMBOBOX);
    pCurr->mpCHookCtrl->SetFuncHook(HookSyncGLCanvasCtrl, _HOOK_COMBOBOX);

    //pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_COMBOBOX);
    
    HideAndShowView(i);
}

void CameraPathGUI_cls::
SetMNumAdv(wxBoxSizer *pBoxSizer){
    MultiInputPanel_cls* pCurr =
	MultiInputPanel_cls::Initialize(&mpCMNumAdv, this);

    pCurr->mItemWidth = -1;
    pCurr->CombinePointerButton(btnPanelHide, _("[+]View Advanced Setting"));
    
#ifdef WIN32
    pCurr->mTextWidth = 90;
#else
    pCurr->mTextWidth = 100;
#endif
    pCurr->mComboWidth = 120;
    
    int boxWidth = 3;
    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Colors"));
    ColorCurveBtn_cls *pCBtnBG = new ColorCurveBtn_cls
	(pCurr, MainGLCanvas_cls::mspCClrCurveBG, TRUE, FALSE);
    ColorCurveBtn_cls *pCBtnOut = new ColorCurveBtn_cls
	(pCurr, MainGLCanvas_cls::mspCClrCurveOut, TRUE, FALSE);
    ColorCurveBtn_cls *pCBtnTag = new ColorCurveBtn_cls
	(pCurr, TagObjectCtrl_cls::mspCClrCurveTag, TRUE, FALSE);
    
    pCurr->CombinePointer(_ITEM_WXPANEL, pCBtnBG);
    pCurr->CombinePointer(_ITEM_WXPANEL, pCBtnOut);
    pCurr->CombinePointer(_ITEM_WXPANEL, pCBtnTag);
    
    pCurr->mItemWidth = 80;
    pCurr->mComboWidth = 80;
    boxWidth = 3;
    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Grid and Width"));
    wxString strListGrid[3] = {
	_("Grid"), _("Grid+Circle"), _("Polar")};
    mIndexGrid = pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &GpCurrentGLPanel->mCGridType.mBody, 
	 3, strListGrid);
    
    pCurr->mItemWidth = 80;
    pCurr->CombinePointer(_ITEM_FLOAT, &MainGLCanvas_cls::msCLineWidth.mBody);
    pCurr->CombinePointer(_ITEM_FLOAT_SLIDER, &MainGLCanvas_cls::msCLineWidth.mBody);
    pCurr->SetRange(_ITEM_LASTITEM, 1, 10);
    
    pCurr->mItemWidth = -1;
    wxString listFBOSize[7] = {wxT("256x256"), wxT("512x512"), wxT("1024x1024"), wxT("2048x2048"), wxT("4096x4096"), wxT("8192x8192"), wxT("16384x16384")};
    mIndexFBOSize = pCurr->CombinePointerList
	(_ITEM_UINT_LIST, 
	 &MainGLCanvas_cls::msCFBOSizeIndex.mBody,
	 7, listFBOSize, _("Buffer Size"));
    
    wxString listTagFont[5] = {wxT("HELVETICA_10"), wxT("HELVETICA_12"), wxT("HELVETICA_18"), wxT("TIMES_ROMAN_10"), wxT("TIMES_ROMAN_24")};
    mIndexTagFont = pCurr->CombinePointerList
	(_ITEM_UINT_LIST, 
	 &TagObjectCtrl_cls::msCTagFontIndex.mBody,
	 5, listTagFont, _("Tag Font"));

    /*
    boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth);
    mIndexShowOtherCam = pCurr->CombinePointer
	(_GP_BOOL, &GpCurrentGLPanel->mCbShowOtherCam.mbBody, _("view point"));
    mIndexXOR = 
	pCurr->CombinePointer(_GP_BOOL, &GpCurrentGLPanel->mCbXOR.mbBody, _("Draw item in XOR"));
    */
#ifdef WIN32
    pCurr->mMinimumWidth = 280;
#else
    pCurr->mMinimumWidth = 320;
#endif
    pCurr->FinishSetting(pBoxSizer);

    pCurr->mpCHookCtrl->SetFuncHook(FuncHookShow, _HOOK_SHOW);
    pCurr->mpCHookCtrl->SetFuncHook(FuncHookAssign, _HOOK_ASSIGN);
    pCurr->mpCHookCtrl->SetFuncHook
	(FuncHookHideButton, _HOOK_BUTTON, pCurr->mVCItemList[0].mId);
}

void CameraPathGUI_cls::
FuncHookHideButton(void *pPanel, void *pArg){
    MultiInputPanel_cls* pCurr = ((MultiInputPanel_cls*)pPanel);
    pCurr->TogglePanelOpenStatus(1);
    mspCThis->ResizeDialog();
}

void CameraPathGUI_cls::
SetMNumPath(wxBoxSizer *pBoxSizer){
    MultiInputPanel_cls* pCurr =
	MultiInputPanel_cls::Initialize(&mpCMNumPath, this);

    pCurr->mUpdateBit = 
	_COMMAND_UPDATE_DISPLAY_CAMCONF | _COMMAND_UPDATE_DISPLAY_GL |
	_COMMAND_UPDATE_DISPLAY_IMGSEQ;

    pCurr->mItemWidth = -1;
    pCurr->CombinePointerButton(btnPanelHide, _("[+]Camera Setting"));
    
    int boxWidth = 3;
    //pCurr->mDefaultButton = _MN_NO_BUTTONS;
#ifdef WIN32
    pCurr->mTextWidth = 90;
#else
    pCurr->mTextWidth = 100;
    //pCurr->mComboWidth = 108;
#endif
    pCurr->mItemWidth = 80;
    //pCurr->mItemWidth = -1;
    pCurr->mComboWidth = 100;



    /*
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, _("Active Cam"));
    wxString strListView[1] = {wxT("Cam0")};
    mIndexActivePath = pCurr->CombinePointer
	(_GP_UNSIGNED_INT_LIST, &mspCParent->mActivePath,
	 _("Active Path"), TRUE, 1, strListView, 0);
    pCurr->CombinePointerButton(buttonCtrlAddPath, wxT("Add"));
    pCurr->CombinePointerButton(buttonCtrlDeletePath, wxT("Delete"));
    */

    pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("Camera Geom."));
    
    wxString strGeometry[4] = {
	_("Perspective"), wxT("Orthographic"), _("Fisheye"), _("Panoramic")};
    int typeSelect = 4/* - (FBO_cls::msbSupported == FALSE) * 2*/;
    pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &mpCCtrl->mpCCameraPathLinked->mCCamGeometry.mBody,
	 typeSelect, strGeometry);
    
    pCurr->mItemWidth = 100;
    pCurr->CombinePointer(_ITEM_BOOL,
			  &mpCCtrl->mpCCameraPathLinked->mCbLookAnchor.mbBody, wxT("Look Anchor"));
    
    pCurr->mItemWidth = 80;
    mIndexAnglePath = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("View Angle"));
    pCurr->CombinePointer
	(_ITEM_DOUBLE_SLIDER, &mpCCtrl->mpCCameraPathLinked->mCVAngle.mBody);
    pCurr->SetRange(_ITEM_LASTITEM, 1, 180);
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCCameraPathLinked->mCVAngle.mBody);

    pCurr->mItemWidth = 80;
    mIndexFishEyeOffset = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Fisheye Offset"));
    pCurr->CombinePointer
	(_ITEM_FLOAT_SLIDER, &mpCCtrl->mpCCameraPathLinked->mCFishEyeOffset.mBody);
    pCurr->SetRange(_ITEM_LASTITEM, -90, 90);
    pCurr->CombinePointer(_ITEM_FLOAT, &mpCCtrl->mpCCameraPathLinked->mCFishEyeOffset.mBody);
    
    int indexTranslate = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Translae"));
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCCameraPathLinked->mpCTranslate->mVCParam[X].mInterpolatedValue);
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCCameraPathLinked->mpCTranslate->mVCParam[Y].mInterpolatedValue);
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCCameraPathLinked->mpCTranslate->mVCParam[Z].mInterpolatedValue);

    int indexRotation = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Rotation"));
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCCameraPathLinked->mpCRotation->mVCParam[X].mInterpolatedValue);
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCCameraPathLinked->mpCRotation->mVCParam[Y].mInterpolatedValue);
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCCameraPathLinked->mpCRotation->mVCParam[Z].mInterpolatedValue);

    int indexAnchor = pCurr->CombinePointer(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, wxT("Anchor"));
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCAnchorPathLinked->mpCTranslate->mVCParam[X].mInterpolatedValue);
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCAnchorPathLinked->mpCTranslate->mVCParam[Y].mInterpolatedValue);
    pCurr->CombinePointer(_ITEM_DOUBLE, &mpCCtrl->mpCAnchorPathLinked->mpCTranslate->mVCParam[Z].mInterpolatedValue);
    
    /*
    boxWidth = 1;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Point Sky"));
    mIndexPointSky = pCurr->CombinePointer
	(_GP_BOOL, &mspCParent->GetActivePath()->mCbSky.mbBody);
    //pCurr->CombinePointer(_GP_STATIC_TEXT, NULL);
    //pCurr->CombinePointer(_GP_STATIC_TEXT, NULL);
    
    boxWidth = 3;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, wxT("Sky Vector"));
    mIndexSky = pCurr->CombinePointer
	(_GP_DOUBLE, &mspCParent->GetActivePath()->mCSky.mpInterpolatedValue[0]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCSky.mpInterpolatedValue[1]);
    pCurr->CombinePointer(_GP_DOUBLE, &mspCParent->GetActivePath()->mCSky.mpInterpolatedValue[2]);
    */
    
#ifdef WIN32
    pCurr->mMinimumWidth = 280;
#else
    pCurr->mMinimumWidth = 320;
#endif
    pCurr->FinishSetting(pBoxSizer);

    pCurr->SetColorForXYZ(indexTranslate+1);
    pCurr->SetColorForXYZ(indexRotation+1);
    pCurr->SetColorForXYZ(indexAnchor+1);

    pCurr->mpCHookCtrl->SetFuncHook(FuncHookShow, _HOOK_SHOW);
    pCurr->mpCHookCtrl->SetFuncHook(FuncHookAssign, _HOOK_ASSIGN);
    pCurr->mpCHookCtrl->SetFuncHook
	(FuncHookHideButton, _HOOK_BUTTON, pCurr->mVCItemList[0].mId);
    /*
    pCurr->SetFuncHook(FrameMain_cls::FuncHookParamChanged, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookAssign, _MN_HOOK_ASSIGN);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_COMBOBOX);
    pCurr->SetFuncHook(&FuncHookRefresh, _MN_HOOK_CHECKBOX);
    pCurr->SetFuncHook(&FuncHookButton, _MN_HOOK_BUTTON, _MN_HOOK_ID_ANY);
    */
}

void CameraPathGUI_cls::
ReconnectView(){
    mpCBook1->SetSelection(GpCurrentGLPanel->mSelfIndex);

    mpCCtrl->mActiveView = GpCurrentGLPanel->mSelfIndex; 
    
    // Adṽ̕N͕Kv FIX_ME !!

    /*
    pCurr->UpdateValues();
    */

    HideAndShowAll();
}

void CameraPathGUI_cls::
ReconnectViewAdv(){
    MultiInputPanel_cls*  pCurr = mpCMNumAdv;
    MainGLCanvas_cls *pCGLCanvasLink = GetPCActiveGLCanvas();

    pCurr->mVCItemList[mIndexGrid].
	CombinePointer(_ITEM_UINT_LIST, &pCGLCanvasLink->mCGridType.mBody);

    pCurr->UpdateValues();
}

MainGLCanvas_cls* CameraPathGUI_cls::
GetPCActiveGLCanvas(){
    return mpCCtrl->mpCParentFrame->
	GetPCGLCanvas(mpCBook1->GetSelection());
}

bool CameraPathGUI_cls::
HideAndShowView(int i){
    static int sbFishEye[5] = {-1, -1, -1, -1, -1};
    MultiInputPanel_cls* pCurr = (MultiInputPanel_cls*)mpCMNumView[i];
    MainGLCanvas_cls *pCGLCanvasLink = 
	mpCCtrl->mpCParentFrame->GetPCGLCanvas(i);
    
    bool bChange = FALSE;
    bool bFishEye = 
	(pCGLCanvasLink->mCGLGeometry.mBody == _GEOM_FISHEYE);
    if (sbFishEye[i] != (int)bFishEye) bChange = TRUE;
    sbFishEye[i] = (int)bFishEye;

    if (bChange == FALSE) return FALSE;
    pCurr->ShowItem(mIndexDistortion, bFishEye);
    pCurr->Layout(); 
    return TRUE;
}

void CameraPathGUI_cls::
HideAndShowAll(){
    bool bChange = FALSE;
    for (int i = 0; i < _GLCANVAS_INDEX_MAX; i++) 
	bChange |= HideAndShowView(i);
    if (bChange == FALSE) return;
    
    mpCBook1->Layout();

    int minSize = 0;
    for (int i = 0; i < mpCBook1->GetPageCount(); i++){
	wxSize size = mpCMNumView[i]->GetSizer()->GetMinSize();
			 
	if (minSize < size.GetHeight())
	    minSize = size.GetHeight();
    }
    mpCBook1->SetSizeHints(-1, minSize + 32);
    GetSizer()->Layout();
    
    ResizeDialog();
}

void CameraPathGUI_cls::
OnNotebookChanged(wxNotebookEvent& event){
    //mspCParent->mActiveView = event.GetSelection();
    ReconnectViewAdv();
}

void CameraPathGUI_cls::
OnButtonViewDefault(wxCommandEvent& event){
    int i = mpCCtrl->mActiveView;
    MainGLCanvas_cls *pCGLCanvasLink = 
	mpCCtrl->mpCParentFrame->GetPCGLCanvas(i);
    int type = pCGLCanvasLink->mCDisplayMode.mBody;
    switch(type){
    case 0:{
	pCGLCanvasLink->SetCameraToDefaultPos();
	pCGLCanvasLink->mCCameraPos.UpdateYPR();
	break; }
    }
    UpdateValuesAll();
    
    /*
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshGL);
    wxGetTopLevelParent(mspCThis->mpCMNumView[0])->
	GetParent()->AddPendingEvent(com);
    */
    SendEventToTopParent
	(mpCMNumView[0], wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshGL);
}

void CameraPathGUI_cls::
FuncHookShow(void *pPanel, void *pArg){
    mspCThis->ResizeDialog();
}

void CameraPathGUI_cls::
FuncHookAssign(void *pPanel, void *pArg){
    mspCThis->mpCCtrl->mpCCameraPathLinked->mpCTranslate->
	SetKeyFrameIfChanged();
    mspCThis->mpCCtrl->mpCCameraPathLinked->mpCRotation->
	SetKeyFrameIfChanged();
    mspCThis->mpCCtrl->mpCAnchorPathLinked->mpCTranslate->
	SetKeyFrameIfChanged();

    mspCThis->SendEventToTopParent
	(mspCThis->mpCMNumView[0], wxEVT_COMMAND_MENU_SELECTED,
	 menuEditRefreshGL);
    /*
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshGL);
    wxGetTopLevelParent(mspCThis->mpCMNumView[0])->
	GetParent()->AddPendingEvent(com);
    */
}

void CameraPathGUI_cls::
HookHideAndShowView(void *pPanel, void *pArg){
    mspCThis->HideAndShowAll();
}

void CameraPathGUI_cls::
HookSyncGLCanvasCtrl(void *pPanel, void *pArg){
    mspCThis->SendEventToTopParent
	(mspCThis->mpCMNumView[0], wxEVT_COMMAND_MENU_SELECTED,
	 menuEditSyncGLCtrl);
    /*
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditSyncGLCtrl);
    wxGetTopLevelParent(mspCThis->mpCMNumView[0])->
	GetParent()->AddPendingEvent(com);
    */
}

void CameraPathGUI_cls::
UpdateValuesAll(){
    HideAndShowAll();
    for (int i = 0; i < _GLCANVAS_INDEX_MAX; i++){
	mpCMNumView[i]->UpdateValues();
    }
    mpCMNumAdv->UpdateValues();
    mpCMNumPath->UpdateValues();
}

// --------

CameraPathCtrl_cls::
CameraPathCtrl_cls(FrameMain_cls* pCParentFrame, 
		   StereoGram_cls* pCStereoGram, 
		   CameraPath_cls* pCCameraPath,
		   AnchorPath_cls* pCAnchorPath, wxString strBitmap){
    mpCGUI = NULL;
    mpCParentFrame = pCParentFrame;
    mpCCameraPathLinked = pCCameraPath;
    mpCAnchorPathLinked = pCAnchorPath;
    mstrBitmap = strBitmap;
    mActiveView = 0;

    InitGUI(pCParentFrame);
}

void CameraPathCtrl_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new CameraPathGUI_cls(this, pC,
				   /*mCScreenRect.GetPosition(),*/
				   /*mCScreenRect.GetSize()*/
				   wxDP, wxDefaultSize);
    SetPDialog(mpCGUI);
}

void CameraPathCtrl_cls::
UpdateItems(bool bForce){
    if (mpCGUI == NULL) return;

    mpCGUI->UpdateValuesAll();

    /*
    double doubleIndex = GpCurrentGLPanel->mpCParentFrameLink->
	GetPCTimeCtrl()->GetDoubleIndex();
    GetActivePath()->mCPosition.GetPInterpolatedValue(doubleIndex);
    mpCGUI->mpCMNumPath->UpdateValues();
    mpCGUI->mpCMNumView->UpdateValues();
    mpCGUI->mpCMNumAdv->UpdateValues();
    */
}

CameraPathGUI_cls* CameraPathGUI_cls::mspCThis = 0;
