// -*- c -*-
// "clsLightSource.h"

#include "clsLightSource.h"

#include "../POVRAY/clsPovRay.h"
#include "../GLCANVAS/clsMainGLCanvas.h"

#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "../UNDO/clsUndoItemConstParam.h"

#include "../COMMON/CommonDefinitions.h"

#define LOOP_RGBA for (int rgba = 0; rgba < 4; rgba++)

LightSRC_cls::
LightSRC_cls(int selfIndex, wxString strTag) : SceneObj_cls(strTag) {
    
    mSelfIndex = selfIndex;
    mObjType   = _SCENEOBJ_LIGHT;
    mObjIndex  = _SCENEOBJ_IDBASE_LIGHT + selfIndex;
    mUpdateBit = _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_LIGHT;
    
    if (mSelfIndex == 1){
	SetInitValueTrans(-1.0, -1.0, -1.41421356);
	SetInitValueRot  (0, -45, 45);
    }
    else{
	SetInitValueTrans(1.0, 1.0, 1.41421356);
	SetInitValueRot  (0, 45, 225);
    }

    mCConfList.Clear();
    mCConfList.Add(&mCLightType, strTag + wxT("Type="), wxT("3"));
    mCConfList.Add(&mCSpotAngle, strTag + wxT("SpotAngle="), wxT("35"));
    mCConfList.Add(&mCFadeType, strTag  + wxT("Fade="), wxT("0"));
    mCConfList.Add(&mCFadeDistance, strTag + wxT("FadeDistance="), wxT("1"));
    mCConfList.Add(&mCbCastShadow, strTag + wxT("CastShadow="), wxT("1"));
    mCConfList.Add(&mCbAreaLight, strTag + wxT("LightArea="), wxT("0"));
    mCConfList.Add(&mCAreaSize, strTag + wxT("LightAreaSize="), wxT("0.2"));
    mCConfList.Add(&mCAreaQuality, strTag + wxT("LightAreaQuality="), wxT("5"));
    mCConfList.Add(&mCTarget, strTag + wxT("Target="), wxT("0"));
    mCConfList.Add(&mCbVisible, strTag + wxT("Visible="), wxT("1"));

    wxString listTag[1] = {wxT("Strength")};
    int updateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL |
	_COMMAND_UPDATE_DISPLAY_LIGHT ;
    mpCLightStrength = new ParamGroup_cls(strTag + wxT(" Strength") , 1, listTag, updateBit, FALSE);
    
    if (mSelfIndex == 1){
	double pLightInit[1] = {0.1};
	mpCLightStrength->SetKeyFrame(pLightInit, TRUE);
    }
    else{
	double pLightInit[1] = {1};
	mpCLightStrength->SetKeyFrame(pLightInit, TRUE);
    }
};

LightSRC_cls::
~LightSRC_cls(){
    delete mpCLightStrength;
}

void LightSRC_cls::
CopyFromP(LightSRC_cls *pCFrom){
    CopyFromP_SceneObj(pCFrom);

    int newValue = pCFrom->mCLightType.mBody;
    int oldValue = mCLightType.mBody;
    GCUndoList.PushNewCommand
	(new CommandSetValue_cls(&mCLightType.mBody, newValue),
	 new CommandSetValue_cls(&mCLightType.mBody, oldValue));
    GCUndoList.ExecCurrent();

    newValue = pCFrom->mCFadeType.mBody;
    oldValue = mCFadeType.mBody;
    GCUndoList.PushNewCommand
	(new CommandSetValue_cls(&mCFadeType.mBody, newValue),
	 new CommandSetValue_cls(&mCFadeType.mBody, oldValue));
    GCUndoList.ExecCurrent();

    newValue = pCFrom->mCSpotAngle.mBody;
    oldValue = mCSpotAngle.mBody;
    GCUndoList.PushNewCommand
	(new CommandSetValue_cls(&mCSpotAngle.mBody, newValue, mUpdateBit),
	 new CommandSetValue_cls(&mCSpotAngle.mBody, oldValue, mUpdateBit));
    GCUndoList.ExecCurrent();

    mpCLightStrength->CopyFromP(pCFrom->mpCLightStrength);
}

void LightSRC_cls::
GetLightInfo(double doubleIndex, float pLightPos[4], float pLightPhiTheta[2],
	     bool bSV, SceneObj_cls *pCScene){
    std::vector < double > VPos;
    std::vector < double > VRot = 
	mpCRotation->GetInterpolatedValue(doubleIndex, 0);
    
    if (bSV == TRUE){
	VPos = GetVDirectionVector(doubleIndex);
	/*
	VPos.push_back(cos(3.14159*VRot[1]/180) * cos(3.14159*(VRot[2]/180+1)));
	VPos.push_back(cos(3.14159*VRot[1]/180) * sin(3.14159*(VRot[2]/180+1)));
	VPos.push_back(sin(3.14159*VRot[1]/180));
	*/
	VPos.push_back(0);
    }
    else{
	VPos = mpCTranslate->GetInterpolatedValue(doubleIndex, 0);
	VPos.push_back(1);
    }

    for (int i = 0; i < 4; i++)	pLightPos[i] = VPos[i];

    if (pCScene != NULL){
	float fpos[3];
	LOOP_XYZ fpos[xyz] = VPos[xyz];
	pCScene->CalcRelativeVector(fpos, doubleIndex);
	LOOP_XYZ VPos[xyz] = fpos[xyz];
    }

    float phiRadi = VRot[2];
    float thetaRadi = VRot[1];

    thetaRadi = atan2(VPos[Z], sqrt(VPos[X]*VPos[X] + VPos[Y]*VPos[Y]));
    phiRadi = atan2(VPos[Y], VPos[X]);

    //pLightPhiTheta[0] = VRot[2];
    //pLightPhiTheta[1] = VRot[1];
    
    pLightPhiTheta[0] = phiRadi * 180 / 3.14159 + 180;
    pLightPhiTheta[1] = thetaRadi * 180 / 3.14149; 
}

void LightSRC_cls::
GetLightColor(double doubleIndex, float  pLightColor[4]){
    std::vector < double > VStr = 
	mpCLightStrength->GetInterpolatedValue(doubleIndex, 0);
    for (int rgb = 0; rgb < 3; rgb++) pLightColor[rgb] = VStr[0];
    pLightColor[3] = 0;
}

void LightSRC_cls::
GLSetAttenuation(int GLLightType){
    switch(mCFadeType.mBody){
    case _LIGHT_FADE_OFF:
	glLightf(GLLightType, GL_CONSTANT_ATTENUATION, 1);
	glLightf(GLLightType, GL_LINEAR_ATTENUATION, 0);
	glLightf(GLLightType, GL_QUADRATIC_ATTENUATION, 0);
	break;
    case _LIGHT_FADE_1: {
	float factor = 1.0 / mCFadeDistance.mBody;
	glLightf(GLLightType, GL_CONSTANT_ATTENUATION, 1);
	glLightf(GLLightType, GL_LINEAR_ATTENUATION, factor);
	glLightf(GLLightType, GL_QUADRATIC_ATTENUATION, 0);
    	break; }
    case _LIGHT_FADE_2: {
	float factor = 1.0 / mCFadeDistance.mBody / mCFadeDistance.mBody;
	glLightf(GLLightType, GL_CONSTANT_ATTENUATION, 1);
	glLightf(GLLightType, GL_LINEAR_ATTENUATION, 0);
	glLightf(GLLightType, GL_QUADRATIC_ATTENUATION, factor);
	break; }
    }
}

void LightSRC_cls::
GLSetSpotEffect(int GLLightType, double doubleIndex, float pos[4]){
    if (mCLightType.mBody != _LIGHT_SPOT){
	glLightf(GLLightType, GL_SPOT_CUTOFF, 180.0);
	return; 
    }
    
    float lightDir[4] = {0, 0, -1, 0};
    float lightZero[3] = {0, 0, 0};

    switch(mCTarget.mBody){
    case _SCENEOBJ_TARGET_NONE: 
    case _SCENEOBJ_TARGET_ORIGIN: {
	std::vector < double> VLightDir = GetVDirectionVector(doubleIndex);
	LOOP_XYZ lightDir[xyz] = -VLightDir[xyz];
	break; }
    case _SCENEOBJ_TARGET_ANCHOR:
	break;
    }
    
    glLightf(GLLightType, GL_SPOT_CUTOFF, mCSpotAngle.mBody);//
    glLightfv(GLLightType, GL_SPOT_DIRECTION, lightDir);
    glLightfv(GLLightType, GL_AMBIENT, lightZero);
    glLightf(GLLightType, GL_SPOT_EXPONENT, 0);//35*300.0f/pCLightParam->mCSpotAngle.mBody/pCLightParam->mCSpotAngle.mBody);
}

void LightSRC_cls::
GLSetLightSourceComp(int GLLightIndex, 
		     double doubleIndex, bool bAmbient, bool bDiffuse){
    if (mCbVisible.mbBody == FALSE) return;
    
    GLuint GLLightType;
    switch(GLLightIndex){
    case 0: GLLightType = GL_LIGHT0; break;
    case 1: GLLightType = GL_LIGHT1; break;
    case 2: GLLightType = GL_LIGHT2; break;
    case 3: GLLightType = GL_LIGHT3; break;
    case 4: GLLightType = GL_LIGHT4; break;
    case 5: GLLightType = GL_LIGHT5; break;
    case 6: GLLightType = GL_LIGHT6; break;
    }

    glEnable(GLLightType);

    float pos[4], phiTheta[2], clr[4] = {0, 0, 0, 0};
    GetLightInfo(doubleIndex, pos, phiTheta, (mCLightType.mBody == _LIGHT_PARALLEL), NULL);
    GetLightColor(doubleIndex, clr);
    if (mCFadeType.mBody != _LIGHT_FADE_OFF)
	for (int i = 0; i < 3; i++) clr[i] *= 2;

    glLightfv(GLLightType, GL_POSITION, pos);

    GLSetAttenuation(GLLightType);
    GLSetSpotEffect(GLLightType, doubleIndex, pos);

    GLfloat pZero[4] = {0, 0, 0, 1};
    GLfloat pOne[4] = {1, 1, 1, 1};
    if (bAmbient == TRUE)
	glLightfv(GLLightType, GL_AMBIENT , pOne);
    else
	glLightfv(GLLightType, GL_AMBIENT , pZero);

    if (bDiffuse == TRUE){
	glLightfv(GLLightType, GL_DIFFUSE , clr);
	glLightfv(GLLightType, GL_SPECULAR, clr);
    }
    else{
	glLightfv(GLLightType, GL_DIFFUSE , pZero);
	glLightfv(GLLightType, GL_SPECULAR, pZero);
    }
}


void LightSRC_cls::
GLSetLightSource(double doubleIndex, bool bEnable, int shadowMode){
    bEnable &= mCbVisible.mbBody;

    GLuint GLLightType;
    switch(mSelfIndex){
    case 0: GLLightType = GL_LIGHT0; break;
    case 1: GLLightType = GL_LIGHT1; break;
    case 2: GLLightType = GL_LIGHT2; break;
    case 3: GLLightType = GL_LIGHT3; break;
    case 4: GLLightType = GL_LIGHT4; break;
    case 5: GLLightType = GL_LIGHT5; break;
    case 6: GLLightType = GL_LIGHT6; break;
    }

    if (bEnable == FALSE){
	glDisable(GLLightType);
	GLfloat pZero[4] = {0, 0, 0, 0};
	GLfloat pOne[4] = {1, 1, 1, 0};
	glLightfv(GLLightType, GL_AMBIENT , pZero);
	glLightfv(GLLightType, GL_DIFFUSE , pZero);
	glLightfv(GLLightType, GL_SPECULAR, pZero);
	return;
    }

    glEnable(GLLightType);
    
    float pos[4], phiTheta[2], clr[4] = {0, 0, 0, 0};
    GetLightInfo(doubleIndex, pos, phiTheta, (mCLightType.mBody == _LIGHT_PARALLEL), NULL);
    GetLightColor(doubleIndex, clr);
    if (mCFadeType.mBody != _LIGHT_FADE_OFF)
	for (int i = 0; i < 3; i++) clr[i] *= 2;
    
    glLightfv(GLLightType, GL_POSITION, pos);

    GLSetAttenuation(GLLightType);
    GLSetSpotEffect(GLLightType, doubleIndex, pos);
    
    if ((mSelfIndex == 0) && (shadowMode != 0) && (mCbCastShadow.mbBody == TRUE))
	LOOP_RGBA clr[rgba] = 0;
    
    GLfloat lightColorAmbient[4] = {0, 0, 0, 0};
    if (mSelfIndex == 0) 
	for (int i = 0; i < 3; i++) lightColorAmbient[i] = 1.0f;
    
    glLightfv(GLLightType, GL_AMBIENT , lightColorAmbient);
    glLightfv(GLLightType, GL_DIFFUSE , clr);
    glLightfv(GLLightType, GL_SPECULAR, clr);
}

void LightSRC_cls::
GLDraw(bool bHittest){
    if (mCbVisible.mbBody == FALSE) return;

    glDisable(GL_LIGHTING);

    double matrix[16];
    mCCurrentObj.GetGLMatrix(matrix);
    GLfloat matrixB[16];
    glGetFloatv(GL_TRANSPOSE_MODELVIEW_MATRIX, matrixB);
    double scale = -matrixB[11];
    
    glEnable(GL_LINE_STIPPLE);
    glColor4f(0.7f, 0.7f, 0.2f, 1.0f);

    // ⏕
    if (bHittest == FALSE){
	glLineStipple(1,0xf0f0);
	glBegin(GL_LINES); 
	glVertex3d(mCCurrentObj.mPos[X], mCCurrentObj.mPos[Y], 0);
	glVertex3d(mCCurrentObj.mPos[X], mCCurrentObj.mPos[Y], mCCurrentObj.mPos[Z]);
	glEnd();

	if (mCLightType.mBody != _LIGHT_POINT){ 
	    glLineStipple(1,0xff00);
	    switch(mCTarget.mBody){
	    case _SCENEOBJ_TARGET_NONE:
		glPushMatrix();
		glMultMatrixd(matrix);
		glBegin(GL_LINES); // 
		glVertex3d(0, 0, 0);
		glVertex3d(0.2*scale, 0, 0);
		glEnd();
		glPopMatrix();
		break;
	    case _SCENEOBJ_TARGET_ORIGIN:
		glBegin(GL_LINES);
		glVertex3dv(mCCurrentObj.mPos);
		glVertex3d(0, 0, 0);
		glEnd();
		break;
	    }
	}
    }
    
    glDisable(GL_LINE_STIPPLE);

    if (bHittest == FALSE || mbSelected == FALSE){
	glPushMatrix(); 
	for (int j = 0; j < 3; j++){
	    matrix[0+j] = matrixB[0+j];
	    matrix[4+j] = matrixB[4+j];
	    matrix[8+j] = matrixB[8+j];
	}
	
	glMultMatrixd(matrix); 
	glScaled(scale*0.02, scale*0.02, scale*0.02);
	
	glBegin(GL_LINES);
	for (int i = 0; i < 360; i+=10){
	    double rad = 3.14159 * i / 180;
	    glVertex3d(cos(rad), sin(rad), 0);
	}
	glEnd();
	glBegin(GL_LINE_STRIP);
	for (int i = 0; i <= 360; i+=10){
	    double rad = 3.14159 * i / 180;
	    glVertex3d(0.7 * cos(rad), 0.7 * sin(rad), 0);
	}
	glEnd();
	glPopMatrix();
    }

    if (bHittest == FALSE) GLDrawPath();
}

wxString LightSRC_cls::
GetStrPovAttenuation(){
    wxString strLight;
    switch(mCFadeType.mBody){
    case _LIGHT_FADE_OFF:
	break;
    case _LIGHT_FADE_1: {
	strLight << wxT("\tfade_distance ") << mCFadeDistance.mBody << ENDL
		 << wxT("\tfade_power 1") << ENDL;
    	break; }
    case _LIGHT_FADE_2:
	strLight << wxT("\tfade_distance ") << mCFadeDistance.mBody << ENDL
		 << wxT("\tfade_power 2") << ENDL;
	break;
    }
    return strLight;
}

wxString LightSRC_cls::
GetStrPovAreaLight(float parallelFactor){
    wxString strLight;

    float quality = mCAreaQuality.mBody;
    float size = mCAreaSize.mBody;
    if (mCLightType.mBody == _LIGHT_PARALLEL) size *= parallelFactor;

    strLight << wxT("\tarea_light ") 
	     << wxT("\tx * ") << GetShortString(size) 
	     << wxT(", y * ") << GetShortString(size)
	     << wxT(",") << GetShortString(quality) 
	     << wxT(",") << GetShortString(quality) << ENDL;
    // ŕȂ
    if (quality < 6)
	strLight << wxT("\tcircular orient\n\tadaptive 1") << ENDL ;
    else
	strLight << wxT("\tcircular orient\n\tadaptive 2") << ENDL ;
    /*
    if (mCbAreaJitter.mbBody == TRUE)
    strLight << wxT("\tjitter") << ENDLWX;
    */
    return strLight;
}

wxString LightSRC_cls::
GetStrPovLight(double doubleIndex){

    float parallelFactor = 1000;

    float pLightPosForSV[4];
    float pLightPhiTheta[2];
    float pLightColor[4];

    bool  bParallel = (mCLightType.mBody == _LIGHT_PARALLEL) ? TRUE : FALSE;
    GetLightInfo(doubleIndex, pLightPosForSV, pLightPhiTheta, bParallel, NULL);
    GetLightColor(doubleIndex, pLightColor);

    if (bParallel == TRUE)
	LOOP_XYZ pLightPosForSV[xyz] *= parallelFactor; //!!

    wxString strLight;
    if (mCbVisible.mbBody == FALSE) return strLight;
    strLight << wxT("light_source {") << ENDL
	     << PovRayCtrl_cls::PovOut3f(wxT("\t"), pLightPosForSV) << ENDL
	     << PovRayCtrl_cls::PovOut3f(wxT("\tcolor"), pLightColor) << ENDL;
    if (mCbCastShadow.mbBody == FALSE)
	strLight << wxT("\tshadowless") << ENDL;

    switch(mCLightType.mBody){
    case _LIGHT_PARALLEL:
	strLight << wxT("\tparallel") << ENDL
		 << wxT("\tpoint_at <0, 0, 0>") << ENDL;
	break;
    case _LIGHT_SPOT: {
	float lightPointAt[3] = {0, 0, -1};
	std::vector < double > VLightDir = GetVDirectionVector(doubleIndex);
	LOOP_XYZ lightPointAt[xyz] = pLightPosForSV[xyz] - VLightDir[xyz];
	strLight << wxT("\tspotlight") << ENDL
		 << PovRayCtrl_cls::PovOut3f(wxT("\tpoint_at"), lightPointAt) << ENDL
		 << wxT("\tradius ") << mCSpotAngle.mBody << ENDL 
		 << wxT("\tfalloff ") << mCSpotAngle.mBody * 1.2 << ENDL;
	break; }
    }

    strLight << GetStrPovAttenuation();
    if (mCbAreaLight.mbBody == TRUE)
	strLight << GetStrPovAreaLight(parallelFactor);

    strLight << wxT("}") << ENDL;

    return strLight;
}

wxString LightSRC_cls::
Serialize(){
    wxString strOut;
    strOut << wxT("---LightSRC ") << mSelfIndex << ENDL;
    strOut << mCConfList.SerializeAll();
    strOut << mpCTranslate->Serialize();
    strOut << mpCRotation->Serialize();
    strOut << mpCLightStrength->Serialize();
    return strOut;
}

void LightSRC_cls::
ReadStrLine(std::wistream &istream, wxString strLineTag){
    // ł---LightSRC n ̃C͓ǂ񂾁B
    std::wstring wstrLine;
    wxString strLine;
    while ( std::getline (istream, wstrLine)){
	strLine = wstrLine.c_str();
	if (strLine.Find(wxT("ParamGroup=")) != wxNOT_FOUND) break;
	mCConfList.ReadStrLineAll(strLine);
    }
    mpCTranslate->ReadStrLine(istream, strLine);
    std::getline (istream, wstrLine);
    strLine = wstrLine.c_str();
    mpCRotation->ReadStrLine(istream, strLine);
    std::getline (istream, wstrLine);
    strLine = wstrLine.c_str();
    mpCLightStrength->ReadStrLine(istream, strLine);
}
