// "clsLightSourceGUI.h"

#include "clsTexPropertyGUI.h"

#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "../UNDO/clsUndoItemConstParam.h"

#include "../WIDGETCTRL/clsMultiInputPanel.h"
#include "../WIDGETCTRL/clsHookCtrl.h"
#include "../WIDGETCTRL/EX/clsStaticTextEX.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#include "../../clsFrameMain.h"

#include "../../wxWidgetEnum.h"

TexPropertyGUI_cls::
TexPropertyGUI_cls(wxWindow *pC, wxBitmap bitmap) :
    mstrTitle(_("Texture::")), 
    wxDialogEXKey(pC, wxID_ANY, _("Texture"), wxDefaultPosition, wxDefaultSize, bitmap) {
    mspCThis = this;

    wxBoxSizer *pBoxSizerSrc   = new wxBoxSizer(wxVERTICAL);

    mDummyShader    = 2;
    mDummyShininess = 10;
    mDummyTransmitType = 0;
    mDummyAlpha = 1;
    mDummyReflectionType = 0;
    mDummyReflectionMin = 0.2;
    mDummyReflectionMax = 0.5;
    mDummyIOR = 1;
    mpCTextureLinked = NULL;

    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	mDummyNormalPattern[i] = 0;
	mDummyNormalDegree[i] = 0.3;
	mDummyNormalScale[i] = 1;
    }

    SetMNumParam();

    mpCBook1 = new wxNotebook(this, ID_BOOK_TEXTURE1);
    
    SetMNumBump(mpCBook1);
    SetMNumMirror(mpCBook1);
    SetMNumInterior(mpCBook1);
    
    mpCBook1->AddPage(mpCMInputBump, _("Bump Map"));
    mpCBook1->AddPage(mpCMInputMirror, _("Mirror"));
    mpCBook1->AddPage(mpCMInputInterior, _("Interior"));
    
    //mpCBookMirror->AddPage(mpCMNumOptions, _("Options"));
    
    pBoxSizerSrc->Add(mpCMInputTexture, 0, wxEXPAND);
    pBoxSizerSrc->Add (mpCBook1, 0, wxEXPAND);

    ReconnectTexture(NULL);
    
    SetSizerAndFit(pBoxSizerSrc);
}

TexPropertyGUI_cls::
~TexPropertyGUI_cls(){
    if (mpCTextureLinked != NULL)
	mpCTextureLinked->mpCLinkedToMe = NULL;
}

void TexPropertyGUI_cls::
SetMNumParam(){
    MultiInputPanel_cls* pCurr = 
	MultiInputPanel_cls::Initialize(&mpCMInputTexture, this);

    pCurr->mUpdateBit = 
	_COMMAND_UPDATE_DISPLAY_GL |_COMMAND_UPDATE_DISPLAY_TEXTURE | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM; 

    pCurr->mTextWidth = 90;
    pCurr->mComboWidth = -1;

    int boxWidth = 1;
    mIndexShader = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Shader"));
    
    wxString strListShader[4] = 
	{_("original"), _("Gouroud"), _("Phong"), _("Toon")};

    pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &mDummyShader, 4, strListShader);

    mIndexShininess = 
	pCurr->SetInputAndSlider(_ITEM_INTEGER_SPIN, &mDummyShininess, 
				 _("Shininess"));
    pCurr->SetRange(_ITEM_LASTITEM, 0, 100, 1);
    
    mIndexTransmit = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Transmit Type"));
    
    wxString strListTransmit[3] = 
	{_("no transmit"), _("transmit"), _("filtered")};
    pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &mDummyTransmitType, 3, strListTransmit);

    mIndexAlpha = 
	pCurr->SetInputAndSlider(_ITEM_FLOAT, &mDummyAlpha, 
				 _("Alpha"));
    pCurr->SetRange(_ITEM_LASTITEM, 0, 100, 100);
    
    pCurr->FinishSetting(NULL);

    //pCurr->mpCHookCtrl->SetFuncHook(&FuncHookAssign, _HOOK_ASSIGN);
    pCurr->mpCHookCtrl->SetFuncHook(&HookHideAndShow, _HOOK_COMBOBOX);
}

void TexPropertyGUI_cls::
SetMNumMirror(wxWindow *pCWindow){
    MultiInputPanel_cls* pCurr = 
	MultiInputPanel_cls::Initialize(&mpCMInputMirror, pCWindow);

    pCurr->mTextWidth = 90;
    pCurr->mComboWidth = -1;
    int boxWidth = 2;
    wxString reflexType[3] = 
	{_("no mirror"), _("mirror like"), _("metal like")};

    boxWidth = 1;
    mIndexReflexType = pCurr->CombinePointer
	(_ITEM_BOXSIZERHORIZONTAL, &boxWidth, _("Reflex Type"));
    
    pCurr->CombinePointerList
	(_ITEM_UINT_LIST, &mDummyReflectionType, 3, reflexType);
    
    boxWidth = 2;
    mIndexReflexMin = pCurr->SetInputAndSlider
	(_ITEM_FLOAT, &mDummyReflectionMin,_("Reflex min"));
    pCurr->SetRange(_ITEM_LASTITEM, 0, 100, 100);

    mIndexReflexMax = pCurr->SetInputAndSlider
	(_ITEM_FLOAT, &mDummyReflectionMax, _("Reflex max"));
    pCurr->SetRange(_ITEM_LASTITEM, 0, 100, 100);
    
   
    pCurr->FinishSetting(NULL);
    
    //pCurr->SetFuncHook(&HookAssign, _MN_HOOK_ASSIGN);
    //pCurr->SetFuncHook(&HookCombo, _MN_HOOK_COMBOBOX);

    pCurr->mVCItemList[mIndexReflexType].
	mpCTextLabel->SetToolTip(_("Metal like reflection is affected by the color of material itself."));
    pCurr->mVCItemList[mIndexReflexMin].
	mpCTextLabel->SetToolTip(_("Reflection when viewed from a direction parallel to its normal."));
    pCurr->mVCItemList[mIndexReflexMax].
	mpCTextLabel->SetToolTip(_("Reflection when viewed from shallow angle. Many materials are more reflective in shallow angle."));

    ReconnectMirror(NULL);
    HideAndShowMirror();

    pCurr->mpCHookCtrl->SetFuncHook(&HookHideAndShow, _HOOK_COMBOBOX);
}

void TexPropertyGUI_cls::
SetMNumBump(wxWindow *pCWindow){
    MultiInputPanel_cls* pCurr = 
	MultiInputPanel_cls::Initialize(&mpCMInputBump, pCWindow);

    pCurr->mTextWidth = -1;
    pCurr->mComboWidth = 80;

    wxString normalPatternList[12] = 
	{_("no normal pattern"), wxT("agate"), _("bozo"), wxT("bumps"), wxT("brick"), wxT("cells"), _("checker"), wxT("crackle"), /*wxT("cylindrical"), */wxT("dents"), wxT("facets"), /*wxT("mandel"), wxT("julia"), wxT("hexagon"),*/ wxT("marble"), wxT("ridged_mf")};
    
    int boxWidth = 2;
    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	pCurr->mItemWidth = -1;
	mIndexNormalPattern[i] = pCurr->CombinePointerList
	    (_ITEM_UINT_LIST, &mDummyNormalPattern[i], 
	     12, normalPatternList, _("Normal"));
	mIndexNormalDegree[i] = pCurr->SetInputAndSlider
	    (_ITEM_FLOAT, &mDummyNormalDegree[i], _("degree"));
	pCurr->SetRange(_ITEM_LASTITEM, 0, 100, 100);
	mIndexNormalScale[i] = pCurr->SetInputAndSlider
	    (_ITEM_FLOAT, &mDummyNormalScale[i], _("scale"));
	pCurr->SetRange(_ITEM_LASTITEM, 0, 1000, 100);
    }

    pCurr->FinishSetting(NULL);

    ReconnectBump(NULL);
    HideAndShowBump();

    pCurr->mpCHookCtrl->SetFuncHook(&HookHideAndShow, _HOOK_COMBOBOX);
}

void TexPropertyGUI_cls::
SetMNumInterior(wxWindow *pCWindow){
    MultiInputPanel_cls* pCurr = 
	MultiInputPanel_cls::Initialize(&mpCMInputInterior, pCWindow);
    pCurr->mTextWidth = 80;
    pCurr->mItemWidth = -1;
    int boxWidth = 2;

    /*
    mIndexFadeDistance = 
	pCurr->SetInputAndSlider(_ITEM_INTEGER_SPIN, &mDummyFadeDistance, 
				 _("Fade Distance"));
    */
    mIndexIOR = 
	pCurr->SetInputAndSlider(_ITEM_FLOAT, &mDummyIOR, 
				 _("IOR"));
    pCurr->SetRange(_ITEM_LASTITEM, 0, 200, 100);

    pCurr->FinishSetting(NULL);
    
    ReconnectInterior(NULL);
    HideAndShowInterior();
}
    

void TexPropertyGUI_cls::
RecalcSize(){
    int minSize = 0;
    for (int i = 0; i < mpCBook1->GetPageCount(); i++){
	mpCBook1->GetPage(i)->Layout();
	wxSize size = mpCBook1->GetPage(i)->GetBestSize();
	if (minSize < size.GetHeight())
	    minSize = size.GetHeight();
    }
    mpCBook1->SetSizeHints(-1, minSize + 30);
}

void TexPropertyGUI_cls::
ReconnectBase(SceneObjTexture_cls *pCTexture){
    ReconnectMirror(pCTexture);
    ReconnectBump(pCTexture);
    ReconnectInterior(pCTexture);
    HideAndShowTexture();
    HideAndShowMirror();
    HideAndShowBump();
    HideAndShowInterior();
    RecalcSize();
}

void TexPropertyGUI_cls::
ReconnectTexture(SceneObjTexture_cls *pCTexture, wxString strCustomName){
    MultiInputPanel_cls* pCurr = mpCMInputTexture;

    if (pCTexture == NULL){
	SetTitle(mstrTitle + _("Not selected"));
	pCurr->mVCItemList[mIndexShader+1].CombinePointer
	    (_ITEM_UINT_LIST, &mDummyShader);
	pCurr->mVCItemList[mIndexShininess+1].CombinePointer
	    (_ITEM_INTEGER_SPIN, &mDummyShininess);
	pCurr->mVCItemList[mIndexShininess+2].CombinePointer
	    (_ITEM_INTEGER_SLIDER, &mDummyShininess);
	pCurr->mVCItemList[mIndexTransmit+1].CombinePointer
	    (_ITEM_UINT_LIST, &mDummyTransmitType);

	pCurr->ReconnectInputAndSlider
	    (mIndexAlpha, _ITEM_FLOAT, &mDummyAlpha);
	
	pCurr->UpdateValues();
	pCurr->EnableItem(mIndexShader+1, FALSE);
	pCurr->EnableItem(mIndexShininess+1, FALSE);
	pCurr->EnableItem(mIndexShininess+2, FALSE);
	pCurr->EnableItem(mIndexTransmit+1, FALSE);
	pCurr->EnableItem(mIndexAlpha, FALSE);
	//pCurr->EnableItem(mIndexAlpha+2, FALSE);
	
	ReconnectBase(pCTexture);
	return;
    }

    SetTitle(mstrTitle + strCustomName);
    pCurr->mVCItemList[mIndexShader+1].CombinePointer
	(_ITEM_UINT_LIST, &(pCTexture->mCShaderType.mBody));
    pCurr->mVCItemList[mIndexShininess+1].CombinePointer
	(_ITEM_INTEGER_SPIN, &(pCTexture->mCShininess.mBody));
    pCurr->mVCItemList[mIndexShininess+2].CombinePointer
	(_ITEM_INTEGER_SLIDER, &(pCTexture->mCShininess.mBody));
    pCurr->mVCItemList[mIndexTransmit+1].CombinePointer
	(_ITEM_UINT_LIST, &(pCTexture->mCTransmitType.mBody));

    pCurr->ReconnectInputAndSlider
	(mIndexAlpha, _ITEM_FLOAT, &(pCTexture->mCAlpha.mBody));
    /*
    pCurr->mVCItemList[mIndexAlpha+1].CombinePointer
	(_ITEM_FLOAT, &(pCTexture->mCAlpha.mBody));
    pCurr->mVCItemList[mIndexAlpha+2].CombinePointer
	(_ITEM_FLOAT_SLIDER, &(pCTexture->mCAlpha.mBody));
    */

    pCurr->EnableItem(mIndexShader+1, TRUE);
    pCurr->EnableItem(mIndexShininess+1, TRUE);
    pCurr->EnableItem(mIndexShininess+2, TRUE);
    pCurr->EnableItem(mIndexTransmit+1, TRUE);

    pCurr->EnableItem(mIndexAlpha, TRUE);
    pCurr->EnableItem(mIndexAlpha+2, TRUE);

    pCurr->UpdateValues();

    if (mpCTextureLinked != NULL) 
	mpCTextureLinked->mpCLinkedToMe = NULL;
    mpCTextureLinked = pCTexture;
    pCTexture->mpCLinkedToMe = this;

    ReconnectBase(pCTexture);
}

void TexPropertyGUI_cls::
ReconnectMirror(SceneObjTexture_cls *pCTexture){
    MultiInputPanel_cls* pCurr = mpCMInputMirror;

    if (pCTexture == NULL){
	pCurr->mVCItemList[mIndexReflexType+1].CombinePointer
	    (_ITEM_UINT_LIST, &mDummyReflectionType);

	pCurr->ReconnectInputAndSlider
	    (mIndexReflexMin, _ITEM_FLOAT, &mDummyReflectionMin);
	pCurr->ReconnectInputAndSlider
	    (mIndexReflexMax, _ITEM_FLOAT, &mDummyReflectionMax);
	
	/*
	pCurr->mVCItemList[mIndexReflexMin+1].CombinePointer
	    (_ITEM_FLOAT, &mDummyReflectionMin);
	pCurr->mVCItemList[mIndexReflexMin+2].CombinePointer
	    (_ITEM_FLOAT_SLIDER, &mDummyReflectionMin);
	
	pCurr->mVCItemList[mIndexReflexMax+1].CombinePointer
	    (_ITEM_FLOAT, &mDummyReflectionMax);
	pCurr->mVCItemList[mIndexReflexMax+2].CombinePointer
	    (_ITEM_FLOAT_SLIDER, &mDummyReflectionMax);
	*/
	
	pCurr->UpdateValues();
	pCurr->EnableItem(mIndexReflexType, FALSE);
	pCurr->EnableItem(mIndexReflexMin, FALSE);
	pCurr->EnableItem(mIndexReflexMax, FALSE);
	return;
    }

    pCurr->mVCItemList[mIndexReflexType+1].CombinePointer
	(_ITEM_UINT_LIST, &(pCTexture->mCReflectType.mBody));
    pCurr->ReconnectInputAndSlider
	(mIndexReflexMin, _ITEM_FLOAT, &(pCTexture->mCReflectMin.mBody));
    pCurr->ReconnectInputAndSlider
	(mIndexReflexMax, _ITEM_FLOAT, &(pCTexture->mCReflectMax.mBody));
    
    /*
    pCurr->mVCItemList[mIndexReflexMin+1].CombinePointer
	(_ITEM_FLOAT_SLIDER, &(pCTexture->mCReflectMin.mBody));
    pCurr->mVCItemList[mIndexReflexMin+2].CombinePointer
	(_ITEM_FLOAT, &(pCTexture->mCReflectMin.mBody));
    pCurr->mVCItemList[mIndexReflexMax+1].CombinePointer
	(_ITEM_FLOAT_SLIDER, &(pCTexture->mCReflectMax.mBody));
    pCurr->mVCItemList[mIndexReflexMax+2].CombinePointer
	(_ITEM_FLOAT, &(pCTexture->mCReflectMax.mBody));
    */
    pCurr->EnableItem(mIndexReflexType, TRUE);
    pCurr->EnableItem(mIndexReflexMin, TRUE);
    pCurr->EnableItem(mIndexReflexMax, TRUE);
    
    pCurr->UpdateValues();
}

void TexPropertyGUI_cls::
ReconnectBump(SceneObjTexture_cls *pCTexture){
    MultiInputPanel_cls* pCurr = mpCMInputBump;

    if (pCTexture == NULL){
	for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	    pCurr->mVCItemList[mIndexNormalPattern[i]].CombinePointer
		(_ITEM_UINT_LIST, &mDummyNormalPattern[i]);
	    pCurr->EnableItem(mIndexNormalPattern[i], FALSE);
	    
	    pCurr->ReconnectInputAndSlider
		(mIndexNormalDegree[i], _ITEM_FLOAT, &mDummyNormalDegree[i]);
	    pCurr->ReconnectInputAndSlider
		(mIndexNormalScale[i], _ITEM_FLOAT, &mDummyNormalScale[i]);
	    
	    pCurr->EnableItem(mIndexNormalDegree[i], FALSE);
	    pCurr->EnableItem(mIndexNormalScale[i], FALSE);
	}
	pCurr->UpdateValues();
	return;
    }

    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	pCurr->mVCItemList[mIndexNormalPattern[i]].CombinePointer
	    (_ITEM_UINT_LIST, &(pCTexture->mCNormalPattern[i].mBody));
	pCurr->EnableItem(mIndexNormalPattern[i], TRUE);
	
	pCurr->ReconnectInputAndSlider
	    (mIndexNormalDegree[i], _ITEM_FLOAT, &(pCTexture->mCNormalDegree[i].mBody));
	pCurr->ReconnectInputAndSlider
	    (mIndexNormalScale[i], _ITEM_FLOAT, &(pCTexture->mCNormalScale[i].mBody));
	pCurr->EnableItem(mIndexNormalDegree[i], TRUE);
	pCurr->EnableItem(mIndexNormalScale[i], TRUE);
    }
    pCurr->UpdateValues();
}

void TexPropertyGUI_cls::
ReconnectInterior(SceneObjTexture_cls *pCTexture){
    MultiInputPanel_cls* pCurr = mpCMInputInterior;

    if (pCTexture == NULL){
	pCurr->ReconnectInputAndSlider
	    (mIndexIOR, _ITEM_FLOAT, &mDummyIOR);
	pCurr->UpdateValues();
	pCurr->EnableItem(mIndexIOR, FALSE);
	return;
    }
    
    pCurr->ReconnectInputAndSlider
	(mIndexIOR, _ITEM_FLOAT, &(pCTexture->mCIOR.mBody));
    pCurr->EnableItem(mIndexIOR, TRUE);
    
    pCurr->UpdateValues();
}

void TexPropertyGUI_cls::
HideAndShowTexture(){
    MultiInputPanel_cls* pCurr = mpCMInputTexture;
    bool bValid = TRUE;
    if (mpCTextureLinked == NULL) bValid = FALSE;
    bool bTransmit = (pCurr->mVCItemList[mIndexTransmit+1].GetValueInt() != 0);
    pCurr->ShowItem(mIndexAlpha, bTransmit & bValid);
}

void TexPropertyGUI_cls::
HideAndShowMirror(){
    MultiInputPanel_cls* pCurr = mpCMInputMirror;
    bool bValid = TRUE;
    if (mpCTextureLinked == NULL) bValid = FALSE;
    bool bMirror = (pCurr->mVCItemList[mIndexReflexType+1].GetValueInt() != 0);
    pCurr->ShowItem(mIndexReflexMin, bMirror & bValid);
    pCurr->ShowItem(mIndexReflexMax, bMirror & bValid);
}

void TexPropertyGUI_cls::
HideAndShowBump(){
    MultiInputPanel_cls* pCurr = mpCMInputBump;
    bool bValid = TRUE;
    if (mpCTextureLinked == NULL) bValid = FALSE;
    for (int i = 0; i < _TEX_NORMAL_MAX; i++){
	bool bNormal = (pCurr->mVCItemList[mIndexNormalPattern[i]].GetValueInt() != 0);
	pCurr->ShowItem(mIndexNormalDegree[i], bNormal & bValid);
	pCurr->ShowItem(mIndexNormalScale[i], bNormal & bValid);
    }
    
}

void TexPropertyGUI_cls::
HideAndShowInterior(){
    MultiInputPanel_cls* pCurr = mpCMInputInterior;
    MultiInputPanel_cls* pCurrTex = mpCMInputTexture;
    bool bValid = TRUE;
    if (mpCTextureLinked == NULL) bValid = FALSE;
    bool bTransmit = (pCurrTex->mVCItemList[mIndexTransmit+1].GetValueInt() != 0);
    pCurr->ShowItem(mIndexIOR, bTransmit & bValid);
}

void TexPropertyGUI_cls::
HookHideAndShow(void *pArg, void *pEvent){
    MultiInputPanel_cls* pCPanel = (MultiInputPanel_cls*)pArg;

    if (pCPanel == mspCThis->mpCMInputMirror){
	mspCThis->HideAndShowMirror();
	mspCThis->RecalcSize();
	mspCThis->ResizeDialog();
    }
    
    if (pCPanel == mspCThis->mpCMInputBump){
	mspCThis->HideAndShowBump();
	mspCThis->RecalcSize();
	mspCThis->ResizeDialog();
    }

    if (pCPanel == mspCThis->mpCMInputTexture){
	mspCThis->HideAndShowTexture();
	mspCThis->HideAndShowInterior();
	mspCThis->RecalcSize();
	mspCThis->ResizeDialog();
    }
}

// ------

TexPropertyCtrl_cls::
TexPropertyCtrl_cls(FrameMain_cls* pCParentFrame, wxString strBitmap){
    mpCGUI = NULL;
    mpCParentFrameLink = pCParentFrame;
    //mpBitmapLink = mpCParentFrameLink->mpBitmapTexture;
    mstrBitmap = strBitmap;
    
    InitGUI(pCParentFrame);
}

TexPropertyCtrl_cls::
~TexPropertyCtrl_cls(){
}

void TexPropertyCtrl_cls::
InitGUI(wxWindow *pC){
    mpCGUI = new TexPropertyGUI_cls(pC, ART(mstrBitmap)
				    /*mCScreenRect.GetPosition(),*/
				    /*mCScreenRect.GetSize()*/
				    /*wxDefaultPosition, wxDefaultSize*/);
    SetPDialog(mpCGUI);
}

void TexPropertyCtrl_cls::
UpdateItems(bool bForce){
    if (mpCGUI == NULL) return;
}

void TexPropertyCtrl_cls::
ReconnectTexture(SceneObjTexture_cls *pCTexture, wxString strCustomName){
    mpCGUI->ReconnectTexture(pCTexture, strCustomName);
    mpCGUI->ResizeDialog();
}

TexPropertyGUI_cls* TexPropertyGUI_cls::mspCThis = NULL;
