// -*- c++ -*-
// (setq c-basic-offset 4)

#include "clsTimeCtrl.h"

#include "clsTimeCtrlGUI.h"
#include "clsTimeCtrlGUIAA.h"
#include "clsTimeCtrlGUIAtoi.h"
#include "clsTimeCtrlGUIButtons.h"
#include "clsTimeCtrlGUITics.h"
#include "clsTimeCtrlGUIZoom.h"
#include "clsPlayCtrlGUI.h"

#include "../MAINCANVAS/clsParamGUICtrl.h"
#include "../PARAMETER/clsParameters.h"
#include "../PARAMETER/clsParamGroup.h"
#include "../PARAMETER/clsParamGroupGUI.h"
#include "../PARTICLEDATA/clsParticleDataSeqCtrl.h"
#include "../UNDO/clsUndoManager.h"
extern UndoManager_cls GCUndoList;
#include "../UNDO/clsUndoItemTimeCtrl.h"
#include "../WIDGETCTRL/clsBitmapPanel.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#include "../../clsMain.h"
#include "../../clsFrameMain.h"
#include "../../wxWidgetEnum.h"
#include "../../wxWidgetEnumMenu.h"

#define wxDefPos wxDefaultPosition

MotionBlurCtrl_cls::
MotionBlurCtrl_cls(TimeCtrl_cls *pCTimeCtrl){
    mpCTimeCtrlLink = pCTimeCtrl;
    
    mbUseMotionBlur = FALSE; // _ŌĂ邩ǂ
    mBlurStep = 0;

    mBlurLength  = 0.5;
    mBlurQuality = 4;
}

void MotionBlurCtrl_cls::
BeginMotionBlur(int direction){
    mbUseMotionBlur = TRUE;
    if (direction == +1)
	mBlurStep = -1; 
    else
	mBlurStep = mBlurQuality;
}

void MotionBlurCtrl_cls::
EndMotionBlur(){
    mbUseMotionBlur = FALSE;
}

bool MotionBlurCtrl_cls::
Increment(int direction){
    if (direction == +1){
	mBlurStep++;
	if (mBlurStep == mBlurQuality){
	    mBlurStep = 0;
	    return TRUE;
	}
    }
    else{
	mBlurStep--;
	if (mBlurStep == -1){
	    mBlurStep = mBlurQuality - 1;
	    return TRUE;
	}
    }
    return FALSE;
}

bool MotionBlurCtrl_cls::
IsInitBluring(){
    if (mpCTimeCtrlLink->mPlayDirection == +1)
	if (mBlurStep == 0) return TRUE;
    if (mpCTimeCtrlLink->mPlayDirection == -1)
	if (mBlurStep == mBlurQuality - 1) return TRUE;
    return FALSE;
}

bool MotionBlurCtrl_cls::
IsLastBluring(){
    if (mbUseMotionBlur == FALSE) return TRUE;
    if (mpCTimeCtrlLink->mPlayDirection == +1)
	if (mBlurStep == mBlurQuality - 1) return TRUE;
    if (mpCTimeCtrlLink->mPlayDirection == -1)
	if (mBlurStep == 0) return TRUE;
    return FALSE;
}

bool MotionBlurCtrl_cls::
IsOutputConditionOK(){
    if (mbUseMotionBlur == FALSE) return TRUE;
    return IsLastBluring();
}

double MotionBlurCtrl_cls::
GetFactor(){
    return mBlurLength * mBlurStep / mBlurQuality;
}

float MotionBlurCtrl_cls::
GetAlpha(){
    return 1.0f / mBlurQuality;
}

TimeCtrl_cls::
TimeCtrl_cls(Main_cls *pCMain) : mCBlurCtrl(this)//TimeRemap(1, L"TimeRemap")
{
    mpCMainLink = pCMain;
    mFrame = 0;
    mInterpStep = 0;
    mCustomRegionMin = 0;
    mCustomRegionMax = 99;

    mbModeShy = FALSE;
    mbPlayTex = FALSE; 
    mbUseCustomWA = FALSE;
    mPlayMode = _TIME_CTRL_PLAY_FULL;
    mPlayDirection = +1;
    mPrefetchRange = 3;
    mClockLastDraw = 0;
    mWaitDuration = 30;
    mbUndoGroupOpened = FALSE;

    mCConfListTime.Add(&mCMaxFrame, L"TimeMaxFrame=", L"100");
    mCConfListTime.Add(&mCWAMin, L"TimeWorkAreaMin=", L"0");
    mCConfListTime.Add(&mCWAMax, L"TimeWorkAreaMax=", L"100");
    mCConfListTime.Add(&mCAAMin, L"TimeActiveAreaMin=", L"0");
    mCConfListTime.Add(&mCAAMax, L"TImeActiveAreaMax=", L"100");

    mCConfListTime.Add(&mCbUseTimeRemap, L"TimeUseTimeRemap=", L"0");
    mCConfListTime.Add(&mCbUseMotionBlur, L"TimeUseMotionBlur=", L"0");
    mCConfListTime.Add(&mCAnimStep, L"TimeAnimStep=", L"1");
    mCConfListTime.Add(&mCAnimInterp, L"TimeAnimInterp=", L"1");
    mCConfListTime.Add(&mCbRepeat, L"TimeRepeat=", L"0");
    
    mCConfListTime.Add(&mCSceneLengthType, L"TimeSceneLengthType=", L"0");
    
    //TimeLinePanel_cls::msVBSnippingTag.resize(99);
    int updateBit = _COMMAND_UPDATE_DISPLAY_GL | _COMMAND_UPDATE_DISPLAY_TL | _COMMAND_UPDATE_DISPLAY_SB | _COMMAND_UPDATE_DISPLAY_FILTER_PARAM;
    wxString listTag[1] = {wxT("Time")};
    mpCTimeRemap = new ParamGroup_cls(wxT("Time Remap"), 1, listTag, updateBit, FALSE);
    
    mCStopWatchWait.Start();
}

TimeCtrl_cls::
~TimeCtrl_cls(){
    delete mpCTimeRemap;
}

void TimeCtrl_cls::
InstantiateGUIA(wxWindow *pCParent){
  
    mpCGUIAA = new TimeCtrlGUIAA_cls	
	(pCParent, this, wxID_ANY, wxDefPos, wxSize(-1, 12));
    mpCGUITics = new TimeCtrlGUITics_cls
	(pCParent, this, wxID_ANY, wxDefPos, wxSize(-1, 16));
    mpCGUIButtons = new TimeCtrlGUIButtons_cls
	(pCParent, this, wxID_ANY, wxDefPos, wxSize(_TL_LEFT_WIDTH, 16));
    mpCGUIZoom = new TimeCtrlGUIZoom_cls
	(pCParent, this, wxID_ANY, wxDefPos, wxSize(_TL_LEFT_WIDTH, 12));
    mpCGUIAtoi = new TimeCtrlGUIAtoi_cls	
	(pCParent, this, wxID_ANY, wxDefPos, wxSize(_TL_LEFT_WIDTH, 24));
    mpCGUI = new TimeCtrlGUI_cls
	(pCParent, this, wxID_ANY, wxDefPos, wxSize(-1, 24));
}

void TimeCtrl_cls::
InstantiateGUIB(wxWindow *pCParent){
    mpCGUIPlay = new PlayCtrlGUI_cls
	(pCParent, this, wxID_ANY, wxDefaultPosition, wxSize(-1, -1));
}

void TimeCtrl_cls::
BeginMotionBlur(){
    mCBlurCtrl.BeginMotionBlur(mPlayDirection);
}

void TimeCtrl_cls::
HideAndShowRemap(){
    if (mCbUseTimeRemap.mbBody == TRUE)
	mpCTimeRemap->mpCGUI->Show();
    else{
	mpCTimeRemap->HideItems();
	mpCTimeRemap->mpCGUI->Hide();
    }
}

void TimeCtrl_cls::SetValueCommand(int target, int valueNew, int valueOld){
    GCUndoList.PushNewCommand
	(new CommandSetTimeCtrlValue_cls(this, target, valueNew),
	 new CommandSetTimeCtrlValue_cls(this, target, valueOld));
    GCUndoList.ExecCurrent();
}

void TimeCtrl_cls::MoveFrameCommand(int dframe){
    int newframe = CalcNextFrame(_TIME_CTRL_CURRENT, dframe);
    if (newframe == mFrame) return;
    SetValueCommand(_COMMAND_TARGET_FRAME, newframe, mFrame);
}

void TimeCtrl_cls::MoveInterpStepCommand(int dstep){
    int newStep = CalcNextStep(_TIME_CTRL_CURRENT, dstep);
    int newFrame = mFrame;
    CalcNextFrameFromStep(&newFrame, &newStep);
    if (newFrame != mFrame || newStep != mInterpStep){
	GCUndoList.OpenBlock();
	SetValueCommand(_COMMAND_TARGET_INTERPSTEP, newStep, mInterpStep);
	SetValueCommand(_COMMAND_TARGET_FRAME, newFrame, mFrame);
	GCUndoList.CloseBlock();
    }
}

void TimeCtrl_cls::MoveFrameCommandLarge(int dframe){
    int newframe;
    int newstep = 0;
    if (dframe < 0) newframe = mCWAMin.mBody;
    else            newframe = mCWAMax.mBody;
    
    if (newframe == mFrame && newstep == mInterpStep) return;
    GCUndoList.OpenBlock();
    SetValueCommand(_COMMAND_TARGET_INTERPSTEP, newstep, mInterpStep);
    SetValueCommand(_COMMAND_TARGET_FRAME, newframe, mFrame);
    GCUndoList.CloseBlock();
}

void TimeCtrl_cls::MoveFrameCommandSnip(int direction){
    int newframe;
    int newstep = 0;
    if (direction > 0) newframe = GetSnipNext();
    else               newframe = GetSnipPrev();
    GCUndoList.OpenBlock();
    SetValueCommand(_COMMAND_TARGET_INTERPSTEP, newstep, mInterpStep);
    SetValueCommand(_COMMAND_TARGET_FRAME, newframe, mFrame);
    GCUndoList.CloseBlock();
}

void TimeCtrl_cls::SetFrameCommand(int frame, int frameAtClick){
    if (frameAtClick == _TIME_CTRL_CURRENT) frameAtClick = mFrame;
    SetValueCommand(_COMMAND_TARGET_FRAME, frame, frameAtClick);
}

void TimeCtrl_cls::SetFrameFromDoubleCommand(double doubleIndex){
    unsigned int newFrame = GetIndexFromDouble(doubleIndex);
    unsigned int newstep = 0;
    if (newFrame == mFrame && newstep == mInterpStep) return;
    GCUndoList.OpenBlock();
    SetValueCommand(_COMMAND_TARGET_INTERPSTEP, newstep, mInterpStep);
    SetValueCommand(_COMMAND_TARGET_FRAME, newFrame, mFrame);
    GCUndoList.CloseBlock();
}

void TimeCtrl_cls::SetInterpStepCommand(int step){
    SetValueCommand(_COMMAND_TARGET_INTERPSTEP, step, mInterpStep);
}

void TimeCtrl_cls::SetWAMinCommand(int min, int minAtClick){
    if (minAtClick == _TIME_CTRL_CURRENT) minAtClick = mCWAMin.mBody;
    SetValueCommand(_COMMAND_TARGET_WAMIN, min, minAtClick);
}

void TimeCtrl_cls::SetWAMaxCommand(int max, int maxAtClick){
    if (maxAtClick == _TIME_CTRL_CURRENT) maxAtClick = mCWAMax.mBody;
    SetValueCommand(_COMMAND_TARGET_WAMAX, max, maxAtClick);
}

void TimeCtrl_cls::SetAAMinCommand(int min, int minAtClick){
    if (minAtClick == _TIME_CTRL_CURRENT) minAtClick = mCAAMin.mBody;
    SetValueCommand(_COMMAND_TARGET_AAMIN, min, minAtClick);
}

void TimeCtrl_cls::SetAAMaxCommand(int max, int maxAtClick){
    if (maxAtClick == _TIME_CTRL_CURRENT) maxAtClick = mCAAMax.mBody;
    SetValueCommand(_COMMAND_TARGET_AAMAX, max, maxAtClick);
}

void TimeCtrl_cls::SetRepeatCommand(int repeat, int repeatPrev){
    SetValueCommand(_COMMAND_TARGET_REPEAT, repeat, repeatPrev);
}

void TimeCtrl_cls::SetAnimStepCommand(int animStep, int animStepPrev){
    GCUndoList.OpenBlock();
    SetValueCommand(_COMMAND_TARGET_ANIMSTEP, animStep, animStepPrev);
    if (mFrame != RoundFrame(mFrame))
	SetValueCommand(_COMMAND_TARGET_FRAME, RoundFrame(mFrame), mFrame);
    GCUndoList.CloseBlock();
}

void TimeCtrl_cls::SetAnimInterpCommand(int animInterp, int animInterpPrev){
    SetValueCommand(_COMMAND_TARGET_ANIMINTERP, animInterp, animInterpPrev);
}

void TimeCtrl_cls::SetPrefetchCommand(int prefetch, int prefetchPrev){
    SetValueCommand(_COMMAND_TARGET_PREFETCH, prefetch, prefetchPrev);
}

void TimeCtrl_cls::SetMaxFrameCommand(int maxFrame){
    SetValueCommand(_COMMAND_TARGET_MAXFRAME, maxFrame, mCMaxFrame.mBody);
}

void TimeCtrl_cls::ExpandAACommand(bool bExpand){
    int minAtClick = mCAAMin.mBody;
    int maxAtClick = mCAAMax.mBody;
    int depthMin = (mFrame -  minAtClick);
    int depthMax = (maxAtClick - mFrame);
    int newMin, newMax;
    if (bExpand == FALSE){
	newMin = mFrame - depthMin * 2; 
	newMax = mFrame + depthMax * 2;
    }
    else{
	newMin = mFrame - depthMin / 2; 
	newMax = mFrame + depthMax / 2;
    }
    
    if (newMin < 0) newMin = 0;
    if (newMin > mFrame) newMin = mFrame;

    if (newMax > mCMaxFrame.mBody) newMax = mCMaxFrame.mBody;
    if (newMax < mFrame) newMax = mFrame;

    if (newMin == minAtClick &&
	newMax == maxAtClick) return;

    GCUndoList.OpenBlock();
    SetAAMinCommand(newMin, minAtClick);
    SetAAMaxCommand(newMax, maxAtClick);
    GCUndoList.CloseBlock();
}

void TimeCtrl_cls::SetFrameNative(int frame, int updateBit){
    mFrame = frame;
    Parameters_cls::msCurrentDoubleIndex = GetDoubleIndex();
    CommandItem_cls::UpdateBit(updateBit);
}

void TimeCtrl_cls::SetInterpStepNative(int step, int updateBit){
    mInterpStep = step;
    CommandItem_cls::UpdateBit(updateBit);
}

void TimeCtrl_cls::SetWAMinNative(int min){
    mCWAMin.mBody = min;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

void TimeCtrl_cls::SetWAMaxNative(int max){
    mCWAMax.mBody = max;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

void TimeCtrl_cls::SetAAMinNative(int min){
    mCAAMin.mBody = min;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

void TimeCtrl_cls::SetAAMaxNative(int max){
    mCAAMax.mBody = max;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

void TimeCtrl_cls::SetRepeatNative(int repeat){
    mCbRepeat.mbBody = (bool)repeat;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

void TimeCtrl_cls::SetAnimStepNative(int animStep){
    mCAnimStep.mBody = animStep;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

void TimeCtrl_cls::SetAnimInterpNative(int animInterp){
    mCAnimInterp.mBody = animInterp;
    Parameters_cls::SetFrameNumber(mCMaxFrame.mBody, mCAnimInterp.mBody);
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

void TimeCtrl_cls::SetPrefetchNative(int prefetch){
    mPrefetchRange = prefetch;
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuEditRefreshTL);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

void TimeCtrl_cls::SetMaxFrameNative(int maxFrame){
    mCMaxFrame.mBody = maxFrame;
    CheckAfterMaxFrameChange();
    Parameters_cls::SetFrameNumber(maxFrame, mCAnimInterp.mBody);
    RefreshGUI();
}

void TimeCtrl_cls::
CheckAfterMaxFrameChange(){
    static unsigned int prevMaxFrame = 99;
    if (mCMaxFrame.mBody != prevMaxFrame){
	if (prevMaxFrame == mCAAMax.mBody)
	    mCAAMax.mBody = mCMaxFrame.mBody;
	if (prevMaxFrame == mCWAMax.mBody)
	    mCWAMax.mBody = mCMaxFrame.mBody;
	if (mCAAMax.mBody > mCMaxFrame.mBody)
	    mCAAMax.mBody = mCMaxFrame.mBody;
	if (mCWAMax.mBody > mCMaxFrame.mBody)
	    mCWAMax.mBody = mCMaxFrame.mBody;
	prevMaxFrame = mCMaxFrame.mBody;
    }
    
    mCustomRegionMin = 0; 
    mCustomRegionMax = mCMaxFrame.mBody;
    
    if (mFrame > (int)mCMaxFrame.mBody) mFrame = 0;
    
    //TimeLinePanel_cls::msVBSnippingTag.resize(mCMaxFrame.mBody);
}

void TimeCtrl_cls::
RefreshGUI(){
    mpCGUI->Refresh(FALSE);
    mpCGUIAA->Refresh(FALSE);
    mpCGUIAtoi->Refresh(FALSE);
    mpCGUITics->Refresh(FALSE);
    mpCGUIPlay->UpdateDisplayedValues(); 

}

void TimeCtrl_cls::
SetWorkingArea(int maxIndex){
    mCWAMin.mBody = mCAAMin.mBody = 0;
    mCWAMax.mBody = mCAAMax.mBody = maxIndex;
}

void TimeCtrl_cls::
CheckWorkingArea(){
    if (mCWAMax.mBody > mCMaxFrame.mBody)
	mCWAMax.mBody = mCMaxFrame.mBody;
    if (mCWAMin.mBody > mCWAMax.mBody){
	unsigned int max = mCWAMax.mBody;
	mCWAMax.mBody = mCWAMin.mBody;
	mCWAMin.mBody = max;
    }

    if (mCAAMax.mBody > mCMaxFrame.mBody)
	mCAAMax.mBody = mCMaxFrame.mBody;
    if (mCAAMin.mBody > mCAAMax.mBody){
	unsigned int max = mCAAMax.mBody;
	mCAAMax.mBody = mCAAMin.mBody;
	mCAAMin.mBody = max;
    }
    
    if (mFrame < (int)mCAAMin.mBody) 
	mFrame = mCAAMin.mBody;
    if (mFrame > (int)mCAAMax.mBody)
	mFrame = mCAAMax.mBody;

    mFrame = CalcNextFrame(_TIME_CTRL_CURRENT, 0);
}

void TimeCtrl_cls::
RemapParameterKeys(double rate){
    Parameters_cls::SetFrameNumber(mCMaxFrame.mBody, mCAnimStep.mBody);
    GCUndoList.OpenBlock();
    mpCMainLink->mpCParamGUICtrlLink->RemapKeys(rate);
    //mspCDF3SeqCtrlLink->mpCGUI->mpCPanelList->RemapKeys(rate);
    GCUndoList.CloseBlock();
}

void TimeCtrl_cls::
SetBUseTimeRemap(bool bUse){
    /*
    mCbUseTimeRemap.mbBody = bUse;
    if (mCbUseTimeRemap.mbBody == TRUE){
	// \[Xłlast = (double)mMaxIndexǖȂH
	double init = 0, last = (double)mCMaxFrame.mBody;//, zero = 0;

	mCTimeRemap.mCbSelectable.mbBody = TRUE;
	mspCDF3SeqCtrlLink->mpCGUI->ShowItem(wxT("[Time Remap]"));
	
	GCUndoList.OpenBlock();
	mCTimeRemap.InitSet(&init);

	KeyFrame_cls CKey(1);//, CKeySlave(0);
	CKey.Set(GetDoubleIndex(last), _PARAM_HERMITE_AUTO, &last);
	mCTimeRemap.SetKey(CKey);
	
	GCUndoList.CloseBlock();
    }
    else{
	mCTimeRemap.mCbSelectable.mbBody = FALSE;
	mspCDF3SeqCtrlLink->mpCGUI->ShowItem(wxT("[Time Remap]"), FALSE);
    }
    mspCDF3SeqCtrlLink->mpCGUI->UpdateItemsVisivility();
    UpdateDF3SeqGUI(TRUE);
    */
}

int  TimeCtrl_cls::
GetExpectedPrefetchMax(){
    int averageMB = mpCMainLink->GetAverageMemoryMB();
    if (averageMB == -1) return -1;
    if (averageMB == 0) averageMB = 1;
    int memoryLimit = mpCMainLink->GetMemoryLimitMB();
    return memoryLimit / averageMB;
}

unsigned int TimeCtrl_cls::
GetRemappedIndex(int frame){
    //if (frame == _TIME_CTRL_CURRENT) frame = mFrame;
    int remappedIndex = frame;
    if (mCbUseTimeRemap.mbBody == TRUE)
	//remappedIndex = frame;
	/*
	remappedIndex = (int)(*mCTimeRemap.GetPInterpolatedValue
			      (GetDoubleIndex(frame)) + 1e-5);
	*/
	remappedIndex = (int)(mpCTimeRemap->GetInterpolatedValue
			      (GetDoubleIndex(frame), _PARAM_USE_COMMON)[0] + 1e-5);
    else
	if (frame == _TIME_CTRL_CURRENT) remappedIndex = mFrame;
    if (remappedIndex < 0) remappedIndex = 0;
    return remappedIndex;
}

double TimeCtrl_cls::
GetRemappedIndexDouble(int frame, int step){
    double remappedIndex;
    if (frame == _TIME_CTRL_CURRENT) frame = mFrame;
    if (step  == _TIME_CTRL_CURRENT) step = mInterpStep;
    if (mCbUseTimeRemap.mbBody == FALSE){
	remappedIndex = frame;
	remappedIndex += (double)step / mCAnimInterp.mBody;
    }
    else{
	double stepDouble = step;
	if (mCBlurCtrl.mbUseMotionBlur == TRUE)
	    stepDouble += mCBlurCtrl.GetFactor();

	double doubleIndex = GetDoubleIndex(frame, stepDouble);

	remappedIndex = mpCTimeRemap->GetInterpolatedValue
	    (doubleIndex, _PARAM_USE_COMMON)[0];
	
	if (remappedIndex - (int)remappedIndex > 0.99999)
	    remappedIndex = (int)(remappedIndex + 0.00001);
    }

    if (remappedIndex < 0) remappedIndex = 0;
    return remappedIndex;
}

double TimeCtrl_cls::
GetRemappedInterpStepDouble(int frame, int step){
    double remappedStep;
    if (frame == _TIME_CTRL_CURRENT) frame = mFrame;
    if (step  == _TIME_CTRL_CURRENT) step = mInterpStep;
    if (mCbUseTimeRemap.mbBody == FALSE){
	if (mCBlurCtrl.mbUseMotionBlur == TRUE)
	    remappedStep = ((double)step + mCBlurCtrl.GetFactor()) / mCAnimInterp.mBody;
	else
	    remappedStep = (double)step / mCAnimInterp.mBody;
    }
    else{ // FIX_ME!!
	double index = GetRemappedIndex();
	double doubleIndex = GetRemappedIndexDouble();
	//double doubleIndexScaled = doubleIndex * mCMaxFrame.mBody;
	remappedStep = doubleIndex - index;
	if (mCBlurCtrl.mbUseMotionBlur == TRUE)
	    remappedStep = ((double)step + mCBlurCtrl.GetFactor()) / mCAnimInterp.mBody;
    }
    return remappedStep;
}

int TimeCtrl_cls::
GetInterval(int frameNumber){
    int interval = 1;
    if (frameNumber > 1000)     interval = 10;
    else if (frameNumber > 500) interval = 5;
    else if (frameNumber > 200) interval = 2;
    return interval;
}

double TimeCtrl_cls::
GetDoubleIndex(int frame, double step){
    if (mCMaxFrame.mBody == 0) return 0;
    if (frame == _TIME_CTRL_CURRENT){
	frame = mFrame;
	step  = mInterpStep;
	double baseTime = ((double)frame + (double)step/mCAnimInterp.mBody) / (mCMaxFrame.mBody);
	if (mCBlurCtrl.mbUseMotionBlur == TRUE){
	    double blurFactor = mCBlurCtrl.GetFactor();
	    double blurOffset = blurFactor / mCAnimInterp.mBody / mCMaxFrame.mBody;
	    baseTime += blurOffset;
	}
	if (baseTime > 1) baseTime = 1;
	return baseTime;
	
    }
    
    double time = ((double)frame + (double)step/mCAnimInterp.mBody) / mCMaxFrame.mBody;
    if (time > 1) time = 1;
    return time;
}

double TimeCtrl_cls::
GetDoubleIndexWAMin(){
    if (mCMaxFrame.mBody == 0) return 0;
    return (double)mCWAMin.mBody / mCMaxFrame.mBody;
}

double TimeCtrl_cls::
GetDoubleIndexWAMax(){
    if (mCMaxFrame.mBody == 0) return 0;
    return (double)mCWAMax.mBody / mCMaxFrame.mBody;
}

double TimeCtrl_cls::
GetDoubleIndexAAMin(){
    if (mCMaxFrame.mBody == 0) return 0;
    return (double)mCAAMin.mBody / mCMaxFrame.mBody;
}

double TimeCtrl_cls::
GetDoubleIndexAAMax(){
    if (mCMaxFrame.mBody == 0) return 0;
    return (double)mCAAMax.mBody / mCMaxFrame.mBody;
}

int TimeCtrl_cls::
GetOffsetForImgSeq(int frame, int step){
    if (frame == -1) frame = mFrame;
    if (step == -1) step = mInterpStep;
    int tmp = frame - mCWAMin.mBody; // uintłȂȂ悤int
    int offset = tmp / mCAnimStep.mBody * mCAnimInterp.mBody + step;
    return offset;
}

int TimeCtrl_cls::
RoundFrame(int frame){
    int roundFrame = frame + mCAnimStep.mBody / 2;
    roundFrame /= mCAnimStep.mBody;
    roundFrame *= mCAnimStep.mBody;
    while (roundFrame > (int)mCMaxFrame.mBody)
	roundFrame -= mCAnimStep.mBody;
    return roundFrame;
}

void TimeCtrl_cls::
AdjustFrameByRange(int *pFrame, int *pInterpStep,
		   int increment, 
		   bool bForceWA, bool bCheckRepeat){

    

    if (*pFrame > (int)mCMaxFrame.mBody){
	if (increment < 0) 
	    if (bCheckRepeat && mCbRepeat.mbBody)
		*pFrame = mCMaxFrame.mBody;
	    else
		*pFrame = 0;
	else
	    if (bCheckRepeat && mCbRepeat.mbBody)
		*pFrame = 0;
	    else
		*pFrame = mCMaxFrame.mBody;
    }

    int WAMin = mCWAMin.mBody, WAMax = mCWAMax.mBody;
    if (mbUseCustomWA == TRUE){
	WAMin = mCustomWAMin;
	WAMax = mCustomWAMax;
    }
    if (bForceWA == TRUE){
	if (*pFrame >= WAMax){
	    *pFrame = WAMax; 
	    *pInterpStep = 0;
	}
	if (*pFrame < WAMin) {
	    *pFrame = WAMin; 
	    *pInterpStep = 0;
	}
    }

    // ǂ?
    *pFrame /= mCAnimStep.mBody;
    *pFrame *= mCAnimStep.mBody;

    //return frame;
}

unsigned int TimeCtrl_cls::
CalcNextFrame(int frame, int increment, bool bForceWA, bool bCheckRepeat){
    if (frame == _TIME_CTRL_CURRENT) frame = mFrame;
    int nextFrame = frame + mCAnimStep.mBody * increment;
    int nextStep = mInterpStep;

    AdjustFrameByRange(&nextFrame, &nextStep, increment, bForceWA, bCheckRepeat);
    
    return nextFrame;
}

int TimeCtrl_cls::
CalcNextStep(int step, int dstep){
    if (step == _TIME_CTRL_CURRENT) step = mInterpStep;
    int nextStep = step + dstep;
    return nextStep;
}

void TimeCtrl_cls::
CalcNextFrameFromStep(int *pFrame, int *pStep, bool bForceWA, bool bCheckRepeat){
    int frame = *pFrame; int step = *pStep;
    int increment = 0;
    if (frame == _TIME_CTRL_CURRENT) frame = mFrame;
    if (step == -1) {
	increment = -1;
	frame -= mCAnimStep.mBody; 
	step = mCAnimInterp.mBody - 1; 
    }
    if (step >= mCAnimInterp.mBody){
	increment = +1;
	frame += mCAnimStep.mBody;
	step = 0;
    }
    //frame = 
    AdjustFrameByRange(&frame, &step, increment, bForceWA, bCheckRepeat);
    
    *pFrame = frame;
    *pStep = step;
    
    //return (increment != 0);
}

unsigned int TimeCtrl_cls::
GetIndexFromDouble(double doubleIndex){
    if (mCMaxFrame.mBody == 0) return 0;
    return (doubleIndex * mCMaxFrame.mBody) + 0.1;
}

bool TimeCtrl_cls::
IsTheFrameReady(int frame, bool bUseInterpolate){
    double indexDouble = GetRemappedIndexDouble(frame);
    int index = (int)indexDouble;
    double indexmod = indexDouble - index;
    if (bUseInterpolate == TRUE) indexmod = 1;
    return mpCMainLink->IsTheIndexReady(index, indexmod);

}

bool TimeCtrl_cls::
IsTheFrameLoaded(int frame){
    int index = GetRemappedIndex(frame);
    return mpCMainLink->IsTheIndexLoaded(index);
}

bool TimeCtrl_cls::
IsWorkingArea(int frame){
    if (frame < (int)mCWAMin.mBody) return FALSE;
    if (frame > (int)mCWAMax.mBody) return FALSE;
    if (frame > (int)mCMaxFrame.mBody) return FALSE;

    return TRUE;
}

bool TimeCtrl_cls::
IsPlayConditionSatisfied(int frame){
    // FIX_ME !!
    return IsWorkingArea(frame);
}

void TimeCtrl_cls::
SetIndexWithinWA(){
    if (mFrame > (int)mCWAMax.mBody) mFrame = mCWAMax.mBody;
    if (mFrame < (int)mCWAMin.mBody) mFrame = mCWAMin.mBody;
}

bool TimeCtrl_cls::
OpenCustomRegionDialog(wxWindow *pC){
    /*
    MultiNumericalInputs_cls *pCurr;

    pCurr = new MultiNumericalInputs_cls (&pCurr, pC, wxID_ANY, wxT("Render Area"));
    pCurr->SetDefaultButtonType(_MN_OK_CANCEL_DEFAULT);
    pCurr->mTextWidth = 80;
    int boxWidth = 2;
    pCurr->CombinePointer(_GP_BOXSIZERHORIZONTAL, &boxWidth, 
			  _("init/fin"));
    pCurr->CombinePointer(_GP_UINT_SPIN, &mCustomRegionMin);
    pCurr->SetRange(0, mCMaxFrame.mBody);
    pCurr->CombinePointer(_GP_UINT_SPIN, &mCustomRegionMax);
    pCurr->SetRange(0, mCMaxFrame.mBody);

    bool bRet = (pCurr->FinishSettingDLG(TRUE, _MN_DLG_MODAL, _MN_DLG_MAKEGEOM)
		 == wxID_OK);
    
    delete pCurr;
    */
    
    return TRUE;
}

void TimeCtrl_cls::
EmptySnippingTag(){
    if (mVbSnippingTag.size() != mCMaxFrame.mBody + 1)
	mVbSnippingTag.resize(mCMaxFrame.mBody + 1);
    for (unsigned int i = 0; i < mVbSnippingTag.size(); i++)
	mVbSnippingTag[i] = FALSE;
}

void TimeCtrl_cls::
UpdateSnippingTag(){
    EmptySnippingTag();
    mpCMainLink->mpCPDataSeqCtrl->ClearSnipping();
    mpCParamGUICtrlLink->UpdateSnippingTag(&mVbSnippingTag);
}

int TimeCtrl_cls::
GetSnipNext(){
    unsigned int i = mFrame + 1;
    int nextStep = mFrame;
    for (; i <= mCMaxFrame.mBody; i++)
	if (mVbSnippingTag[i] == TRUE) {
	    nextStep = i;
	    break;
	}
    if (mFrame == nextStep) nextStep = mCMaxFrame.mBody;

    return nextStep;
}

int TimeCtrl_cls::
GetSnipPrev(){
    int i = mFrame - 1;
    int prevStep = mFrame;
    for (; i >= 0; i--)
	if (mVbSnippingTag[i] == TRUE) {
	    prevStep = i; 
	    break;
	}
    if (mFrame == prevStep) prevStep = 0;
    
    return prevStep;
}

void TimeCtrl_cls::
SetWAInHere(){
    mCWAMin.mBody = mFrame;
    CheckWorkingArea();
}

void TimeCtrl_cls::
SetWAOutHere(){
    mCWAMax.mBody = mFrame;
    CheckWorkingArea();
}

void TimeCtrl_cls::
SetIndex(unsigned int index){
    mFrame = index;
    mFrame = CalcNextFrame(_TIME_CTRL_CURRENT, 0);
    CheckWorkingArea();
}

void TimeCtrl_cls::
StartPlayCommand(int direction){
    if (mbPlayTex == FALSE){
	// ĐJn̂݋L^Ƃ
	GCUndoList.PushNewCommand(new CommandStartAnim_cls(this, direction),
				  new CommandStopAnim_cls(this, mFrame));
	GCUndoList.ExecCurrent();
    }
    else{
	mCStopWatchA.Pause();
	ProcStopPlaying();
    }
    mpCGUIPlay->UpdateButtonColor();
}

void TimeCtrl_cls::
StartPlay(int direction){
    mbPlayTex = TRUE;
    mPlayDirection = direction;
    int nextStep = CalcNextFrame(_TIME_CTRL_CURRENT, direction);
    if (IsPlayConditionSatisfied(nextStep) == FALSE){
	MoveFrameCommand(direction);
	mpCGUIPlay->RefreshGLPanels();
	return;
    }
    mCStopWatchA.Start();
}

void TimeCtrl_cls::
StopAnimationAt(int frame){
    mbPlayTex = FALSE;
    SetFrameNative(frame);
}

void TimeCtrl_cls::
ProcStopPlaying(){
    mbPlayTex = FALSE;
    mbUseCustomWA = FALSE;
    mCBlurCtrl.EndMotionBlur();
    
    // ̈[ł̏IJԂ̏ꍇ
    if (mCbRepeat.mbBody == TRUE && 
	((mPlayDirection > 0 && mFrame == mCWAMax.mBody) ||
	 (mPlayDirection < 0 && mFrame == mCWAMin.mBody))){
	mbPlayTex = TRUE;
	mpCMainLink->mpCFrame->mbRepeatPlay = TRUE;
	if (mPlayDirection > 0) mFrame = mCWAMin.mBody;
	else                    mFrame = mCWAMax.mBody;
    }

    mpCGUIPlay->UpdateButtonColor();
	
    // Aԕۑ̃{^̐FXV邽߂ɂx
    wxCommandEvent com(wxEVT_COMMAND_MENU_SELECTED, menuPlayStop);
    mpCMainLink->GetPTopLevelWindow()->AddPendingEvent(com);
}

bool TimeCtrl_cls::
ProcPlaying(){
    bool bCalcNextStep = TRUE;
    int nextStep;
    int nextFrame;

    if (mCBlurCtrl.mbUseMotionBlur == TRUE)
	if (mCBlurCtrl.Increment(mPlayDirection) == FALSE)
	    bCalcNextStep = FALSE;
    
    if (bCalcNextStep == TRUE){
	nextStep = CalcNextStep(_TIME_CTRL_CURRENT, mPlayDirection);
	nextFrame = mFrame;
	CalcNextFrameFromStep(&nextFrame, &nextStep);
	if (nextFrame == mFrame && nextStep == mInterpStep){
	    ProcStopPlaying();
	    return FALSE;
	}
    }
    else{
	nextStep = mInterpStep;
	nextFrame = mFrame;
    }
    
    while(TRUE){
	if ((mCStopWatchWait.Time() - mClockLastDraw) >= mWaitDuration) break;
	wxMilliSleep(3);
    }
    
    mClockLastDraw = mCStopWatchWait.Time();
    
    SetFrameNative(nextFrame);
    SetInterpStepNative(nextStep);
    mpCGUIPlay->RefreshGLPanels();
    
    if (mFrame == mCMaxFrame.mBody) {
	ProcStopPlaying();
	return FALSE;
    }

    return TRUE;
}

// -----

void TimeCtrl_cls::
ReadStrLine(std::wistream &istream, wxString strLine){
    if (mCConfListTime.ReadStrLineAll(strLine) == TRUE){
	Parameters_cls::SetFrameNumber(mCMaxFrame.mBody, mCAnimInterp.mBody);
	mpCGUIPlay->UpdateDisplayedValues();
    }

    if (strLine == wxT("---TimeRemap")){
	std::wstring wstrLine;
	wxString strLine;
	while ( std::getline (istream, wstrLine)){
	    strLine = wstrLine.c_str();
	    if (strLine.Find(wxT("ParamGroup=")) != wxNOT_FOUND) break;
	    //mCConfList.ReadStrLineAll(strLine);
	}
	mpCTimeRemap->ReadStrLine(istream, strLine);
    }

    /*
    std::wstring::size_type loc = strLine.find(L"--TimeRemap");
    if (loc != std::wstring::npos){
	std::wstring strLine;
	std::getline(istream, strLine);
	mCTimeRemap.ReadStrLine(istream, strLine);
    }
    */
}

wxString TimeCtrl_cls::
Serialize(){
    wxString sOut;
    sOut << mCConfListTime.SerializeAll();
    if (mCbUseTimeRemap.mbBody == TRUE){
	sOut << wxT("---TimeRemap\n");
	sOut << mpCTimeRemap->Serialize();
    }
    return sOut;
}

//DF3SeqGUI_cls* TimeCtrl_cls::mspCDF3SeqGUI = NULL;
//DF3SeqCtrl_cls* TimeCtrl_cls::mspCDF3SeqCtrlLink = NULL;
