
#ifndef _UNDO_MANAGER_H_
#define _UNDO_MANAGER_H_

#include <vector>
#include <stdio.h>

#include <wx/wx.h>

#define _COMMAND_UNDO true
#define _COMMAND_REDO false

#define _COMMAND_TYPE_NORMAL 0
#define _COMMAND_TYPE_BLOCK_OPEN 1
#define _COMMAND_TYPE_BLOCK_CLOSE 2

#define _COMMAND_UPDATE_DISPLAY_ATTRIB 1
#define _COMMAND_UPDATE_DISPLAY_GL     2
#define _COMMAND_UPDATE_DISPLAY_TL     4
#define _COMMAND_UPDATE_DISPLAY_IMGSEQ 8
#define _COMMAND_UPDATE_DISPLAY_CAMCONF 16
#define _COMMAND_UPDATE_DISPLAY_COLOR 32
#define _COMMAND_UPDATE_DISPLAY_EXPORTER 64
#define _COMMAND_UPDATE_DISPLAY_DRAWOPTIONS 128
#define _COMMAND_UPDATE_DISPLAY_LIGHT 256
#define _COMMAND_UPDATE_DISPLAY_EXPOSURE 512
#define _COMMAND_UPDATE_DISPLAY_STEREO 1024
#define _COMMAND_UPDATE_DISPLAY_FILTER 2048
#define _COMMAND_UPDATE_DISPLAY_FILTER_PARAM 4096
#define _COMMAND_UPDATE_SNAPSHOT_DATA  8192
#define _COMMAND_UPDATE_DISPLAY_TEXTURE 16384
#define _COMMAND_UPDATE_DISPLAY_SB 32768

class CommandItem_cls;

class CommandItem_cls {
public:
    CommandItem_cls();
    virtual ~CommandItem_cls();
    
    virtual void Execute(bool bFlag = _COMMAND_REDO) = 0;
    
    void UpdateBit();
    static void UpdateBit(long long updateBit);
    int mCommandType;
    wxString mstrUndoTag;
    wxString mstrID; // p[^̃^O
    long long mUpdateBit;
};

// ----------- UNDO BLOCK -------------

class ComBlockOpen : public CommandItem_cls {
public:
    ComBlockOpen(wxString strUndoTag = wxT(""));
    void Execute(bool bFlag = _COMMAND_REDO);
};

class ComBlockClose : public CommandItem_cls {
public:
    ComBlockClose(wxString strUndoTag = wxT(""));
    void Execute(bool bFlag = _COMMAND_REDO);
};

// -----------------------------------

class UndoManager_cls {
 public:
    UndoManager_cls();
    ~UndoManager_cls();

    std::vector < CommandItem_cls* > mVPCRedoList;
    std::vector < CommandItem_cls* > mVPCUndoList;
    unsigned int mCurrentIndex;

    int  mBlockDepth;
    long long mUpdateAccumulateBit;

    wxString GetStrList();
    
    void Clear();
    void RemoveItem(unsigned int i);
    void OpenBlock(wxString strTag = wxT(""));
    void CloseBlock(wxString strTag = wxT(""));
    void SendEventToFrame(int comID);
    void ExecUpdate();

    void PushNewCommand(CommandItem_cls* pComRedo, CommandItem_cls* pComUndo);

    int  GetRedoCommandType(int level = 1);
    int  GetUndoCommandType(int level = 1); 
    void SkipCurrent();
    void ExecCurrent();

    void UndoCurrent();
};

#endif
