// -*- c -*- 
// "clsBitmapPanel.h"

#include "clsBitmapPanel.h"
#include "../WIDGETCTRL/WX/myArtProv.h"

#include "../COMMON/CommonDefinitions.h"

BEGIN_EVENT_TABLE(BitmapPanel_cls, wxPanel)
     EVT_PAINT       (BitmapPanel_cls::OnPaint)
     EVT_MOUSE_EVENTS(BitmapPanel_cls::OnMouse)
END_EVENT_TABLE()

BitmapPanel_cls::
BitmapPanel_cls(wxWindow *parent, wxWindowID id, 
	       wxPoint pos, wxSize size) :
    wxPanel(parent, id, pos, size, wxNO_BORDER),
    mpBitmapDisable(NULL)
{
    mbDeleteAtDestroyDisable = FALSE;
    mbButton = TRUE;
    mbBorder = FALSE;
    mState = 0;
    mbPushed = 0;
    mBoarderX = 0;
    mBoarderY = 0;

    Connect(GetId(), wxEVT_KEY_DOWN,
	    wxKeyEventHandler(BitmapPanel_cls::OnKeyDown));

    mpCMouseEventTarget = NULL;
}

BitmapPanel_cls::
~BitmapPanel_cls(){
    for (int i = 0; i < mVPBitmap.size(); i++)
	if (mVbDeleteAtDestroy[i] == TRUE) delete mVPBitmap[i];
    
    if (mpBitmapDisable != NULL &&
	mbDeleteAtDestroyDisable == TRUE) delete mpBitmapDisable;
}

void BitmapPanel_cls::
SetAspect(int orient, float aspect){
    wxSize CSize = GetSize();
    if (orient == wxVERTICAL){
	CSize.SetWidth(CSize.GetHeight()*aspect);
	SetMinSize(CSize);
	SetMaxSize(CSize);
    }
    if (orient == wxHORIZONTAL){
	CSize.SetHeight(CSize.GetWidth()*aspect);
	SetMinSize(CSize);
	SetMaxSize(CSize);
    }
}

void BitmapPanel_cls::
SetBoarder(int boarderX, int boarderY){
    mBoarderX = boarderX;
    mBoarderY = boarderY;
}

void BitmapPanel_cls::
SetMouseEventTarget(wxWindow *pCTgt){
    mpCMouseEventTarget = pCTgt;
}

void BitmapPanel_cls::
AddBitmapP(wxBitmap* pBitmap, bool bDeleteAtDestroy){
    mVPBitmap.push_back(pBitmap);
    mVbDeleteAtDestroy.push_back(bDeleteAtDestroy);
}

void BitmapPanel_cls::
SetBitmapPDisable(wxBitmap *pBitmap, bool bDeleteAtDestroy){
    mpBitmapDisable = pBitmap;
    mbDeleteAtDestroyDisable = bDeleteAtDestroy;
}

void BitmapPanel_cls::
AddBitmapStr(wxString strBitmap){
    mVstrBitmap.push_back(strBitmap);
}

void BitmapPanel_cls::
SetBitmapStrDisable(wxString strBitmap){
    mstrBitmapDisable = strBitmap;
}

void BitmapPanel_cls::
UpdateDisplayUseP(wxPaintDC& dc){
    
    if (mVPBitmap.size() == 0) return;
    if ((IsEnabled() == FALSE) && (mpBitmapDisable != NULL)){
	dc.DrawBitmap(*mpBitmapDisable, mBoarderX, mBoarderY, TRUE);
	return;
    }

    if (mState >= mVPBitmap.size()) return;

    if (mVPBitmap[mState]->IsOk() == TRUE)
	dc.DrawBitmap(*mVPBitmap[mState], mBoarderX, mBoarderY, TRUE);
    else if (mbBorder == TRUE){
	wxSize size = GetClientSize();
	dc.SetPen(wxPen(wxColour(128, 128, 128)));
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	dc.DrawRectangle(0, 0, size.GetWidth()-1, size.GetHeight()-1);
    }
}

void BitmapPanel_cls::
UpdateDisplay(wxPaintDC& dc){

    wxSize size = GetClientSize();
    size.IncBy(-mBoarderX*2, -mBoarderY*2);
    if (mVPBitmap.size() != 0) return UpdateDisplayUseP(dc);
    
    if (mVstrBitmap.size() == 0) return;
    if ((IsEnabled() == FALSE) && (mstrBitmapDisable.IsEmpty() == FALSE)){
	dc.DrawBitmap(ARTSIZE(mstrBitmapDisable, size), mBoarderX, mBoarderY, TRUE);
	return;
    }

    if (mState >= mVstrBitmap.size()) return;
    if (mVstrBitmap[mState].IsEmpty() == FALSE)
	dc.DrawBitmap(ARTSIZE(mVstrBitmap[mState], size), mBoarderX, mBoarderY, TRUE);
    else if (mbBorder == TRUE){
	dc.SetPen(wxPen(wxColour(128, 128, 128)));
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	dc.DrawRectangle(0, 0, size.GetWidth()-1, size.GetHeight()-1);
    }
}

void BitmapPanel_cls::
OnPaint(wxPaintEvent& event){
    wxPaintDC dc(this);
    UpdateDisplay(dc);
    event.Skip();
}

void BitmapPanel_cls::
OnMouse(wxMouseEvent& event){
    if (event.Leaving() == TRUE){
	mbPushed = FALSE;
	SetBackgroundColour(wxColour(wxNullColour));
	Refresh(TRUE);
    }

    if (event.ButtonDown(wxMOUSE_BTN_LEFT)){
	mbPushed = TRUE;
	if (mbButton == TRUE)
	    ;//SetBackgroundColour(wxColour(wxT("GREY")));
	Refresh(TRUE);
    }

    if ((event.ButtonUp(wxMOUSE_BTN_LEFT) && mbPushed == TRUE) ||
	event.ButtonDClick(wxMOUSE_BTN_LEFT)){
	mState++;
	
	if (mVPBitmap.size() != 0){
	    if (mState >= mVPBitmap.size()) mState = 0;
	}
	else
	    if (mState >= mVstrBitmap.size()) mState = 0;

	mbPushed = FALSE;
	SetBackgroundColour(wxColour(wxNullColour));
	Refresh(TRUE);
	if (mpCMouseEventTarget == NULL)
	    SendMouseEvent(GetParent(), wxEVT_LEFT_UP, GetId());
	else
	    SendMouseEvent(mpCMouseEventTarget, wxEVT_LEFT_UP, GetId());
    }
    event.Skip();
}

wxWindow* BitmapPanel_cls::
GetPWindow(){
    return this;
}

void BitmapPanel_cls::
SetStatus(unsigned int status){
    mState = status;
    Refresh();
}

void BitmapPanel_cls::
OnKeyDown(wxKeyEvent& event){
    ProcessKeyEvent(event);
}
